/*************************************************************************
* NASA Exoplanet Science Institute (NExScI) -- http://nexsci.caltech.edu *
*	Copyright (c) 1997-2009 California Institute of Technology       *
*                     U.S. Government Sponsorship                        *
* Under contract with the National Aeronautics and Space Administration  *
**************************************************************************/
// nbCalib -- calibration of narrow-band visibility amplitudes for PTI
// and KI -- Andy Boden 1997 through 2003

// now wrapped by a Perl script.  This C++ core of the nbCalib app has
// been renamed to wbcEngine.cc -- David A. Imel 2004

// *************************
// nbNullCalib version  RLA/JSH 03-27-2008


#include <config.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h> 
#include "getopt.h"

#include "angles.h"
#include "Vector.h"

#include "timeUtil.h"

#include "KI.h"
#include "baseline.h"
#include "coords.h"

#include "nbNullStarList.h"

#include <map>
#include <string>
#include <iostream>
using namespace std;

#define LINE_LENGTH 2048

#define MAX_TARGETS 2000
#define MAX_PSEUDONYMS 10
#define MAX_CALIBRATORS 2000

// Maximum number of spectral channels
#define MAX_CHANNELS 64

#define LAMBDA 2.2e-6
#define MJD_2000 51544.0

// Delay Check tolerance -- approximately 3 arcminutes
#define DELAY_CHECK_TOLERANCE 0.10


// Global declarations

// Create the default baseline objects.  Use the form of the
// constructor with delay bias -- we'll be checking delay values
baseline KI_K1K2(KI_LONGITUDE,KI_LATITUDE,
		 KI_Baseline_K1K2,KI_K1K2_DELAY_BIAS,"KI_K1K2");

baseline *defBaseline = &KI_K1K2;
char baselineFile[LINE_LENGTH];

// Calibration Target Names, Counts, and Positions (arrays -- sorry)
char *targetName[MAX_TARGETS][MAX_PSEUDONYMS];
int nTargets,nTargetNames[MAX_TARGETS];
Vector *targetV[MAX_TARGETS];
coords *targetCoord[MAX_TARGETS];


// Calibrator Names, Count, and Positions  (arrays -- sorry)
int nCalibrators;
char calName[MAX_CALIBRATORS][132];
Vector *calibV[MAX_CALIBRATORS];
coords *calibCoord[MAX_TARGETS];
double calSize[MAX_CALIBRATORS],sigmaSize[MAX_CALIBRATORS];


// Program Operational Options.  These are used to control the code's
// processing behavior.  Global variables are the best mechanism I've
// come up with to handle this situation...

// Calibration time window for picking up calibration scans.
// units are in days, so 1/24 is +/- 1 hour time window
double calibTimeWindow = (1.0 / 24.0);

int usingTempErrorProp = 1;
double timeErrorPropCoeff = (1.0 * 24.0);

// Propagating null errors across the sky.  See sysNullEnsemble.
// Units are in inverse radians.
int usingSkyErrorProp = 1;
double angleErrorPropCoeff = (1.0 / (15.0 * DEG_TO_RAD));

// Minimum possible scan uncertainty
int usingMinimumScanUncertainty = 0;
int usingMinimumCalUncertainty = 0;
double minScanUncertainty = 0.01;
double minCalUncertainty = 0.01;


// Thresholds for reporting/rejecting calibration errors (dimensionless).

// Threshold for reporting calibration errors -- 2.0 sigma disagreement
double chiWarning = 2.0 * 2.0;
// Threshold for rejecting the calibration -- 3.0 sigma disagreement
double chiReject = 3.0 * 3.0;

// Verbosity variable to control reporting Added 7/01 A.B.
#define MUTE 0
#define NORMAL 1
#define EXCESSIVE 2
int verbose = NORMAL; // Default to certain errors


// Variable for delay checking
int delayCheck = 1;
double delayCheckTol = DELAY_CHECK_TOLERANCE;


/************************************************************************
*  printHelpMessage -- local to calib.cc (A.B. 4/13/98)			*
*									*
*  A.  Purpose:	The function printHelpMessage prints a help message     *
*       for the calib program to stderr and then exits the program.	*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description: None					*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration printHelpMessage  *********************/
void printHelpMessage()
{
/**********  begin printHelpMessage  ************************************/
  (void) fprintf(stdout,"Usage: nbNullCalib calScript nullsumfile1 [nullsumfile2...] [> cal.out]\n\n");
  (void) fprintf(stdout,(char *) "Calibrate KI nuller data\n\n");
  (void) fprintf(stdout,"Options:\n");
  (void) fprintf(stdout,(char *) "  -help: print (this) help message and exit (overrides all other options)\n");
  (void) fprintf(stdout,(char *) "  -version: print version information and exit\n");
  (void) fprintf(stdout,"  -calTimeWindow argument: set calibration scan acceptance time window (hr)\n");
  (void) fprintf(stdout,"  -[no]timeWeighting: set use/nonuse of time proximity weighting of calibration scans\n");
  (void) fprintf(stdout,"  -timeErrorVal argument: if timeWeighting, set calibration scan time weighting value (hr)\n");
  (void) fprintf(stdout,"  -[no]angleProxWeighting: set use/nonuse of sky angle proximity weighting of calibration scans\n");
  (void) fprintf(stdout,"  -angleErrorVal argument: if angleProxWeighting, set calibration scan angle weighting value (deg)\n");
  (void) fprintf(stdout,"  -[no]minScanUncertainty argument: sets [non]use of minimum possible scan uncertainty of value argument (V2 units)\n");
  (void) fprintf(stdout,(char *) "  -[no]minCalUncertainty argument: sets [non]use of minimum possible cal uncertainty of value argument (V2 units)\n");
  (void) fprintf(stdout,"  -chiWarning argument: set calibrator inconsistency warning chi threshold \n");
  (void) fprintf(stdout,"  -chiReject argument: set calibrator inconsistency scan rejection chi threshold \n");
  (void) fprintf(stdout,(char *) "  -[no]delayCheck[=argument]: set use/nonuse of delay value checking \n\t (optional delay tolerance argument) \n");
  (void) fprintf(stdout,"  -verbose|normal|quiet|shutUp: control volume of information, warning, and error messages\n");
  (void) fprintf(stdout,(char *) "  -baseline <file>: Use the baseline supplied by <file>. Normal behavior is to look for a\n");
  (void) fprintf(stdout,(char *) "                    yyyyddd.nullbaseline file for each yyyyddd.nullspec file given on the command\n");
  (void) fprintf(stdout,(char *) "                    line. If that fails then look for yyyyddd.nullbline. If neither is found then\n");
  (void) fprintf(stdout,(char *) "                    nbNullCalib uses a built in default baseline. The option allows the user to\n");
  (void) fprintf(stdout,(char *) "                    override that search by providing the filename directly. Note that when\n");
  (void) fprintf(stdout,(char *) "                    multiple SPEC files are given in combination with this option all of them\n");
  (void) fprintf(stdout,(char *) "                    use the supplied <file>.\n");

  (void) fprintf(stdout,"  note:  command-line options may be used in any order, interspersed with non-options\n");
  (void) fprintf(stdout,"   and abbreviated to uniqueness (e.g. -h, -noang, -incoh).\n");
  (void) fprintf(stdout,(char *) "  Find more nbNullCalib info at http://nexsci.caltech.edu/software\n");

  (void) exit(-1);


/**********  end printHelpMessage  **************************************/
}


/************************************************************************
*  processCmdLineOpts -- local to calib.cc (A.B. 4/12/98)		*
*									*
*  A.  Purpose:	The function processCmdLineOpts processes the command   *
*       line invocation options used with calib for this run.		*
*									*
*  B.  Background and use: processCmdLineOpts uses getopt_long_only to  *
*       process the command line options the user has added to the      *
*       invocation.  getopt_long_only is used because it allows the     *
*       arbitrarily intersperse command-line options along with the     *
*       filename command-line arguments: it permutes the command line   *
*       as it processes so when finished all the non-options (i.e. the  *
*       filename arguments) are at the end (for simple processing by    *
*       the main routine).				                *
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        argc, argv -- the standard command-line argument variables.    *
*	Output -							*
*        argv's contents are permuted so as to move all non-option      *
*         arguments to the end.                                         *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the index of the first     *
*       non-option on the command-line, which is presumably the name    *
*       of the calibration script file (int).              		*
*									*
*************************************************************************/
/**********  function declaration processCmdLineOpts  *******************/
int processCmdLineOpts(int argc, char **argv)
{
/**********  local declarations processCmdLineOpts  *********************/
  // Valid command-line options
  int retVal,optIndex,argsProcessed = 0;
  static struct option longOptions[] = 
  {
    {((char *) "help"),no_argument,0,0},                     // Option 0
    {((char *) "version"),no_argument,0,0},                  // Option 1
    {((char *) "timeWeighting"),no_argument,0,0},            // Option 2
    {((char *) "notimeWeighting"),no_argument,0,0},          // Option 3
    {((char *) "angleProxWeighting"),no_argument,0,0},       // Option 4
    {((char *) "noangleProxWeighting"),no_argument,0,0},     // Option 5
    {((char *) "calTimeWindow"),required_argument,0,0},      // Option 6
    {((char *) "timeErrorVal"),required_argument,0,0},       // Option 7
    {((char *) "angleErrorVal"),required_argument,0,0},      // Option 8
    {((char *) "minScanUncertainty"),required_argument,0,0}, // Option 9
    {((char *) "nominScanUncertainty"),no_argument,0,0},     // Option 10
    {((char *) "chiWarning"),required_argument,0,0},         // Option 11
    {((char *) "chiReject"),required_argument,0,0},          // Option 12
    {((char *) "quiet"),no_argument,0,0},                    // Option 13
    {((char *) "normal"),no_argument,0,0},                   // Option 14
    {((char *) "verbose"),no_argument,0,0},                  // Option 15
    {((char *) "shutUp"),no_argument,0,0},                   // Option 16
    {((char *) "delayCheck"),optional_argument,0,0},         // Option 17
    {((char *) "nodelayCheck"),no_argument,0,0},             // Option 18
    {((char *) "baseline"),required_argument,0,0},           // Option 19 - baseline file override
    {((char *) "minCalUncertainty"),required_argument,0,0}, // Option 20
    {((char *) "nominCalUncertainty"),no_argument,0,0},     // Option 21

    {0,0,0,0}
  };

  extern char *optarg;
  extern int optind,opterr,optopt;


/**********  begin processCmdLineOpts  **********************************/

  // Print welcome/version
  if (verbose)
    (void) fprintf(stdout,(char *) "## This is nbNullCalib v%s\n\n",(char *) VERSION);

  do {
    if ((retVal = getopt_long_only(argc,argv,"",longOptions,&optIndex)) != EOF) {
      if (verbose)
	(void) printf("# option %s\n",longOptions[optIndex].name);
      argsProcessed++;
      switch (optIndex) {

      case 0: {
	(void) printHelpMessage();
	break;
      } // end printing help message

      case 1: {
	(void) fprintf(stdout,(char *) "nbNullCalib v%s\n\n",(char *) VERSION);
	(void) exit(-1);
	break;
      } // end printing help message

      case 2: {
	usingTempErrorProp = 1;
	break;
      } // end setting usingTempErrorProp -- option 2
	
      case 3: {
	usingTempErrorProp = 0;
	break;
      } // end setting NOT usingTempErrorProp -- option 3

      case 4: {
	usingSkyErrorProp = 1;
	break;
      } // end setting usingSkyErrorProp -- option 4

      case 5: {
	usingSkyErrorProp = 0;
	break;
      } // end setting NOT usingSkyErrorProp -- option 5

      case 6: {
	calibTimeWindow = atof(optarg);
	calibTimeWindow /= 24.0;
	break;
      } // end setting calibTilmWindow -- option 6

      case 7: {
	timeErrorPropCoeff = atof(optarg);
	timeErrorPropCoeff *= 24.0;
	break;
      } // end setting timeErrorPropCoeff -- option 7

      case 8: {
	angleErrorPropCoeff = atof(optarg);
	angleErrorPropCoeff *= DEG_TO_RAD;
	angleErrorPropCoeff = 1.0 / angleErrorPropCoeff;
	break;
      } // end setting angleErrorPropCoeff -- option 8

      case 9: {
	usingMinimumScanUncertainty = 1;
	minScanUncertainty = atof(optarg);
	break;
      } // end setting usingMinimumScanUncertainty/minScanUncertainty  -- option 9

      case 10: {
	usingMinimumScanUncertainty = 0;
	break;
      } // end setting usingMinimumScanUncertainty -- option 10

      case 11: {
	chiWarning = atof(optarg);
	chiWarning *= chiWarning;
	break;
      } // end setting chiWarning -- option 11

      case 12: {
	chiReject = atof(optarg);
	chiReject *= chiReject;
	break;
      } // end setting chiReject -- option 12

      case 13: {
	// be quiet!!!
	verbose = MUTE;
	break; 
      } // end shutting up -- option 13

      case 14: {
	// normal (normal setup, serious warnings and errors)
	verbose = NORMAL;
	break; 
      } // setting normal reporting -- option 14

      case 15: {
	// verbose (everything)
	verbose = EXCESSIVE;
	break; 
      } // end setting excessive reporting -- option 15

      case 16: {
	// SHUT UP!!!
	verbose = MUTE;
	break; 
      } // end setting shutUp -- option 16

      case 17: {
	delayCheck = 1;
	if (optarg) {
	  delayCheckTol = atof(optarg);
	} // end setting delay check tolerance
	break;
      } // end setting delayCheck -- option 17

      case 18: {
	delayCheck = 0;
	break;
      } // end setting delayCheck -- option 18

      case 19: {
	strncpy(baselineFile, optarg, sizeof(baselineFile)-1);
	baselineFile[sizeof(baselineFile)] = '\0';
	break;
      } // option 19

      case 20: {
        usingMinimumCalUncertainty = 1;
        minCalUncertainty = atof(optarg);
        break;
      } // end setting usingMinimumCalUncertainty/minCalUncertainty -- option 20

      case 21: {
        usingMinimumCalUncertainty = 0;
        break;
      } // end setting usingMinimumCalUncertainty -- option 21

      } // end switch
    } // end if
  } while (retVal != EOF);

  if ((argsProcessed) && (verbose))
    (void) printf("# %d command line arguments processed\n",argsProcessed);

  return optind;
    
/**********  end processCmdLineOpts  ************************************/
}


/************************************************************************
*  parseCoordLine -- local to calib.cc (A.B. 11/98)			*
*									*
*  A.  Purpose:	The function parse coordinate line parses the           *
*       input coordinate line for the target/calibrator and extracts    *
*       the pertinent information					*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns a pointer to the           *
*       coordinate object created by as a result of this line		*
*       (coords *).                                                     *
*									*
*************************************************************************/
coords *parseCoordLine(char *line,double *calSize,double *sigmaSize)
{
/**********  local declarations parseCoordLine  *************************/
  char *separator = ((char *) "#"),*tmpstr;
  char decStr[256], *minusSign = ((char *) "-");
  double ra,dec;
  float rah,ram,ras,decd,decm,decs,pmRA,pmDec,plx;
  float cSize,sSize;
  coords *wrk;


/**********  begin parseCoordLine  **************************************/
  (void) strtok(line,separator); // Use strtok to stomp on any comments

  int rn = sscanf(line,(char *) "%lf%f%f%s%f%f%f%f%f%f%f",
		  &ra,&ram,&ras,decStr,&decm,&decs,&pmRA,&pmDec,&plx,
		  &cSize,&sSize);

  dec = atof(decStr);
  int decSign = 1; // handle negative declinations (stupid)!!!
  if (strncmp(decStr,minusSign,1) == 0) // If the first character is a minus...
    decSign = -1;

  switch (rn)
    {
    case 11: // We got the whole smack
      {
	// template 15 21 06.880 +24 57 28.872 -0.029 -0.019 0.00587 1.05  0.1
	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec,pmRA,pmDec,plx);

	if (calSize)
	  *calSize = cSize;
	if (sigmaSize)
	  *sigmaSize = sSize;

	break;
      }

    case 9: // We got full coordinates but no calibrator info
      {
	// template 15 34 41.268 +26 42 52.895  0.135 -0.089 0.04365

	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec,pmRA,pmDec,plx);

	// If we were _expecting_ calibrator information, generate an error
	if (calSize && sigmaSize)
	  {
	    (void) fprintf(stderr,(char *) "Fatal error parsing calibration script\n");
	    exit(-1);
	  }

	break;
      } // end rn = 9

    case 8: // We got positions and calibrator info
      {
	// Template 16 01 02.662 +33 18 12.634 0.70 0.1
	// Note this case is degenerate with coords, pm, no plx, but I
	// have to assume something...
	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec);

	// The calibrator info got read into the pm variables!!!
	if (calSize)
	  *calSize = pmRA;
	if (sigmaSize)
	  *sigmaSize = pmDec;

	break;
      } // end rn = 8

    case 6: // We got positions but no calibrator info
      {
	// Template 16 01 02.662 +33 18 12.634 
	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec);

	// If we were _expecting_ calibrator information, generate an error
	if (calSize && sigmaSize)
	  {
	    (void) fprintf(stderr,(char *) "Fatal error parsing calibration script\n");
	    exit(-1);
	  }

	break;
      } // end rn = 6

    case 4: // We got decimal degrees and calibrator information
      {
	// Template 331.75 25.33 1.0 0.05
	ra *= DEG_TO_RAD;
	dec *= DEG_TO_RAD;
	wrk = new coords(ra,dec);

	if (calSize)
	  *calSize = cSize;
	if (sigmaSize)
	  *sigmaSize = sSize;

	break;
      } // end rn = 4

    case 2: // We got decimal degrees and no calibrator information
      {
	// Template 331.75 25.33
	ra *= DEG_TO_RAD;
	dec *= DEG_TO_RAD;
	wrk = new coords(ra,dec);

	// If we were _expecting_ calibrator information, generate an error
	if (calSize && sigmaSize)
	  {
	    (void) fprintf(stderr,(char *) "Fatal error parsing calibration script\n");
	    exit(-1);
	  }

	break;
      } // end rn = 2

    } // end switch (rn)

  return wrk;

/**********  end parseCoordLine  ****************************************/
}


/************************************************************************
*  readScript -- local to calib.cc (7/27/97)				*
*									*
*  A.  Purpose:	The function readScript reads the input calibration     *
*       script and parses the input on the input.			*
*									*
*  B.  Background and use: readScript reads the input user calibration  *
*       script, parses the contents, and loads a sequence of global     *
*       variables used in subsequent processing.                        *
*         First up are the calibration targets.  The code reads the     *
*       names of the calibration targets to be used in this calibration *
*       run, along with the possibility of pseudonyms.  (Astronomy is   *
*       nothing if not replete with pseudonyms.)  Optionally, target    *
*       positions are read for two reasons: one, to weight the impact   *
*       of calibrators by proximity, and two, to compute the baseline   *
*       projection on the target (u-v, HA) during the output stage --   *
*       this is added as a convenience for some users, as it is         *
*       slightly orthogonal to the main mission of calib.               *
*         Next up are the calibrators.  These are (of course) presumed  *
*       to be single stars, and their names, coordinates, angular sizes *
*       and uncertainties are read into global arrays.                  *
*         Mod 11/10/98 AB -- extended read logic to read full           *
*       astrometric state (pos, PM, plx) for all objects.               *
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        filename -- the (pointer to the) character string containing   *
*         the name of the calibration script to process (char *).       *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*                  nTargets, nTargetNames, targetName, targetV;         *
*                  nCalibrators, calName, calSize, sigmaSize, calibV.   *
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration readScript  ***************************/
void readScript(char *filename)
{
/**********  local declarations readScript  *****************************/
  char line[LINE_LENGTH];
  char *separator = ((char *) ":,;| \t\n"),*tmpstr,*tName;
  int rn;
  float rah,ram,ras,decd,decm,decs;
  double ra,dec;
  FILE *fptr;
  
/**********  begin readScript *******************************************/
  // Open/check on the health of the input script filename
  if (!(fptr = fopen(filename,"r")))
    {
      (void) fprintf(stderr," Fatal Error -- Can't open calibration script file %s\n",filename);
      exit(-1);
    }

  // Read the input information on the target stars
  tmpstr = fgets(line,LINE_LENGTH,fptr);
  while ((!feof(fptr))
	 && (nTargets < MAX_TARGETS)
	 && (strncmp(line,"---",3) != 0))
    if (line[0] != '#') // If this isn't a comment
      {
	// Get the target star name(s)
	nTargetNames[nTargets] = 0;
	while ((tName = strtok(tmpstr,separator))
	       && (nTargetNames[nTargets] < MAX_PSEUDONYMS))
	  if (tName[0] != '#')
	    {
	      tmpstr = NULL;
	      targetName[nTargets][nTargetNames[nTargets]] = strdup(tName);
	      nTargetNames[nTargets]++;
	    } // end parsing target names

	// Write out an informational message on the names
	if (verbose)
	  (void) printf("# Calibrating star %s ",targetName[nTargets][0]);
	if (nTargetNames[nTargets] > 1)
	  {
	    if (verbose)
	      (void) printf("(and pseudonyms ");
	    for (int tmp = 1; tmp < nTargetNames[nTargets]; tmp++)
	      if (verbose)
		(void) printf("%s ",targetName[nTargets][tmp]);
	    if (verbose)
	      (void) printf(") ");
	  } // end pseudonym loop

	// Check to see if a position has been given for the target
	(void) fgets(line,LINE_LENGTH,fptr);
	if ( (line[0] != '#')
	      && (((rn = sscanf(line,"%f%f%f%f%f%f",
			       &rah,&ram,&ras,&decd,&decm,&decs)) == 6)
		  || ((rn = sscanf(line,"%lf%lf",&ra,&dec)) == 2)))
	  {
	    targetCoord[nTargets] = parseCoordLine(line,0,0);
	    targetV[nTargets]
	      = new Vector(targetCoord[nTargets]->topocentricPos(MJD_2000,KI_LATITUDE,KI_LONGITUDE,KI_ALTITUDE));

	    if (verbose)
	      (void) printf(" at %s\n",targetCoord[nTargets]->print());

	    // If we were successful in interpreting this line as
	    // coordinates, read in the next one (presumably the next
	    // target name(s) or comment
	    tmpstr = fgets(line,LINE_LENGTH,fptr);

	  } // end read/decode target coordinates
	else
	  {
	    targetCoord[nTargets] = (coords *) NULL;
	    targetV[nTargets] = (Vector *) NULL;
	  }

	if (verbose)
	  (void) printf("\n");
	nTargets++;
      } // end processing target/while target loop
    else // This was a comment, echo it and get next line
      {
	if (verbose > NORMAL)
	  printf("%s",line);
	tmpstr = fgets(line,LINE_LENGTH,fptr);
      } // end comment processing

  if (feof(fptr))
    {
      (void) fprintf(stderr," Fatal Error -- premature eof encountered in %s\n",filename);
      exit(-1);
    }
  else
    if (verbose)
      (void) printf("#  with respect to ");


  // Get the calibrators
  // Read the next line for the number of calibrators
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,"%d",&nCalibrators) != 1)
    {
      (void) fprintf(stderr," Fatal Error -- Error reading calibration star information from %s\n");
      exit(-1);
    }
  else
    if (verbose)
      (void) printf("%d calibrators: \n",nCalibrators);

  int i = 0;
  while (!feof(fptr) && (i < nCalibrators)) // loop over calibrator read
    {
      (void) fgets(line,LINE_LENGTH,fptr);
      // Check the line for a comment
      if (line[0] != '#')
	{
	  // Check the read for format
	  if ((rn = sscanf(line,"%s",&calName[i])) == 1)
	    {

	      // Skip over the calibrator name
	      int tmp = 0;
	      while (line[tmp] != ' ')
		tmp++;

	      calibCoord[i] = parseCoordLine(&line[tmp],&calSize[i],&sigmaSize[i]);
	      calibV[i]
		= new Vector(calibCoord[i]->topocentricPos(MJD_2000,KI_LATITUDE,KI_LONGITUDE,KI_ALTITUDE));
	      if (verbose)
		(void) printf("#  star %s  at %s  size %g  error %g",
			      calName[i],calibCoord[i]->print(),calSize[i],sigmaSize[i]);

	      calSize[i] *= MAS_TO_RAD;
	      sigmaSize[i] *= MAS_TO_RAD;

	      if (nTargets == 1)
		if (targetV[0])
		  if (verbose)
		    (void) printf("  %3.2g degrees away from target",
				  angle_between(calibV[i],targetV[0]) * RAD_TO_DEG);

	      if (verbose)
		(void) printf("\n");
	      i++; // increment calibrator counter
	    } // end successful read of calibrator name/info

	} // end check comment field
      else // otherwise echo the comment
	if (verbose > NORMAL)
	  printf("%s",line);
    } // end i/read calibrator loop

  if (i != nCalibrators)
    {
      if (verbose)
	{
	  (void) printf("# Warning: only %d calibrators read from %s (%d expected)\n",
			i,filename,nCalibrators);
	  (void) fprintf(stderr,"# Warning: only %d calibrators read from %s (%d expected)\n",
			 i,filename,nCalibrators);
	}
      nCalibrators = i;
    }

  // Close the script file
  (void) fclose(fptr);

/**********  end readScript  ********************************************/
}


/************************************************************************
*  printRunDirectives -- local to calib.cc (A.B. 4/11/98)		*
*									*
*  A.  Purpose:	The function printRunDirectives prints a report of the  *
*       calibration directives for this run.				*
*									*
*  B.  Background and use: As calib acquires options, it becomes        *
*       increasingly necessary to document the settings of the program  *
*       for a given dataset.  This function dumps a short report on the *
*       settings used in this run to stdout (where the rest of the      *
*       output is also going). 						*
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration printRunDirectives  *******************/
void printRunDirectives()
{

/**********  begin printRunDirectives  **********************************/
  // Reports on some operational (used to be preprocessor) directives
  (void) fprintf(stdout,(char *) "\n# nbNullCalib version %s\n",(char *) VERSION);
  (void) fprintf(stdout,(char *) "# nbNullCalib calibration code settings for this run: \n");

  (void) printf("# > Using calibration scans within +/- %g hr of the target scan\n",
		calibTimeWindow * 24.0);

  if (usingTempErrorProp)
    {
      (void) printf("# > Weighting calibration scans WRT temporal proximity,");
      (void) printf(" variance doubles in %g hr\n",timeErrorPropCoeff / 24.0);
    }
  else
    (void) printf("# > No weighting of calibration scans WRT temporal proximity\n");

  if (usingSkyErrorProp)
    {
      (void) printf("# > Weighting calibration scans WRT sky proximity,");
      (void) printf(" variance doubles in %g deg\n",RAD_TO_DEG / angleErrorPropCoeff);
    }
  else
    (void) printf("# > No weighting of calibration scans WRT sky proximity\n");

  if (usingMinimumScanUncertainty)
    {
      (void) printf("# > Applying minimum calibrated scan uncertainty of %g\n",minScanUncertainty);
    }
  else
    (void) printf("# > No minimum calibrated scan uncertainty\n");

  if (usingMinimumCalUncertainty)
    {
      (void) printf((char *) "# > Applying minimum calibrated cal uncertainty of %g\n",minCalUncertainty);
    }
  else
    (void) printf((char *) "# > No minimum calibrated cal uncertainty\n");

  (void) printf("# > Flagging calibration warnings when multiple calibrators disagree by more than %g sigma\n",
		sqrt(chiWarning));
  (void) printf("# > Rejecting the scan when multiple calibrators disagree by more than %g sigma\n",
		sqrt(chiReject));

  if (delayCheck)
    (void) printf((char *) "# > Comparing model/measured delays with %3g m tolerance\n",
		  delayCheckTol);
  else
    (void) printf((char *) "# > Not comparing model/measured delays\n");
 
 // added with new option by hjb 5/2005 ... just following along
  if(baselineFile[0] != '\0') 
    (void) printf((char *) "# > Using baseline file %s\n", baselineFile);

  (void) printf((char *) "\n\n");


/**********  end printRunDirectives  ************************************/
}

/************************************************************************
*  blineInsert -- local to nbNullCalib.cc (jsh 2008-08)                 *
*                                                                       *
*  A.  Purpose: The function blineInsert adds baselines from file fptr  *
*      to the bline map. Used by readBaseline().                        *
*                                                                       *
*  B.  Background and use:                                              *
*                                                                       *
*  C.  Called by:                                                       *
*                                                                       *
*  D.  Input/output files:                                              *
*                                                                       *
*  E.  Parameter description:                                           *
*       Input -                                                         *
*       Output -                                                        *
*                                                                       *
*  F.  Detail function description:                                     *
*             1.  Local Functions:                                      *
*             2.  Externals:                                            *
*             3.  Globals and Nonlocals used:                           *
*             4.  Global and Nonlocal variables modified:               *
*             5.  Function flow is as follows:                          *
*             6.  Functions called by                                   *
*                                                                       *
*  G.  Function value: This function returns no useable value.          *
*                                                                       *
*************************************************************************/
void blineInsert(FILE *fptr, map<string,baseline*> &blines, char work[LINE_LENGTH]) {
    while(1) {
        baseline *bl = new baseline(fptr);
        if(! bl->error())
           blines.insert(map<string,baseline*>::value_type(bl->baselineName(), bl));
        else
           break;
    }

    if(blines.empty()) {
        (void) printf("# Error: no valid baselines read from %s\n", work);
    } else {
        map<string,baseline*>::iterator pos;
        for(pos = blines.begin(); pos != blines.end(); ++pos)
          (void) printf("# Found baseline %s in %s \n", pos->first.c_str(), work);
    }

/**********  end blineInsert  ************************************/
}


/************************************************************************
*  readBaseline -- local to nbCalib.cc (A.B. 3/2000)			*
*									*
*  A.  Purpose:	The function readBaseline instantiates a baseline       *
*       object to go along with the nullsum file being read/to be read. *
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
//baseline *readBaseline(char *specFilename, char *baselinef)
// updated function interface 2/06, hjb
// updated 2008-08, jsh
map<string,baseline*> readBaseline(char *specFilename, char *baselinef)
{
/**********  local declarations readBaseline  ***************************/
  char work[LINE_LENGTH], *ptr;

  FILE *fptr = NULL;
  baseline *bLine = NULL;
  map<string,baseline*> blines;
  
/**********  begin readBaseline  ****************************************/
  // One.  Extract trailing suffix and append default baseline suffix
  (void) strncpy(work,specFilename,LINE_LENGTH);
  if ((ptr = strstr(work,(const char *) ".nullspec")) != NULL)
    (void) strcpy(ptr, (const char *) ".nullbaseline\0");

  // One.Five hjb 5/2005
  // working in the baseline override
  // the idea is not to disturb any of the original code for this function
  // so we'll take the user supplied file name, check that, and if it's readable,
  // slip it into the "work" variable used by the rest of the fucntion. if it's 
  // not readable, we're going to fall back the the original search strategy for
  // finding a baseline
  if(baselinef[0] != '\0') {
    // make sure the file is available, accessible, around for reading
    // if not, emit warnings and do nothing, thereby falling back to the 
    // standard search for baseline file (nearly transparent to user)
    if((fptr = fopen(baselinef,"r")) == NULL) {
      (void) printf("# Error reading user supplied baseline file %s\n", baselinef);
      (void) printf("# Attempting to open baseline file\n");
    } else {
      // copy baseline into function's idea of a "work" file, protect agnst overflow
      (void) strncpy(work, baselinef, sizeof(work));
      work[sizeof(work)] = '\0';
      blineInsert(fptr, blines, work);
      if(!blines.empty())
         return blines;
    }
  }

  // if user supplied baseline file is missing or not valid, try to read from
  // the xxx.nullbaseline file
  if((fptr == NULL) || (blines.empty())) {
      if((fptr = fopen(work, "r")) == NULL) {
         (void) printf("# Error trying to open baseline file %s\n", work);
      }
      else {
         blineInsert(fptr, blines, work);
         if(!blines.empty())
             return blines;
      }
  }

  // if xxx.nullbaseline file is missing or not valid, try to read from
  // the xxx.nullbline file
  if((fptr == NULL) || (blines.empty())) {
      (void) printf("# Attempting to open bline file\n");
      if( (ptr = strstr(work, (const char *) ".nullbaseline")) != NULL)
        (void) strcpy(ptr, (const char *) ".nullbline\0");
      if((fptr = fopen(work, "r")) == NULL) {
        (void) printf("# Error trying to open bline file %s\n", work);
        (void) printf("# Error Could not read any baseline specification\n");
        return blines;
      }
      blineInsert(fptr, blines, work);
  }
  return blines;

  /**********  end readBaseline  ******************************************/
}


/************************************************************************
*  readSpVline -- local to calib.cc (A.B. 12/18/97)			*
*									*
*  A.  Purpose:	The function readSpVline scans the input text line for  *
*       nChannels of spectrometer null information.			*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns a one (1) on successful    *
*       completion, and a zero (0) otherwise (int).                     *
*									*
*************************************************************************/
/**********  function declaration readSpVline  **************************/
int readSpVline(char *line,int nSkip,int nChannels,
		double *nPhot,double *lambda,
		double *nullVal,double *nullErr,double *ratio)
{
/**********  local declarations readSpVline  ****************************/

  char *separator = ((char *) " \n"),*wkstr = line,*tmpstr;
  int i,j;

/**********  begin readSpVline  *****************************************/
  /* Prototype spec line

   new NULL line format:

NULLSPEC  08/30/2007/15:09:12  15.136667  tau_ceti  10.0  -40.0  0  200.0 0 0 OUT N1-2 0 KI_K1K2  10  8.491  1001.7 0.0371  0.0371  1.0 0.0 0.0 0.0   8.938  776.4 0.0464  0.0464  1.0 0.0 0.0 0.0 9.523  599 0.0729  0.0729  1.0 0.0 0.0 0.0  9.979  594.7 0.0847  0.0847  1.0 0.0 0.0 0.0 10.465  518.4 0.0963  0.0963  1.0 0. 0 0.0 0.0 10.94  400.3 0.1108  0.1108  1.0 0.0 0.0 0.0 11.411  268.5 0.121  0.121  1.0 0.0 0.0 0.0 11.938  175.7 0.1405  0.1405  1.0 0.0 0.0 0.0 12.378  125.1 0.1702  0.1702  1.0 0.0 0.0 0.0  12.746  76.4 0.2154  0.2154  1.0 0.0 0.0 0.0
  */

  // Using strtok technology.  Skip past the stuff we already have.
  tmpstr = strtok(wkstr,separator);
  for (i = 1; i < nSkip; i++)
    tmpstr = strtok(NULL,separator);

  // Now read each spectrometer channel's information.
  for (i = 0; i < nChannels; i++)
    {
      switch (nSkip)
	{
        case 16 :  // Null spec file format
          {
            lambda[i] = atof(tmpstr);
            tmpstr = strtok(NULL,separator);
            nPhot[i] = atof(tmpstr);
            tmpstr = strtok(NULL,separator);
            nullVal[i] = atof(tmpstr);
            tmpstr = strtok(NULL,separator);
            nullErr[i] = atof(tmpstr);
            tmpstr = strtok(NULL,separator);
            ratio[i] = atof(tmpstr);
            tmpstr = strtok(NULL,separator);
            ratio[i] = atof(tmpstr);
            tmpstr = strtok(NULL,separator);
            ratio[i] = atof(tmpstr);
            tmpstr = strtok(NULL,separator);
            ratio[i] = atof(tmpstr);
            tmpstr = strtok(NULL,separator);
            break;
          } // end 16

	} // end switch

    } // end i loop
	
  // If we get here we're okay -- return good-to-go
  return 1;

/**********  end readSpVline  *******************************************/
}


/************************************************************************
*  readSpecFile -- local to calib.cc (A.B. 7/27/97)			*
*									*
*  A.  Purpose:	The function readSpecFile reads the input nullspec file *
*       and extracts the relevant information for storage in the        *
*       linked-list star data structure.				*
*									*
*  B.  Background and use:						*
*         readSpecFile is designed to be called multiple times with     *
*       multiple specfile arguments to construct the database to be     *
*       used in this calibration run.                                   *
*         Input to the routine consists of the name of the specfile to  *
*       be read and parsed, and the (pointer to the) database of        *
*       null information that is being constituted.  On output    	*
*       information from scans successfully read-in from the input      *
*       sumfile have been entered into the database.  All this          *
*       database manipulation is accomplished via a set of class        *
*       libraries described in nbStarList.h (and nbStarList.cc).        *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        filename -- the (pointer to the) nullspecfile filename to be   *
*         processed (char *).                                           *
*        starList -- the (pointer to the) starlist database object      *
*         being constructed for later use (StarList *).                 *
*	Output -							*
         starList -- contents modified.                                 *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the number of valid data   *
*       lines that were successfully read from the input file (int).    *
*									*
*************************************************************************/
int readSpecFile(char *filename,char *baselinef, StarList *starList)
{
/**********  local declarations readSpecFile  ***************************/

  char line[LINE_LENGTH],starName[132],blDesig[132];
  int month,day,year,hour,minute,second,nRead =0,
    nChannels = 0,dataLines = 0,frameRate=0,calFlag=0;
  double jd,ut,delay,longDelay,jitter;
  char numCycles[10],percentGatedCycles[132],duration[10],mode[10],portsUsed[10],producer[10];
 
  FILE *fptr = NULL;
  line[0] = '\0';
  blDesig[0] = '\0';
  starName[0] = '\0';

  // Try to open this spectrometer file
  if (!(fptr = fopen(filename, (char *) "r"))) {
    if (verbose)
      (void) printf("# >> Error opening spectrometer file %s\n",filename);
    return 0;
  }

  // Read/retrieve the baseline corresponding to this spec file (or use
  // default baseline if missing or error)
  // baseline *thisBaseline = readBaseline(filename, baselinef);

  // hjb 2/06 ... use the new constructor in libbaseline which allows for comments
  // and multi-record baseline files
  // in the future this needs update to allow for time associate of baseline to data record
  // baseline *thisBaseline = readBaseline(filename, baselinef);

  map<string,baseline*> blines = readBaseline(filename, baselinef);

  if(blines.empty()) {
    (void) printf("# Warning ... continuing with built-in baseline values\n");
  } else {
    map<string,baseline*>::iterator pos;
    for(pos = blines.begin(); pos != blines.end(); ++pos) {
      (void) printf("# Using baseline %s ", pos->second->baselineName());
      (void) printf("(%f, %f, %f, %f) \n", pos->second->localBaseline()[0],
 		    pos->second->localBaseline()[1], pos->second->localBaseline()[2],
		    pos->second->bias());
    }
  }

  /* new NULL line format:
NULLSPEC  08/30/2007/15:09:12  15.136667  tau_ceti  10.0  -40.0  0  200.0 0 0 OUT N1-2 0 KI_K1K2  10  8.491  1001.7 0.0371  0.0371  1.0 0.0 0.0 0.0   8.938  776.4 0.0464  0.0464  1.0 0.0 0.0 0.0 9.523  599 0.0729  0.0729  1.0 0.0 0.0 0.0  9.979  594.7 0.0847  0.0847  1.0 0.0 0.0 0.0 10.465  518.4 0.0963  0.0963  1.0 0. 0 0.0 0.0 10.94  400.3 0.1108  0.1108  1.0 0.0 0.0 0.0 11.411  268.5 0.121  0.121  1.0 0.0 0.0 0.0 11.938  175.7 0.1405  0.1405  1.0 0.0 0.0 0.0 12.378  125.1 0.1702  0.1702  1.0 0.0 0.0 0.0  12.746  76.4 0.2154  0.2154  1.0 0.0 0.0 0.0
  */

  while (!feof(fptr)) {
    (void) fgets(line,LINE_LENGTH,fptr);
    if (!(feof(fptr)) && (line[0] != '#') &&
         ((nRead = sscanf(line,
                             "NULLSPEC %d/%d/%d/%d:%d:%d %lf %s %lf%lf%s%s%s%s%s%s%s%s%d",
                             &month,&day,&year,&hour,&minute,&second,&ut,
                             &starName,&delay,&longDelay,&numCycles,
                             &percentGatedCycles,&duration,&mode,&portsUsed,
                             &producer,&calFlag,
                             &blDesig,&nChannels)) == 19)) {

      // Allocate necessary memory for channel data
      double *flux = new double [nChannels];
      double *lambda = new double[nChannels];
      double *nullLeakage = new double [nChannels];
      double *nullLeakageError = new double [nChannels];
      double *ratio = new double [nChannels];

      // Read the spectrometer data from line
      if ((nRead == 19) && (readSpVline(line,16,nChannels,flux,lambda,nullLeakage,nullLeakageError,ratio))) {

	    dataLines++;

	    // Turn dates and UTs into JDs
	    if (year < 96)
	      year += 100;
	    if (year < 1900)
	      year += 1900;
	    jd = jdFromUTCal(ut,year,month,day);
	    
	    // KI Kvis-format line -- set baseline from designation
	    // hjb updated if clause, to allow for pti records w/ baseline tags, from imel pipeline (2/06)
	    baseline *thisBaseline;
	    if(blDesig[0] != '\0') {
	      map<string,baseline*>::iterator pos = blines.find(blDesig);
	      if( pos == blines.end())
		thisBaseline = defBaseline;
	      else
		thisBaseline = pos->second;
	    } else {
	      if(blines.size() == 1) {
		map<string,baseline*>::iterator pos = blines.begin();
		thisBaseline = pos->second;
	      } else
		thisBaseline = defBaseline;
	    }

	    if(thisBaseline == defBaseline) { 
	      // case of NULL blDesig should fall through to the end
	      if(strncmp(blDesig, (char *) "KI_K1K2",7) == 0)
		thisBaseline = &KI_K1K2;
	      else
		thisBaseline = defBaseline;
	    } // end set baseline from designation

	    // Create a narrow-band data node with this data contents
              NBNulldata *nbNode = new NBNulldata(starName,0,month,day,year,
                                          jd,ut,delay,longDelay,nChannels,
                                          lambda,flux,nullLeakage,nullLeakageError,
                                          thisBaseline);

	      
	      // Find or create a star node corresponding to this star
	      if (!starList->findStar(starName))
		starList->insert(starName);
	      
	      // Insert the data from this line into the linked-list
	      // structure for this star.  Note it is being
	      // insertion-sorted into the list based on jd. 
	      starList->getStarData()->insert(nbNode,jd,INSERT_SORT);

	  } // end read spectrometer line successful/selecting waveband/framerate
    } // end good spectrometer line
  } // end file while loop
  
  (void) fclose(fptr);
  
  return dataLines;

}


/************************************************************************
*  nullPred -- local to calib.cc 					*
*									*
*  A.  Purpose:	The function nullPred computes the expected null value (and *
*       variance) for a given calibrator and projected normalized       *
*       baseline length (aka inverse fringe spacing).			*
*									*
*  B.  Background and use:						*
*         nullPred is pretty simple: it computes the expected null for an   *
*       ideal interferometer observing a simple star -- presumably a    *
*       calibrator.  The ratio of what we got to what we expected is    *
*       an estimate of the system null.	                           	*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        baseline -- the sky-projected baseline length (m) (double).    *
*        lambda -- the wavelength we're working at (m) (double).        *
*        aSize -- the estimated angular size of object (rad) (double).  *
*        aSigma -- the error in aSize (rad) (double).                   *
*	Output -							*
*        varNullPred -- the variance in the output null prediction based    *
*         on the input angular size error (dimensionless) (double *).   *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the expected null value    *
*       (dimensionless) (double).                       		*
*									*
*************************************************************************/
double nullPred(double baseline,double lambda,
	      double aSize,double aSigma,double *varNullPred)
{
/**********  begin nullPred  **********************************************/
  // The fringe spacing on the sky (units of radians)
  double fringeSP = baseline / lambda;
  double arg = PI * fringeSP * aSize;
  double dArgdaSize = PI * fringeSP;
  double v,dvdArg,n;
  double sigmaV,sigmaN;

  // Trap the (annoying) point source case
  if (arg > 0)
    {
      v = 2.0 * j1(arg) / arg;
      dvdArg = 2.0 * (0.5 * arg * (j0(arg) - jn(2,arg)) - j1(arg)) / arg / arg;
      n = (1.0-v)/(1.0+v);
    }
  else
    {
      v = 0.0;
      dvdArg = 0.0;
    }

  // Use the chain rule (otherwise I never get this right).
  double dv2dv = 2.0 * v;

  // RLA - add nulling variance //
  // this first line is the V sigma
  sigmaV = dvdArg * dArgdaSize * aSigma;
  // leakage sigma
  sigmaN = sigmaV * pow(1.0+n,2)/(2.0);
  // now make it a variance
  *varNullPred = pow(sigmaN,2);

  return (n);
/**********  end nullPred  ************************************************/
}


/************************************************************************
*  sysNullSC -- local to calib.cc 					*
*									*
*  modified for nbNullCalib						*
*									*
*  A.  Purpose: The function sysNullSC estimates the system null   	*
*       (and the error in that estimate) with respect to a particular   *
*       calibrator at a particular jd.					*
*									*
*  B.  Background and use:						*
*         sysNullSC mechanizes the estimate of the system null at  	*
*       a particular epoch based on a particular calibrator.  In the    *
*       calib scheme, each calibrator is considered independently (this *
*       routine), and then these results are (weighted) averaged to     *
*       derive the system null estimate for the target (done      	*
*       in sysNullEnsemble).                                            *
*         sysNullSC does its work by extracting all calibrator scans    *
*       within a given time window (CALIWINDOW) around the input      	*
*       time.  From each calibrator scan a system null estimate   	*
*       (and error estiamte) is derived.  These system null       	*
*       estimates are then (weighted) averaged to determine the best    *
*       estimate for the system null at the input time.  Note:    	*
*       these system null estimates _can_ be weighted with        	*
*       respect to temporal proximity to the desired output time.       *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the Julian Data for the system null estimate.	      	*
*        calibData -- the (pointer to the) data list object for this    *
*         calibrator (dataList<NBdata> *).                              *
*        calSize -- the estimated angular size for this calibrator      *
*         (rad) (double).                                               *
*        sigmaSize -- the estimated error in calSize (rad) (double).    *
*        star -- the (pointer to the) unit 3-vector sky position for    *
*         this calibrator (dimensionless) (Vector *).                   *
*	Output -							*
*        varSysNull -- the estimated error variance in the system       *
*         null estimate (dimensionless) (double *).    		        *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the estimate of system     *
*       null at the time of interest derived from a     		*
*       particular calibrator (dimensionless) (double).		        *
*									*
*************************************************************************/
/**********  function declaration sysNullSC  *****************************/
int sysNullSC(double jd,
	     dataList<NBNulldata> *calibData,char *calName,
	     double calSize,double sigmaSize,
	     Vector *star,int *nChannels,
	     double *sysNullEst,double *varsysNullSC)
{
/**********  local declarations sysNullSC  *******************************/

  int i,calHits = 0;
  double weight;
  double jdScan,utScan,delayScan;
  double weightSum[MAX_CHANNELS] = {0.0};
  double expNull[MAX_CHANNELS],varExpNull[MAX_CHANNELS];
  double NullAvg[MAX_CHANNELS] = {0.0};

/**********  begin sysNullSC  ********************************************/
  // Loop over the designated calibration targets
  (void) calibData->gotoHead();
  int numNodes = calibData->numberNodes();
  int nNode = 0;

  if (numNodes)
    {
      // Skip forward to the interesting time
      while((nNode < numNodes)
	     && ((jd - calibData->getCurrent()->get_jd()) > calibTimeWindow)) {
	calibData->iterate();
        nNode++;
      }

      // Extract relevant calibration information from this
      // calibration source
      while((nNode <= numNodes)
	     && (fabs(jd - calibData->getCurrent()->get_jd()) < calibTimeWindow))
	{
	  baseline *Baseline = calibData->getCurrent()->get_baseline();
          delayScan = calibData->getCurrent()->get_delay() - calibData->getCurrent()->get_longDelay();
          jdScan = calibData->getCurrent()->get_jd();
          utScan = calibData->getCurrent()->get_ut();

	      calHits++;
	      
	      if (delayCheck)
		{
		  // New 11/30/97.  Check the delay value for this calibration object
		  double tmp = delayScan - Baseline->delay(jdScan,*star);
		  if (fabs(tmp) > delayCheckTol)
		    {
		      if (verbose)
			{
			  (void) printf("# Warning: at %10.4f, %4g m measured/expected delay mismatch on calibrator %s\n",jdScan,tmp,calName);
			  (void) fprintf(stderr,"Warning: at %10.4f, %4g m measured/expected delay mismatch on calibrator %s\n",jdScan,tmp,calName);
			} // verbose reporting check
		    } // end delay value check
		} // end delayCheck

	      for (i = 0; i < (*nChannels); i++) // Spectral channel loop
		{
                   double NullScan = calibData->getCurrent()->get_nullLeak(i);
		   double NullScanSigma = calibData->getCurrent()->get_nullSigma(i); 
		   // convert sigma to variance
                   double NullScanVar = pow(NullScanSigma,2);

		  // Compute the expected null value at this epoch, sky
		  // position, baseline projection, wavelength...
		  expNull[i] = nullPred(Baseline->skyProj(jdScan,*star),
				    calibData->getCurrent()->get_lambda(i) * 1e-6,
				    calSize,sigmaSize,&varExpNull[i]);

		  if (usingTempErrorProp)
		    {
		      // If defined, propagate the error in the system null
		      //  to the time of the target observation.  Assuming linear
		      //  (quadratic) increase in the uncertainty (variance).
                      NullScanVar *= 1.0 + pow(timeErrorPropCoeff * fabs(jd - jdScan),1.0);
		    } // end usingTempErrorProp

               // Check min scan uncertainty - RLA 3/31/08 //
                    if((usingMinimumCalUncertainty) && (NullScanSigma < minCalUncertainty)) {
                    NullScanSigma = minCalUncertainty;
	            NullScanVar = pow(NullScanSigma,2);	
                    if(verbose)
                      (void) printf((char *) "# Minimum scan uncertainty applied to calibrator at %11.5f\n",jdScan);
                  }

		  // Weighting.  Weight by variance, jitter
	          weight = 1.0 / (NullScanVar + varExpNull[i]);	
		  weightSum[i] += weight;
                  NullAvg[i] += weight * (NullScan - expNull[i]);
		} // channels loop
                nNode++;
                calibData->iterate();
	} // calibration scan
      
      if (calHits) // If we've found good calibration hits
	{
	  for (i = 0; i < (*nChannels); i++)
	    {
              varsysNullSC[i] = 1.0 / weightSum[i];
              sysNullEst[i] = NullAvg[i] / weightSum[i];
	    } // end spectral channel
	  return 1;
	} // end good calibration
      else // Otherwise, return a stupid value to flag calib 
	return 0;

    } // end calibData node count check
  else // no data for this calibrator
    return 0;

/**********  end sysNullSC  **********************************************/
}


/************************************************************************
*  sysNullEnsemble -- local to calib.cc 				*
*									*
*  A.  Purpose:	sysNullEnsemble computes the system null estimate  	*
*       from the full ensemble of calibrators selected for this         *
*       calibration run.						*
*									*
*  B.  Background and use:						*
*         sysNullEnsemble mechanizes the computation of an average      *
*       system null estimate from the composite of single         	*
*       calibrator system null estimates (computed in sysNullSC).  	*
*       sysNullEnsemble also performs system null estimate         	*
*       inner-consistency checks among multiple calibrators.            *
*         In the case of multiple calibrators, the cross-consistency    *
*       of derived system null estimates is checked, and          	*
*       mismatches at two levels are flagged for the user as a          *
*       diagnostic that there might be something wrong with either the  *
*       input calibrator information or the calibration object itself.  *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the Julian Data for the system null estimate       	*
*         (double).                                                     *
*        targV -- the (pointer to the) unit 3-vector sky position of    *
*         the target (dimensionless) (Vector *).  Note: this is okay    *
*         if NULL, it just means there can be no angular proximity      *
*         weighting in the system null estimate computation.      	*
*        starList -- the (pointer to the) null database object    	*
*         (StarList *).                                                 *
*	Output -							*
*        sysNull -- the composite estimated system null at the  	*
*         desired time (dimensionless) (double *).                      *
*        varSysNull -- the estimated error variance in sysNull          *
*         (dimensionless) (double *).                                   *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration sysNullEnsemble  ***********************/
int sysNullEnsemble(double jd,Vector *targV,
		   StarList *starList,
		   int nChannels,double *sysNull,double *varSysNull)
{
/**********  local declarations sysNullEnsemble  *************************/
  int i,j;
  double sNull[MAX_CALIBRATORS][MAX_CHANNELS],
    varSNull[MAX_CALIBRATORS][MAX_CHANNELS];

/**********  begin sysNullEnsemble  **************************************/
  // First Loop over the calibrators.
  // Determine the estimated system null at the time of
  // this scan based on each calibrator...
  for (i = 0; i < nCalibrators; i++)
    if (starList->findStar(calName[i]))
      {
	dataList<NBNulldata> *calibData = starList->getStarData();
	// Evaluate the calibrator apparent position
	// Bug alert: No way to ensure that this baseline is valid for this time (quant effects small)
	baseline *Baseline = calibData->getCurrent()->get_baseline();
	*(calibV[i]) = calibCoord[i]->topocentricPos(jd,Baseline->latitude(),
						     Baseline->longitude(),
						     Baseline->altitude());

	// Check/flag failed calibration attemps
	if (!(sysNullSC(jd,
		       calibData,calName[i],calSize[i],sigmaSize[i],
		       calibV[i],&nChannels,sNull[i],varSNull[i])))
	  sNull[i][0] = -99.0;
      } // end findStar check/calibrator loop
    else
      sNull[i][0] = -99.0;

  // Second loop over calibrators
  // Compute the weighted average of the system null
  // calibrations, and check the interconsistency of the
  // calibrations at this time.
  int calHits = 0;
  double weight,weightSum[MAX_CHANNELS] = {0.0},nullSum[MAX_CHANNELS] = {0.0};
  for (i = 0; i < nCalibrators; i++)
    if (sNull[i][0] > -90.0) // if there was a non-trivial calibration
      {
	calHits++;

	// Loop over spectrometer channels
	for (j = 0; j < nChannels; j++)
	  {

	    // Add angular proximity weighting as desired
	    if (usingSkyErrorProp)
	      if (targV)
		varSNull[i][j] *= 1.0 + angleErrorPropCoeff * angle_between(calibV[i],targV);

	    weightSum[j] += weight = 1.0 / varSNull[i][j];
	    nullSum[j] += weight * sNull[i][j];
	
	    // Check for calibration inconsistencies
	    for (int k = i + 1; k < nCalibrators; k++)
	      if (sNull[k][0] > -90.0)
		{
		  double chi = pow(sNull[i][j] - sNull[k][j],2.0)
		    / (varSNull[i][j] + varSNull[k][j]);
		  // Inconsistencies are flagged at two thresholds.  The
		  // first is just warning...
		  if (chi > chiWarning)
		    {
		      if (chi < chiReject)
			{
			  if (verbose)
			    {
			      (void) printf("# Warning: at %10.4f sysNull estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
					    jd,calName[i],calName[k],sqrt(chi));
			      (void) fprintf(stderr,"Warning: at %10.4f system null estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
					     jd,calName[i],calName[k],sqrt(chi));
			    } // end verbose check
			} // end warning
		      else // The second is an outright rejection of the scan...
			{
			  (void) printf("# Reject: at %10.4f sysNull estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
					jd,calName[i],calName[k],sqrt(chi));
			  (void) fprintf(stderr,"Reject: at %10.4f system null estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
					 jd,calName[i],calName[k],sqrt(chi));
			  // return a null-calibration
			  return 0;
			}
		    } // end chi test
		} // end cross-calibration check (k) loop
	  } // end channel (j) loop 
      } // end weight-average loop

  // If we've gotten a good calibration, compute the system null.
  if (calHits)
    for (j = 0; j < nChannels; j++)
      {
	sysNull[j] = nullSum[j] / weightSum[j];
	varSysNull[j] = 1.0 / weightSum[j];
      } // end (j) channel loop/calibration hit check

  // Return the number of calibrators used in this system calibration
  return calHits;

/**********  end sysNullEnsemble  ****************************************/
}


/************************************************************************
*  calib -- local to calib.cc (A.B. 7/27/97)				*
*									*
*  modified for nbNullCalib						*
*									*
*									*
*  A.  Purpose:	The function calib drives the calibration of null data  *
*       output by Mark's program.					*
*									*
*  B.  Background and use:						*
*         Other than database construction, calib is where it all       *
*       happens.  calib loops over the targets to be calibrated, the    *
*       list of pseudonyms for each target, and the scans recorded      *
*       under each pseudonym.  A system null estimate  			*
*       is derived from the average of system null estimates      	*
*       computed from multiple independent calibration objects          *
*       (sysNullEnsemble).                                              *
*         In the case of multiple calibrators, the cross-consistency    *
*       of derived system null estimates is checked, and          	*
*       mismatches at two levels are flagged for the user as a          *
*       diagnostic that there might be something wrong with either the  *
*       input calibrator information or the calibration object itself.  *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        starList -- the (pointer to the) null database object    	*
*         (StarList *).                                                 *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration calib  ********************************/
void calib(StarList *starList)
{
/**********  local declarations calib  **********************************/
  int i,month,day,year,nChannels,numScans = 0;
  double jdAvg,utAvg,delayAvg;
  double sNull[MAX_CALIBRATORS],varSNull[MAX_CALIBRATORS];

/**********  begin calib  ***********************************************/
  // Output a status line indicating the column contents in the
  // calibfile format
  if (verbose)
    {
      (void) printf("# Calibration data format:\n");
      (void) printf("# name\t#NullPts\tMJD\tm/d/y/h:m:s\tUT\tDelay(m)\t#Chan\tWL(um)\tCalNull\tErr\tRawNull\tErr\tSysNull\tErr...\t#Cal\tU(m)\tV(m)\tHA(hr)\tBLName\n");
    }

  // Now do the calibration

  // Loop over the calibration targets
  for (int nTarg = 0; nTarg < nTargets; nTarg++)
    {
      numScans = 0;
      (void) printf("\n# Target %s:\n",targetName[nTarg][0]);

      // Loop over the pseudonyms for each target
      for (int pseudonym = 0; pseudonym < nTargetNames[nTarg]; pseudonym++)
	// If we've got hits on the target under this name...
	if (starList->findStar(targetName[nTarg][pseudonym]))
	  {
	    int nDataName = 0;
	    dataList<NBNulldata> *targetData = starList->getStarData();
	    (void) targetData->gotoHead();
	    if (targetData->numberNodes())
	      do {
		int st,nCal=0;
		double sNullm[MAX_CHANNELS],varSNullm[MAX_CHANNELS];
		baseline *Baseline = targetData->getCurrent()->get_baseline();

                // # Pts hardcoded to 1
                st = 1;

		
		// Compute the null data for the calibration
		// target at this time.  If we've got a non-trivial
		// number of target scans...
                delayAvg = targetData->getCurrent()->get_delay() - targetData->getCurrent()->get_longDelay();
                month = targetData->getCurrent()->get_month();
                day = targetData->getCurrent()->get_day();
                year = targetData->getCurrent()->get_year();
                jdAvg = targetData->getCurrent()->get_jd();
                utAvg = targetData->getCurrent()->get_ut();
                delayAvg = targetData->getCurrent()->get_delay();
                nChannels = targetData->getCurrent()->get_nChannels();

		  // Compute the ensemble system null estimate
		  // at this time.  If we've got a good null
		  // calibration...
		  // store the current pointer, so we can reset after calibration
		  NBNulldata* resetNode = targetData->getCurrent();
		  if (nCal = sysNullEnsemble(jdAvg,targetV[nTarg],
					    starList,
					    nChannels,sNullm,varSNullm))
		    {
		      // New 11/6/97  Check the delay value for the target object
		      if ((targetCoord[nTarg]) && (delayCheck))
			{
			  // Evaluate the instantaneous target apparent position
			  Vector targ
			    = targetCoord[nTarg]->topocentricPos(jdAvg,
								 Baseline->latitude(),
								 Baseline->longitude(),
								 Baseline->altitude());

			  double delay = Baseline->delay(jdAvg,targ);
			  double delayDiff = delayAvg - delay;
			  if (fabs(delayDiff) > delayCheckTol)
			    {
			      if (verbose)
				{
				  (void) printf("# Warning: at %10.4f, %g m measured/expected delay mismatch on target %s\n",jdAvg,delayDiff,targetName[nTarg][pseudonym]);
				  (void) fprintf(stderr,"Warning: at %10.4f, %g m measured/expected delay mismatch on target %s\n",jdAvg,delayDiff,targetName[nTarg][pseudonym]);
				} // end verbose check
			      
			    } // end delay value check
			  
			} // end target position check
		      
		      // Some time machinations
		      int hr = (int) utAvg;
		      double frac = utAvg - ((double) hr);
		      int min = (int) (frac * 60.0);
		      frac -= ((double) min / 60.0);
		      int sec = (int) (frac * 3600.0);
		      
		      // Output the header information -- names, times, delay, # spec channels
		      (void) printf("%s\t%d\t%11.5f\t%d/%d/%d/",
				    targetName[nTarg][pseudonym],
				    st,jdAvg,month,day,year);

		      // Mucho tedious logic to get the time format correct
		      // (it wouldn't be so tedius if we used printf the way god intended)
		      if (hr < 10)
			(void) printf((char *) "0%1d:",hr);
		      else
			(void) printf((char *) "%2d:",hr);
		      
		      if (min < 10)
			(void) printf((char *) "0%1d:",min);
		      else
			(void) printf((char *) "%2d:",min);
		      
		      if (sec < 10)
			(void) printf((char *) "0%1d",sec);
		      else
			(void) printf((char *) "%2d",sec);
		      
		      (void) printf((char *) "\t%g\t%g\t%d\t",
				    utAvg,delayAvg,nChannels);
		      
		      // Loop over each spectrometer channel, writing out the WL,
		      //  calibrated, raw, and system null...
		      for (int chan = 0; chan < nChannels; chan++)
			{
                        double nullAvg = targetData->getCurrent()->get_nullLeak(chan);
                        double nullSig = targetData->getCurrent()->get_nullSigma(chan);

			// combine errors in quadrature => add variances
                        double varTNull = pow(nullSig,2) + varSNullm[chan];
                        double scanSigma = sqrt(varTNull);
                        if((usingMinimumScanUncertainty) && (scanSigma < minScanUncertainty)) {
                            scanSigma = minScanUncertainty;
                        }
			  
			  (void) printf((char *) "%g\t%g\t%g\t%g\t%g\t%g\t%g\t",
					targetData->getCurrent()->get_lambda(chan),
                                      nullAvg - sNullm[chan],scanSigma,
//					nullAvg - sNullm[chan],sqrt(varTNull),
                                        nullAvg,nullSig,
					sNullm[chan],sqrt(varSNullm[chan]));
			} // end channel loop
		      
		      (void) printf("%d",nCal);
		      
		      // And if we've got a position on the
		      // target, output u-v coordinates and the
		      // like to help our consumers...
		      if (targetCoord[nTarg])
			{
			  Vector targ // The instantaneous apparent direction of target
			    = targetCoord[nTarg]->topocentricPos(jdAvg,
								 Baseline->latitude(),
								 Baseline->longitude(),
								 Baseline->altitude());
			  
			  Vector SV = Baseline->skyProjV(jdAvg,targ);
			  
			  (void) printf((char *) "\t%10.6f\t%10.6f\t%g",SV[0],SV[1],
					Baseline->hourAngle(jdAvg,targ));
			}
		      // Print out the baseline name
		      (void) printf((char *) "\t%s\n",Baseline->baselineName());
		      (void) printf((char *) "\n");
		      numScans++;
		    } // end successful target scan/calibration
		  else { // Otherwise, calibration failure...notify the user.
		      (void) printf("# No suitable calibration found for %10.4f\n",jdAvg);
		  }
		  nDataName++;
		  targetData->resetCurrent(resetNode);
		  targetData->iterate();
        } while(nDataName < targetData->numberNodes());


	    (void) printf("# %d calibrated scans output under name %s\n",
			numScans,targetName[nTarg][pseudonym]);

	  } // end psedonym list/data check
	
      (void) printf("# %d total calibrated scans output on target %s\n",
		    numScans,targetName[nTarg][0]);
    } // end nTarg loop
/**********  end calib  *************************************************/
}


/**********  main routine calib  ****************************************/
main (int argc,char **argv)
{
  // initialize baselineFile global to blank
  baselineFile[0] = '\0';

  // Process any command-line options
  int argIndex = processCmdLineOpts(argc,argv);

  // Trap trivial/bogus invocations
  if (argc < 3)
    (void) printHelpMessage();

  // Read the calibration script on the stars we're going to care about
  (void) readScript(argv[argIndex]);

  // Report to the user on the run directives (if verbose)
  if (verbose)
    (void) printRunDirectives();

  // Initialize the data structure and read in the data from the
  // summary file(s)
  StarList *starList = new StarList();
  for (int i = argIndex + 1; i < argc; i++)
    {
      int nData = readSpecFile(argv[i],baselineFile, starList);
      if (verbose)
	(void) printf("# %d data lines read from %s\n",nData,argv[i]);
    }

  // Now we do the calibration.  Output is to stdout (to be redirected
  // as desired.  Isn't unix great...?)
  (void) calib(starList);

/**********  end program calib  ****************************************/
}
