/*************************************************************************
* NASA Exoplanet Science Institute (NExScI) -- http://nexsci.caltech.edu *
*	Copyright (c) 1997-2009 California Institute of Technology       *
*                     U.S. Government Sponsorship                        *
* Under contract with the National Aeronautics and Space Administration  *
**************************************************************************/
// nbCalib -- calibration of narrow-band visibility amplitudes for PTI
// and KI -- Andy Boden 1997 through 2003

// now wrapped by a Perl script.  This C++ core of the nbCalib app has
// been renamed to wbcEngine.cc -- David A. Imel 2004

#include <config.h>


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h> 
#include "getopt.h"

#include "angles.h"
#include "Vector.h"

#include "timeUtil.h"

#include "PTI.h"
#include "KI.h"
#include "baseline.h"
#include "coords.h"

#include "nbStarList.h"

#include <map>
#include <string>
#include <iostream>
using namespace std;

#define LINE_LENGTH 2048
//#define SPEC_LINE_LENGTH 50000
#define SPEC_LINE_LENGTH 30000

#define MAX_TARGETS 2000
#define MAX_PSEUDONYMS 10
#define MAX_CALIBRATORS 2000

// Maximum number of spectral channels
#define MAX_CHANNELS 400

#define LAMBDA 2.2e-6
#define MJD_2000 51544.0

// Delay Check tolerance -- approximately 3 arcminutes
#define DELAY_CHECK_TOLERANCE 0.10


// Global declarations

// Create the default baseline objects.  Use the form of the
// constructor with delay bias -- we'll be checking delay values

// PTI
baseline PTI_NS(PTI_LONGITUDE,PTI_LATITUDE,
		PTI_Baseline_NS,PTI_NS_DELAY_BIAS,"PTI_NS");
baseline PTI_NW(PTI_LONGITUDE,PTI_LATITUDE,
		PTI_Baseline_NW,PTI_NW_DELAY_BIAS,"PTI_NW");
baseline PTI_SW(PTI_LONGITUDE, PTI_LATITUDE,
		PTI_Baseline_SW, PTI_SW_DELAY_BIAS, "PTI_SW");

// KI
baseline KI_K1K2(KI_LONGITUDE,KI_LATITUDE,
		 KI_Baseline_K1K2,KI_K1K2_DELAY_BIAS,"KI_K1K2");
baseline KI_S1S2(KI_LONGITUDE,KI_LATITUDE,
		 KI_Baseline_S1S2,KI_S1S2_DELAY_BIAS,"KI_S1S2");
baseline KI_K1K2S(KI_LONGITUDE,KI_LATITUDE,
                 KI_Baseline_K1K2,KI_K1K2_DELAY_BIAS,"KI_K1K2S");
baseline KI_Camera_FT_PRIM(KI_LONGITUDE,KI_LATITUDE,
                 KI_Baseline_K1K2,KI_K1K2_DELAY_BIAS,"KI_Camera_FT_PRIM");
baseline KI_Camera_FT_SEC(KI_LONGITUDE,KI_LATITUDE,
                 KI_Baseline_K1K2,KI_K1K2_DELAY_BIAS,"KI_Camera_FT_SEC");
baseline KI_Camera_FT_LBAND(KI_LONGITUDE,KI_LATITUDE,
                 KI_Baseline_K1K2,KI_K1K2_DELAY_BIAS,"KI_Camera_FT_LBAND");

//baseline *defBaseline = &PTI_NS;
baseline *defBaseline = &KI_K1K2;
string baselineFile;
string fluxBiasFile;

// Star Stuff

// Calibration Target Names, Counts, and Positions (arrays -- sorry)
char *targetName[MAX_TARGETS][MAX_PSEUDONYMS];
int nTargets,nTargetNames[MAX_TARGETS];
Vector *targetV[MAX_TARGETS];
coords *targetCoord[MAX_TARGETS];


// Calibrator Names, Count, and Positions  (arrays -- sorry)
int nCalibrators;
char calName[MAX_CALIBRATORS][132];
Vector *calibV[MAX_CALIBRATORS];
coords *calibCoord[MAX_TARGETS];
double calSize[MAX_CALIBRATORS],sigmaSize[MAX_CALIBRATORS];


// Program Operational Options.  These are used to control the code's
// processing behavior.  Global variables are the best mechanism I've
// come up with to handle this situation...

// Used to indicate whether the program processes the incoherent data
// (default) or coherent data.
int usingIncoherentV2 = 1;

// Flags to indicate whether we try to correct for phase jitter and
// intensity mismatches.
int usingJitterCorrection = 1;
double jitterCoeff = 0.04;  // Default phase jitter correction coeff
int usingRatioCorrection = 0;

// Calibration time window for picking up calibration scans.
// units are in days, so 1/24 is +/- 1 hour time window
// #define CALIB_WINDOW (1.0 / 24.0)
double calibTimeWindow = (1.0 / 24.0);

// Propagation coefficient for system visibility error.
// A doubling of the sysVisSC error estimate in an hour.  See sysVisSC.
// Units are in inverse days, so 24 is 1 hr.
// #define PROP_COEFF (1.0 * 24.0)
int usingTempErrorProp = 1;
double timeErrorPropCoeff = (1.0 * 24.0);

// Propagating visibility errors across the sky.  See sysVisEnsemble.
// Units are in inverse radians.
// #define ANGLE_COEF (1.0 / (15.0 * DEG_TO_RAD))
int usingSkyErrorProp = 1;
double angleErrorPropCoeff = (1.0 / (15.0 * DEG_TO_RAD));

// Minimum possible scan uncertainty
int usingMinimumScanUncertainty = 0;
double minScanUncertainty = 0.01;

// Using root-N scaling on error bars
int useRootN = 0;

// Thresholds for reporting/rejecting calibration errors (dimensionless).
// Threshold for reporting calibration errors -- 2.0 sigma disagreement
// #define CHI_WARNING 2.0
double chiWarning = 2.0 * 2.0;
// Threshold for rejecting the calibration -- 3.0 sigma disagreement
// #define CHI_REJECT 9.0
double chiReject = 3.0 * 3.0;

// scanThreshold -- time used to define sample points that synthesized
// into a scan (units are in days because the cut is made in JD space)
// AFB & RLA 2/2002
double scanThreshold = 150.0 / 3600.0 / 24.0;

// Band selection (K,H,all)  Added 11/99  A.B.
char usingWL = 'A';  // Default to all wavebands
double K_wl = 2.2, K_width = 0.2;
double H_wl = 1.6, H_width = 0.1;


// Frame rate selection (50, 100, 0 -- all)  Added 11/99 A.B.
double usingFR = 0;  // Default to all frame rates


// Verbosity variable to control reporting Added 7/01 A.B.
#define MUTE 0
#define NORMAL 1
#define EXCESSIVE 2
int verbose = NORMAL; // Default to certain errors



// Variable for delay checking
int delayCheck = 1;
double delayCheckTol = DELAY_CHECK_TOLERANCE;

// values for FluxBias correction
int usingFluxBiasCorrection = 0;

typedef struct {
  float WLFluxBias;
  float SpecFluxBias;
  float NZero;
  float Normflux;
  int usingDefaults; // set to 1 initially, becomes 0 if cameraName read from fluxBiasFile
} fluxBiasStruct;

// KI_K1K2 will use ftPrim defaults
// fluxBiasStruct k1k2 = {0.063,0.16,20.0,1000.0,1};
fluxBiasStruct k1k2s = {0.0,0.0,20.0,1000.0,1};
fluxBiasStruct ftPrim = {0.063,0.16,20.0,1000.0,1};
fluxBiasStruct ftSec = {0.0,0.0,20.0,1000.0,1};
fluxBiasStruct ftLband = {0.0,0.0,20.0,1000.0,1};

map<string,fluxBiasStruct> fluxBiasMap;

// values to control which spec lines are used
int numSpecChans = 0;
int starCount = 0; // number of target & calibrator names in calscript file
char *calscriptStars[MAX_CALIBRATORS+MAX_TARGETS]; // names of targets & calibrators
int firstSpecFile = 0; // flag for first spec file read
map<int,int> nightSpecChans;  // julian_dayOfYear/numSpecChans pair


/************************************************************************
*  printHelpMessage -- local to calib.cc (A.B. 4/13/98)			*
*									*
*  A.  Purpose:	The function printHelpMessage prints a help message     *
*       for the calib program to stderr and then exits the program.	*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description: None					*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration printHelpMessage  *********************/
void printHelpMessage()
{
/**********  begin printHelpMessage  ************************************/
  (void) fprintf(stdout,"Usage: nbCalib calScript specfile1 [specfile2...] [> cal.out]\n\n");
  (void) fprintf(stdout,(char *) "Calibrate PTI/KI narrowband V2 data\n\n");
  (void) fprintf(stdout,"Options:\n");
  (void) fprintf(stdout,(char *) "  -help: print (this) help message and exit (overrides all other options)\n");
  (void) fprintf(stdout,(char *) "  -version: print version information and exit\n");
  (void) fprintf(stdout,(char *) "  -Kband|Hband|allband: calibrate specific band data \n");
  (void) fprintf(stdout,(char *) "  -frameRate 50|100|0: calibrate specific frame rate data (Hz -- 0 all) \n");
  (void) fprintf(stdout,"  -numSpecChans argument: use data with this number of spec channels \n");
  (void) fprintf(stdout,"  -defaultFile argument: read calibration default values from a file\n");
  (void) fprintf(stdout,(char *) "  -[no]jitterCorrection[=argument]: set [non]use of jitter correction \n\t (optional correction coefficient argument)\n");
  (void) fprintf(stdout,"  -[no]ratioCorrection: set [non]use of ratio correction\n");
  (void) fprintf(stdout,"  -[in]coherentV2: set use of [in]coherent V2\n");
  (void) fprintf(stdout,"  -calTimeWindow argument: set calibration scan acceptance time window (hr)\n");
  (void) fprintf(stdout,"  -[no]timeWeighting: set use/nonuse of time proximity weighting of calibration scans\n");
  (void) fprintf(stdout,"  -timeErrorVal argument: if timeWeighting, set calibration scan time weighting value (hr)\n");
  (void) fprintf(stdout,"  -[no]angleProxWeighting: set use/nonuse of sky angle proximity weighting of calibration scans\n");
  (void) fprintf(stdout,"  -angleErrorVal argument: if angleProxWeighting, set calibration scan angle weighting value (deg)\n");
  (void) fprintf(stdout,"  -[no]minScanUncertainty argument: sets [non]use of minimum possible scan uncertainty of value argument (V2 units)\n");
  (void) fprintf(stdout,(char *) "  -[no]rootNWeighting : sets use/nonuse of root-Nsamples weighting in computed scan errors\n");
  (void) fprintf(stdout,(char *) "  -scanThreshold argument: time window used to define maximum scan duration (sec)\n");
  (void) fprintf(stdout,"  -chiWarning argument: set calibrator inconsistency warning chi threshold \n");
  (void) fprintf(stdout,"  -chiReject argument: set calibrator inconsistency scan rejection chi threshold \n");
  (void) fprintf(stdout,(char *) "  -[no]delayCheck[=argument]: set use/nonuse of delay value checking \n\t (optional delay tolerance argument) \n");
  (void) fprintf(stdout,"  -verbose|normal|quiet|shutUp: control volume of information, warning, and error messages\n");
  (void) fprintf(stdout,"  -fluxBiasCorrection [=file]: corrects flux dependent V2 Biases using the biases provided by <file> or default corrections if no file is specified\n");
  (void) fprintf(stdout,(char *) "  -baseline <file>: Use the baseline supplied by <file>. Normal behavior is to look for a\n");
  (void) fprintf(stdout,(char *) "                    yyyyddd.baseline file for each yyyyddd.spec file given on the command\n");
  (void) fprintf(stdout,(char *) "                    line. If that fails then look for yyyyddd.bline. If neither is found then\n");
  (void) fprintf(stdout,(char *) "                    nbCalib uses a built in default baseline. The option allows the user to\n");
  (void) fprintf(stdout,(char *) "                    override that search by providing the filename directly. Note that when\n");
  (void) fprintf(stdout,(char *) "                    multiple SPEC files are given in combination with this option all of them\n");
  (void) fprintf(stdout,(char *) "                    use the supplied <file>.\n");


  (void) fprintf(stdout,"  note:  command-line options may be used in any order, interspersed with non-options\n");
  (void) fprintf(stdout,"   and abbreviated to uniqueness (e.g. -h, -noang, -incoh).\n");
  (void) fprintf(stdout,(char *) "  Find more nbCalib info at http://msc.caltech.edu/software\n");

  (void) exit(-1);



/**********  end printHelpMessage  **************************************/
}


/************************************************************************
*  readCalDefs -- local to calib.cc (A.B. 4/13/98)			*
*									*
*  A.  Purpose:	The function readCalDefs reads the calibration code     *
*       defaults file to establish the default settings for the run.    *
*									*
*  B.  Background and use: The function readCalDefs gives a .ini-style  *
*       interface to the initialization of calib.  There are a set of   *
*       predefined default values built directly into calib, then the   *
*       user can create a default file that can override those values.	*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        filename -- the (pointer to the) calibration code defaults     *
*         filename to be processed (char *).                            *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration readCalDefs  **************************/
void readCalDefs(char *filename)
{
/**********  local declarations readCalDefs  ****************************/
  char line[LINE_LENGTH],*tmpstr;
  FILE *fptr;

/**********  begin readCalDefs  *****************************************/
  if (!(fptr = fopen(filename,"r")))
    {
      (void) fprintf(stderr," Error -- can't open calibration-defaults file %s,",filename);
      (void) fprintf(stderr," proceeding with previous settings\n");
      (void) printf("# Error -- can't open calibration-defaults file %s,",filename);
      (void) printf(" proceeding with previous settings\n");
      return;
    }

  // Read the defaults file
  // coherent/incoherent switch
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,"%d",&usingIncoherentV2) != 1)
    {
      (void) fprintf(stderr," Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }

  // jitter switch
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,"%d",&usingJitterCorrection) != 1)
    {
      (void) fprintf(stderr," Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }

  // ratio correction switch
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,"%d",&usingRatioCorrection) != 1)
    {
      (void) fprintf(stderr," Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }

  // calibration time window
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,"%lf",&calibTimeWindow) != 1)
    {
      (void) fprintf(stderr," Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }
  calibTimeWindow /= 24.0; // Convert hours to days

  // time weighting switch
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,"%d",&usingTempErrorProp) != 1)
    {
      (void) fprintf(stderr," Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }

  // calibrator time weighting
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,"%lf",&timeErrorPropCoeff) != 1)
    {
      (void) fprintf(stderr," Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }
  timeErrorPropCoeff *= 24.0; // Converting hours to inverse days
  

  // sky proximity switch
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,"%d",&usingSkyErrorProp) != 1)
    {
      (void) fprintf(stderr," Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }

  // calibrator sky proximity weighting
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,"%lf",&angleErrorPropCoeff) != 1)
    {
      (void) fprintf(stderr," Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }
  angleErrorPropCoeff *= DEG_TO_RAD;  // Converting to degrees inverse radians
  angleErrorPropCoeff = 1.0 / angleErrorPropCoeff;

  // chiWarning
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,"%lf",&chiWarning) != 1)
    {
      (void) fprintf(stderr," Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }
  chiWarning *= chiWarning; // Convert to chi-squared

  // chiReject
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,"%lf",&chiReject) != 1)
    {
      (void) fprintf(stderr," Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }
  chiReject *= chiReject; // Convert to chi-squared

  // Print-out successful reading report, close the defaults file, and return.
  (void) printf("# calibration defaults read from %s\n",filename);
  (void) fclose(fptr);

/**********  end readCalDefs  *******************************************/
}



/************************************************************************
*  processCmdLineOpts -- local to calib.cc (A.B. 4/12/98)		*
*									*
*  A.  Purpose:	The function processCmdLineOpts processes the command   *
*       line invocation options used with calib for this run.		*
*									*
*  B.  Background and use: processCmdLineOpts uses getopt_long_only to  *
*       process the command line options the user has added to the      *
*       invocation.  getopt_long_only is used because it allows the     *
*       arbitrarily intersperse command-line options along with the     *
*       filename command-line arguments: it permutes the command line   *
*       as it processes so when finished all the non-options (i.e. the  *
*       filename arguments) are at the end (for simple processing by    *
*       the main routine).				                *
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        argc, argv -- the standard command-line argument variables.    *
*	Output -							*
*        argv's contents are permuted so as to move all non-option      *
*         arguments to the end.                                         *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the index of the first     *
*       non-option on the command-line, which is presumably the name    *
*       of the calibration script file (int).              		*
*									*
*************************************************************************/
/**********  function declaration processCmdLineOpts  *******************/
int processCmdLineOpts(int argc, char **argv)
{
/**********  local declarations processCmdLineOpts  *********************/
  // Valid command-line options
  int retVal,optIndex,argsProcessed = 0;
  static struct option longOptions[] = 
  {
    {((char *) "help"),no_argument,0,0},                     // Option 0
    {((char *) "version"),no_argument,0,0},                  // Option 1
    {((char *) "jitterCorrection"),optional_argument,0,0},   // Option 2
    {((char *) "nojitterCorrection"),no_argument,0,0},       // Option 3
    {((char *) "ratioCorrection"),no_argument,0,0},          // Option 4
    {((char *) "noratioCorrection"),no_argument,0,0},        // Option 5
    {((char *) "coherentV2"),no_argument,0,0},               // Option 6
    {((char *) "incoherentV2"),no_argument,0,0},             // Option 7
    {((char *) "timeWeighting"),no_argument,0,0},            // Option 8
    {((char *) "notimeWeighting"),no_argument,0,0},          // Option 9
    {((char *) "angleProxWeighting"),no_argument,0,0},       // Option 10
    {((char *) "noangleProxWeighting"),no_argument,0,0},     // Option 11
    {((char *) "calTimeWindow"),required_argument,0,0},      // Option 12
    {((char *) "timeErrorVal"),required_argument,0,0},       // Option 13
    {((char *) "angleErrorVal"),required_argument,0,0},      // Option 14
    {((char *) "minScanUncertainty"),required_argument,0,0}, // Option 15
    {((char *) "nominScanUncertainty"),no_argument,0,0},     // Option 16
    {((char *) "chiWarning"),required_argument,0,0},         // Option 17
    {((char *) "chiReject"),required_argument,0,0},          // Option 18
    {((char *) "defaultFile"),required_argument,0,0},        // Option 19
    {((char *) "Kband"),no_argument,0,0},                    // Option 20
    {((char *) "Hband"),no_argument,0,0},                    // Option 21
    {((char *) "allband"),no_argument,0,0},                  // Option 22
    {((char *) "frameRate"),required_argument,0,0},          // Option 23
    {((char *) "quiet"),no_argument,0,0},                    // Option 24
    {((char *) "normal"),no_argument,0,0},                   // Option 25
    {((char *) "verbose"),no_argument,0,0},                  // Option 26
    {((char *) "shutUp"),no_argument,0,0},                   // Option 27
    {((char *) "scanThreshold"),required_argument,0,0},      // Option 28
    {((char *) "delayCheck"),optional_argument,0,0},         // Option 29
    {((char *) "nodelayCheck"),no_argument,0,0},             // Option 30
    {((char *) "rootNWeighting"),no_argument,0,0},           // Option 31
    {((char *) "norootNWeighting"),no_argument,0,0},         // Option 32
    {((char *) "baseline"),required_argument,0,0},           // Option 33 - baseline file override

// JSH IMPORTANT NOTE
// I'm not familiar enough with Perl to understand why, but when specifying
// an optional string, you MUST use an "=" on the command line, otherwise it
// core dumps. The "=" is not necessary for required arguments, but for some
// reason, it is for optional ones. So for fluxBiasCorrection, if specifying a
// filename, the format MUST be: -fluxBiasCorrection=fluxBiasFilename
// I need to look into this further
    {((char *) "fluxBiasCorrection"), optional_argument, 0, 0}, // Option 34 - correct flux dependent bias
    {((char *) "nofluxBiasCorrection"), no_argument, 0, 0}, // Option 35
    {((char *) "numSpecChans"),required_argument,0,0},       // Option 36
    {0,0,0,0}
  };

  extern char *optarg;
  extern int optind,opterr,optopt;


/**********  begin processCmdLineOpts  **********************************/

  // Print welcome/version
  if (verbose)
    (void) fprintf(stdout,(char *) "## This is nbCalib v%s\n\n",(char *) VERSION);

  do {
    if ((retVal = getopt_long_only(argc,argv,"",longOptions,&optIndex)) != EOF) {
      if (verbose)
	(void) printf("# option %s\n",longOptions[optIndex].name);
      argsProcessed++;
      switch (optIndex) {

      case 0: {
	(void) printHelpMessage();
	break;
      } // end printing help message

      case 1: {
	(void) fprintf(stdout,(char *) "nbCalib v%s\n\n",(char *) VERSION);
	(void) exit(-1);
	break;
      } // end printing help message

      case 2: {
	usingJitterCorrection = 1;
	if (optarg) {
	  jitterCoeff = atof(optarg);
	} // end setting jitter correction coefficient
	break;
      } // end setting usingJitterCorrection -- option 2

      case 3: {
	usingJitterCorrection = 0;
	break;
      } // end setting NOT usingJitterCorrection -- option 3

      case 4: {
	usingRatioCorrection = 1;
	break;
      } // end setting usingRatioCorrection -- option 4

      case 5: {
	usingRatioCorrection = 0;
	break;
      } // end setting NOT usingRatioCorrection -- option 5

      case 6: {
	usingIncoherentV2 = 0;
	break;
      } // end setting NOT usingIncoherentV2 -- option 6

      case 7: {
	usingIncoherentV2 = 1;
	break;
      } // end setting usingIncoherentV2 -- option 7
	
      case 8: {
	usingTempErrorProp = 1;
	break;
      } // end setting usingTempErrorProp -- option 8
	
      case 9: {
	usingTempErrorProp = 0;
	break;
      } // end setting NOT usingTempErrorProp -- option 9

      case 10: {
	usingSkyErrorProp = 1;
	break;
      } // end setting usingSkyErrorProp -- option 10

      case 11: {
	usingSkyErrorProp = 0;
	break;
      } // end setting NOT usingSkyErrorProp -- option 11

      case 12: {
	calibTimeWindow = atof(optarg);
	calibTimeWindow /= 24.0;
	break;
      } // end setting calibTilmWindow -- option 12

      case 13: {
	timeErrorPropCoeff = atof(optarg);
	timeErrorPropCoeff *= 24.0;
	break;
      } // end setting timeErrorPropCoeff -- option 13

      case 14: {
	angleErrorPropCoeff = atof(optarg);
	angleErrorPropCoeff *= DEG_TO_RAD;
	angleErrorPropCoeff = 1.0 / angleErrorPropCoeff;
	break;
      } // end setting angleErrorPropCoeff -- option 14

      case 15: {
	usingMinimumScanUncertainty = 1;
	minScanUncertainty = atof(optarg);
	break;
      } // end setting usingMinimumScanUncertainty/minScanUncertainty  -- option 15

      case 16: {
	usingMinimumScanUncertainty = 0;
	break;
      } // end setting usingMinimumScanUncertainty -- option 16

      case 17: {
	chiWarning = atof(optarg);
	chiWarning *= chiWarning;
	break;
      } // end setting chiWarning -- option 17

      case 18: {
	chiReject = atof(optarg);
	chiReject *= chiReject;
	break;
      } // end setting chiReject -- option 18

      case 19: {
	(void) readCalDefs(optarg);
	break;
      } // end reading defaults file -- option 19

      case 20: {
	usingWL = 'K';
	break;
      } // end selecting K-band data -- option 20

      case 21: {
	usingWL = 'H';
	break;
      } // end selecting H-band data -- option 21

      case 22: {
	usingWL = 'A';
	break;
      } // end selecting all bands -- option 22

      case 23: {
	usingFR = atoi(optarg);
	break;
      } // end setting usingFR -- option 23

      case 24: {
	// be quiet!!!
	verbose = MUTE;
	break; 
      } // end shutting up -- option 24

      case 25: {
	// normal (normal setup, serious warnings and errors)
	verbose = NORMAL;
	break; 
      } // setting normal reporting -- option 25

      case 26: {
	// verbose (everything)
	verbose = EXCESSIVE;
	break; 
      } // end setting excessive reporting -- option 26

      case 27: {
	// SHUT UP!!!
	verbose = MUTE;
	break; 
      } // end setting shutUp -- option 27

      case 28: {
	scanThreshold = atof(optarg);
	scanThreshold /= 3600.0 * 24.0; // Convert to units of days
	break;
      } // end setting scanThreshold -- option 28

      case 29: {
	delayCheck = 1;
	if (optarg) {
	  delayCheckTol = atof(optarg);
	} // end setting delay check tolerance
	break;
      } // end setting delayCheck -- option 29

      case 30: {
	delayCheck = 0;
	break;
      } // end setting delayCheck -- option 30

      case 31: {
	useRootN = 1;
	break;
      } // end setting useRootN -- option 31

      case 32: {
	useRootN = 0;
	break;
      } // end setting useRootN -- option 32

      case 33: {
        baselineFile.assign(optarg);
	break;
      }

      case 34: {
	usingFluxBiasCorrection = 1;
	if(optarg)
            fluxBiasFile.assign(optarg);
	break;
      }

      case 35: {
	usingFluxBiasCorrection = 0;
	break;
      }

      case 36: {
        numSpecChans = atoi(optarg);
        break;
      } // end setting numSpecChans -- option 36


      } // end switch
    } // end if
  } while (retVal != EOF);

  if ((argsProcessed) && (verbose))
    (void) printf("# %d command line arguments processed\n",argsProcessed);

  return optind;
    
/**********  end processCmdLineOpts  ************************************/
}


/************************************************************************
*  parseCoordLine -- local to calib.cc (A.B. 11/98)			*
*									*
*  A.  Purpose:	The function parse coordinate line parses the           *
*       input coordinate line for the target/calibrator and extracts    *
*       the pertinent information					*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns a pointer to the           *
*       coordinate object created by as a result of this line		*
*       (coords *).                                                     *
*									*
*************************************************************************/
coords *parseCoordLine(char *line,double *calSize,double *sigmaSize)
{
/**********  local declarations parseCoordLine  *************************/
  char *separator = ((char *) "#"),*tmpstr;
  char decStr[256], *minusSign = ((char *) "-");
  double ra,dec;
  float rah,ram,ras,decd,decm,decs,pmRA,pmDec,plx;
  float cSize,sSize;
  coords *wrk;


/**********  begin parseCoordLine  **************************************/
  (void) strtok(line,separator); // Use strtok to stomp on any comments

  int rn = sscanf(line,(char *) "%lf%f%f%s%f%f%f%f%f%f%f",
		  &ra,&ram,&ras,decStr,&decm,&decs,&pmRA,&pmDec,&plx,
		  &cSize,&sSize);

  dec = atof(decStr);
  int decSign = 1; // handle negative declinations (stupid)!!!
  if (strncmp(decStr,minusSign,1) == 0) // If the first character is a minus...
    decSign = -1;

  switch (rn)
    {
    case 11: // We got the whole smack
      {
	// template 15 21 06.880 +24 57 28.872 -0.029 -0.019 0.00587 1.05  0.1
	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec,pmRA,pmDec,plx);

	if (calSize)
	  *calSize = cSize;
	if (sigmaSize)
	  *sigmaSize = sSize;

	break;
      }

    case 9: // We got full coordinates but no calibrator info
      {
	// template 15 34 41.268 +26 42 52.895  0.135 -0.089 0.04365

	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec,pmRA,pmDec,plx);

	// If we were _expecting_ calibrator information, generate an error
	if (calSize && sigmaSize)
	  {
	    (void) fprintf(stderr,(char *) "Fatal error parsing calibration script\n");
	    exit(-1);
	  }

	break;
      } // end rn = 9

    case 8: // We got positions and calibrator info
      {
	// Template 16 01 02.662 +33 18 12.634 0.70 0.1
	// Note this case is degenerate with coords, pm, no plx, but I
	// have to assume something...
	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec);

	// The calibrator info got read into the pm variables!!!
	if (calSize)
	  *calSize = pmRA;
	if (sigmaSize)
	  *sigmaSize = pmDec;

	break;
      } // end rn = 8

    case 6: // We got positions but no calibrator info
      {
	// Template 16 01 02.662 +33 18 12.634 
	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec);

	// If we were _expecting_ calibrator information, generate an error
	if (calSize && sigmaSize)
	  {
	    (void) fprintf(stderr,(char *) "Fatal error parsing calibration script\n");
	    exit(-1);
	  }

	break;
      } // end rn = 6

    case 4: // We got decimal degrees and calibrator information
      {
	// Template 331.75 25.33 1.0 0.05
	ra *= DEG_TO_RAD;
	dec *= DEG_TO_RAD;
	wrk = new coords(ra,dec);

	if (calSize)
	  *calSize = cSize;
	if (sigmaSize)
	  *sigmaSize = sSize;

	break;
      } // end rn = 4

    case 2: // We got decimal degrees and no calibrator information
      {
	// Template 331.75 25.33
	ra *= DEG_TO_RAD;
	dec *= DEG_TO_RAD;
	wrk = new coords(ra,dec);

	// If we were _expecting_ calibrator information, generate an error
	if (calSize && sigmaSize)
	  {
	    (void) fprintf(stderr,(char *) "Fatal error parsing calibration script\n");
	    exit(-1);
	  }

	break;
      } // end rn = 2

    } // end switch (rn)

  return wrk;

/**********  end parseCoordLine  ****************************************/
}


/************************************************************************
*  readScript -- local to calib.cc (7/27/97)				*
*									*
*  A.  Purpose:	The function readScript reads the input calibration     *
*       script and parses the input on the input.			*
*									*
*  B.  Background and use: readScript reads the input user calibration  *
*       script, parses the contents, and loads a sequence of global     *
*       variables used in subsequent processing.                        *
*         First up are the calibration targets.  The code reads the     *
*       names of the calibration targets to be used in this calibration *
*       run, along with the possibility of pseudonyms.  (Astronomy is   *
*       nothing if not replete with pseudonyms.)  Optionally, target    *
*       positions are read for two reasons: one, to weight the impact   *
*       of calibrators by proximity, and two, to compute the baseline   *
*       projection on the target (u-v, HA) during the output stage --   *
*       this is added as a convenience for some users, as it is         *
*       slightly orthogonal to the main mission of calib.               *
*         Next up are the calibrators.  These are (of course) presumed  *
*       to be single stars, and their names, coordinates, angular sizes *
*       and uncertainties are read into global arrays.                  *
*         Mod 11/10/98 AB -- extended read logic to read full           *
*       astrometric state (pos, PM, plx) for all objects.               *
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        filename -- the (pointer to the) character string containing   *
*         the name of the calibration script to process (char *).       *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*                  nTargets, nTargetNames, targetName, targetV;         *
*                  nCalibrators, calName, calSize, sigmaSize, calibV.   *
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration readScript  ***************************/
void readScript(char *filename)
{
/**********  local declarations readScript  *****************************/
  char line[LINE_LENGTH];
  char *separator = ((char *) ":,;| \t\n"),*tmpstr,*tName;
  int rn;
  float rah,ram,ras,decd,decm,decs;
  double ra,dec;
  FILE *fptr;
  
/**********  begin readScript *******************************************/
  // Open/check on the health of the input script filename
  if (!(fptr = fopen(filename,"r")))
    {
      (void) fprintf(stderr," Fatal Error -- Can't open calibration script file %s\n",filename);
      exit(-1);
    }

  // Read the input information on the target stars
  tmpstr = fgets(line,LINE_LENGTH,fptr);
  while ((!feof(fptr))
	 && (nTargets < MAX_TARGETS)
	 && (strncmp(line,"---",3) != 0))
    if (line[0] != '#') // If this isn't a comment
      {
	// Get the target star name(s)
	nTargetNames[nTargets] = 0;
	while ((tName = strtok(tmpstr,separator))
	       && (nTargetNames[nTargets] < MAX_PSEUDONYMS))
	  if (tName[0] != '#')
	    {
	      tmpstr = NULL;
	      targetName[nTargets][nTargetNames[nTargets]] = strdup(tName);
	      nTargetNames[nTargets]++;
	    } // end parsing target names

	// Write out an informational message on the names
	if (verbose)
	  (void) printf("# Calibrating star %s ",targetName[nTargets][0]);
	if (nTargetNames[nTargets] > 1)
	  {
	    if (verbose)
	      (void) printf("(and pseudonyms ");
	    for (int tmp = 1; tmp < nTargetNames[nTargets]; tmp++)
	      if (verbose)
		(void) printf("%s ",targetName[nTargets][tmp]);
	    if (verbose)
	      (void) printf(") ");
	  } // end pseudonym loop

	// Check to see if a position has been given for the target
	(void) fgets(line,LINE_LENGTH,fptr);
	if ( (line[0] != '#')
	      && (((rn = sscanf(line,"%f%f%f%f%f%f",
			       &rah,&ram,&ras,&decd,&decm,&decs)) == 6)
		  || ((rn = sscanf(line,"%lf%lf",&ra,&dec)) == 2)))
	  {
	    targetCoord[nTargets] = parseCoordLine(line,0,0);
	    targetV[nTargets]
	      = new Vector(targetCoord[nTargets]->topocentricPos(MJD_2000,PTI_LATITUDE,PTI_LONGITUDE,PTI_ALTITUDE));

	    if (verbose)
	      (void) printf(" at %s\n",targetCoord[nTargets]->print());

	    // If we were successful in interpreting this line as
	    // coordinates, read in the next one (presumably the next
	    // target name(s) or comment
	    tmpstr = fgets(line,LINE_LENGTH,fptr);

	  } // end read/decode target coordinates
	else
	  {
	    targetCoord[nTargets] = (coords *) NULL;
	    targetV[nTargets] = (Vector *) NULL;
	  }

	if (verbose)
	  (void) printf("\n");
	nTargets++;
      } // end processing target/while target loop
    else // This was a comment, echo it and get next line
      {
	if (verbose > NORMAL)
	  printf("%s",line);
	tmpstr = fgets(line,LINE_LENGTH,fptr);
      } // end comment processing

  if (feof(fptr))
    {
      (void) fprintf(stderr," Fatal Error -- premature eof encountered in %s\n",filename);
      exit(-1);
    }
  else
    if (verbose)
      (void) printf("#  with respect to ");



  // Get the calibrators
  // Read the next line for the number of calibrators
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,"%d",&nCalibrators) != 1)
    {
      (void) fprintf(stderr," Fatal Error -- Error reading calibration star information from %s\n");
      exit(-1);
    }
  else
    if (verbose)
      (void) printf("%d calibrators: \n",nCalibrators);

  int i = 0;
  while (!feof(fptr) && (i < nCalibrators)) // loop over calibrator read
    {
      (void) fgets(line,LINE_LENGTH,fptr);
      // Check the line for a comment
      if (line[0] != '#')
	{
	  // Check the read for format
	  if ((rn = sscanf(line,"%s",&calName[i])) == 1)
	    {

	      // Skip over the calibrator name
	      int tmp = 0;
	      while (line[tmp] != ' ')
		tmp++;

	      calibCoord[i] = parseCoordLine(&line[tmp],&calSize[i],&sigmaSize[i]);
	      calibV[i]
		= new Vector(calibCoord[i]->topocentricPos(MJD_2000,PTI_LATITUDE,PTI_LONGITUDE,PTI_ALTITUDE));
	      if (verbose)
		(void) printf("#  star %s  at %s  size %g  error %g",
			      calName[i],calibCoord[i]->print(),calSize[i],sigmaSize[i]);

	      calSize[i] *= MAS_TO_RAD;
	      sigmaSize[i] *= MAS_TO_RAD;

	      if (nTargets == 1)
		if (targetV[0])
		  if (verbose)
		    (void) printf("  %3.2g degrees away from target",
				  angle_between(calibV[i],targetV[0]) * RAD_TO_DEG);

	      if (verbose)
		(void) printf("\n");
	      i++; // increment calibrator counter
	    } // end successful read of calibrator name/info

	} // end check comment field
      else // otherwise echo the comment
	if (verbose > NORMAL)
	  printf("%s",line);
    } // end i/read calibrator loop

  if (i != nCalibrators)
    {
      if (verbose)
	{
	  (void) printf("# Warning: only %d calibrators read from %s (%d expected)\n",
			i,filename,nCalibrators);
	  (void) fprintf(stderr,"# Warning: only %d calibrators read from %s (%d expected)\n",
			 i,filename,nCalibrators);
	}
      nCalibrators = i;
    }

  // Close the script file
  (void) fclose(fptr);

/**********  end readScript  ********************************************/
}


/************************************************************************
*  printRunDirectives -- local to calib.cc (A.B. 4/11/98)		*
*									*
*  A.  Purpose:	The function printRunDirectives prints a report of the  *
*       calibration directives for this run.				*
*									*
*  B.  Background and use: As calib acquires options, it becomes        *
*       increasingly necessary to document the settings of the program  *
*       for a given dataset.  This function dumps a short report on the *
*       settings used in this run to stdout (where the rest of the      *
*       output is also going). 						*
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration printRunDirectives  *******************/
void printRunDirectives()
{
/**********  begin printRunDirectives  **********************************/
  // Reports on some operational (used to be preprocessor) directives
  (void) fprintf(stdout,(char *) "\n# nbCalib version %s\n",(char *) VERSION);
  (void) fprintf(stdout,(char *) "# nbCalib calibration code settings for this run: \n");

  switch (usingWL)
    {
    case 'A' :
      {
	(void) fprintf(stdout,"# > Calibrating all wavebands\n");
	break;
      }
    case 'K' :
      {
	(void) fprintf(stdout,"# > Calibrating only K-band (rejecting others at read)\n");
	break;
      }
    case 'H' :
      {
	(void) fprintf(stdout,"# > Calibrating only H-band (rejecting others at read)\n");
	break;
      }
    } // end waveband switch

  if (usingFR == 0)
    (void) fprintf(stdout,"# > Calibrating all frame rates\n");
  else
    (void) fprintf(stdout,"# > Calibrating %fHz frame rate (%dms frame time) data (rejecting others at read)\n",
		   usingFR,((int) rint(1000.0 / (usingFR))));
    

  if (usingIncoherentV2)
    (void) printf("# > Calibrating incoherent spectrometer V2 data\n");
  else
    (void) printf("# > Calibrating coherent spectrometer V2 data\n");

  if (usingJitterCorrection)
    (void) printf((char *) "# > Applying jitter correction, coefficient %g\n",
		  jitterCoeff);

  if (usingRatioCorrection)
    (void) printf("# > Applying ratio correction\n");

  (void) printf("# > Using calibration scans within +/- %g hr of the target scan\n",
		calibTimeWindow * 24.0);

  if (usingTempErrorProp)
    {
      (void) printf("# > Weighting calibration scans WRT temporal proximity,");
      (void) printf(" variance doubles in %g hr\n",timeErrorPropCoeff / 24.0);
    }
  else
    (void) printf("# > No weighting of calibration scans WRT temporal proximity\n");

  if (usingSkyErrorProp)
    {
      (void) printf("# > Weighting calibration scans WRT sky proximity,");
      (void) printf(" variance doubles in %g deg\n",RAD_TO_DEG / angleErrorPropCoeff);
    }
  else
    (void) printf("# > No weighting of calibration scans WRT sky proximity\n");

  if (usingMinimumScanUncertainty)
    {
      (void) printf("# > Applying minimum calibrated scan uncertainty of %g\n",minScanUncertainty);
    }
  else
    (void) printf("# > No minimum calibrated scan uncertainty\n");

  if (useRootN)
    (void) printf((char *) "# > Weighting computed scan errors by root-Nsamples\n");
  else
    (void) printf((char *) "# > Not weighting computed scan errors by root-Nsamples\n");

  (void) printf((char *) "# > Using %g s scan time threshold\n",(scanThreshold * 3600.0 * 24.0));

  (void) printf("# > Flagging calibration warnings when multiple calibrators disagree by more than %g sigma\n",
		sqrt(chiWarning));
  (void) printf("# > Rejecting the scan when multiple calibrators disagree by more than %g sigma\n",
		sqrt(chiReject));

  if (delayCheck)
    (void) printf((char *) "# > Comparing model/measured delays with %3g m tolerance\n",
		  delayCheckTol);
  else
    (void) printf((char *) "# > Not comparing model/measured delays\n");
 
 // added with new option by hjb 5/2005 ... just following along
  if(!baselineFile.empty())
    (void) printf((char *) "# > Using baseline file %s\n", baselineFile.c_str());
  if(usingFluxBiasCorrection) {
      if(!fluxBiasFile.empty())
         (void) printf((char *) "# > Using Flux Bias Coefficients from file %s\n", fluxBiasFile.c_str());
      else
         (void) printf((char *) "# > Using built-in Flux Bias Coefficients.\n");
  } // if(usingFluxBiasCorrection)

  (void) printf((char *) "\n\n");


/**********  end printRunDirectives  ************************************/
}

/************************************************************************
*  blineInsert -- local to nbCalib.cc (jsh 8/2008)                      *
*                                                                       *
*  A.  Purpose: The function blineInsert adds baselines from file fptr  *
*      to the bline map. Used by readBaseline().                        *
*                                                                       *
*  B.  Background and use:                                              *
*                                                                       *
*  C.  Called by:                                                       *
*                                                                       *
*  D.  Input/output files:                                              *
*                                                                       *
*  E.  Parameter description:                                           *
*       Input -                                                         *
*       Output -                                                        *
*                                                                       *
*  F.  Detail function description:                                     *
*             1.  Local Functions:                                      *
*             2.  Externals:                                            *
*             3.  Globals and Nonlocals used:                           *
*             4.  Global and Nonlocal variables modified:               *
*             5.  Function flow is as follows:                          *
*             6.  Functions called by                                   *
*                                                                       *
*  G.  Function value: This function returns no useable value.          *
*                                                                       *
*************************************************************************/
void blineInsert(FILE *fptr, map<string,baseline*> &blines, char work[LINE_LENGTH]) {
    while(1) {
        baseline *bl = new baseline(fptr);
        if(! bl->error())
           blines.insert(map<string,baseline*>::value_type(bl->baselineName(), bl));
        else
           break;
    }

    if(blines.empty()) {
        (void) printf("# Error: no valid baselines read from %s\n", work);
    } else {
        map<string,baseline*>::iterator pos;
        for(pos = blines.begin(); pos != blines.end(); ++pos)
          (void) printf("# Found baseline name %s in baseline file %s \n", pos->first.c_str(), work);
    }

/**********  end blineInsert  ************************************/
}


/************************************************************************
*  readBaseline -- local to nbCalib.cc (A.B. 3/2000)			*
*									*
*  A.  Purpose:	The function readBaseline instantiates a baseline       *
*       object to go along with the spec file being read/to be read.    *
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
//baseline *readBaseline(char *specFilename, char *baselinef)
// updated function interface 2/06, hjb
// updated 8/08, jsh
map<string,baseline*> readBaseline(char *specFilename, char *baselinef)
{
/**********  local declarations readBaseline  ***************************/
  char work[LINE_LENGTH], *ptr;

  FILE *fptr = NULL;
  baseline *bLine = NULL;
  map<string,baseline*> blines;
  
/**********  begin readBaseline  ****************************************/
  // One.  Extract trailing suffix and append default baseline suffix
  (void) strncpy(work,specFilename,LINE_LENGTH);
  if ((ptr = strstr(work,(const char *) ".spec")) != NULL)
    (void) strcpy(ptr, (const char *) ".baseline\0");

  // One.Five hjb 5/2005
  // working in the baseline override
  // the idea is not to disturb any of the original code for this function
  // so we'll take the user supplied file name, check that, and if it's readable,
  // slip it into the "work" variable used by the rest of the fucntion. if it's 
  // not readable, we're going to fall back the the original search strategy for
  // finding a baseline
  if(baselinef[0] != '\0') {
    // make sure the file is available, accessible, around for reading
    // if not, emit warnings and do nothing, thereby falling back to the 
    // standard search for baseline file (nearly transparent to user)
    if((fptr = fopen(baselinef,"r")) == NULL) {
      (void) printf("# Error reading user supplied baseline file %s\n", baselinef);
      (void) printf("# Attempting to open baseline file\n");
    } else {
      // copy baseline into function's idea of a "work" file, protect against overflow
      (void) strncpy(work, baselinef, sizeof(work));
      work[sizeof(work)] = '\0';
      blineInsert(fptr, blines, work);
      if(!blines.empty())
         return blines;
    }
  }

  // if user supplied baseline file is missing or not valid, try to read from
  // the xxx.baseline file
  if((fptr == NULL) || (blines.empty())) {
      if((fptr = fopen(work, "r")) == NULL) {
         (void) printf("# Error trying to open baseline file %s\n", work);
      }
      else {
         blineInsert(fptr, blines, work);
         if(!blines.empty())
             return blines;
      }
  }

  // if xxx.baseline file is missing or not valid, try to read from
  // the xxx.bline file
  if((fptr == NULL) || (blines.empty())) {
      (void) printf("# Attempting to open bline file\n");
      if( (ptr = strstr(work, (const char *) ".baseline")) != NULL)
        (void) strcpy(ptr, (const char *) ".bline\0");
      if((fptr = fopen(work, "r")) == NULL) {
        (void) printf("# Error trying to open bline file %s\n", work);
        (void) printf("# Error Could not read any baseline specification\n");
        return blines;
      }
      blineInsert(fptr, blines, work);
  }
  return blines;

  // Two.  Check to see we can in fact read the file.  If so, create a
  // baseline object and return a pointer to it
//   if ((fptr = fopen(work,"r")) != NULL) {
//     (void) fclose(fptr);
//     baseline *thisBaseline = new baseline(work);
//     if (!thisBaseline->error()) {
//       if (verbose > NORMAL)
// 	(void) printf("## reading baseline model from %s (%s)\n",
// 		      work,thisBaseline->baselineName());
//       return thisBaseline;
//     } else {
//       if (verbose > NORMAL)
// 	(void) printf("\n## Error detected in %s",work);
//       delete thisBaseline;

//       // Attempt to find/read the RT system baseline file...
//       (void) strncpy(work,specFilename,LINE_LENGTH);
//       if ((ptr = strstr(work,(const char *) ".spec")) != NULL)
// 	(void) strcpy(ptr, (const char *) ".baseline\0");
      
//       if ((fptr = fopen(work,"r")) != NULL) {
// 	(void) fclose(fptr);
// 	baseline *thisBaseline = new baseline(work);
// 	if (!thisBaseline->error()) {
// 	  if (verbose > NORMAL)
// 	    (void) printf("; reading baseline model from %s (%s)\n",
// 			  work,thisBaseline->baselineName());
// 	  return thisBaseline;
// 	} else {
// 	  // The .baseline file failed as well -- return default baseline
// 	  if (verbose > NORMAL)
// 	    (void) printf(" and %s; provisionally using default baseline (%s)\n",
// 			  work,defBaseline->baselineName());
// 	  delete thisBaseline;
// 	  return defBaseline;
// 	} // failed .baseline file
// 	// end finding .baseline file
//       } else {
// 	// Couldn't _find_ .baseline file
// 	if (verbose > NORMAL)
// 	  (void) printf(" and %s not found; provisionally using default baseline (%s)\n",
// 			work,defBaseline->baselineName());
// 	return defBaseline;
//       }
//     } // end couldn't parse/error in .bline file
//     // end reading .bline baseline file
//   } else { // Three.  Couldn't _find_ .bline file -- try .baseline
//     if (verbose > NORMAL)
//       (void) printf("## %s not found",work);
//     // Attempt to find/read the RT system baseline file...
//     (void) strncpy(work,specFilename,LINE_LENGTH);
//     if ((ptr = strstr(work,(const char *) ".spec")) != NULL)
//       (void) strcpy(ptr, (const char *) ".baseline\0");
    
//     if ((fptr = fopen(work,"r")) != NULL) {
//       (void) fclose(fptr);
//       baseline *thisBaseline = new baseline(work);
//       if (!thisBaseline->error()) {
// 	if (verbose > NORMAL)
// 	  (void) printf("; reading baseline model from %s (%s)\n",
// 			work,thisBaseline->baselineName());
// 	return thisBaseline;
//       } else {
// 	// The .baseline file failed as well -- return default baseline
// 	if (verbose > NORMAL)
// 	  (void) printf(" and %s has errors; provisionally using default baseline (%s)\n",
// 			work,defBaseline->baselineName());
// 	delete thisBaseline;
// 	return defBaseline;
//       } // failed .baseline file
//     } else { 
//       // Couldn't _find_ .baseline file
//       if (verbose > NORMAL)
// 	(void) printf(" and %s not found; provisionally using default baseline (%s)\n",
// 		      work,defBaseline->baselineName());
//       return defBaseline;
//     }
//   } // end couldn't find .bline

  /**********  end readBaseline  ******************************************/
}


/************************************************************************
*  readSpVline -- local to calib.cc (A.B. 12/18/97)			*
*									*
*  A.  Purpose:	The function readSpVline scans the input text line for  *
*       nChannels of spectrometer visibility information.		*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns a one (1) on successful    *
*       completion, and a zero (0) otherwise (int).                     *
*									*
*************************************************************************/
/**********  function declaration readSpVline  **************************/
int readSpVline(char *line,int nSkip,int nChannels,
		double *nPhot,double *lambda,
		double *v2c,double *v2i,double *ratio,double *frameRate,
		// hjb 2/06, added blDesig to interface, but not used for
		// KI spec files
		char *blDesig)
{
/**********  local declarations readSpVline  ****************************/
#define CHANNEL_READ 5
#define DEF_FRAME_RATE 100

  char *separator = ((char *) " \n"),*wkstr = line,*tmpstr;
  int i,j;

/**********  begin readSpVline  *****************************************/
  /* Prototype spec line

SPEC: 11/06/97      1596   2.161944    -0.52  7   2.195  2.385  0.624  1.974  1.079  7.858  2.320  0.695  0.892  1.035  9.525  2.254  0.697  0.817  1.045  9.896  2.189  0.718  0.858  1.063 11.470  2.124  0.661  0.914  1.062  7.782  2.058  0.648  0.859  1.056  2.860  1.993  0.503  0.642  1.080

Added frame rate in 2001 A.B.
SPEC: 05/25/101 HDC121107    3.711736   -4.1739532  5 35.531  2.396  0.568  0.750  1.000 68.477  2.299  0.555  0.731  1.002 76.750  2.203  0.548  0.724  1.005 66.785  2.106  0.560  0.758  1.008 31.513  2.009  0.520  0.740  1.004  50 


*/

  // Using strtok technology.  Skip past the stuff we already have.
  tmpstr = strtok(wkstr,separator);
  for (i = 1; i < nSkip; i++)
    tmpstr = strtok(NULL,separator);

  // Now read each spectrometer channel's information.
  for (i = 0; i < nChannels; i++)
    {
      switch (nSkip)
	{
	case 7 :  // PTI vis spec file format
	  {
	    nPhot[i] = atof(tmpstr);
	    tmpstr = strtok(NULL,separator);
	    lambda[i] = atof(tmpstr);
	    tmpstr = strtok(NULL,separator);
	    v2c[i] = atof(tmpstr);
	    tmpstr = strtok(NULL,separator);
	    v2i[i] = atof(tmpstr);
	    tmpstr = strtok(NULL,separator);
	    ratio[i] = atof(tmpstr);
	    tmpstr = strtok(NULL,separator);
	    break;
	  } // end 7/PTI format

	case 12 :  // Kvis spec file format
	  {
	    lambda[i] = atof(tmpstr);
	    tmpstr = strtok(NULL,separator);
	    nPhot[i] = atof(tmpstr);
	    tmpstr = strtok(NULL,separator);
	    v2i[i] = atof(tmpstr);
	    tmpstr = strtok(NULL,separator);
	    v2c[i] = atof(tmpstr);
	    tmpstr = strtok(NULL,separator);
	    ratio[i] = atof(tmpstr);
	    tmpstr = strtok(NULL,separator);
	    break;
	  } // end 12
	} // end switch

    } // end i loop
	
  // return 0; // read error condition

  // Look to see whether we can read a frame rate as the last thing on
  // the line

  //  if (frameRate) {
  // if ((!tmpstr)
	// hjb 2/06 ... try to get the baseline desig too
  //	|| (!sscanf(tmpstr,"%d %s",frameRate, blDesig)))
  //   *frameRate = DEF_FRAME_RATE;
  //}

  if(tmpstr != NULL) {
    sscanf(tmpstr, "%lf", frameRate);
    tmpstr = strtok(NULL, separator);
    if(tmpstr != NULL) {
      sscanf(tmpstr, "%s", blDesig);
    }
  }

  // If we get here we're okay -- return good-to-go
  return 1;

/**********  end readSpVline  *******************************************/
}


/************************************************************************
*  readSpecFile -- local to calib.cc (A.B. 7/27/97)			*
*									*
*  A.  Purpose:	The function readSpecFile reads the input spec file     *
*       and extracts the relevant information for storage in the        *
*       linked-list star data structure.				*
*									*
*  B.  Background and use:						*
*         readSpecFile is designed to be called multiple times with     *
*       multiple specfile arguments to construct the database to be     *
*       used in this calibration run.                                   *
*         Input to the routine consists of the name of the specfile to  *
*       be read and parsed, and the (pointer to the) database of        *
*       visibility information that is being constituted.  On output    *
*       information from scans successfully read-in from the input      *
*       specfile have been entered into the database.  All this         *
*       database manipulation is accomplished via a set of class        *
*       libraries described in nbStarList.h (and nbStarList.cc).        *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        filename -- the (pointer to the) specfile filename to be       *
*         processed (char *).                                           *
*        starList -- the (pointer to the) starlist database object      *
*         being constructed for later use (StarList *).                 *
*	Output -							*
         starList -- contents modified.                                 *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the number of valid data   *
*       lines that were successfully read from the input file (int).    *
*									*
*************************************************************************/
int readSpecFile(char *filename,char *baselinef,char *fluxcoeffs, StarList *starList)
{
/**********  local declarations readSpecFile  ***************************/
#define MIN_JITTER 0.0
#define MAX_JITTER 3.0
#define MIN_RATIO 0.99
#define MAX_RATIO 5.0

  char line[SPEC_LINE_LENGTH],starName[132],blDesig[132],camera[256];
  int month,day,year,hour,minute,second,nRead =0,
    nChannels = 0,dataLines = 0,calFlag=0;
  double jd,ut,delay,longDelay,jitter,frameRate=0.0;
  int jd_doy = 0; // julian day of year

  FILE *fptr = NULL;
  FILE *fptr_coeff = NULL;
  line[0] = '\0';
  blDesig[0] = '\0';
  starName[0] = '\0';

  // Try to open this spectrometer file
  if (!(fptr = fopen(filename, (char *) "r"))) {
    if (verbose)
      (void) printf("# >> Error opening spectrometer file %s\n",filename);
    return 0;
  }

  // create a list of targets & calibrators names from the calscript file
  if(firstSpecFile == 0) { // list only needs to be created once
      for(int j=0; j<nCalibrators; j++) {
          calscriptStars[starCount] = calName[j];
          starCount++;
      }
      for(int j=0; j<nTargets; j++) {
         for(int k=0; k<=nTargetNames[nTargets]; k++) {
            calscriptStars[starCount] = targetName[j][k];
            starCount++;
         }
      }
      firstSpecFile = 1;
  } // if(firstSpecFile == 0)

  // Read/retrieve the baseline/camera corresponding to this spec file (or use
  // default baseline if missing or error)
  // baseline *thisBaseline = readBaseline(filename, baselinef);

  // hjb 2/06 ... use the new constructor in libbaseline which allows for comments
  // and multi-record baseline files
  // in the future this needs update to allow for time associate of baseline to data record
  // baseline *thisBaseline = readBaseline(filename, baselinef);
  map<string,baseline*> blines = readBaseline(filename, baselinef);

  if(blines.empty()) {
    (void) printf("# Warning ... continuing with built-in baseline values\n");
    if(usingFluxBiasCorrection) {
        (void) printf("# Warning: Because default baselines are being used, no flux bias correction will be applied.\n");
        usingFluxBiasCorrection = 0;
    }
  } else {
    map<string,baseline*>::iterator pos;
    for(pos = blines.begin(); pos != blines.end(); ++pos) {
      (void) printf("# Using baseline %s ", pos->second->baselineName());
      (void) printf("(%f, %f, %f, %f) \n", pos->second->localBaseline()[0],
 		    pos->second->localBaseline()[1], pos->second->localBaseline()[2], pos->second->bias());
    }
    (void) printf("\n");
  }

    // Create a map of baselines/camera names that are found in the spec file
    // but which do not exist in the baselineFile. This is used to print
    // warnings a single time rather than once for each spec line that has
    // a baseline/camera name missing in the baseline file
    map<string,int> blWarningMap;

    // Create a map of camera names that are found in the spec file but which
    // do not exist in the fluxBiasFile. This is used to print warnings a single
    // time for each unique camera name in the spec file that does not exist in
    // the flux file
    map<string,int> camWarningMap;

  /* 12/17/97.  Okay, we're reading Mark's spectrometer file format, which
     according to the Ben Lane looks like this:

>    date     starname     time(hrs) del(m) #ch nphot  wl     V2c    V2i    ratio  nphot... 
SPEC: 10/04/97 HDC197572   4.771667  -6.96  7   2.798  2.385  0.650  0.778  1.068  7.207  2.320  0.682  0.862  1.110  8.195  2.254  0.726  0.816  1.124  8.766  2.189  0.620  0.670  1.102  9.165  2.124  0.593  0.843  1.130  5.820  2.058  0.635  0.731  1.162  1.699  1.993  0.807  0.456  1.034

  One extra tidbit.  I've added a check to see if a data line has
  been commented out by a pound sign.  This allows for some degree
  of user intervention in the instance of bogus data.

  */

  /* new KI line format:

SPEC 11/07/2001/08:03:21 8.055854 none       -12.530247  40.000000  1.186  500  0 KI_K1K2      4 2.400  481.3 0.737 0.512 1.200 2.300  574.9 0.697 0.480 1.200 2.200  593.3 0.653 0.440 1.200 2.100  440.3 0.617 0.413 1.200 

  */

  while (!feof(fptr)) {
    (void) fgets(line,SPEC_LINE_LENGTH,fptr);
    if (!(feof(fptr)) && (line[0] != '#') &&
	(((nRead = sscanf(line,
			  "SPEC: %d/%d/%d %s %lf%lf%d",
			  &month,&day,&year,&starName,&ut,&delay,&nChannels)) == 7)
	 || ((nRead = sscanf(line,
			     "SPEC %d/%d/%d/%d:%d:%d %lf %s %lf%lf%lf%lf%d%s%d",
			     &month,&day,&year,&hour,&minute,&second,&ut,
			     &starName,&delay,&longDelay,&jitter,&frameRate,&calFlag,
			     &blDesig,&nChannels)) == 15))) {

          int skipLine = 0;

// for each spec line,
//   if starName matches a target/calib
//      get the numSpecChans and doy
//	if there isn't already a doy/numSpecChans map member, then create one
//      compare current doy/numSpecChans pair just read to the map's doy/numSpecChans
//      you can only have one numSpecChan per doy
//      if the doy == current doy && numSpecChan != currentNumSpecChan
//          exit program with error message
//          numSpecChan changed during the night


            // Turn dates and UTs into JDs
            if (year < 96)
              year += 100;
            if (year < 1900)
              year += 1900;
            jd = jdFromUTCal(ut,year,month,day);
            jd_doy = (int) jd;	

	    //  JSH 08/05/2009 
	    // if starName matches a target/calib get the numSpecChans and doy
	    // if there isn't already a doy/numSpecChans map member, create one
	    // compare the doy/numSpecChans pair just read to the map's doy/numSpecChans
	    // you can only have one numSpecChan per doy
	    // if the doy == current doy && numSpecChan != currentNumSpecChan
	    //   exit program with error message numSpecChan changed during the night

            // look for lines with given starName
            for(int j=0; j<starCount; j++) {
                if(strcmp(calscriptStars[j], starName) == 0) {
                    // if no numSpecChans specified on command line, then check jd_doy/numSpecChan pair
                    if(numSpecChans == 0) {
                        // you only care about changing spec channels for the same jd_doy.
		        map<int,int>::iterator pos = nightSpecChans.find(jd_doy);
		        if((pos != nightSpecChans.end()) && (pos->second != nChannels)) {
                            (void) printf("Fatal Error reading spectrometer file %s. The number of spec channels has changed from %d to %d for star %s at time %.2d/%.2d/%d/%.2d:%.2d:%.2d\n", filename, pos->second, nChannels, starName, month, day, year, hour, minute, second);
                            (void) exit(-1);
                        }
                        else { // if a pair doesn't exist, create one
	 	            nightSpecChans.insert(make_pair(jd_doy,nChannels));
                        } 
                    } // end if (numSpecChans == 0)
                    else if (nChannels != numSpecChans) {
                       skipLine = 1;
		    }
               } // end if calscriptStars[j] == starName
            } // end for

	if(skipLine == 1)
           continue;


      // Allocate necessary memory for channel data
      double *nPhot = new double [nChannels];
      double *lambda = new double[nChannels];
      double *v2c = new double [nChannels];
      double *v2i = new double [nChannels];
      double *ratio = new double [nChannels];

      // Read the spectrometer data from line
      char *dummyp = NULL; // 
      if (((nRead == 7) && (readSpVline(line,7,nChannels,nPhot,lambda,v2c,v2i,ratio,&frameRate,blDesig)))
	  || ((nRead == 15) && (readSpVline(line,12,nChannels,nPhot,lambda,v2c,v2i,ratio,0, dummyp))))
	if ((usingWL == 'A')  // Added logic to read only specific bands  A.B. 11/99
	    || ((usingWL == 'K') && (fabs(lambda[0] - K_wl) < K_width))
	    || ((usingWL == 'H') && (fabs(lambda[0] - H_wl) < H_width))   )
	  if ((usingFR == 0)  // Checking frame rate logic.  A.B. 11/99
	      || (usingFR == frameRate)) {
	    dataLines++;

	    // KI Kvis-format line -- set baseline from designation
	    // hjb updated if clause, to allow for pti records w/ baseline tags, from imel pipeline (2/06)
	    // if( (strncmp(line,"SPEC:",5) != 0) &&
	    baseline *thisBaseline;
	    int missingBl = 0;
	    if(blDesig[0] != '\0') {
	      map<string,baseline*>::iterator pos = blines.find(blDesig);
	      if( pos == blines.end()) {
		thisBaseline = defBaseline;
		missingBl= 1;
              }	
	      else
		thisBaseline = pos->second;
	    } else {
	      if(blines.size() == 1) {
		map<string,baseline*>::iterator pos = blines.begin();
		thisBaseline = pos->second;
	      } 
	      else {
		thisBaseline = defBaseline;
		missingBl = 1;
              }
	    }

	    // this checks addresses, not actual member equality.
	    if(thisBaseline == defBaseline) {
	      // case of NULL blDesig should fall through to the end
	      if(strncmp(blDesig, (char *) "PTI_NS",6) == 0)
		thisBaseline = &PTI_NS;
	      else if(strncmp(blDesig, (char *) "PTI_NW",6) == 0)
		thisBaseline = &PTI_NW;
	      // hjb added PTI_SW baseline 2/06
	      else if(strncmp(blDesig, (char *) "PTI_SW", 6) == 0)
		thisBaseline = &PTI_SW;
	      else if(strcmp(blDesig, (char *) "KI_K1K2S") == 0)
                thisBaseline = &KI_K1K2S;
	      else if(strcmp(blDesig, (char *) "KI_K1K2") == 0)
		thisBaseline = &KI_K1K2;
	      else if(strcmp(blDesig, (char *) "KI_S1S2") == 0)
		thisBaseline = &KI_S1S2;
              // jsh added Camera_FT baselines 11/10
              else if(strcmp(blDesig, (char *) "KI_Camera_FT_PRIM") == 0)
                thisBaseline = &KI_Camera_FT_PRIM;
              else if(strcmp(blDesig, (char *) "KI_Camera_FT_SEC") == 0)
                thisBaseline = &KI_Camera_FT_SEC;
              else if(strcmp(blDesig, (char *) "KI_Camera_FT_LBAND") == 0)
                thisBaseline = &KI_Camera_FT_LBAND;
	      else
		thisBaseline = defBaseline;
 
              if(missingBl == 1) {
                 map<string,int>::iterator pos = blWarningMap.find(blDesig);

                 // only print warning the first time unique blDesig encountered
                 if(pos == blWarningMap.end()) {
                    // the "1" in the pair is arbitrary
                    blWarningMap.insert(std::make_pair(blDesig,1));
                    if(baselinef[0] != '\0')
                       (void) printf("# Warning: Baseline name %s from SPEC record in file %s not found in baseline file %s.\n", blDesig, filename, baselinef);
                    (void) printf("# Warning: Using built-in default %s baseline components E:%f N:%f U:%f C:%f for U,V and delay calculations.\n\n", thisBaseline->baselineName(), thisBaseline->localBaseline()[0], thisBaseline->localBaseline()[1],  thisBaseline->localBaseline()[2], thisBaseline->bias());
                    (void) printf("\n");
                 }
              } // missingBl == 1
            } // end set baseline/camera from designation

	    if (nRead == 7) // PTI format
	      longDelay = jitter = 0;

	    int kickout = 0;
	    if(strncmp(line,"SPEC:",5) != 0) { // KI Kvis-format line
	      
	      // hjb 06/07 
	      // jitter and ratio correction moved below (outside) the if KI statement
	      // so that those corrections can be applied to PTI data also

	      //  -- apply jitter and ratio
	      // 	      if (usingJitterCorrection) {
	      // 		if ((jitter > MIN_JITTER) && (jitter < MAX_JITTER))
	      // 		  for (int i = 0; i < nChannels; i++) {
	      // 		    double jitterCorrection = exp(jitterCoeff * jitter * jitter);
	      // 		    v2c[i] *= jitterCorrection;
	      // 		    v2i[i] *= jitterCorrection;
	      // 		  } else {
	      // 		  if (verbose)
	      // 		    (void) fprintf(stdout,(char *) "# Warning: jitter %g out of range; data line rejected\n",jitter);
	      // 		  kickout = 1; // jitter failed range test
	      // 		}
	      // 	      } // end usingJitterCorrection
	      
	      // 	      if (usingRatioCorrection) {
	      // 		for (int i = 0; (i < nChannels) && (!kickout); i++)
	      // 		  if ((ratio[i] > MIN_RATIO) && (ratio[i] < MAX_RATIO)) {
	      // 		    v2c[i] *= ratio[i];
	      // 		    v2i[i] *= ratio[i];
	      // 		  } else {
	      // 		    if (verbose)
	      // 		      (void) fprintf(stdout,(char *) "# Warning: ratio %g out of range for %s at UT %f (channel %d); Data not used\n",ratio[i], starName, ut, i);
	      // 		    // ratio failed range test
	      // 		    kickout = 1;
	      // 		  } // end ratio check
	      
	      // 	      } // end usingRatioCorrection

	      if(usingJitterCorrection) {
		if( (jitter > MIN_JITTER) && (jitter < MAX_JITTER)) {
		   for(int i = 0; i < nChannels; i++) {
		    double jitterCorrection = exp(jitterCoeff * jitter * jitter);
		    v2c[i] *= jitterCorrection;
		    v2i[i] *= jitterCorrection;
                   }
		  } else {
		  if(verbose)
		    (void) fprintf(stdout,(char *) "# Warning: jitter %g out of range for %s at UT %f; Data not used\n",
				   jitter, starName, ut);
		    kickout = 1; // jitter failed range test
		}
	      } // end usingJitterCorrection
		      
	      if(usingFluxBiasCorrection) {

	         float SpecFluxBiasCoeff;
                 float Nzero;
                 float NormFlux;

                 // blDesig must match name in blines which contains KI_ prefix
		 // fluxBiasMap does not (usually) have the "KI_" prefix. If KI_
 		 // prefix is present, strip it from blDesig to get camera name
                 if (sscanf(blDesig, "KI_%s", camera) == 0)
                    strcpy(camera, blDesig);
                 map<string, fluxBiasStruct>::iterator pos = fluxBiasMap.find(camera);
                 if(pos != fluxBiasMap.end()) {
                    SpecFluxBiasCoeff = pos->second.SpecFluxBias;
                    Nzero = pos->second.NZero;
                    NormFlux = pos->second.Normflux;

                    // Skip these warnings if cameraName read from fluxBiasFile
                    // This only apply if using built-in defaults
                    if(pos->second.usingDefaults == 1) {
                        map<string,int>::iterator pos2 = camWarningMap.find(camera);
                        // only print warning the first time camName encountered
                        if(pos2 == camWarningMap.end()) {
                           // the "1" in the pair is arbitrary
                           camWarningMap.insert(std::make_pair(camera,1));
                           if((strcmp(camera,"KI_K1K2") == 0) || (strcmp(camera,"K1K2") == 0))
                              (void) printf("# Warning: Using built-in primary camera defaults for %s: SpecFluxBias = %f  Nzero = %f  NormFlux = %f\n\n", camera, SpecFluxBiasCoeff, Nzero, NormFlux);
                           else
                              (void) printf("# Warning: Using built-in camera defaults for %s: SpecFluxBias = %f  Nzero = %f  NormFlux = %f\n\n", camera, SpecFluxBiasCoeff, Nzero, NormFlux);
                        } // if(pos2 == camWarningMap.end())
                    } // if(pos->second.usingDefaults == 1)

  		    // received 6/07 from rla/mc
    		    // visibilities should be corrected for flux bias by the formula
		    // correction = 1 - k log10 ( (Ni+Nzero)/(Nnorm + Nzero))
		    // where Ni is this record's flux (spec or wl), Nzero about 20dN (80 phot) and the normalization
		    // level set to 1000 dN
		    // k_spec = SpecFluxBiasCoeff ~= 0.16
	    
		    for(int i = 0; i < nChannels; i++) {
		      double spec_correction = 1 - SpecFluxBiasCoeff * log10( (nPhot[i] + Nzero) / (NormFlux + Nzero) );
         	      if(v2c[i] != -999)
		         v2c[i] *= spec_correction;
	              if(v2i[i] != -999)
		         v2i[i] *= spec_correction;
		    }
                 } // if(pos != fluxBiasMap.end())
	         else { // camera name not found in fluxBiasMap
                    map<string,int>::iterator pos2 = camWarningMap.find(camera);
                    // only print warnings once for each unique cameraName
                    if(pos2 == camWarningMap.end()) {
                       // the "1" in the pair is arbitrary
                       camWarningMap.insert(std::make_pair(camera,1));

 		       if(fluxcoeffs[0] != '\0') {
                          (void) printf("# Warning: Baseline/Camera name %s in SPEC record of file %s not found in flux bias coefficients file %s. No flux bias correction will be applied.\n\n", camera, filename, fluxcoeffs);
                       }
                       else {
                          (void) printf("# Warning: Baseline/Camera name %s in SPEC record of file %s not recognized in default list of flux bias coefficients. No flux bias correction will be applied.\n\n", camera, filename);
                       }
                    } // if (pos2 == camWarningMap.end())
                 } // else
              } // if(usingFluxBiasCorrection)
            } // KI-specific corrections

	    // moved ratio correction down here  hjb 6/07
	    // as originally coded, these corrections could only apply to KI data.
	    // per rla, should be allowed for pti 
	    if((usingRatioCorrection) && (!kickout)) {
              int badRatios = 0;   
              for(int i = 0; (i < nChannels); i++) {
		if( (ratio[i] > MIN_RATIO) && (ratio[i] < MAX_RATIO)) {
                  if(v2c[i] != -999)
		     v2c[i] *= ratio[i];
                  if(v2i[i] != -999)
		     v2i[i] *= ratio[i];
		} else {
		  // ratio failed range test
		  // check each spec channel independently
		  badRatios++;
		  v2c[i] = -999;
		  v2i[i] = -999;
		} // end ratio check
              } // end for i<nChannels
              if((badRatios > 0) && (verbose)){
                   (void) fprintf(stdout,(char *) "# Warning: ratio correction value(s) out of range(%.2f-%.2f) for %d spec channel(s) for %s at UT %f. Out of range data not used.\n", MIN_RATIO, MAX_RATIO, badRatios, starName, ut);
              }
              // all ratios are bad, kick the sample out	
              if(badRatios == nChannels)
                  kickout = 1;
	    } // end usingRatioCorrection

	    // Create a narrow-band data node with this data contents
            if (!kickout) {
	      NBdata *nbNode = new NBdata(starName,0,month,day,year,
					  jd,ut,delay,longDelay,jitter,
					  nChannels,
					  lambda,nPhot,v2c,v2i,ratio,
					  frameRate,thisBaseline, blDesig);
	      // Find or create a star node corresponding to this star
	      if (!starList->findStar(starName))
		starList->insert(starName);
	      
	      // Insert the data from this line into the linked-list
	      // structure for this star.  Note it is being
	      // insertion-sorted into the list based on jd.  The proper
	      // functioning of avgV2Data depends on the list being
	      // time-sorted.
	      starList->getStarData()->insert(nbNode,jd,INSERT_SORT);
            } else 
              dataLines--; // conversion failure	
        } // end read spectrometer line successful/selecting waveband/framerate
    } // end good spectrometer line
  } // end file while loop
  
  (void) fclose(fptr);
  
  return dataLines;

}


/************************************************************************
*  avgV2Data -- local to calib.cc (A.B. 7/27/97)			*
*									*
*  A.  Purpose:	The function avgV2Data computes the average of the V2   *
*       data output by Mark's vis program for a single scan.            *
*									*
*  B.  Background and use:						*
*         avgV2Data sounds simple -- you're just averaging (and         *
*       computing sample standard deviations) for a sequence of scans.  *
*       But in point of fact avgV2Data is in fact a little trickier     *
*       in that it is the agent charged with walking the constituent    *
*       linked list using the functionality in the linked list template *
*       class library dataList (dataList.h).  On input, the current     *
*       pointer in the starData list points to the first record in the  *
*       sequence for a particular scan.  avgV2Data then walks the list  *
*       acculating pertinent sums on all the records for that scan, as  *
*       judged by being within a time window (SCAN_THRESH) of this      *
*       first record.  On output, current should now be pointing to the *
*       first record in the next scan on the target.  Clearly this      *
*       methodology requires the record list to be arranged in          *
*       increasing time order.  This is how the database is constructed *
*       in readSumFile (new data records are inserted into the list     *
*       using the INSERT_SORT option).                                  *
*         On output averages and variances are compted for the V2, and  *
*       averages are computed for jd, ut, jitter, and ratioC.           *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        starData -- the (pointer to the) linked-list object for the    *
*         object we're working on (dataList<NBdata>).                   *
*	Output -							*
*        jdAvg, utAvg, v2Avg, jitterAvg, ratioCAvg -- the averaged      *
*         quantities for this scan (various dimensions) (double *).     *
*        v2Var -- the estimated error variance in v2Avg                 *
*         (dimensionless) (double *).                                   *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the number of data entries *
*       found in this scan (int).                                       *
*									*
*************************************************************************/
int avgV2data(dataList<NBdata> *starData,
	      int *month,int *day,int *year,int *nChannels,
	      double *jdAvg,double *utAvg,double *delayAvg,
	      double *wlAvg,double *v2Avg,double *v2Var,
	      double *jitterAvg,double *ratioCAvg,double *frameRate)
{
/**********  local declarations avgV2data  ******************************/

  string firstBaseline = starData->getCurrent()->get_baseline()->baselineName();
  double firstUT = starData->getCurrent()->get_ut();
  double firstJD = starData->getCurrent()->get_jd();
  *month = starData->getCurrent()->get_month();
  *day = starData->getCurrent()->get_day();
  *year = starData->getCurrent()->get_year();
  *nChannels = starData->getCurrent()->get_nChannels();

  int i,sumTerms = 0;
  int specSumTerms[MAX_CHANNELS] = {0};
  double myFrameRate=0.0;
  double sumWL[MAX_CHANNELS] = {0.0};
  double sumV2[MAX_CHANNELS] = {0.0};
  double sumV2_2[MAX_CHANNELS] = {0.0};
  double sumJD = 0.0;
  double sumUT = 0.0;
  double sumDelay = 0.0;
  double sumJitter = 0.0;
  double sumRatioC = 0.0;

  int alreadyWarned = 0;

/**********  begin avgV2data  *******************************************/
  // Retrieve the data for this scan (accumulate the appropriate
  // sums).  Subtle but important.  Note that this loop is doing the
  // linked-list iteration.  When it kicks out all the records within
  // SCAN_THRESH of the first record on routine invocation has been read.
  do
    {
      for (i = 0; i < (*nChannels); i++)
	{
	  sumWL[i] += starData->getCurrent()->get_lambda(i);
	  if (usingIncoherentV2) // switch between incoherent and coherent
	    { // Incoherent
	      // if spec chan has passed previous validity check
              if(starData->getCurrent()->get_v2i(i) != -999) {
	          sumV2[i] += starData->getCurrent()->get_v2i(i);
	          sumV2_2[i] += starData->getCurrent()->get_v2i(i)
		      * starData->getCurrent()->get_v2i(i);
                  specSumTerms[i]++;
	      }
	    } // end incoherent
	  else
	    { // Coherent
               // if spec chan has passed previous validity check	
               if(starData->getCurrent()->get_v2c(i) != -999) {
	           sumV2[i] += starData->getCurrent()->get_v2c(i);
	           sumV2_2[i] += starData->getCurrent()->get_v2c(i)
		        * starData->getCurrent()->get_v2c(i);
                   specSumTerms[i]++;
               }
	    } // end coherent
	} // end i loop

      sumDelay += starData->getCurrent()->get_delay()
	- starData->getCurrent()->get_longDelay();
      sumJD += starData->getCurrent()->get_jd();
      sumUT += starData->getCurrent()->get_ut();
      // sumJitter += starData->getCurrent()->get_jitter();
      sumRatioC += starData->getCurrent()->get_ratio(i);
      myFrameRate = starData->getCurrent()->get_frameRate();
      sumTerms++;

      if(!alreadyWarned) {
          string currentBaseline = starData->getCurrent()->get_baseline()->baselineName();
          if(currentBaseline != firstBaseline) {
              (void) printf((char *) "\n# Warning: Baseline name changed from %s to %s within the averaging block starting at %f. Only Baseline %s will be used for this average.\n", firstBaseline.c_str(), currentBaseline.c_str(), firstUT, firstBaseline.c_str());
              alreadyWarned = 1;
          }
      } // if(!alreadyWarned)
    } // do
  while ((starData->iterate())
	 && (fabs(starData->getCurrent()->get_jd() - firstJD) < scanThreshold));

  // Compute the output values.

  double denom = 1.0 / (double) sumTerms;
  // For each of the spectral channels...
  for (i = 0; i < (*nChannels); i++)
    {
      double specDenom = 0.0;
      wlAvg[i] = sumWL[i] * denom;
      if(specSumTerms[i] != 0) {
          specDenom = 1.0 / (double) specSumTerms[i];
          v2Avg[i] = sumV2[i] * specDenom;
          v2Var[i] = (sumV2_2[i] - (sumV2[i] * sumV2[i] * specDenom)) * specDenom;
          if (useRootN)          // Add root-N scaling (N in variance) to
   	     v2Var[i] *= specDenom;   //  computed error bars A.B. 7.02
      }
      else { // if no valid v2 values for this spec chan then avg=-999
          v2Avg[i] = -999;
          v2Var[i] = -999;
      }
    } // end i loop

  *delayAvg = sumDelay * denom;
  *jdAvg = sumJD * denom;
  *utAvg = sumUT * denom;
  *jitterAvg = sumJitter * denom;
  *ratioCAvg = sumRatioC * denom;
  *frameRate = myFrameRate;

  // Return the number of records we processed for this scan
  return sumTerms;

/**********  end avgV2data  *********************************************/
}


/************************************************************************
*  v2Pred -- local to calib.cc (A.B. 7/27/97)				*
*									*
*  A.  Purpose:	The function v2Pred computes the expected v2 value (and *
*       variance) for a given calibrator and projected normalized       *
*       baseline length (aka inverse fringe spacing).			*
*									*
*  B.  Background and use:						*
*         v2Pred is pretty simple: it computes the expected V2 for an   *
*       ideal interferometer observing a simple star -- presumably a    *
*       calibrator.  The ratio of what we got to what we expected is    *
*       an estimate of the system visibility.                           *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        baseline -- the sky-projected baseline length (m) (double).    *
*        lambda -- the wavelength we're working at (m) (double).        *
*        aSize -- the estimated angular size of object (rad) (double).  *
*        aSigma -- the error in aSize (rad) (double).                   *
*	Output -							*
*        varV2Pred -- the variance in the output V2 prediction based    *
*         on the input angular size error (dimensionless) (double *).   *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the expected V2 value      *
*       (dimensionless) (double).                       		*
*									*
*************************************************************************/
double v2Pred(double baseline,double lambda,
	      double aSize,double aSigma,double *varV2Pred)
{
/**********  begin v2Pred  **********************************************/
  // The fringe spacing on the sky (units of radians)
  double fringeSP = baseline / lambda;
  double arg = PI * fringeSP * aSize;
  double dArgdaSize = PI * fringeSP;
  double v,dvdArg;

  // Trap the (annoying) point source case
  if (arg > 0)
    {
      v = 2.0 * j1(arg) / arg;
      dvdArg = 2.0 * (0.5 * arg * (j0(arg) - jn(2,arg)) - j1(arg)) / arg / arg;
    }
  else
    {
      v = 1.0;
      dvdArg = 0.0;
    }

  // Use the chain rule (otherwise I never get this right).
  double dv2dv = 2.0 * v;

  *varV2Pred = dv2dv * dvdArg * dArgdaSize * aSigma;
  *varV2Pred *= *varV2Pred;

  return (v * v);
/**********  end v2Pred  ************************************************/
}


/************************************************************************
*  sysVisSC -- local to calib.cc (A.B. 7/27/97)				*
*									*
*  A.  Purpose: The function sysVisSC estimates the system visibility   *
*       (and the error in that estimate) with respect to a particular   *
*       calibrator at a particular jd.					*
*									*
*  B.  Background and use:						*
*         sysVisSC mechanizes the estimate of the system visibility at  *
*       a particular epoch based on a particular calibrator.  In the    *
*       calib scheme, each calibrator is considered independently (this *
*       routine), and then these results are (weighted) averaged to     *
*       derive the system visibility estimate for the target (done      *
*       in sysVisEnsemble).                                             *
*         sysVisSC does its work by extracting all calibrator scans     *
*       within a given time window (CALIWINDOW) around the input      *
*       time.  From each calibrator scan a system visibility estimate   *
*       (and error estiamte) is derived.  These system visibility       *
*       estimates are then (weighted) averaged to determine the best    *
*       estimate for the system visibility at the input time.  Note:    *
*       these system visibility estimates _can_ be weighted with        *
*       respect to temporal proximity to the desired output time.       *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the Julian Data for the system visibility estimate.      *
*        calibData -- the (pointer to the) data list object for this    *
*         calibrator (dataList<NBdata> *).                              *
*        calSize -- the estimated angular size for this calibrator      *
*         (rad) (double).                                               *
*        sigmaSize -- the estimated error in calSize (rad) (double).    *
*        star -- the (pointer to the) unit 3-vector sky position for    *
*         this calibrator (dimensionless) (Vector *).                   *
*	Output -							*
*        varSysVis -- the estimated error variance in the system        *
*         visibility (squared) estimate (dimensionless) (double *).     *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the estimate of system     *
*       visibility (squared) at the time of interest derived from a     *
*       particular calibrator (dimensionless) (double).		        *
*									*
*************************************************************************/
/**********  function declaration sysVisSC  *****************************/
int sysVisSC(double jd,double targFrameRate,
	     dataList<NBdata> *calibData,char *calName,
	     double calSize,double sigmaSize,
	     Vector *star,int *nChannels,
	     double *sysVisEst,double *varsysVisSC)
{
/**********  local declarations sysVisSC  *******************************/
// Propagation coefficient for system visibility error.
// A doubling of the sysVisSC error estimate in an hour.
#define PROP_COEFF (1.0 * 24.0)

  int i,month,day,year,calHits = 0;
  double jdAvg = 0.0,utAvg = 0.0,calFrameRate=0.0;
  double V2Avg[MAX_CHANNELS] = {0.0};
  double V2sig[MAX_CHANNELS] = {0.0};
  double weight,w2;
  double weightSum[MAX_CHANNELS] = {0.0};
  double weightSum2[MAX_CHANNELS] = {0.0};
  double weightSumVar[MAX_CHANNELS] = {0.0};
  double expV2[MAX_CHANNELS],varExpV2[MAX_CHANNELS];
  double jdScan,utScan,delayScan,jitterScan,ratioCScan;
  double WLScan[MAX_CHANNELS],V2Scan[MAX_CHANNELS],V2ScanVar[MAX_CHANNELS];

/**********  begin sysVisSC  ********************************************/
  // Loop over the designated calibration targets
  (void) calibData->gotoHead();
  if (calibData->numberNodes())
    {
      // Skip forward to the interesting time
      while (!calibData->endOfList()
	     && ((jd - calibData->getCurrent()->get_jd()) > calibTimeWindow))
	calibData->iterate();

      // Extract relevant calibration information from this
      // calibration source
      while ((!calibData->endOfList())
	     && (fabs(jd - calibData->getCurrent()->get_jd()) < calibTimeWindow))
	{
	  baseline *Baseline = calibData->getCurrent()->get_baseline();
	  if ((avgV2data(calibData,&month,&day,&year,nChannels,
			 &jdScan,&utScan,&delayScan,
			 WLScan,V2Scan,V2ScanVar,
			 &jitterScan,&ratioCScan,&calFrameRate) > 1)
	      && (calFrameRate == targFrameRate)) // require frame rate match
	    {
	      calHits++;
	      if (delayCheck)
		{
		  // New 11/30/97.  Check the delay value for this calibration object
		  // double tmp = delayScan - PTI_NSbase.delay(jdScan,*star);
		  double tmp = delayScan - Baseline->delay(jdScan,*star);
		  if (fabs(tmp) > delayCheckTol)
		    {
		      if (verbose)
			{
			  (void) printf("# Warning: at %10.4f, %4g m measured/expected delay mismatch on calibrator %s\n",jdScan,tmp,calName);
			  (void) fprintf(stderr,"Warning: at %10.4f, %4g m measured/expected delay mismatch on calibrator %s\n",jdScan,tmp,calName);
			} // verbose reporting check
		    } // end delay value check
		} // end delayCheck

	      for (i = 0; i < (*nChannels); i++) // Spectral channel loop
		{
                  if((V2Scan[i] != -999) || (V2ScanVar[i] != -999)){
		    // Compute the expected V2 value at this epoch, sky
		    // position, baseline projection, wavelength...
		    expV2[i] = v2Pred(Baseline->skyProj(jdScan,*star),
				    WLScan[i] * 1e-6,
				    calSize,sigmaSize,&varExpV2[i]);

     		    if (usingTempErrorProp)
		    {
		      // If defined, propagate the error in the system visibility
		      //  to the time of the target observation.  Assuming linear
		      //  (quadratic) increase in the uncertainty (variance).
		      V2ScanVar[i] *= 1.0 + pow(timeErrorPropCoeff * fabs(jd - jdScan),1.0);
		    } // end usingTempErrorProp

		    // Weighting.  Weight by variance, jitter
  		    weight = 1.0 / V2ScanVar[i];
		    // weight *= exp(- jitterScan * jitterScan);
		    weightSum[i] += weight;
		    weightSum2[i] += w2 = weight * weight;

		    V2Avg[i] += weight * (V2Scan[i] / expV2[i]);
		    weightSumVar[i] += w2 * (V2ScanVar[i] / expV2[i] / expV2[i]);
		    weightSumVar[i] += w2 * pow((V2Scan[i] / expV2[i] / expV2[i]),2.0) * varExpV2[i];
                  } // if ((V2Scan[i] != -999) || (V2ScanVar[i] != -999))
		} // channels loop
	    } else { // end non-trivial calibration scan
	    (void) printf("# Warning: Calibration scan on %s, near %f, has only 1 SPEC rec. Skipping scan.\n", calName, utScan);
	  }
	} // calibration scan
      
      if (calHits) // If we've found good calibration hits
	{
	  for (i = 0; i < (*nChannels); i++)
	    {
              if(V2Avg[i] != 0) {
	         varsysVisSC[i] = weightSumVar[i] / weightSum2[i];
	         sysVisEst[i] = V2Avg[i] / weightSum[i];
	      }
              else {
                 varsysVisSC[i] = -999;
                 sysVisEst[i] = -999;
              }
	    } // end spectral channel
	  return 1;
	} // end good calibration
      else // Otherwise, return a stupid value to flag calib 
	return 0;

    } // end calibData node count check
  else // no data for this calibrator
    return 0;

/**********  end sysVisSC  **********************************************/
}


/************************************************************************
*  sysVisEnsemble -- local to calib.cc (A.B. 10/25/97)			*
*									*
*  A.  Purpose:	sysVisEnsemble computes the system visibility estimate  *
*       from the full ensemble of calibrators selected for this         *
*       calibration run.						*
*									*
*  B.  Background and use:						*
*         sysVisEnsemble mechanizes the computation of an average       *
*       system visibility estimate from the composite of single         *
*       calibrator system visibility estimates (computed in sysVisSC).  *
*       sysVisEnsemble also performs system visibility estimate         *
*       inner-consistency checks among multiple calibrators.            *
*         In the case of multiple calibrators, the cross-consistency    *
*       of derived system visibility estimates is checked, and          *
*       mismatches at two levels are flagged for the user as a          *
*       diagnostic that there might be something wrong with either the  *
*       input calibrator information or the calibration object itself.  *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the Julian Data for the system visibility estimate       *
*         (double).                                                     *
*        targV -- the (pointer to the) unit 3-vector sky position of    *
*         the target (dimensionless) (Vector *).  Note: this is okay    *
*         if NULL, it just means there can be no angular proximity      *
*         weighting in the system visibility estimate computation.      *
*        starList -- the (pointer to the) visibility database object    *
*         (StarList *).                                                 *
*	Output -							*
*        sysVis -- the composite estimated system visibility at the     *
*         desired time (dimensionless) (double *).                      *
*        varSysVis -- the estimated error variance in sysVis            *
*         (dimensionless) (double *).                                   *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration sysVisEnsemble  ***********************/
int sysVisEnsemble(double jd,Vector *targV,
		   double targFrameRate,StarList *starList,
		   int nChannels,double *sysVis,double *varSysVis)
{
/**********  local declarations sysVisEnsemble  *************************/
  int i,j;
  static double sV2[MAX_CALIBRATORS][MAX_CHANNELS],
    varSV2[MAX_CALIBRATORS][MAX_CHANNELS];

/**********  begin sysVisEnsemble  **************************************/
  // First Loop over the calibrators.
  // Determine the estimated system visibility at the time of
  // this scan based on each calibrator...
  for (i = 0; i < nCalibrators; i++)
    if (starList->findStar(calName[i]))
      {
	dataList<NBdata> *calibData = starList->getStarData();
	// Evaluate the calibrator apparent position
	// Bug alert: No way to ensure that this baseline/camera is valid for this time (quant effects small)
	baseline *Baseline = calibData->getCurrent()->get_baseline();
	*(calibV[i]) = calibCoord[i]->topocentricPos(jd,Baseline->latitude(),
						     Baseline->longitude(),
						     Baseline->altitude());

	// Check/flag failed calibration attempts
	if (!(sysVisSC(jd,targFrameRate,
		       calibData,calName[i],calSize[i],sigmaSize[i],
		       calibV[i],&nChannels,sV2[i],varSV2[i])))
	  sV2[i][0] = -1.0;
      } // end findStar check/calibrator loop
    else
      sV2[i][0] = -1.0;

  // Second loop over calibrators
  // Compute the weighted average of the system visibility
  // calibrations, and check the interconsistency of the
  // calibrations at this time.
  int calHits = 0;
  double weight,w2,weightSum[MAX_CHANNELS] = {0.0},v2Sum[MAX_CHANNELS] = {0.0};
  for (i = 0; i < nCalibrators; i++) {
    if (sV2[i][0] > 0.0) // if there was a non-trivial calibration
      {
	calHits++;

	// Loop over spectrometer channels
	for (j = 0; j < nChannels; j++)
	  {
            if((sV2[i][j] != -999) && (varSV2[i][j] != -999)){
	       // Add angular proximity weighting as desired
	       if (usingSkyErrorProp)
	         if (targV)
	   	   varSV2[i][j] *= 1.0 + angleErrorPropCoeff * angle_between(calibV[i],targV);

  	       weightSum[j] += weight = 1.0 / varSV2[i][j];
	       w2 = weight * weight;
	       v2Sum[j] += weight * sV2[i][j];
	
	       // Check for calibration inconsistencies
	       for (int k = i + 1; k < nCalibrators; k++) {
	         if ((sV2[k][0] > 0.0) && (sV2[k][j] != -999) && (varSV2[k][j] != -999))
		   {
		     double chi = pow(sV2[i][j] - sV2[k][j],2.0)
		       / (varSV2[i][j] + varSV2[k][j]);
		     // Inconsistencies are flagged at two thresholds.  The
		     // first is just warning...
		     if (chi > chiWarning)
		       {
		         if (chi < chiReject)
		   	 {
			    if (verbose)
			      {
			        (void) printf("# Warning: at %10.4f sysVis estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
					    jd,calName[i],calName[k],sqrt(chi));
			        (void) fprintf(stderr,"Warning: at %10.4f system visibility estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
					     jd,calName[i],calName[k],sqrt(chi));
			       } // end verbose check
			   } // end warning
		           else // The second is an outright rejection of the scan...
			   {
			      (void) printf("# Reject: at %10.4f sysVis estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
					jd,calName[i],calName[k],sqrt(chi));
			      (void) fprintf(stderr,"Reject: at %10.4f system visibility estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
					 jd,calName[i],calName[k],sqrt(chi));
			      // return a null-calibration
			      return 0;
			   }
		       } // end chi test
                     } // end if(sV2[k][0] != -999 && varSV2[k][j] != -999)
                 } // end cross-calibration check (k) loop
             } // end if(sV2[i][j] != -999 && varSV2[i][j] != -999)
	  } // end channel (j) loop 
      } // end weight-average loop
  } // end for loop

  // If we've gotten a good calibration, compute the system visibility.
  if (calHits)
    for (j = 0; j < nChannels; j++)
      {
         if((v2Sum[j] != 0) && (weightSum[j] != 0)) {
	    sysVis[j] = v2Sum[j] / weightSum[j];
	    varSysVis[j] = 1.0 / weightSum[j];
         }
         else {
            sysVis[j] = -999;
            varSysVis[j] = -999;
         }
      } // end (j) channel loop/calibration hit check

  // Return the number of calibrators used in this system calibration
  return calHits;

/**********  end sysVisEnsemble  ****************************************/
}



/************************************************************************
*  calib -- local to calib.cc (A.B. 7/27/97)				*
*									*
*  A.  Purpose:	The function calib drives the calibration of V2 data    *
*       output by Mark's vis program.					*
*									*
*  B.  Background and use:						*
*         Other than database construction, calib is where it all       *
*       happens.  calib loops over the targets to be calibrated, the    *
*       list of pseudonyms for each target, and the scans recorded      *
*       under each pseudonym.  For each of these scans the average      *
*       data is computed (avgV2Data), and a system visibility estimate  *
*       is derived from the average of system visibility estimates      *
*       computed from multiple independent calibration objects          *
*       (sysVisEnsemble).                                               *
*         In the case of multiple calibrators, the cross-consistency    *
*       of derived system visibility estimates is checked, and          *
*       mismatches at two levels are flagged for the user as a          *
*       diagnostic that there might be something wrong with either the  *
*       input calibrator information or the calibration object itself.  *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        starList -- the (pointer to the) visibility database object    *
*         (StarList *).                                                 *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration calib  ********************************/
void calib(StarList *starList)
{
/**********  local declarations calib  **********************************/
  int i,month,day,year,nChannels,numScans = 0;
  double jdAvg,utAvg,delayAvg,jitterAvg,ratioCAvg,targFrameRate=0.0;
  double WLAvg[MAX_CHANNELS] = {0.0};
  double V2Avg[MAX_CHANNELS] = {0.0};
  double V2Var[MAX_CHANNELS] = {0.0};
  double sV2[MAX_CALIBRATORS],varSV2[MAX_CALIBRATORS],sV4;

/**********  begin calib  ***********************************************/
  // Output a status line indicating the column contents in the
  // calibfile format
  if (verbose)
    {
      (void) printf("# Calibration data format:\n");
      (void) printf("# name\t#VisPts\tMJD\tm/d/y/h:m:s\tUT\tDelay(m)\t#Chan\tWL(um)\tCalVis\tErr\tRawVis\tErr\tSysVis\tErr...\t#Cal\tU(m)\tV(m)\tHA(hr)\tBL/CameraName\n");
    }

  // Now do the calibration

  // Loop over the calibration targets
  for (int nTarg = 0; nTarg < nTargets; nTarg++)
    {
      numScans = 0;
      (void) printf("\n# Target %s:\n",targetName[nTarg][0]);

      // Loop over the pseudonyms for each target
      for (int pseudonym = 0; pseudonym < nTargetNames[nTarg]; pseudonym++)
	// If we've got hits on the target under this name...
	if (starList->findStar(targetName[nTarg][pseudonym]))
	  {
	    int nDataName = 0;
	    dataList<NBdata> *targetData = starList->getStarData();
	    (void) targetData->gotoHead();
	    if (targetData->numberNodes())
	      do {
		int st,nCal;
		double tFrameRate = targetData->getCurrent()->get_frameRate();
		double sV2m[MAX_CHANNELS],varSV2m[MAX_CHANNELS];
		baseline *Baseline = targetData->getCurrent()->get_baseline();
		
		// Compute the averaged V2 data for the calibration
		// target at this time.  If we've got a non-trivial
		// number of target scans...
		if ((st = avgV2data(targetData,&month,&day,&year,
				    &nChannels,&jdAvg,&utAvg,&delayAvg,
				    WLAvg,V2Avg,V2Var,
				    &jitterAvg,&ratioCAvg,
				    &targFrameRate)) > 1) {
		  // Compute the ensemble system visibility estimate
		  // at this time.  If we've got a good visibility
		  // calibration...
		  // store the current pointer, so we can reset after calibration
		  // neccessary for --doCalibrators option
		  NBdata* resetNode = targetData->getCurrent();
		  if (nCal = sysVisEnsemble(jdAvg,targetV[nTarg],
					    targFrameRate,starList,
					    nChannels,sV2m,varSV2m))
		    {
		      // New 11/6/97  Check the delay value for the target object
		      if ((targetCoord[nTarg]) && (delayCheck))
			{
			  // Evaluate the instantaneous target apparent position
			  Vector targ
			    = targetCoord[nTarg]->topocentricPos(jdAvg,
								 Baseline->latitude(),
								 Baseline->longitude(),
								 Baseline->altitude());

			  double delay = Baseline->delay(jdAvg,targ);
			  double delayDiff = delayAvg - delay;
			  if (fabs(delayDiff) > delayCheckTol)
			    {
			      if (verbose)
				{
				  (void) printf("# Warning: at %10.4f, %g m measured/expected delay mismatch on target %s\n",jdAvg,delayDiff,targetName[nTarg][pseudonym]);
				  (void) fprintf(stderr,"Warning: at %10.4f, %g m measured/expected delay mismatch on target %s\n",jdAvg,delayDiff,targetName[nTarg][pseudonym]);
				} // end verbose check
			      
			    } // end delay value check
			  
			} // end target position check
		      
		      // Some time machinations
		      int hr = (int) utAvg;
		      double frac = utAvg - ((double) hr);
		      int min = (int) (frac * 60.0);
		      frac -= ((double) min / 60.0);
		      int sec = (int) (frac * 3600.0);
		      
		      // Output the header information -- names, times, delay, # spec channels
		      (void) printf("%s\t%d\t%11.5f\t%d/%d/%d/",
				    targetName[nTarg][pseudonym],
				    st,jdAvg,month,day,year);

		      // Mucho tedious logic to get the time format correct
		      // (it wouldn't be so tedius if we used printf the way god intended)
		      if (hr < 10)
			(void) printf((char *) "0%1d:",hr);
		      else
			(void) printf((char *) "%2d:",hr);
		      
		      if (min < 10)
			(void) printf((char *) "0%1d:",min);
		      else
			(void) printf((char *) "%2d:",min);
		      
		      if (sec < 10)
			(void) printf((char *) "0%1d",sec);
		      else
			(void) printf((char *) "%2d",sec);
		      
		      (void) printf((char *) "\t%g\t%g\t%d\t",
				    utAvg,delayAvg,nChannels);
		      
		      // Loop over each spectrometer channel, writing out the WL,
		      //  calibrated, raw, and system v2...
		      for (int chan = 0; chan < nChannels; chan++)
			{
                         double varTV2;
                         if((V2Avg[chan] == -999) || (V2Var[chan] == -999) || (sV2m[chan] == -999) || (varSV2m[chan] == -999)) {
                            (void) printf((char *) "%g\t%s\t%s\t%s\t%s\t%s\t%s\t",
                                        WLAvg[chan],
                                        "NULL","NULL",
                                        "NULL","NULL",
                                        "NULL","NULL");

		         } else {
			    varTV2 = V2Var[chan] / sV2m[chan] / sV2m[chan]
  			       + pow(V2Avg[chan] / sV2m[chan] / sV2m[chan],2.0) * varSV2m[chan];

			    double scanSigma = sqrt(varTV2);
                            if((usingMinimumScanUncertainty) && (scanSigma < minScanUncertainty)) {
                                 scanSigma = minScanUncertainty;
                            }

			    (void) printf((char *) "%g\t%g\t%g\t%g\t%g\t%g\t%g\t",
					WLAvg[chan],
					V2Avg[chan] / sV2m[chan],scanSigma,
//					V2Avg[chan] / sV2m[chan],sqrt(varTV2),
					V2Avg[chan],sqrt(V2Var[chan]),
					sV2m[chan],sqrt(varSV2m[chan]));
                         }
			} // end channel loop
		      
		      (void) printf("%d",nCal);
		      
		      // And if we've got a position on the
		      // target, output u-v coordinates and the
		      // like to help our consumers...
		      if (targetCoord[nTarg])
			{
			  Vector targ // The instantaneous apparent direction of target
			    = targetCoord[nTarg]->topocentricPos(jdAvg,
								 Baseline->latitude(),
								 Baseline->longitude(),
								 Baseline->altitude());
			  
			  Vector SV = Baseline->skyProjV(jdAvg,targ);
			  
			  (void) printf((char *) "\t%10.6f\t%10.6f\t%g",SV[0],SV[1],
					Baseline->hourAngle(jdAvg,targ));
			}
		      // Print out the baseline/camera name
		      if(targetData->getCurrent()->get_cameraName()[0] != '\0')
	                  (void) printf((char *) "\t%s\n",targetData->getCurrent()->get_cameraName());
                      else // PTI has no camera name so use baseline
			  (void) printf((char *) "\t%s\n", targetData->getCurrent()->get_baseline()->baselineName());
		      (void) printf((char *) "\n");
		      numScans++;
		      nDataName++;
		    } // end successful target scan/calibration
		  else { // Otherwise, calibration failure...notify the user.
		    if (st > 1) {
		      (void) printf("# No suitable calibration found for %10.4f\n",jdAvg);
		    }
		  }
		  targetData->resetCurrent(resetNode);
		} else {
		  (void) printf("# Warning: Target scan on %s, near UT %f, has only 1 SPEC record. Will not calibrate.\n", targetName[nTarg][pseudonym], utAvg);
		}
		

	      } // end walking the target list
	      while (!targetData->endOfList());

	    (void) printf("# %d calibrated scans output under name %s\n",
			  nDataName,targetName[nTarg][pseudonym]);

	  } // end psedonym list/data check
	
      (void) printf("# %d total calibrated scans output on target %s\n",
		    numScans,targetName[nTarg][0]);
    } // end nTarg loop
/**********  end calib  *************************************************/
}

int readFluxBiasFile(char *fluxcoeffs) {
  FILE *fptr_coeff = NULL;
  map<string, fluxBiasStruct>::iterator pos;

     if(fluxcoeffs[0] != '\0') {
       if( !(fptr_coeff = fopen(fluxcoeffs, (char *) "r"))) {
          (void) printf((char *) "# >> Error opening FluxBiasCoefficients File %s\n", fluxcoeffs);
          (void) printf((char *) "# Correcting for flux dependent bias with built in values\n");
       } else {
          // sscanf the line out of the coefficients file
          // Line format:
          // Baseline/CameraName WLFluxBiasCoeff SpecFluxBiasCoeff NzeroFlux NormFlux
	  // NOTE: WLFluxBiasCoeff is not used for nbCalib, but in order to be
	  // able to use the same fluxBiasCoeff file for both nbCalib and wbCalib,
	  // it is required to be included for nbCalib.

          // Baseline/CameraName can be KI_K1K2, KI_K1K2S, Camera_FT_PRIM,
	  // Camera_FT_SEC, Camera_FT_LBAND
          // If we put the defaults in a file, it would look like:
          // Camera_FT_PRIM 0.063 0.16 20 1000
          char cameraName[256];
 	  float wlcoeff, speccoeff, nzero, norm;
          int ret;
          char *line, line2[1024], linebuf[1024];

          while(fgets(linebuf, sizeof(linebuf), fptr_coeff) != NULL) {
             strcpy(line2, linebuf);
             line = line2;
             if(strncmp(line,"#",1) == 0)
                 continue;

             // keeping wlcoeff so you can use the same Flux file for both
	     // wbCalib & nbCalib
	     if( (ret = sscanf(line, "%s %f %f %f %f", cameraName, &wlcoeff, &speccoeff, &nzero, &norm)) != 5) {
                (void) printf((char *) "# Error - Could not read flux dependent bias coefficients from file %s for the following line:\n", fluxcoeffs);
                (void) printf((char *) "#       %s\n", line);
                (void) printf((char *) "# Line format should be:\n");
                (void) printf((char *) "Baseline/Camera Name(string) WLFluxBiasCoeff(float) SpecFluxBiasCoeff(float) NzeroFlux(float) NormFlux(float)\n");
                (void) printf((char *) "# e.g. Camera_FT_PRIM 0.063 0.16 20 1000\n");
                (void) printf((char *) "# Please fix the line and re-run or run without -fluxBiasCorrection to use the built-in default values.\n\n");
             } else {
                // strip off any "KI_" prefix from the baseline/camera names
                char cam[256];
                if(sscanf(cameraName, "KI_%s", cam) == 1) {
                   strcpy(cameraName, cam);
                }

		// look for the baseline/camera name in the current map
                pos = fluxBiasMap.find(cameraName);

                // if not found, add it to the map
                if(pos == fluxBiasMap.end()) {
                   fluxBiasStruct newBias;
                   newBias.WLFluxBias = wlcoeff;
                   newBias.SpecFluxBias = speccoeff;
                   newBias.NZero = nzero;
                   newBias.Normflux = norm;
                   newBias.usingDefaults = 0;
                   fluxBiasMap.insert(std::make_pair(cameraName,newBias));
                }
                else {
                   pos->second.WLFluxBias = wlcoeff;
                   pos->second.SpecFluxBias = speccoeff;
                   pos->second.NZero = nzero;
                   pos->second.Normflux = norm;
                   pos->second.usingDefaults = 0;
                }
                (void) printf((char *) "# > Found camera name %s in %s\n", cameraName, fluxcoeffs);
	     }
            } // while(fgets...)
          fclose(fptr_coeff);

          (void) printf((char *) "# > Correcting flux dependent bias with the following values from %s:\n", fluxcoeffs);

          (void) printf((char *) "# >%15s %11s %9s %10s\n", "Baseline/Camera Name", "Spec_Coeff", "Nzero", "NormFlux");
          for(pos=fluxBiasMap.begin(); pos != fluxBiasMap.end(); ++pos) {
              if(pos->second.usingDefaults == 0)
                 (void) printf((char *) "%20s %12.4f %10.1f %10.1f\n", pos->first.c_str(), pos->second.SpecFluxBias, pos->second.NZero, pos->second.Normflux);
          }
          printf("\n");

          return 0;
       }
     } // if(fluxcoeffs[0] != '\0')

     else { // using default flux bias coeffs
        (void) printf((char *) "# > No flux bias coefficients file specified with -fluxBiasCorrection.\n");
     }

     (void) printf((char *) "# >     Correcting flux dependent bias with the following built-in default values:\n");
     (void) printf((char *) "# >%15s %11s %9s %10s\n", "Baseline/Camera Name", "Spec_Coeff", "Nzero", "NormFlux");
     for(pos=fluxBiasMap.begin(); pos != fluxBiasMap.end(); ++pos) {
          (void) printf((char *) "%20s %12.4f %10.1f %10.1f\n", pos->first.c_str(), pos->second.SpecFluxBias, pos->second.NZero, pos->second.Normflux);
     }
     printf("\n");

     return 0;
} // readFluxBiasFile

/**********  main routine calib  ****************************************/
main (int argc,char **argv)
{
  char baseFile[LINE_LENGTH];
  char fluxFile[LINE_LENGTH];

  // initialize baselineFile and fluxBiasFile globals to blank
  baselineFile.clear();
  fluxBiasFile.clear();

  // Process any command-line options
  int argIndex = processCmdLineOpts(argc,argv);

  // copy baseline & flux file strings into c-strings to pass to functions  
  strcpy(fluxFile, fluxBiasFile.c_str());
  strcpy(baseFile, baselineFile.c_str());

  // Trap trivial/bogus invocations
  if (argc < 3)
    (void) printHelpMessage();

  // Read the calibration script on the stars we're going to care
  // about
  (void) readScript(argv[argIndex]);

  // Report to the user on the run directives (if verbose)
  if (verbose)
    (void) printRunDirectives();

  if(usingFluxBiasCorrection) {

      // KI_K1K2 will use the default values for FT_PRIM
      //fluxBiasMap.insert(std::make_pair("K1K2",k1k2));
      fluxBiasMap.insert(std::make_pair("K1K2",ftPrim));
      fluxBiasMap.insert(std::make_pair("K1K2S",k1k2s));
      fluxBiasMap.insert(std::make_pair("Camera_FT_PRIM",ftPrim));
      fluxBiasMap.insert(std::make_pair("Camera_FT_SEC",ftSec));
      fluxBiasMap.insert(std::make_pair("Camera_FT_LBAND",ftLband));
      readFluxBiasFile(fluxFile);
  }

  // Initialize the data structure and read in the data from the
  // spec file(s)
  StarList *starList = new StarList();
  for (int i = argIndex + 1; i < argc; i++)
    {
      int nData = readSpecFile(argv[i], baseFile, fluxFile, starList);

      if (verbose)
	(void) printf("# %d data lines read from %s\n",nData,argv[i]);
    }

  // Now we do the calibration.  Output is to stdout (to be redirected
  // as desired.  Isn't unix great...?)
  (void) calib(starList);

/**********  end program calib  ****************************************/
}
