/**************************************************************************
*  NASA Exoplanet Science Institute (NExScI) -- http://nexsci.caltech.edu *
*	Copyright (c) 1997-2009 California Institute of Technology        *
*                     U.S. Government Sponsorship                         *
* Under contract with the National Aeronautics and Space Administration   *
***************************************************************************/

// wbCalib -- calibration of wide-band visibility amplitudes for PTI
// and KI -- Andy Boden 1997 through 2003

// now wrapped by a Perl script.  This C++ core of the wbCalib app has
// been renamed to wbcEngine.cc -- David A. Imel 2004

// *************************
// nullCalib version  RLA 12-10-07

#include <config.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include "getopt.h"

#include "angles.h"
#include "Vector.h"

#include "timeUtil.h"

#include "KI.h"
#include "baseline.h"

#include "coords.h"

#include "nullStarList.h"
#include <map>
#include <string>
#include <iostream>
using namespace std;

#define LINE_LENGTH 255

#define MAX_TARGETS 2000
#define MAX_PSEUDONYMS 10
#define MAX_CALIBRATORS 2000

#define LAMBDA 2.18e-6

#define MJD_2000 51544.5

// Delay Check tolerance -- approximately 3 arcminutes
#define DELAY_CHECK_TOLERANCE 0.10


// Global declarations

// Create the default baseline objects.  Use the form of the
// constructor with delay bias -- we'll be checking delay values
baseline KI_K1K2(KI_LONGITUDE,KI_LATITUDE,
		 KI_Baseline_K1K2,KI_K1K2_DELAY_BIAS,"KI_K1K2");

baseline *defBaseline = &KI_K1K2;
char baselineFile[LINE_LENGTH];

// Star Stuff

// Calibration Target Names, Counts, and Positions (arrays -- sorry)
char *targetName[MAX_TARGETS][MAX_PSEUDONYMS];
int nTargets,nTargetNames[MAX_TARGETS];
Vector *targetV[MAX_TARGETS];  // Going to phase this out...
coords *targetCoord[MAX_TARGETS];


// Calibrator Names, Count, and Positions  (arrays -- sorry)
int nCalibrators;
char calName[MAX_CALIBRATORS][132];
Vector *calibV[MAX_CALIBRATORS]; // Hoping to phase this out...
coords *calibCoord[MAX_TARGETS];
double calSize[MAX_CALIBRATORS],sigmaSize[MAX_CALIBRATORS];



// Program Operational Options.  These are used to control the code's
// processing behavior.  Global variables are the best mechanism I've
// come up with to handle this situation...

// Calibration time window for picking up calibration scans.
// units are in days, so 1/24 is +/- 1 hour time window
// #define CALIB_WINDOW (1.0 / 24.0)
double calibTimeWindow = (1.0 / 24.0);

// Propagation coefficient for system null error.
// A doubling of the sysNullSC error estimate in an hour.  See sysNullSC.
// Units are in inverse days, so 24 is 1 hr.
// #define PROP_COEFF (1.0 * 24.0)
int usingTempErrorProp = 1;
double timeErrorPropCoeff = (1.0 * 24.0);

// Propagating null errors across the sky.  See sysNullEnsemble.
// Units are in inverse radians.
// #define ANGLE_COEF (1.0 / (15.0 * DEG_TO_RAD))
int usingSkyErrorProp = 1;
double angleErrorPropCoeff = (1.0 / (15.0 * DEG_TO_RAD));

// Minimum possible scan uncertainty
int usingMinimumScanUncertainty = 0;
int usingMinimumCalUncertainty = 0;
double minScanUncertainty = 0.01;
double minCalUncertainty = 0.01;

// DoCalibrators flag:  treats the calibrators as target stars
int useDoCalibrators = 0;

// Thresholds for reporting/rejecting calibration errors (dimensionless).
// Threshold for reporting calibration errors -- 2.0 sigma disagreement
// #define CHI_WARNING 1.0
double chiWarning = 2.0 * 2.0;
// Threshold for rejecting the calibration -- 3.0 sigma disagreement
// #define CHI_REJECT 9.0
double chiReject = 3.0 * 3.0;

// Verbosity variable to control reporting Added 7/01 A.B.
#define MUTE 0
#define NORMAL 1
#define EXCESSIVE 2
int verbose = NORMAL; // Default to certain errors

// Variable for delay checking
int delayCheck = 1;
double delayCheckTol = DELAY_CHECK_TOLERANCE;


/************************************************************************
*  printHelpMessage -- local to calib.cc (A.B. 4/13/98)			*
*									*
*  A.  Purpose:	The function printHelpMessage prints a help message     *
*       for the calib program to stderr and then exits the program.	*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description: None					*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration printHelpMessage  *********************/
void printHelpMessage()
{

/**********  begin printHelpMessage  ************************************/
  (void) fprintf(stdout,(char *) "Usage: nullCalib calScript sumfile1 [sumfile2...] [> cal.out]\n\n");
  (void) fprintf(stdout,(char *) "Calibrate KI nuller data\n\n");
  (void) fprintf(stdout,(char *) "Options:\n\n");
  (void) fprintf(stdout,(char *) "  -help: print (this) help message and exit (overrides all other options)\n");
  (void) fprintf(stdout,(char *) "  -version: print version information and exit\n");
  (void) fprintf(stdout,(char *) "  -calTimeWindow argument: set calibration scan acceptance time window (hr)\n");
  (void) fprintf(stdout,(char *) "  -[no]timeWeighting: set use/nonuse of time proximity weighting of calibration scans\n");
  (void) fprintf(stdout,(char *) "  -timeErrorVal argument: if timeWeighting, set calibration scan time weighting value (hr)\n");
  (void) fprintf(stdout,(char *) "  -[no]angleProxWeighting: set use/nonuse of sky angle proximity weighting of calibration scans\n");
  (void) fprintf(stdout,(char *) "  -angleErrorVal argument: if angleProxWeighting, set calibration scan angle weighting value (deg)\n");
  (void) fprintf(stdout,(char *) "  -[no]minScanUncertainty argument: sets [non]use of minimum possible scan uncertainty of value argument (V2 units)\n");
  (void) fprintf(stdout,(char *) "  -[no]minCalUncertainty argument: sets [non]use of minimum possible cal uncertainty of value argument (V2 units)\n");
  (void) fprintf(stdout,(char *) "  -chiWarning argument: set calibrator inconsistency warning chi threshold \n");
  (void) fprintf(stdout,(char *) "  -chiReject argument: set calibrator inconsistency scan rejection chi threshold \n");
  (void) fprintf(stdout,(char *) "  -[no]delayCheck[=argument]: set use/nonuse of delay value checking \n\t (optional delay tolerance argument) \n");
  (void) fprintf(stdout,(char *) "  -verbose|normal|quiet|shutUp: control volume of information, warning, and error messages\n");
  (void) fprintf(stdout,(char *) "  -doCalibrators:  doCalibrators development debugging option\n");
  (void) fprintf(stdout,(char *) "  -baseline <file>: Use the baseline supplied by <file>. Normal behavior is to look for a\n");
  (void) fprintf(stdout,(char *) "                    yyyyddd.nullbaseline file for each yyyyddd.nullsum file given on the command\n");
  (void) fprintf(stdout,(char *) "                    line. If that fails then look for yyyyddd.nullbline. If neither is found then\n");
  (void) fprintf(stdout,(char *) "                    nullCalib uses a built in default baseline. The option allows the user to\n");
  (void) fprintf(stdout,(char *) "                    override that search by providing the filename directly. Note that when\n");
  (void) fprintf(stdout,(char *) "                    multiple NULLSUM files are given in combination with this option all of them\n");
  (void) fprintf(stdout,(char *) "                    use the supplied <file>.\n");

  (void) fprintf(stdout,(char *) "  note:  command-line options may be used in any order, interspersed with non-options\n");
  (void) fprintf(stdout,(char *) "   and abbreviated to uniqueness (e.g. -h, -noang, -incoh -K).\n");
  (void) fprintf(stdout,(char *) "  Find more nullCalib info at http://nexsci.caltech.edu/software\n");

  (void) exit(-1);


/**********  end printHelpMessage  **************************************/
}

/**********  Little utility for a formatted UT time stamp ******************/

void formatted_time(double ut, char* time_string)
{
  // Some time machinations
  int hr      = (int) ut;
  double frac = ut - ((double) hr);
  int min     = (int) (frac * 60.0);
  frac       -= ((double) min / 60.0);
  int sec     = (int) (frac * 3600.0);
  
  // Mucho tedious logic to get the time format correct

  sprintf(time_string,"%02d:%02d:%02d",hr,min,sec);
}


/******** end utilities ********/

/************************************************************************
*  processCmdLineOpts -- local to calib.cc (A.B. 4/12/98)		*
*									*
*  A.  Purpose:	The function processCmdLineOpts processes the command   *
*       line invocation options used with calib for this run.		*
*									*
*  B.  Background and use: processCmdLineOpts uses getopt_long_only to  *
*       process the command line options the user has added to the      *
*       invocation.  getopt_long_only is used because it allows the     *
*       arbitrarily intersperse command-line options along with the     *
*       filename command-line arguments: it permutes the command line   *
*       as it processes so when finished all the non-options (i.e. the  *
*       filename arguments) are at the end (for simple processing by    *
*       the main routine).				                *
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        argc, argv -- the standard command-line argument variables.    *
*	Output -							*
*        argv's contents are permuted so as to move all non-option      *
*         arguments to the end.                                         *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the index of the first     *
*       non-option on the command-line, which is presumably the name    *
*       of the calibration script file (int).              		*
*									*
*************************************************************************/
/**********  function declaration processCmdLineOpts  *******************/
int processCmdLineOpts(int argc, char **argv)
{
/**********  local declarations processCmdLineOpts  *********************/
  // Valid command-line options
  int retVal,optIndex,argsProcessed = 0;
  static struct option longOptions[] = 
  {
    {((char *) "help"),no_argument,0,0},                     // Option 0
    {((char *) "version"),no_argument,0,0},                  // Option 1
    {((char *) "timeWeighting"),no_argument,0,0},            // Option 2
    {((char *) "notimeWeighting"),no_argument,0,0},          // Option 3
    {((char *) "angleProxWeighting"),no_argument,0,0},       // Option 4
    {((char *) "noangleProxWeighting"),no_argument,0,0},     // Option 5
    {((char *) "calTimeWindow"),required_argument,0,0},      // Option 6
    {((char *) "timeErrorVal"),required_argument,0,0},       // Option 7
    {((char *) "angleErrorVal"),required_argument,0,0},      // Option 8
    {((char *) "minScanUncertainty"),required_argument,0,0}, // Option 9
    {((char *) "nominScanUncertainty"),no_argument,0,0},     // Option 10
    {((char *) "chiWarning"),required_argument,0,0},         // Option 11
    {((char *) "chiReject"),required_argument,0,0},          // Option 12
    {((char *) "quiet"),no_argument,0,0},                    // Option 13
    {((char *) "normal"),no_argument,0,0},                   // Option 14
    {((char *) "verbose"),no_argument,0,0},                  // Option 15
    {((char *) "shutUp"),no_argument,0,0},                   // Option 16
    {((char *) "delayCheck"),optional_argument,0,0},         // Option 17
    {((char *) "nodelayCheck"),no_argument,0,0},             // Option 18
    {((char *) "doCalibrators"),no_argument,0,0},            // Option 19
    {((char *) "baseline"),required_argument,0,0},           // Option 20
    {((char *) "minCalUncertainty"),required_argument,0,0},  // Option 21
    {((char *) "nominCalUncertainty"),no_argument,0,0},      // Option 22
    {0,0,0,0}
  };

  extern char *optarg;
  extern int optind,opterr,optopt;

  /**********  begin processCmdLineOpts  **********************************/

  // Print welcome/version
  if (verbose)
    (void) fprintf(stdout,(char *) "## This is nullCalib v%s\n\n",(char *) VERSION);
  
  do {
    if((retVal = getopt_long_only(argc,argv,(char *) "",longOptions,&optIndex)) != EOF) {
      if (verbose)
	(void) printf((char *) "# option %s\n",longOptions[optIndex].name);
      argsProcessed++;
      switch (optIndex) {

      case 0: {
	(void) printHelpMessage();
	break;
      } // end printing help message -- option 0

      case 1: {
	(void) fprintf(stdout,(char *) "nullCalib v%s\n\n",(char *) VERSION);
	(void) exit(0);
	break;
      } // end printing version message -- option 1

      case 2: {
	usingTempErrorProp = 1;
	break;
      } // end setting usingTempErrorProp -- option 2

      case 3: {
	usingTempErrorProp = 0;
	break;
      } // end setting NOT usingTempErrorProp -- option 3

      case 4: {
	usingSkyErrorProp = 1;
	break;
      } // end setting usingSkyErrorProp -- option 4

      case 5: {
	usingSkyErrorProp = 0;
	break;
      } // end setting NOT usingSkyErrorProp -- option 5

      case 6: {
	calibTimeWindow = atof(optarg);
	calibTimeWindow /= 24.0;
	break;
      } // end setting calibTimeWindow -- option 6

      case 7: {
	timeErrorPropCoeff = atof(optarg);
	timeErrorPropCoeff *= 24.0;
	break;
      } // end setting timeErrorPropCoeff -- option 7

      case 8: {
	angleErrorPropCoeff = atof(optarg);
	angleErrorPropCoeff *= DEG_TO_RAD;
	angleErrorPropCoeff = 1.0 / angleErrorPropCoeff;
	break;
      } // end setting angleErrorPropCoeff -- option 8 

      case 9: {
	usingMinimumScanUncertainty = 1;
	minScanUncertainty = atof(optarg);
	break;
      } // end setting usingMinimumScanUncertainty/minScanUncertainty -- option 9

      case 10: {
	usingMinimumScanUncertainty = 0;
	break;
      } // end setting usingMinimumScanUncertainty -- option 10

      case 11: {
	chiWarning = atof(optarg);
	chiWarning *= chiWarning;
	break;
      } // end setting chiWarning -- option 11

      case 12: {
	chiReject = atof(optarg);
	chiReject *= chiReject;
	break;
      } // end setting chiReject -- option 12

      case 13: {
	// be quiet!!!
	verbose = MUTE;
	break; 
      } // end shutting up -- option 13

      case 14: {
	// normal (normal setup, serious warnings and errors)
	verbose = NORMAL;
	break; 
      } // setting normal reporting -- option 14

      case 15: {
	// verbose (everything)
	verbose = EXCESSIVE;
	break; 
      } // end setting excessive reporting -- option 15

      case 16: {
	// SHUT UP!!!
	verbose = MUTE;
	break; 
      } // end setting shutUp -- option 16

      case 17: {
	delayCheck = 1;
	if (optarg) {
	  delayCheckTol = atof(optarg);
	} // end setting delay check tolerance
	break;
      } // end setting delayCheck -- option 17

      case 18: {
	delayCheck = 0;
	break;
      } // end setting delayCheck -- option 18

      case 19: {
	useDoCalibrators = 1;
	break;
      } // end setting useDoCalibrators -- option 19

      case 20: {
	strncpy(baselineFile, optarg, sizeof(baselineFile)-1);
	baselineFile[sizeof(baselineFile)] = '\0';
	break;
      } // option 20

      case 21: {
        usingMinimumCalUncertainty = 1;
        minCalUncertainty = atof(optarg);
        break;
      } // end setting usingMinimumCalUncertainty/minCalUncertainty -- option 21

      case 22: {
        usingMinimumCalUncertainty = 0;
        break;
      } // end setting usingMinimumCalUncertainty -- option 22
	
      } // end big butt-ugly option switch
    } // end if return value not EOF
  } while (retVal != EOF);

  if (argsProcessed && verbose)
    (void) printf((char *) "# %d command line arguments processed\n",argsProcessed);

  return optind;
  
/**********  end processCmdLineOpts  ************************************/
}


/************************************************************************
*  parseCoordLine -- local to calib.cc (A.B. 11/98)			*
*									*
*  A.  Purpose:	The function parse coordinate line parses the           *
*       input coordinate line for the target/calibrator and extracts    *
*       the pertinent information					*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns a pointer to the           *
*       coordinate object created by as a result of this line		*
*       (coords *).                                                     *
*									*
*************************************************************************/
coords *parseCoordLine(char *line,double *calSize,double *sigmaSize)
{
/**********  local declarations parseCoordLine  *************************/
  char *separator = ((char *) "#"),*tmpstr;
  char decStr[256], *minusSign = ((char *) "-");
  double ra,dec;
  float rah,ram,ras,decd,decm,decs,pmRA,pmDec,plx;
  float cSize,sSize;
  coords *wrk;


/**********  begin parseCoordLine  **************************************/
  (void) strtok(line,separator); // Use strtok to stomp on any comments

  int rn = sscanf(line,(char *) "%lf%f%f%s%f%f%f%f%f%f%f",
		  &ra,&ram,&ras,decStr,&decm,&decs,&pmRA,&pmDec,&plx,
		  &cSize,&sSize);

  dec = atof(decStr);
  int decSign = 1; // handle negative declinations (stupid)!!!
  if (strncmp(decStr,minusSign,1) == 0) // If the first character is a minus...
    decSign = -1;

  switch (rn)
    {
    case 11: // We got the whole smack
      {
	// template 15 21 06.880 +24 57 28.872 -0.029 -0.019 0.00587 1.05  0.1
	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec,pmRA,pmDec,plx);

	if (calSize)
	  *calSize = cSize;
	if (sigmaSize)
	  *sigmaSize = sSize;

	break;
      }

    case 9: // We got full coordinates but no calibrator info
      {
	// template 15 34 41.268 +26 42 52.895  0.135 -0.089 0.04365

	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec,pmRA,pmDec,plx);

	// If we were _expecting_ calibrator information, generate an error
	if (calSize && sigmaSize)
	  {
	    (void) fprintf(stderr,(char *) "Fatal error parsing calibration script\n");
	    exit(-1);
	  }

	break;
      } // end rn = 9

    case 8: // We got positions and calibrator info
      {
	// Template 16 01 02.662 +33 18 12.634 0.70 0.1
	// Note this case is degenerate with coords, pm, no plx, but I
	// have to assume something...
	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec);

	// The calibrator info got read into the pm variables!!!
	if (calSize)
	  *calSize = pmRA;
	if (sigmaSize)
	  *sigmaSize = pmDec;

	break;
      } // end rn = 8

    case 6: // We got positions but no calibrator info
      {
	// Template 16 01 02.662 +33 18 12.634 
	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec);

	// If we were _expecting_ calibrator information, generate an error
	if (calSize && sigmaSize)
	  {
	    (void) fprintf(stderr,(char *) "Fatal error parsing calibration script\n");
	    exit(-1);
	  }

	break;
      } // end rn = 6

    case 4: // We got decimal degrees and calibrator information
      {
	// Template 331.75 25.33 1.0 0.05
	ra *= DEG_TO_RAD;
	dec *= DEG_TO_RAD;
	wrk = new coords(ra,dec);

	if (calSize)
	  *calSize = cSize;
	if (sigmaSize)
	  *sigmaSize = sSize;

	break;
      } // end rn = 4

    case 2: // We got decimal degrees and no calibrator information
      {
	// Template 331.75 25.33
	ra *= DEG_TO_RAD;
	dec *= DEG_TO_RAD;
	wrk = new coords(ra,dec);

	// If we were _expecting_ calibrator information, generate an error
	if (calSize && sigmaSize)
	  {
	    (void) fprintf(stderr,(char *) "Fatal error parsing calibration script\n");
	    exit(-1);
	  }

	break;
      } // end rn = 2

    } // end switch (rn)

  return wrk;

/**********  end parseCoordLine  ****************************************/
}


/************************************************************************
*  readScript -- local to calib.cc (7/27/97)				*
*									*
*  A.  Purpose:	The function readScript reads the input calibration     *
*       script and parses the input on the input.			*
*									*
*  B.  Background and use: readScript reads the input user calibration  *
*       script, parses the contents, and loads a sequence of global     *
*       variables used in subsequent processing.                        *
*         First up are the calibration targets.  The code reads the     *
*       names of the calibration targets to be used in this calibration *
*       run, along with the possibility of pseudonyms.  (Astronomy is   *
*       nothing if not replete with pseudonyms.)  Optionally, target    *
*       positions are read for two reasons: one, to weight the impact   *
*       of calibrators by proximity, and two, to compute the baseline   *
*       projection on the target (u-v, HA) during the output stage --   *
*       this is added as a convenience for some users, as it is         *
*       slightly orthogonal to the main mission of calib.               *
*         Next up are the calibrators.  These are (of course) presumed  *
*       to be single stars, and their names, coordinates, angular sizes *
*       and uncertainties are read into global arrays.                  *
*         Mod 11/10/98 AB -- extended read logic to read full           *
*       astrometric state (pos, PM, plx) for all objects.               *
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        filename -- the (pointer to the) character string containing   *
*         the name of the calibration script to process (char *).       *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*                  nTargets, nTargetNames, targetName, targetV;         *
*                  nCalibrators, calName, calSize, sigmaSize, calibV.   *
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration readScript  ***************************/
void readScript(char *filename)
{
/**********  local declarations readScript  *****************************/
  char line[LINE_LENGTH];
  char *separator = ((char *) ":,;| \t\n"),*tmpstr,*tName;
  int rn;
  float rah,ram,ras,decd,decm,decs;
  double ra,dec;
  FILE *fptr;

  line[0] = '\0';
  
/**********  begin readScript *******************************************/
  // Open/check on the health of the input script filename
  if (!(fptr = fopen(filename,(char *) "r")))
    {
      (void) fprintf(stderr,(char *) " Fatal Error -- Can't open calibration script file %s\n",filename);
      exit(-1);
    }

  // Read the input information on the target stars
  tmpstr = fgets(line,LINE_LENGTH,fptr);
  while ((!feof(fptr))
	 && (nTargets < MAX_TARGETS)
	 && (strncmp(line,(char *) "---",3) != 0))
    if (line[0] != '#') // If this isn't a comment
      {
	// Get the target star name(s)
	nTargetNames[nTargets] = 0;
	while ((tName = strtok(tmpstr,separator))
	       && (nTargetNames[nTargets] < MAX_PSEUDONYMS))
	  if (tName[0] != '#')
	    {
	      tmpstr = NULL;
	      targetName[nTargets][nTargetNames[nTargets]] = strdup(tName);
	      nTargetNames[nTargets]++;
	    } // end parsing target names

	// Write out an informational message on the names
	if (verbose)
	  (void) printf((char *) "# >> Calibrating star %s ",targetName[nTargets][0]);
	if (nTargetNames[nTargets] > 1)
	  {
	    if (verbose)
	      (void) printf((char *) "(and pseudonyms ");
	    for (int tmp = 1; tmp < nTargetNames[nTargets]; tmp++)
	      if (verbose)
		(void) printf((char *) "%s ",targetName[nTargets][tmp]);
	    if (verbose)
	      (void) printf((char *) ") ");
	  } // end pseudonym loop

	// Check to see if a position has been given for the target
	(void) fgets(line,LINE_LENGTH,fptr);
	if ( (line[0] != '#')
	      && (((rn = sscanf(line,(char *) "%f%f%f%f%f%f",
			       &rah,&ram,&ras,&decd,&decm,&decs)) == 6)
		  || ((rn = sscanf(line,(char *) "%lf%lf",&ra,&dec)) == 2)))
	  {
	    targetCoord[nTargets] = parseCoordLine(line,0,0);
	    targetV[nTargets]
	      = new Vector(targetCoord[nTargets]->topocentricPos(MJD_2000,KI_LATITUDE,KI_LONGITUDE,KI_ALTITUDE));

	    if (verbose)
	      (void) printf((char *) " at %s\n",targetCoord[nTargets]->print());

	    // If we were successful in interpreting this line as
	    // coordinates, read in the next one (presumably the next
	    // target name(s) or comment
	    tmpstr = fgets(line,LINE_LENGTH,fptr);

	  } // end read/decode target coordinates
	else
	  {
	    targetCoord[nTargets] = (coords *) NULL;
	    targetV[nTargets] = (Vector *) NULL;
	  }

	if (verbose)
	  (void) printf((char *) "\n");
	nTargets++;
      } // end processing target/while target loop
    else // This was a comment, echo it and get next line
      {
	if (verbose > NORMAL)
	  printf((char *) "%s",line);
	tmpstr = fgets(line,LINE_LENGTH,fptr);
      } // end comment processing

  if (feof(fptr))
    {
      (void) fprintf(stderr,(char *) " Fatal Error -- premature eof encountered in %s\n",filename);
      exit(-1);
    }
  else
    if (verbose)
      (void) printf((char *) "# >> with respect to ");



  // Get the calibrators
  // Read the next line for the number of calibrators
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,(char *) "%d",&nCalibrators) != 1)
    {
      (void) fprintf(stderr,(char *) " Fatal Error -- Error reading calibration star information from %s\n");
      exit(-1);
    }
  else
    if (verbose)
      (void) printf((char *) "%d calibrators: \n",nCalibrators);

  int i = 0;
  while (!feof(fptr) && (i < nCalibrators)) // loop over calibrator read
    {
      (void) fgets(line,LINE_LENGTH,fptr);
      // Check the line for a comment
      if (line[0] != '#')
	{
	  // Check the read for format
	  if ((rn = sscanf(line,(char *) "%s",&calName[i])) == 1)
	    {

	      // Skip over the calibrator name
	      int tmp = 0;
	      while (line[tmp] != ' '        // Check for spaces,
		     && line[tmp] != '\t'    //  tabs,
		     && line[tmp] != '\0')   //  or terminators
		tmp++;

	      calibCoord[i] = parseCoordLine(&line[tmp],&calSize[i],&sigmaSize[i]);
	      calibV[i]
		= new Vector(calibCoord[i]->topocentricPos(MJD_2000,KI_LATITUDE,KI_LONGITUDE,KI_ALTITUDE));
	      if (verbose)
		(void) printf((char *) "# >>  star %s  at %s  size %g  error %g",
			      calName[i],calibCoord[i]->print(),calSize[i],sigmaSize[i]);

	      calSize[i] *= MAS_TO_RAD;
	      sigmaSize[i] *= MAS_TO_RAD;

	      if (nTargets == 1)
		if (targetV[0])
		  if (verbose)
		    (void) printf((char *) "  %3.2g degrees away from target",
				  angle_between(*calibV[i],*targetV[0]) * RAD_TO_DEG);

	      if (verbose)
		(void) printf((char *) "\n");

	      if (useDoCalibrators) {  // treat calibrators as target stars
		// Get the calibration star name (to be treated as a target) -- assume just one name -- no pseudos
		nTargetNames[nTargets] = 0;
		tName = strtok(line,separator);
		if (tName[0] != '#')
		  {
		    tmpstr = NULL;
		    targetName[nTargets][nTargetNames[nTargets]] = strdup(tName);
		    nTargetNames[nTargets]++;
		  } // end parsing target name
		
		// Write out an informational message on the names
		if (verbose)
		  (void) printf((char *) "# >> cal star treated as target:  %s ",targetName[nTargets][0]);
		if (nTargetNames[nTargets] > 1)
		  {
		    if (verbose)
		      (void) printf((char *) "(and pseudonyms ");
		    for (int tmp = 1; tmp < nTargetNames[nTargets]; tmp++)
		      if (verbose)
			(void) printf((char *) "%s ",targetName[nTargets][tmp]);
		    if (verbose)
		      (void) printf((char *) ") ");
		  } // end pseudonym loop
		
		targetCoord[nTargets] = calibCoord[i];
		targetV[nTargets]
		  = new Vector(targetCoord[nTargets]->topocentricPos(MJD_2000,KI_LATITUDE,KI_LONGITUDE,KI_ALTITUDE));
		
		if (verbose)
		  (void) printf((char *) " at %s\n",targetCoord[nTargets]->print());
		
		if (verbose)
		  (void) printf((char *) "\n");
		nTargets++;
	      } // end useDoCalibrators

	      i++; // increment calibrator counter
	    } // end successful read of calibrator name/info

	} // end check comment field
      else // otherwise echo the comment
	if (verbose > NORMAL)
	  printf((char *) "%s",line);
    } // end i/read calibrator loop

  if (i != nCalibrators)
    {
      if (verbose)
	{
	  (void) printf((char *) "# Warning: only %d calibrators read from %s (%d expected)\n",
			i,filename,nCalibrators);
	  (void) fprintf(stderr,(char *) "# Warning: only %d calibrators read from %s (%d expected)\n",
			 i,filename,nCalibrators);
	} // checking verbosity
      nCalibrators = i;
    }

  // Close the script file
  (void) fclose(fptr);

/**********  end readScript  ********************************************/
}


/************************************************************************
*  printRunDirectives -- local to calib.cc (A.B. 4/11/98)		*
*									*
*  A.  Purpose:	The function printRunDirectives prints a report of the  *
*       calibration directives for this run.				*
*									*
*  B.  Background and use: As calib acquires options, it becomes        *
*       increasingly necessary to document the settings of the program  *
*       for a given dataset.  This function dumps a short report on the *
*       settings used in this run to stdout (where the rest of the      *
*       output is also going). 						*
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration printRunDirectives  *******************/
void printRunDirectives()
{

/**********  begin printRunDirectives  **********************************/
  // Reports on some operational (used to be preprocessor) directives
  (void) fprintf(stdout,(char *) "\n# nullCalib version %s\n",(char *) VERSION);
  (void) fprintf(stdout,(char *) "# nullCalib calibration code settings for this run: \n");

  (void) printf((char *) "# > Using calibration scans within +/- %g hr of the target scan\n",
		calibTimeWindow * 24.0);

  if (usingTempErrorProp)
    {
      (void) printf((char *) "# > Weighting calibration scans WRT temporal proximity,");
      (void) printf((char *) " variance doubles in %g hr\n",timeErrorPropCoeff / 24.0);
    }
  else
    (void) printf((char *) "# > No weighting of calibration scans WRT temporal proximity\n");

  if (usingSkyErrorProp)
    {
      (void) printf((char *) "# > Weighting calibration scans WRT sky proximity,");
      (void) printf((char *) " variance doubles in %g deg\n",RAD_TO_DEG / angleErrorPropCoeff);
    }
  else
    (void) printf((char *) "# > No weighting of calibration scans WRT sky proximity\n");

  if (usingMinimumScanUncertainty)
    {
      (void) printf((char *) "# > Applying minimum calibrated scan uncertainty of %g\n",minScanUncertainty);
    }
  else
    (void) printf((char *) "# > No minimum calibrated scan uncertainty\n");

  if (usingMinimumCalUncertainty)
    {
      (void) printf((char *) "# > Applying minimum calibrated cal uncertainty of %g\n",minCalUncertainty);
    }
  else
    (void) printf((char *) "# > No minimum calibrated cal uncertainty\n");

  (void) printf((char *) "# > Flagging calibration warnings when multiple calibrators disagree by more than %g sigma\n",
		sqrt(chiWarning));
  (void) printf((char *) "# > Rejecting the scan when multiple calibrators disagree by more than %g sigma\n",
		sqrt(chiReject));

  if (delayCheck)
    (void) printf((char *) "# > Comparing model/measured delays with %3g m tolerance\n",
		  delayCheckTol);
  else
    (void) printf((char *) "# > Not comparing model/measured delays\n");

  // added with new option by hjb 5/2005 ... just following along
  if(baselineFile[0] != '\0') 
    (void) printf((char *) "# > Using baseline file %s\n", baselineFile);

  (void) printf((char *) "\n\n");


/**********  end printRunDirectives  ************************************/
}

/************************************************************************
*  blineInsert -- local to nullCalib.cc (jsh 2008-08)                   *
*                                                                       *
*  A.  Purpose: The function blineInsert adds baselines from file fptr  *
*      to the bline map. Used by readBaseline().                        *
*                                                                       *
*  B.  Background and use:                                              *
*                                                                       *
*  C.  Called by:                                                       *
*                                                                       *
*  D.  Input/output files:                                              *
*                                                                       *
*  E.  Parameter description:                                           *
*       Input -                                                         *
*       Output -                                                        *
*                                                                       *
*  F.  Detail function description:                                     *
*             1.  Local Functions:                                      *
*             2.  Externals:                                            *
*             3.  Globals and Nonlocals used:                           *
*             4.  Global and Nonlocal variables modified:               *
*             5.  Function flow is as follows:                          *
*             6.  Functions called by                                   *
*                                                                       *
*  G.  Function value: This function returns no useable value.          *
*                                                                       *
*************************************************************************/
void blineInsert(FILE *fptr, map<string,baseline*> &blines, char work[LINE_LENGTH]) {
    while(1) {
        baseline *bl = new baseline(fptr);
        if(! bl->error())
           blines.insert(map<string,baseline*>::value_type(bl->baselineName(), bl));
        else
           break;
    }

    if(blines.empty()) {
        (void) printf("# Error: no valid baselines read from %s\n", work);
    } else {
        map<string,baseline*>::iterator pos;
        for(pos = blines.begin(); pos != blines.end(); ++pos)
          (void) printf("# Found baseline %s in %s \n", pos->first.c_str(), work);
    }

/**********  end blineInsert  ************************************/
}


/************************************************************************
*  readBaseline -- local to wbCalib.cc (A.B. 3/2000)			*
*									*
*  A.  Purpose:	The function readBaseline instantiates a baseline       *
*       object to go along with the sume file being read/to be read.    *
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
//baseline *readBaseline(char *sumFilename, char *baselinef)
// updated function interface 2/06, hjb
// updated 2008-08, jsh
map<string,baseline*> readBaseline(char *sumFilename, char *baselinef)
{
/**********  local declarations readBaseline  ***************************/
  char work[LINE_LENGTH], *ptr;

  FILE *fptr = NULL;
  baseline *bLine = NULL;
  map<string,baseline*> blines;
  
/**********  begin readBaseline  ****************************************/
  // One.  Extract trailing suffix and append default baseline suffix
  (void) strncpy(work,sumFilename,LINE_LENGTH);
  if ((ptr = strstr(work,(const char *) ".nullsum")) != NULL)
    (void) strcpy(ptr, (const char *) ".nullbaseline\0");
  
  // One.Five hjb 5/2005
  // working in the baseline override
  // the idea is not to disturb any of the original code for this function
  // so we'll take the user supplied file name, check that, and if it's readable,
  // slip it into the "work" variable used by the rest of the fucntion. if it's 
  // not readable, we're going to fall back the the original search strategy for
  // finding a baseline
  if(baselinef[0] != '\0') {
    // make sure the file is available, accessible, around for reading
    // if not, emit warnings and do nothing, thereby falling back to the 
    // standard search for baseline file (nearly transparent to user)
    if((fptr = fopen(baselinef,"r")) == NULL) {
      (void) printf("# Error reading user supplied baseline file %s\n", baselinef);
      (void) printf("# Attempting to open baseline file\n");
    } else {
      (void) strncpy(work, baselinef, sizeof(work));
      work[sizeof(work)] = '\0';
      blineInsert(fptr, blines, work);
      if(!blines.empty())
         return blines;
    }
  }

  // if user supplied baseline file is missing or not valid, try to read from
  // the xxx.nullbaseline file
  if((fptr == NULL) || (blines.empty())) {
      if((fptr = fopen(work, "r")) == NULL) {
         (void) printf("# Error trying to open baseline file %s\n", work);
      }
      else {
         blineInsert(fptr, blines, work);
         if(!blines.empty())
             return blines;
      }
  }

  // if xxx.nullbaseline file is missing or not valid, try to read from
  // the xxx.nullbline file
  if((fptr == NULL) || (blines.empty())) {
      (void) printf("# Attempting to open bline file\n");
      if( (ptr = strstr(work, (const char *) ".nullbaseline")) != NULL)
        (void) strcpy(ptr, (const char *) ".nullbline\0");
      if((fptr = fopen(work, "r")) == NULL) {
        (void) printf("# Error trying to open bline file %s\n", work);
        (void) printf("# Error Could not read any baseline specification\n");
        return blines;
      }
      blineInsert(fptr, blines, work);
  }
  return blines;

  // note - if we can open the user supplied file for reading, but we can't create
  // a good baseline out of it, we just fall back to .nullbaseline, not .nullbline. this
  // isn't designed or intentional, but it seems okay, given that if our user supplied
  // file doesn't work, it's hard to imagine the bline could be okay

  /**********  end readBaseline  ******************************************/
}


/************************************************************************
*  readSumFile -- local to calib.cc (A.B. 7/27/97)			*
*									*
*  A.  Purpose:	The function readSumFile reads the input summary file   *
*       and extracts the relevant information for storage in the        *
*       linked-list star data structure.				*
*									*
*  B.  Background and use:						*
*         readSumFile is designed to be called multiple times with      *
*       multiple sumfile arguments to construct the database to be used *
*       in this calibration run.                                        *
*         Input to the routine consists of the name of the sumfile to   *
*       be read and parsed, and the (pointer to the) database of        *
*       null information that is being constituted.  On output    	*
*       information from scans successfully read-in from the input      *
*       sumfile have been entered into the database.  All this          *
*       database manipulation is accomplished via a set of class        *
*       libraries described in StarList.h (and StarList.cc).            *
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files: filename argument				*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        filename -- the (pointer to the) sumfile filename to be        *
*         processed (char *).                                           *
*        starList -- the (pointer to the) starlist database object      *
*         being constructed for later use (StarList *).                 *
*	Output -							*
         starList -- contents modified.                                 *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the number of valid data   *
*       lines that were successfully read from the input file (int).    *
*									*
*************************************************************************/
int readSumFile(char *filename,char *baselinef, StarList *starList)
{
/**********  local declarations readSumFile  ****************************/

  char line[LINE_LENGTH],starName[132];
  char portsUsed[12],version[12],blDesig[132];
  int nRead,month,day,year,hour,minute,second,duration,dataLines=0;
  double jd,ut,delay,longDelay=0.0,nullLeak,nullSigma,wbWavelength;
  
  // The following items are actually double values in the NullSum file.
  // They are not used in nullCalib so they are being read in as strings
  // so that any NULL values will be trapped and won't cause sscanf to abort
  // due to non-essential data being NULL.
  char wlPhot[12],snrXC[12],snr1[12],snr2[12],rnn[12],jnn[12],percentGatedCycles[12];
  // These are actually ints in NullSum that are being read in as strings
  char nCycles[12],mode[12],calFlag[12];

  FILE *fptr = NULL;
  line[0] = '\0';
  starName[0] = '\0';
  version[0] = '\0';
  portsUsed[0] = '\0';
  blDesig[0] = '\0';

  wlPhot[0] = '\0';
  snrXC[0] = '\0';
  snr1[0] = '\0';
  snr2[0] = '\0';
  rnn[0] = '\0';
  jnn[0] = '\0';
  percentGatedCycles[0] = '\0';
  nCycles[0] = '\0';
  mode[0] = '\0';
  calFlag[0] = '\0';

  // Try to open this summary file
  if (!(fptr = fopen(filename,(char *) "r"))) {
    if (verbose)
      (void) printf((char *) "# >> Error opening summary file %s\n",filename);
    return 0;
  }

  // Read/retrieve the baseline corresponding to this sum file (or use
  // default baseline if missing or error)

  // hjb 2/06 ... use the new constructor in libbaseline which allows for comments
  // and multi-record baseline files
  // in the future this needs update to allow for time associate of baseline to data record
  // baseline *thisBaseline = readBaseline(filename, baselinef);
  map<string,baseline*> blines = readBaseline(filename, baselinef);

  if(blines.empty()) {
    (void) printf("# Warning ... continuing with built-in baseline values\n");
  } else {
    map<string,baseline*>::iterator pos;
    for(pos = blines.begin(); pos != blines.end(); ++pos) {
      (void) printf("# Using baseline %s ", pos->second->baselineName());
      (void) printf("(%f, %f, %f, %f) \n", pos->second->localBaseline()[0],
 		    pos->second->localBaseline()[1], pos->second->localBaseline()[2],
		    pos->second->bias());
    }
  }

/* 2008-01  JSH
  nullCalib reads NULLSUM line format:
      date   time(hrs)  star  delay(m) longDelay  flux  nullLeakage  nullSigma  snrXC snr1   snr2  rnn  jnn  wbWavelength  #cycles percentGatedCycles duration  mode  portsUsed  version  calFlag  baseline
NULLSUM 08/30/2007/12:53:43  12.878611  HD209747  10.0 -40.0 1141.4  0.0192  0.0174  33914  15415 15526  22.0 7.0 10.0 430 0.0  0   500 IN+OUT  N1.2  0 KI_K1K2
*/

// All "non-required" items (the ones that are never used anywhere) are read in
// as strings, regardless of their data type in the NullSum file. This is done in
// order to catch any "NULL" values that would cause the sscanf to abort if it
// were expecting a double but got a "NULL" instead. All required values should
// be read as their normal datatype and if a "NULL" is encountered then sscanf
// should abort and that line should be skipped over for processing since required
// data is missing.

  while (!feof(fptr)) {
    (void) fgets(line,LINE_LENGTH,fptr);
    // Read data
    if (!(feof(fptr)) && (line[0] != '#')
	&& (((nRead = sscanf(line,
		       (char *) "NULLSUM %d/%d/%d/%d:%d:%d %lf %s %lf%lf%s%lf%lf%s%s%s%s%s%lf%s%s%d%s%s%s%s%s",
			  &month,&day,&year,&hour,&minute,&second,&ut,&starName,&delay,&longDelay,&wlPhot,&nullLeak,&nullSigma,&snrXC,&snr1,&snr2,&rnn,&jnn,&wbWavelength,&nCycles,&percentGatedCycles,&duration,&mode,&portsUsed,&version,&calFlag,&blDesig)) == 27))) {

	  dataLines++;  // If all tests passed, we've got a data line

	  // Turn dates and UTs into JDs
	  if (year < 96)
	    year += 100;
	  if (year < 1900)
	    year += 1900;
	  jd = jdFromUTCal(ut,year,month,day);
	  
	  // KI Kvis-format line -- set baseline fron designation
	  // if no input baseline model (so using default)
	  baseline *thisBaseline;
	  if(blDesig[0] != '\0') {
	    map<string,baseline*>::iterator pos = blines.find(blDesig);
	    if( pos == blines.end())
	      thisBaseline = defBaseline;
	    else
	      thisBaseline = pos->second;
	  } else {
	    if(blines.size() == 1) 
	      thisBaseline = blines.begin()->second;
	    else
	      thisBaseline = defBaseline;
	  }

	  // this checks addresses, not actual member equality.
	  if( thisBaseline == defBaseline ) {
	    // case of NULL blDesig should fall through to the end
	    if(strncmp(blDesig, (char *) "KI_K1K2",7) == 0)
	      thisBaseline = &KI_K1K2;
	    else
	      thisBaseline = defBaseline;
	  } // end set baseline from designation

	  // Create a data node with this data contents
	    NullData *nullNode = new NullData(starName,0,month,day,year,
					jd,ut,delay,longDelay,
					nullLeak, nullSigma,
					wbWavelength,
					duration,
					thisBaseline);

	    // Find or create a star node corresponding to this star
	    if (!starList->findStar(starName))
	      starList->insert(starName);
	    
	    // Insert the data from this line into the linked-list
	    // structure for this star.  Note it is being
	    // insertion-sorted into the list based on jd. 
	    starList->getStarData()->insert(nullNode,jd,INSERT_SORT);
	    //(void) printf("# Created node for %s, with baseline %s at bias %f\n", nullNode->get_starName(), nullNode->get_baseline()->baselineName(), 
	    //	  nullNode->get_baseline()->bias());

	} // end read/line format/WL check
  } // end while loop

  (void) fclose(fptr); // Close the file

  return dataLines;

}


/************************************************************************
*  nullPred -- local to calib.cc (A.B. 7/27/97)				*
*									*
*   Modified for nullCalib						*
*							 		*
*  A.  Purpose:	The function nullPred computes the expected null value (and *
*       variance) for a given calibrator and projected normalized       *
*       baseline length (aka inverse fringe spacing).			*
*									*
*  B.  Background and use:						*
*         nullPred is pretty simple: it computes the expected null for an   *
*       ideal interferometer observing a simple star -- presumably a    *
*       calibrator.  The ratio of what we got to what we expected is    *
*       an estimate of the system null.		                        *
*									*
*  C.  Called by: sysNullSC						*
*									*
*  D.  Input/output files: none						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        baseline -- the sky-projected baseline length (m) (double).    *
*        lambda -- the wavelength we're working at (m) (double).        *
*        aSize -- the estimated angular size of object (rad) (double).  *
*        aSigma -- the error in aSize (rad) (double).                   *
*	Output -							*
*        varNullPred -- the variance in the output null prediction based  *
*         on the input angular size error (dimensionless) (double *).   *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the expected null value    *
*       (dimensionless) (double).                       		*
*									*
*************************************************************************/
double nullPred(double baseline,double lambda,
	      double aSize,double aSigma,double *varNullPred)
{
/**********  begin nullPred  **********************************************/
  // The fringe spacing on the sky (units of radians)
  double fringeSP = baseline / lambda;
  double arg = PI * fringeSP * aSize;
  double dArgdaSize = PI * fringeSP;
  double v,dvdArg,n;
  double sigmaV,sigmaN;


  // Trap the (annoying) point source case
  if (arg > 0)
    {
      v = 2.0 * j1(arg) / arg;
      dvdArg = 2.0 * (0.5 * arg * (j0(arg) - jn(2,arg)) - j1(arg)) / arg / arg; 
      n = (1.0-v)/(1.0+v);
    }
  else
    {
      v = 0.0;
      dvdArg = 0.0;
    }

  // Use the chain rule (otherwise I never get this right).
  double dv2dv = 2.0 * v;

  // RLA - add nulling variance //
  // this first line is the V sigma
  sigmaV = dvdArg * dArgdaSize * aSigma;
  // leakage sigma
  sigmaN = sigmaV * pow(1.0+n,2)/(2.0);
  // now make it a variance
  *varNullPred = pow(sigmaN,2);  

  return (n);
/**********  end nullPred  ************************************************/
}


/************************************************************************
*  sysNullSC -- local to calib.cc			 		*
*									*
*  A.  Purpose: The function sysNullSC estimates the system null  	*
*       (and the error in that estimate) with respect to a particular   *
*       calibrator at a particular jd.					*
*									*
*  B.  Background and use:						*
*         sysNullSC mechanizes the estimate of the system null at 	*
*       a particular epoch based on a particular calibrator.  In the    *
*       calib scheme, each calibrator is considered independently (this *
*       routine), and then these results are (weighted) averaged to     *
*       derive the system null estimate for the target (done      	*
*       in sysNullEnsemble).                                            *
*         sysNullSC does its work by extracting all calibrator scans    *
*       within a given time window (CALIB_WINDOW) around the input      *
*       time.  From each calibrator scan a system null estimate   	*
*       (and error estiamte) is derived.  These system null       	*
*       estimates are then (weighted) averaged to determine the best    *
*       estimate for the system null at the input time.  Note:    	*
*       these system null estimates _can_ be weighted with        	*
*       respect to temporal proximity to the desired output time.       *
*									*
*  C.  Called by: sysNullEnsemble					*
*									*
*  D.  Input/output files: stdout & stderr				*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the Julian Data for the system null estimate.      	*
*        calibData -- the (pointer to the) data list object for this    *
*         calibrator (dataList<WBdata> *).                              *
*        calSize -- the estimated angular size for this calibrator      *
*         (rad) (double).                                               *
*        sigmaSize -- the estimated error in calSize (rad) (double).    *
*        star -- the (pointer to the) unit 3-vector sky position for    *
*         this calibrator (dimensionless) (Vector *).                   *
*	Output -							*
*        varSysNull -- the estimated error variance in the system       *
*         null estimate (dimensionless) (double *).     		*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the estimate of system     *
*       null at the time of interest derived from a     		*
*       particular calibrator (dimensionless) (double).		        *
*									*
*************************************************************************/
/**********  function declaration sysNullSC  *****************************/
double sysNullSC(double jd,
		dataList<NullData> *calibData,char *calName,
		double calSize,double sigmaSize,
		Vector *star,double *varsysNullSC)
{
/**********  local declarations sysNullSC  *******************************/

  int calHits = 0;
  double nullAvg = 0.0;
  double weight,weightSum = 0.0;
  double expNull,varExpNull;
  double jdScan,utScan,delayScan,wlScan;
  double nullLeak, nullSigma;

/**********  begin sysNullSC  ********************************************/
  // Loop over the designated calibration targets
  (void) calibData->gotoHead();

  int numNodes = calibData->numberNodes();
  int nNode = 0;
  double nullSigmaVar;

  if (calibData->numberNodes())
    {
      // Skip forward to the interesting time
      while((nNode < numNodes)
	     && ((jd - calibData->getCurrent()->get_jd()) > calibTimeWindow)) {
 	  calibData->iterate();
          nNode++;
      }

      // Extract relevant calibration information from this
      // calibration source

        while((nNode <= numNodes)
	     && (fabs(jd - calibData->getCurrent()->get_jd()) < calibTimeWindow))
	{
 	      baseline *Baseline = calibData->getCurrent()->get_baseline();
  	      nullLeak = calibData->getCurrent()->get_nullLeak();
	      nullSigma = calibData->getCurrent()->get_nullSigma();
	      // RLA - convert sigma to variance
	      nullSigmaVar = pow(nullSigma,2);

	      wlScan = calibData->getCurrent()->get_wbWavelength();
              delayScan = calibData->getCurrent()->get_delay() - calibData->getCurrent()->get_longDelay();
              jdScan = calibData->getCurrent()->get_jd();
              utScan = calibData->getCurrent()->get_ut();

	      calHits++;

	      if (delayCheck)
		{
		  // New 11/30/97.  Check the delay value for this calibration object
		  // double delayModel = Baseline->delay(jdScan,*star);
		  double tmp = delayScan - Baseline->delay(jdScan,*star);
		  if (fabs(tmp) > delayCheckTol)
		    if (verbose)
		      {
			(void) printf((char *) "# Warning: at %10.4f, %4g m measured/expected delay mismatch on calibrator %s\n",jdScan,tmp,calName);
			(void) fprintf(stderr,(char *) "Warning: at %10.4f, %4g m measured/expected delay mismatch on calibrator %s\n",jdScan,tmp,calName);
		      } // end verbose/delay value check
		} // end delayCheck

	      // Compute the expected null value at this epoch, sky
	      // position, baseline projection, wavelength...
	      // New 12/20/97 -- use sumfile-averge WL

	      expNull = nullPred(Baseline->skyProj(jdScan,*star),
			     wlScan * 1.0e-6,
			     calSize,sigmaSize,&varExpNull);

	      if (usingTempErrorProp)
		{
		  // If defined, propagate the error in the system null
		  //  to the time of the target observation.  Assuming linear
		  //  (quadratic) increase in the uncertainty (variance).
	          nullSigmaVar *= 1.0 + pow(timeErrorPropCoeff * fabs(jd - jdScan),1.0);
		}

               // Check min scan uncertainty - RLA 3/31/08 //
	       if((usingMinimumCalUncertainty) && (nullSigma < minCalUncertainty)) {
		    nullSigma = minCalUncertainty;
                    nullSigmaVar = pow(nullSigma,2);
                    if(verbose)
                      (void) printf((char *) "# Minimum scan uncertainty applied to calibrator at %11.5f\n",jdScan);
                  }

	      // Weighting.  Weight by variance of data and prediction
              weight = 1.0 / (nullSigmaVar + varExpNull);
	      weightSum += weight;
              nullAvg += weight * (nullLeak - expNull);	
              nNode++;
              calibData->iterate();
	} // calibration scan

      if (calHits) // If we've found good calibration hits
	{
	  *varsysNullSC = 1.0/ weightSum;	
	  return (nullAvg / weightSum);
	}
      else // Otherwise, return a stupid value to flag calib 
	return -99.0;


    } // end calibData node count check
  else // no data for this calibrator
    return -99.0;

/**********  end sysNullSC  **********************************************/
}


/************************************************************************
*  sysNullEnsemble -- local to calib.cc					*
*									*
*  A.  Purpose:	sysNullEnsemble computes the system null estimate  	*
*       from the full ensemble of calibrators selected for this         *
*       calibration run.						*
*									*
*  B.  Background and use:						*
*         sysNullEnsemble mechanizes the computation of an average      *
*       system null estimate from the composite of single         	*
*       calibrator system null estimates (computed in sysNullSC).  	*
*       sysNullEnsemble also performs system null estimate	        *
*       inner-consistency checks among multiple calibrators.            *
*         In the case of multiple calibrators, the cross-consistency    *
*       of derived system null estimates is checked, and        	*
*       mismatches at two levels are flagged for the user as a          *
*       diagnostic that there might be something wrong with either the  *
*       input calibrator information or the calibration object itself.  *
*									*
*  C.  Called by: calib							*
*									*
*  D.  Input/output files: stdout & stderr				*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the Julian Data for the system null estimate 	        *
*         (double).                                                     *
*        targV -- the (pointer to the) unit 3-vector sky position of    *
*         the target (dimensionless) (Vector *).  Note: this is okay    *
*         if NULL, it just means there can be no angular proximity      *
*         weighting in the system null estimate computation.      	*
*        starList -- the (pointer to the) null database object    	*
*         (StarList *).                                                 *
*	Output -							*
*        sysNull -- the composite estimated system null at the	        *
*         desired time (dimensionless) (double *).                      *
*        varSysNull -- the estimated error variance in sysNull          *
*         (dimensionless) (double *).                                   *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration sysNullEnsemble  ***********************/
int sysNullEnsemble(double jd,Vector *targV,
		   StarList *starList,
		   double *sysNull,double *varSysNull)
{
/**********  local declarations sysNullEnsemble  *************************/
  int i;
  double sNull[MAX_CALIBRATORS],varSNull[MAX_CALIBRATORS];

/**********  begin sysNullEnsemble  **************************************/
  // First loop over the calibrators.
  // Determine the estimated system null at the time of
  // this scan based on each calibrator...
	
  for (i = 0; i < nCalibrators; i++)
    if (starList->findStar(calName[i]))
      {
	dataList<NullData> *calibData = starList->getStarData();
	// Evaluate the calibrator apparent position
	// Bug alert: No way to ensure that this baseline is valid for this time (quant effects small)
	baseline *Baseline = calibData->getCurrent()->get_baseline();
	*(calibV[i]) = calibCoord[i]->topocentricPos(jd,Baseline->latitude(),
						     Baseline->longitude(),
						     Baseline->altitude());
	sNull[i] = sysNullSC(jd,
			  calibData,calName[i],calSize[i],
			  sigmaSize[i],calibV[i],&varSNull[i]);

      } // end first loop over calibrators
    else
      sNull[i] = -99.0;

  // Second loop over the calibrators.
  // Compute the weighted average of the system null
  // calibrations, and check the interconsistency of the
  // calibrations at this time.
  int calHits = 0;
  double weight,weightSum = 0.0,nullSum = 0.0;
  for (i = 0; i < nCalibrators; i++)
    if (sNull[i] > -90.0) // if there was a non-trivial calibration
      {
	calHits++;

	// Add angular proximity weighting
	if (usingSkyErrorProp)
	  if (targV)
	    varSNull[i] *= 1.0 + angleErrorPropCoeff * angle_between(*calibV[i],*targV);

	weightSum += weight = 1.0 / varSNull[i];
	nullSum += weight * sNull[i];
	
	// Check for calibration inconsistencies
	for (int j = i + 1; j < nCalibrators; j++)
	  if (sNull[j] > -90.0)
	    {
	      double chi = pow(sNull[i] - sNull[j],2.0) / (varSNull[i] + varSNull[j]);
	      // Inconsistencies are flagged at two thresholds.  The
	      // first is just warning...
	      if (chi > chiWarning)
		{
		  if (chi < chiReject)
		    {
		      if (verbose > NORMAL)
			{
			  (void) printf((char *) "# Warning: at %10.4f sysNull estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
					jd,calName[i],calName[j],sqrt(chi));
			  (void) fprintf(stderr,(char *) "Warning: at %10.4f system null estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
					 jd,calName[i],calName[j],sqrt(chi));
			} // verbose check
		    } // end warning
		  else // The second is an outright rejection of the scan...
		    {
		      (void) printf((char *) "# Reject: at %10.4f system null estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
				    jd,calName[i],calName[j],sqrt(chi));
		      (void) fprintf(stderr,(char *) "Reject: at %10.4f system null estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
				     jd,calName[i],calName[j],sqrt(chi));
		      // return a null-calibration
		      return 0;
		    }
		} // end chi test
	    } // end cross-calibration check loop
      } // end weight-average loop

  // If we've gotten a good calibration, compute the system null.
  if (calHits)
    {
      *sysNull = nullSum / weightSum;
      *varSysNull = 1.0 / weightSum;
    }

  // Return the number of calibrators used in this system calibration
  return calHits;

/**********  end sysNullEnsemble  ****************************************/
}



/************************************************************************
*  calib -- local to calib.cc (A.B. 7/27/97)				*
*									*
*  modified for nullCalib						*
*									*
*  A.  Purpose:	The function calib drives the calibration of data       *
*       output by Mark's program.					*
*									*
*  B.  Background and use:						*
*         Other than database construction, calib is where it all       *
*       happens.  calib loops over the targets to be calibrated, the    *
*       list of pseudonyms for each target, and the scans recorded      *
*       under each pseudonym.                                           *
*         In the case of multiple calibrators, the cross-consistency    *
*       of derived system null estimates is checked, and  	        *
*       mismatches at two levels are flagged for the user as a          *
*       diagnostic that there might be something wrong with either the  *
*       input calibrator information or the calibration object itself.  *
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files: stdout & stderr				*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        starList -- the (pointer to the) null database object    	*
*         (StarList *).                                                 *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration calib  ********************************/
void calib(StarList *starList)
{
/**********  local declarations calib  **********************************/
  int i,month,day,year,numScans = 0;
  double jdAvg,utAvg,delayAvg,wlAvg;
  double nullSigmaVar, nullLeak;


  // Output a status line indicating the column contents in the
  // calibfile format
  if(verbose) {
    (void) printf((char *) "# Calibration data format:\n");
    (void) printf((char *) "# name\t#NullPts\tMJD\tm/d/y/h:m:s\t\tUT\tDelay(m)\tWL(um)\tCalNull\tErr\tRawNull\tErr\tSysNull\tErr\t#Cal\tU(m)\tV(m)\tHA(hr)\tDuration\tBLName\n");
  } // end verbose/template line

  // Now do the calibration

  // Loop over the calibration targets
  for(int nTarg = 0; nTarg < nTargets; nTarg++) {
    numScans = 0;
    (void) printf((char *) "\n# Target %s:\n",targetName[nTarg][0]);

    // Loop over the pseudonyms for each target
    for(int pseudonym = 0; pseudonym < nTargetNames[nTarg]; pseudonym++)
      // If we've got hits on the target under this name...
      if(starList->findStar(targetName[nTarg][pseudonym])) {
	int nDataName = 0;
	dataList<NullData> *targetData = starList->getStarData();
	(void) targetData->gotoHead();
	if(targetData->numberNodes())
	  do {
	    int st, nCal, duration = targetData->getCurrent()->get_duration();
	    double sNullm,varSNullm;
	    baseline *Baseline = targetData->getCurrent()->get_baseline();

            // # Pts hardcoded to 1
            st = 1;

	    // Compute the averaged null data for the calibration
	    // target at this time.  If we've got a non-trivial
	    // number of target scans...
            nullLeak = targetData->getCurrent()->get_nullLeak();
            // RLA -- convert nullSigma to variance
            nullSigmaVar = pow(targetData->getCurrent()->get_nullSigma(),2);

            wlAvg = targetData->getCurrent()->get_wbWavelength();
            delayAvg = targetData->getCurrent()->get_delay() - targetData->getCurrent()->get_longDelay();
            jdAvg = targetData->getCurrent()->get_jd();
            utAvg = targetData->getCurrent()->get_ut();
            month = targetData->getCurrent()->get_month();
            day = targetData->getCurrent()->get_day();
            year = targetData->getCurrent()->get_year();

	      // Compute the ensemble system null estimate
	      // at this time.  If we've got a good null
	      // calibration...
	      // store the current pointer, so we can reset after calibration
	      // this is necessary for --doCalibrators option
	      NullData* resetNode = targetData->getCurrent();

              // if nCal == 0, then chi was rejected
	      if (nCal = sysNullEnsemble(jdAvg,targetV[nTarg], starList, &sNullm,&varSNullm)) {
		// ...compute and output the calibrated scan.
                // nullCalib edit
                // RLA - combine errors in quadrature => add variances
		double varTNull = nullSigmaVar + varSNullm;	
 
		// New 11/6/97  Check the delay value for the target object
		if ((targetCoord[nTarg]) && (delayCheck)) {
		  Vector targ // The instantaneous direction of target
		    = targetCoord[nTarg]->topocentricPos(jdAvg,	Baseline->latitude(), Baseline->longitude(), Baseline->altitude());
		  double delay = Baseline->delay(jdAvg,targ);
		  double delayDiff = delayAvg - delay;

		  if(fabs(delayDiff) > delayCheckTol) {
		    if(verbose) {
		      (void) printf((char *) "# Warning: at %10.4f, %g m measured/expected delay mismatch on target %s\n",
				    jdAvg,delayDiff,targetName[nTarg][pseudonym]);
		      (void) fprintf(stderr,(char *) "Warning: at %10.4f, %g m measured/expected delay mismatch on target %s\n",
				     jdAvg,delayDiff,targetName[nTarg][pseudonym]);
		    } // end delay value/verbose check
		  }
		}

		  // New 4/27/98  Check/apply minimum possible scan uncertainty
		  double scanSigma = sqrt(varTNull);
		  if((usingMinimumScanUncertainty) && (scanSigma < minScanUncertainty)) {
		    scanSigma = minScanUncertainty;
		    if(verbose)
		      (void) printf((char *) "# Minimum scan uncertainty applied at %11.5f\n",jdAvg);
		  }
		
		  char ut_str[LINE_LENGTH];
		  formatted_time(utAvg,ut_str);  // convert utAvg to hr:min:sec with defined format in ut_str

		  // (Finally) Printout the data line
                  // nullCalib edit
		  (void) printf((char *) "%s\t%d\t%11.5f\t%d/%d/%d/%s",
				targetName[nTarg][pseudonym], st,jdAvg,month,day,year,ut_str);

		  (void) printf((char *) "\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%d",
				utAvg,delayAvg, wlAvg,nullLeak - sNullm,scanSigma,
				nullLeak,sqrt(nullSigmaVar), sNullm,sqrt(varSNullm),nCal);

		  // And if we've got a position on the
		  // target, output u-v coordinates and the
		  // like to help our consumers...
		  if(targetCoord[nTarg]) {
		    Vector targ // The instantaneous direction of target
		      = targetCoord[nTarg]->topocentricPos(jdAvg, Baseline->latitude(), Baseline->longitude(), Baseline->altitude());
		    Vector SV = Baseline->skyProjV(jdAvg,targ);
		    (void) printf((char *) "\t%10.6f\t%10.6f\t%g",SV[0],SV[1], Baseline->hourAngle(jdAvg,targ));
		  }

		  // Print out the duration and the baseline name
		  (void) printf((char *) "\t%d\t%s\n", duration,Baseline->baselineName());
		  numScans++;
		} else { // Otherwise, calibration failure...notify the user.
		    (void) printf("# No suitable calibration found for %10.4f\n",jdAvg);
		} 
                nDataName++;
	        targetData->resetCurrent(resetNode);
                targetData->iterate();
        } while(nDataName < targetData->numberNodes());

	// Target list is finished, printout a summary line for this target name
	(void) printf((char *) "# %d calibrated scans output under name %s\n\n\n",
                     numScans,targetName[nTarg][pseudonym]);

      } // end psedonym list/data check
	// Pseudonym list is finished, printout a summary line on this target
    (void) printf((char *) "# %d total calibrated scans output on target %s\n\n\n", numScans,targetName[nTarg][0]);
  }

}


/**********  main routine calib  ****************************************/
main (int argc,char **argv)
{
  // initialize baselineFile global to blank
  baselineFile[0] = '\0';

  // Process any command-line options
  int argIndex = processCmdLineOpts(argc,argv);

  // Trap trivial/bogus invocations
  if (argc < 3)
    (void) printHelpMessage();

  // Read the calibration script on the stars we're going to care
  // about
  (void) readScript(argv[argIndex]);

  // Report to the user on the run directives
  if (verbose)
    (void) printRunDirectives();

  // Initialize the data structure and read in the data from the
  // summary file(s) listed in the command line
  StarList *starList = new StarList();
  for (int i = argIndex + 1; i < argc; i++)
    {
      int nData = readSumFile(argv[i],baselineFile, starList);
      if (verbose)
	(void) printf((char *) "# %d data lines read from %s\n",nData,argv[i]);
    }

  // Now we do the calibration.  Output is to stdout (to be redirected
  // as desired.  Isn't unix great...?)
  (void) calib(starList);

/**********  end program calib  ****************************************/
}
