/*************************************************************************
* NASA Exoplanet Science Institute (NExScI) -- http://nexsci.caltech.edu *
*	Copyright (c) 1997-2009 California Institute of Technology       *
*                     U.S. Government Sponsorship                        *
* Under contract with the National Aeronautics and Space Administration  *
**************************************************************************/

// wbCalib -- calibration of wide-band visibility amplitudes for PTI
// and KI -- Andy Boden 1997 through 2003

// now wrapped by a Perl script.  This C++ core of the wbCalib app has
// been renamed to wbcEngine.cc -- David A. Imel 2004

#include <config.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include "getopt.h"

#include "angles.h"
#include "Vector.h"

#include "timeUtil.h"

#include "PTI.h"
#include "KI.h"
#include "baseline.h"

#include "coords.h"

#include "wbStarList.h"

#include <vector>
#include <map>
#include <string>
#include <iostream>
using namespace std;

#define LINE_LENGTH 255

#define MAX_TARGETS 2000
#define MAX_PSEUDONYMS 10
#define MAX_CALIBRATORS 2000

#define LAMBDA 2.18e-6

#define MJD_2000 51544.5

// Delay Check tolerance -- approximately 3 arcminutes
#define DELAY_CHECK_TOLERANCE 0.10


// Global declarations

// Create the default baseline objects.  Use the form of the
// constructor with delay bias -- we'll be checking delay values

// PTI
baseline PTI_NS(PTI_LONGITUDE,PTI_LATITUDE,
		PTI_Baseline_NS,PTI_NS_DELAY_BIAS,"PTI_NS");
baseline PTI_NW(PTI_LONGITUDE,PTI_LATITUDE,
		PTI_Baseline_NW,PTI_NW_DELAY_BIAS,"PTI_NW");
baseline PTI_SW(PTI_LONGITUDE, PTI_LATITUDE,
		PTI_Baseline_SW, PTI_SW_DELAY_BIAS, "PTI_SW");
// KI
baseline KI_K1K2(KI_LONGITUDE,KI_LATITUDE,
		 KI_Baseline_K1K2,KI_K1K2_DELAY_BIAS,"KI_K1K2");
baseline KI_S1S2(KI_LONGITUDE,KI_LATITUDE,
		 KI_Baseline_S1S2,KI_S1S2_DELAY_BIAS,"KI_S1S2");
baseline KI_K1K2S(KI_LONGITUDE,KI_LATITUDE,
                 KI_Baseline_K1K2,KI_K1K2_DELAY_BIAS,"KI_K1K2S");
baseline KI_Camera_FT_PRIM(KI_LONGITUDE,KI_LATITUDE,
                 KI_Baseline_K1K2,KI_K1K2_DELAY_BIAS,"KI_Camera_FT_PRIM");
baseline KI_Camera_FT_SEC(KI_LONGITUDE,KI_LATITUDE,
                 KI_Baseline_K1K2,KI_K1K2_DELAY_BIAS,"KI_Camera_FT_SEC");
baseline KI_Camera_FT_LBAND(KI_LONGITUDE,KI_LATITUDE,
                 KI_Baseline_K1K2,KI_K1K2_DELAY_BIAS,"KI_Camera_FT_LBAND");

//baseline *defBaseline = &PTI_NS;
baseline *defBaseline = &KI_K1K2;
string baselineFile;
string fluxBiasFile;

// Star Stuff

// Calibration Target Names, Counts, and Positions (arrays -- sorry)
char *targetName[MAX_TARGETS][MAX_PSEUDONYMS];
int nTargets,nTargetNames[MAX_TARGETS];
Vector *targetV[MAX_TARGETS];  // Going to phase this out...
coords *targetCoord[MAX_TARGETS];


// Calibrator Names, Count, and Positions  (arrays -- sorry)
int nCalibrators;
char calName[MAX_CALIBRATORS][132];
Vector *calibV[MAX_CALIBRATORS]; // Hoping to phase this out...
coords *calibCoord[MAX_TARGETS];
double calSize[MAX_CALIBRATORS],sigmaSize[MAX_CALIBRATORS];



// Program Operational Options.  These are used to control the code's
// processing behavior.  Global variables are the best mechanism I've
// come up with to handle this situation...

// Used to indicate whether the program processes spectrometer data
// (default), or white-light data.
int usingSpec = 1;

// Used to indicate whether the program processes the incoherent data
// (default) or coherent data.
int usingIncoherentV2 = 1;

// Flags to indicate whether we try to correct for phase jitter and
// intensity mismatches.
int usingJitterCorrection = 1;
double jitterCoeff = 0.04;  // Default phase jitter correction coeff
int usingRatioCorrection = 0;

// Calibration time window for picking up calibration scans.
// units are in days, so 1/24 is +/- 1 hour time window
// #define CALIB_WINDOW (1.0 / 24.0)
double calibTimeWindow = (1.0 / 24.0);

// Propagation coefficient for system visibility error.
// A doubling of the sysVisSC error estimate in an hour.  See sysVisSC.
// Units are in inverse days, so 24 is 1 hr.
// #define PROP_COEFF (1.0 * 24.0)
int usingTempErrorProp = 1;
double timeErrorPropCoeff = (1.0 * 24.0);

// Propagating visibility errors across the sky.  See sysVisEnsemble.
// Units are in inverse radians.
// #define ANGLE_COEF (1.0 / (15.0 * DEG_TO_RAD))
int usingSkyErrorProp = 1;
double angleErrorPropCoeff = (1.0 / (15.0 * DEG_TO_RAD));

// Minimum possible scan uncertainty
int usingMinimumScanUncertainty = 0;
double minScanUncertainty = 0.01;

// Using root-N scaling on error bars
int useRootN = 0;

// DoCalibrators flag:  treats the calibrators as target stars
int useDoCalibrators = 0;

// Thresholds for reporting/rejecting calibration errors (dimensionless).
// Threshold for reporting calibration errors -- 2.0 sigma disagreement
// #define CHI_WARNING 1.0
double chiWarning = 2.0 * 2.0;
// Threshold for rejecting the calibration -- 3.0 sigma disagreement
// #define CHI_REJECT 9.0
double chiReject = 3.0 * 3.0;

// scanThreshold -- time used to define sample points that synthesized
// into a scan (units are in days because the cut is made in JD space)
// AFB & RLA 2/2002
double scanThreshold = 150.0 / 3600.0 / 24.0;

// WL Scaling (used in systematic error estimation and what-if scenarios)
int usingWLScale = 0;
double wlScale = 1.0;

// Band selection (K,H,all)  Added 11/99  A.B.
char usingWL = 'A';  // Default to all wavebands
double K_wl = 2.2, K_width = 0.2;
double H_wl = 1.6, H_width = 0.1;


// Frame rate selection (50, 100, 0 -- all)  Added 11/99 A.B.
double usingFR = 0;  // Default to all frame rates


// Verbosity variable to control reporting Added 7/01 A.B.
#define MUTE 0
#define NORMAL 1
#define EXCESSIVE 2
int verbose = NORMAL; // Default to certain errors

// Variable for delay checking
int delayCheck = 1;
double delayCheckTol = DELAY_CHECK_TOLERANCE;

// values for FluxBias correction
int usingFluxBiasCorrection = 0;

typedef struct {
  float WLFluxBias;
  float SpecFluxBias;
  float NZero;
  float Normflux;
  int usingDefaults; // set to 1 initially, becomes 0 if cameraName read from fluxBiasFile
} fluxBiasStruct;

// KI_K1K2 will use ftPrim defaults
// fluxBiasStruct k1k2 = {0.063,0.16,20.0,1000.0,1};
fluxBiasStruct k1k2s = {0.0,0.0,20.0,1000.0,1};
fluxBiasStruct ftPrim = {0.063,0.16,20.0,1000.0,1};
fluxBiasStruct ftSec = {0.0,0.0,20.0,1000.0,1};
fluxBiasStruct ftLband = {0.0,0.0,20.0,1000.0,1};

map<string,fluxBiasStruct> fluxBiasMap;


/************************************************************************
*  printHelpMessage -- local to calib.cc (A.B. 4/13/98)			*
*									*
*  A.  Purpose:	The function printHelpMessage prints a help message     *
*       for the calib program to stderr and then exits the program.	*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description: None					*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration printHelpMessage  *********************/
void printHelpMessage()
{

/**********  begin printHelpMessage  ************************************/
  (void) fprintf(stdout,(char *) "Usage: wbCalib calScript sumfile1 [sumfile2...] [> cal.out]\n\n");
  (void) fprintf(stdout,(char *) "Calibrate PTI/KI wideband V2 data\n\n");
  (void) fprintf(stdout,(char *) "Options:\n\n");
  (void) fprintf(stdout,(char *) "  -help: print (this) help message and exit (overrides all other options)\n");
  (void) fprintf(stdout,(char *) "  -version: print version information and exit\n");
  (void) fprintf(stdout,(char *) "  -Kband|Hband|allband: calibrate specific band data \n");
  (void) fprintf(stdout,(char *) "  -frameRate 50|100|0: calibrate specific frame rate data (Hz -- 0 all) \n");
  (void) fprintf(stdout,(char *) "  -defaultFile argument: read calibration default values from a file\n");
  (void) fprintf(stdout,(char *) "  -[no]jitterCorrection[=argument]: set [non]use of jitter correction \n\t (optional correction coefficient argument)\n");
  (void) fprintf(stdout,(char *) "  -[no]ratioCorrection: set [non]use of ratio correction\n");
  (void) fprintf(stdout,(char *) "  -WL: calibrate wide-band (\"white light\") visibilities\n");
  (void) fprintf(stdout,(char *) "  -Spectrometer: calibrate spectrometer composite wide-band visibilities\n");
  (void) fprintf(stdout,(char *) "  -[in]coherentV2: set use of [in]coherent V2\n");
  (void) fprintf(stdout,(char *) "  -calTimeWindow argument: set calibration scan acceptance time window (hr)\n");
  (void) fprintf(stdout,(char *) "  -[no]timeWeighting: set use/nonuse of time proximity weighting of calibration scans\n");
  (void) fprintf(stdout,(char *) "  -timeErrorVal argument: if timeWeighting, set calibration scan time weighting value (hr)\n");
  (void) fprintf(stdout,(char *) "  -[no]angleProxWeighting: set use/nonuse of sky angle proximity weighting of calibration scans\n");
  (void) fprintf(stdout,(char *) "  -angleErrorVal argument: if angleProxWeighting, set calibration scan angle weighting value (deg)\n");
  (void) fprintf(stdout,(char *) "  -[no]minScanUncertainty argument: sets [non]use of minimum possible scan uncertainty of value argument (V2 units)\n");
  (void) fprintf(stdout,(char *) "  -[no]rootNWeighting : sets use/nonuse of root-Nsamples weighting in computed scan errors\n");
  (void) fprintf(stdout,(char *) "  -scanThreshold argument: time window used to define maximum scan duration (sec)\n");
  (void) fprintf(stdout,(char *) "  -chiWarning argument: set calibrator inconsistency warning chi threshold \n");
  (void) fprintf(stdout,(char *) "  -chiReject argument: set calibrator inconsistency scan rejection chi threshold \n");
  (void) fprintf(stdout,(char *) "  -[no]delayCheck[=argument]: set use/nonuse of delay value checking \n\t (optional delay tolerance argument) \n");
  (void) fprintf(stdout,(char *) "  -wlScale argument: scale operating wavelength by factor \n");
  (void) fprintf(stdout,(char *) "  -verbose|normal|quiet|shutUp: control volume of information, warning, and error messages\n");
  (void) fprintf(stdout,(char *) "  -doCalibrators:  doCalibrators development debugging option\n");
  (void) fprintf(stdout,(char *) "  -fluxBiasCorrection [=file]: corrects flux dependent V2 Biases using the biases provided by [file] or default corrections if no file is specified\n");
  (void) fprintf(stdout,(char *) "  -baseline <file>: Use the baseline supplied by <file>. Normal behavior is to look for a\n");
  (void) fprintf(stdout,(char *) "                    yyyyddd.baseline file for each yyyyddd.sum file given on the command\n");
  (void) fprintf(stdout,(char *) "                    line. If that fails then look for yyyyddd.bline. If neither is found then\n");
  (void) fprintf(stdout,(char *) "                    wbCalib uses a built in default baseline. The option allows the user to\n");
  (void) fprintf(stdout,(char *) "                    override that search by providing the filename directly. Note that when\n");
  (void) fprintf(stdout,(char *) "                    multiple SUM files are given in combination with this option all of them\n");
  (void) fprintf(stdout,(char *) "                    use the supplied <file>.\n");

  (void) fprintf(stdout,(char *) "  note:  command-line options may be used in any order, interspersed with non-options\n");
  (void) fprintf(stdout,(char *) "   and abbreviated to uniqueness (e.g. -h, -noang, -incoh -K).\n");
  (void) fprintf(stdout,(char *) "  Find more wbCalib info at http://msc.caltech.edu/software\n");

  (void) exit(-1);



/**********  end printHelpMessage  **************************************/
}

/**********  Little utility for a formatted UT time stamp ******************/

void formatted_time(double ut, char* time_string)
{
  // Some time machinations
  int hr      = (int) ut;
  double frac = ut - ((double) hr);
  int min     = (int) (frac * 60.0);
  frac       -= ((double) min / 60.0);
  int sec     = (int) (frac * 3600.0);
  
  // Mucho tedious logic to get the time format correct

  sprintf(time_string,"%02d:%02d:%02d",hr,min,sec);
}


/******** end utilities ********/

/************************************************************************
*  readCalDefs -- local to calib.cc (A.B. 4/13/98)			*
*									*
*  A.  Purpose:	The function readCalDefs reads the calibration code     *
*       defaults file to establish the default settings for the run.    *
*									*
*  B.  Background and use: The function readCalDefs gives a .ini-style  *
*       interface to the initialization of calib.  There are a set of   *
*       predefined default values built directly into calib, then the   *
*       user can create a default file that can override those values.	*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        filename -- the (pointer to the) calibration code defaults     *
*         filename to be processed (char *).                            *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration readCalDefs  **************************/
void readCalDefs(char *filename)
{
/**********  local declarations readCalDefs  ****************************/
  char line[LINE_LENGTH],*tmpstr;
  FILE *fptr;

/**********  begin readCalDefs  *****************************************/
  if (!(fptr = fopen(filename,(char *) "r")))
    {
      (void) fprintf(stderr,(char *) " Error -- can't open calibration-defaults file %s,",filename);
      (void) fprintf(stderr,(char *) " proceeding with previous settings\n");
      (void) printf((char *) "# Error -- can't open calibration-defaults file %s,",filename);
      (void) printf((char *) " proceeding with previous settings\n");
      return;
    }

  // Read the defaults file
  // WL/spectrometer switch
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,(char *) "%d",&usingSpec) != 1)
    {
      (void) fprintf(stderr,(char *) " Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }

  // coherent/incoherent switch
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,(char *) "%d",&usingIncoherentV2) != 1)
    {
      (void) fprintf(stderr,(char *) " Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }

  // jitter switch
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,(char *) "%d",&usingJitterCorrection) != 1)
    {
      (void) fprintf(stderr,(char *) " Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }

  // ratio correction switch
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,(char *) "%d",&usingRatioCorrection) != 1)
    {
      (void) fprintf(stderr,(char *) " Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }

  // calibration time window
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,(char *) "%lf",&calibTimeWindow) != 1)
    {
      (void) fprintf(stderr,(char *) " Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }
  calibTimeWindow /= 24.0; // Convert hours to days

  // time weighting switch
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,(char *) "%d",&usingTempErrorProp) != 1)
    {
      (void) fprintf(stderr,(char *) " Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }

  // calibrator time weighting
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,(char *) "%lf",&timeErrorPropCoeff) != 1)
    {
      (void) fprintf(stderr,(char *) " Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }
  timeErrorPropCoeff *= 24.0; // Converting hours to inverse days
  

  // sky proximity switch
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,(char *) "%d",&usingSkyErrorProp) != 1)
    {
      (void) fprintf(stderr,(char *) " Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }

  // calibrator sky proximity weighting
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,(char *) "%lf",&angleErrorPropCoeff) != 1)
    {
      (void) fprintf(stderr,(char *) " Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }
  angleErrorPropCoeff *= DEG_TO_RAD;  // Converting to degrees inverse radians
  angleErrorPropCoeff = 1.0 / angleErrorPropCoeff;

  // chiWarning
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,(char *) "%lf",&chiWarning) != 1)
    {
      (void) fprintf(stderr,(char *) " Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }
  chiWarning *= chiWarning; // Convert to chi-squared

  // chiReject
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,(char *) "%lf",&chiReject) != 1)
    {
      (void) fprintf(stderr,(char *) " Fatal error reading calibration defaults file %s\n",filename);
      exit(-1);
    }
  chiReject *= chiReject; // Convert to chi-squared

  // Print-out successful reading report, close the defaults file, and return.
  (void) printf((char *) "# calibration defaults read from %s\n",filename);
  (void) fclose(fptr);

/**********  end readCalDefs  *******************************************/
}



/************************************************************************
*  processCmdLineOpts -- local to calib.cc (A.B. 4/12/98)		*
*									*
*  A.  Purpose:	The function processCmdLineOpts processes the command   *
*       line invocation options used with calib for this run.		*
*									*
*  B.  Background and use: processCmdLineOpts uses getopt_long_only to  *
*       process the command line options the user has added to the      *
*       invocation.  getopt_long_only is used because it allows the     *
*       arbitrarily intersperse command-line options along with the     *
*       filename command-line arguments: it permutes the command line   *
*       as it processes so when finished all the non-options (i.e. the  *
*       filename arguments) are at the end (for simple processing by    *
*       the main routine).				                *
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        argc, argv -- the standard command-line argument variables.    *
*	Output -							*
*        argv's contents are permuted so as to move all non-option      *
*         arguments to the end.                                         *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the index of the first     *
*       non-option on the command-line, which is presumably the name    *
*       of the calibration script file (int).              		*
*									*
*************************************************************************/
/**********  function declaration processCmdLineOpts  *******************/
int processCmdLineOpts(int argc, char **argv)
{
/**********  local declarations processCmdLineOpts  *********************/
  // Valid command-line options
  int retVal,optIndex,argsProcessed = 0;
  static struct option longOptions[] = 
  {
    {((char *) "help"),no_argument,0,0},                     // Option 0
    {((char *) "version"),no_argument,0,0},                  // Option 1
    {((char *) "jitterCorrection"),optional_argument,0,0},   // Option 2
    {((char *) "nojitterCorrection"),no_argument,0,0},       // Option 3
    {((char *) "ratioCorrection"),no_argument,0,0},          // Option 4
    {((char *) "noratioCorrection"),no_argument,0,0},        // Option 5
    {((char *) "coherentV2"),no_argument,0,0},               // Option 6
    {((char *) "incoherentV2"),no_argument,0,0},             // Option 7
    {((char *) "timeWeighting"),no_argument,0,0},            // Option 8
    {((char *) "notimeWeighting"),no_argument,0,0},          // Option 9
    {((char *) "angleProxWeighting"),no_argument,0,0},       // Option 10
    {((char *) "noangleProxWeighting"),no_argument,0,0},     // Option 11
    {((char *) "calTimeWindow"),required_argument,0,0},      // Option 12
    {((char *) "timeErrorVal"),required_argument,0,0},       // Option 13
    {((char *) "angleErrorVal"),required_argument,0,0},      // Option 14
    {((char *) "minScanUncertainty"),required_argument,0,0}, // Option 15
    {((char *) "nominScanUncertainty"),no_argument,0,0},     // Option 16
    {((char *) "chiWarning"),required_argument,0,0},         // Option 17
    {((char *) "chiReject"),required_argument,0,0},          // Option 18
    {((char *) "defaultFile"),required_argument,0,0},        // Option 19
    {((char *) "WL"),no_argument,0,0},                       // Option 20
    {((char *) "Spectrometer"),no_argument,0,0},             // Option 21
    {((char *) "wlScale"),required_argument,0,0},            // Option 22
    {((char *) "Kband"),no_argument,0,0},                    // Option 23
    {((char *) "Hband"),no_argument,0,0},                    // Option 24
    {((char *) "allband"),no_argument,0,0},                  // Option 25
    {((char *) "frameRate"),required_argument,0,0},          // Option 26
    {((char *) "quiet"),no_argument,0,0},                    // Option 27
    {((char *) "normal"),no_argument,0,0},                   // Option 28
    {((char *) "verbose"),no_argument,0,0},                  // Option 29
    {((char *) "shutUp"),no_argument,0,0},                   // Option 30
    {((char *) "scanThreshold"),required_argument,0,0},      // Option 31
    {((char *) "delayCheck"),optional_argument,0,0},         // Option 32
    {((char *) "nodelayCheck"),no_argument,0,0},             // Option 33
    {((char *) "rootNWeighting"),no_argument,0,0},           // Option 34
    {((char *) "norootNWeighting"),no_argument,0,0},         // Option 35
    {((char *) "doCalibrators"),no_argument,0,0},            // Option 36 -- new, to be renamed
    {((char *) "baseline"),required_argument,0,0},           // Option 37 - baseline file override

// JSH IMPORTANT NOTE
// I'm not familiar enough with Perl to understand why, but when specifying
// an optional string, you MUST use an "=" on the command line, otherwise it
// core dumps. The "=" is not necessary for required arguments, but for some
// reason, it is for optional ones. So for fluxBiasCorrection, if specifying a
// filename, the format MUST be: -fluxBiasCorrection=fluxBiasFilename
// I need to look into this further
    {((char *) "fluxBiasCorrection"),optional_argument, 0, 0}, // option 38 - correct Flux dependent Biases
    {((char *) "nofluxBiasCorrection"), no_argument, 0, 0}, //option 39 

    {0,0,0,0}
  };

  extern char *optarg;
  extern int optind,opterr,optopt;

  /**********  begin processCmdLineOpts  **********************************/

  // Print welcome/version
  if (verbose)
    (void) fprintf(stdout,(char *) "## This is wbCalib v%s\n\n",(char *) VERSION);
  
  do {
    if((retVal = getopt_long_only(argc,argv,(char *) "",longOptions,&optIndex)) != EOF) {
      if (verbose)
	(void) printf((char *) "# option %s\n",longOptions[optIndex].name);
      argsProcessed++;
      switch (optIndex) {

      case 0: {
	(void) printHelpMessage();
	break;
      } // end printing help message -- option 0

      case 1: {
	(void) fprintf(stdout,(char *) "wbCalib v%s\n\n",(char *) VERSION);
	(void) exit(0);
	break;
      } // end printing version message -- option 1

      case 2: {
	usingJitterCorrection = 1;
	if (optarg) {
	  jitterCoeff = atof(optarg);
	} // end setting jitter correction coefficient
	break;
      } // end setting usingJitterCorrection -- option 2

      case 3: {
	usingJitterCorrection = 0;
	break;
      } // end setting NOT usingJitterCorrection -- option 3

      case 4: {
	usingRatioCorrection = 1;
	break;
      } // end setting usingRatioCorrection -- option 4

      case 5: {
	usingRatioCorrection = 0;
	break;
      } // end setting NOT usingRatioCorrection -- option 5

      case 6: {
	usingIncoherentV2 = 0;
	break;
      } // end setting NOT usingIncoherentV2 -- option 6

      case 7: {
	usingIncoherentV2 = 1;
	break;
      } // end setting usingIncoherentV2 -- option 7

      case 8: {
	usingTempErrorProp = 1;
	break;
      } // end setting usingTempErrorProp -- option 8

      case 9: {
	usingTempErrorProp = 0;
	break;
      } // end setting NOT usingTempErrorProp -- option 9

      case 10: {
	usingSkyErrorProp = 1;
	break;
      } // end setting usingSkyErrorProp -- option 10

      case 11: {
	usingSkyErrorProp = 0;
	break;
      } // end setting NOT usingSkyErrorProp -- option 11

      case 12: {
	calibTimeWindow = atof(optarg);
	calibTimeWindow /= 24.0;
	break;
      } // end setting calibTimeWindow -- option 12

      case 13: {
	timeErrorPropCoeff = atof(optarg);
	timeErrorPropCoeff *= 24.0;
	break;
      } // end setting timeErrorPropCoeff -- option 13

      case 14: {
	angleErrorPropCoeff = atof(optarg);
	angleErrorPropCoeff *= DEG_TO_RAD;
	angleErrorPropCoeff = 1.0 / angleErrorPropCoeff;
	break;
      } // end setting angleErrorPropCoeff -- option 14

      case 15: {
	usingMinimumScanUncertainty = 1;
	minScanUncertainty = atof(optarg);
	break;
      } // end setting usingMinimumScanUncertainty/minScanUncertainty -- option 15

      case 16: {
	usingMinimumScanUncertainty = 0;
	break;
      } // end setting usingMinimumScanUncertainty -- option 16

      case 17: {
	chiWarning = atof(optarg);
	chiWarning *= chiWarning;
	break;
      } // end setting chiWarning -- option 17

      case 18: {
	chiReject = atof(optarg);
	chiReject *= chiReject;
	break;
      } // end setting chiReject -- option 18

      case 19: {
	(void) readCalDefs(optarg);
	break;
      } // end printing help message -- option 19

      case 20: {
	usingSpec = 0;
	break;
      } // end using WL visibilities -- option 20

      case 21: {
	usingSpec = 1;
	break;
      } // end using spectrometer visibilities -- option 21

      case 22: {
	usingWLScale = 1;
	wlScale = atof(optarg);
	break;
      } // end using spectrometer visibilities -- option 22

      case 23: {
	usingWL = 'K';
	break;
      } // end selecting K-band data -- option 23

      case 24: {
	usingWL = 'H';
	break;
      } // end selecting H-band data -- option 24

      case 25: {
	usingWL = 'A';
	break;
      } // end selecting all bands -- option 25

      case 26: {
	usingFR = atof(optarg);
	break;
      } // end setting angleErrorPropCoeff -- option 26

      case 27: {
	// be quiet!!!
	verbose = MUTE;
	break; 
      } // end shutting up -- option 27

      case 28: {
	// normal (normal setup, serious warnings and errors)
	verbose = NORMAL;
	break; 
      } // setting normal reporting -- option 28

      case 29: {
	// verbose (everything)
	verbose = EXCESSIVE;
	break; 
      } // end setting excessive reporting -- option 29

      case 30: {
	// SHUT UP!!!
	verbose = MUTE;
	break; 
      } // end setting shutUp -- option 30

      case 31: {
	scanThreshold = atof(optarg);
	scanThreshold /= 3600.0 * 24.0; // Convert to units of days
	break;
      } // end setting scanThreshold -- option 31

      case 32: {
	delayCheck = 1;
	if (optarg) {
	  delayCheckTol = atof(optarg);
	} // end setting delay check tolerance
	break;
      } // end setting delayCheck -- option 32

      case 33: {
	delayCheck = 0;
	break;
      } // end setting delayCheck -- option 33

      case 34: {
	useRootN = 1;
	break;
      } // end setting useRootN -- option 34
	
      case 35: {
	useRootN = 0;
	break;
      } // end setting useRootN -- option 35

      case 36: {
	useDoCalibrators = 1;
	break;
      } // end setting useDoCalibrators -- option 36

      case 37: {
        baselineFile.assign(optarg);
	break;
      }

      case 38: {
	usingFluxBiasCorrection = 1;
 	if(optarg)
	   fluxBiasFile.assign(optarg);
	break;
      } // end setting usingFluxBiasCorrection

      case 39: {
	usingFluxBiasCorrection = 0;
	break;
      }
	
      } // end big butt-ugly option switch
    } // end if return value not EOF
  } while (retVal != EOF);

  if (argsProcessed && verbose)
    (void) printf((char *) "# %d command line arguments processed\n",argsProcessed);

  return optind;
  
/**********  end processCmdLineOpts  ************************************/
}


/************************************************************************
*  parseCoordLine -- local to calib.cc (A.B. 11/98)			*
*									*
*  A.  Purpose:	The function parse coordinate line parses the           *
*       input coordinate line for the target/calibrator and extracts    *
*       the pertinent information					*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns a pointer to the           *
*       coordinate object created by as a result of this line		*
*       (coords *).                                                     *
*									*
*************************************************************************/
coords *parseCoordLine(char *line,double *calSize,double *sigmaSize)
{
/**********  local declarations parseCoordLine  *************************/
  char *separator = ((char *) "#"),*tmpstr;
  char decStr[256], *minusSign = ((char *) "-");
  double ra,dec;
  float rah,ram,ras,decd,decm,decs,pmRA,pmDec,plx;
  float cSize,sSize;
  coords *wrk;


/**********  begin parseCoordLine  **************************************/
  (void) strtok(line,separator); // Use strtok to stomp on any comments

  int rn = sscanf(line,(char *) "%lf%f%f%s%f%f%f%f%f%f%f",
		  &ra,&ram,&ras,decStr,&decm,&decs,&pmRA,&pmDec,&plx,
		  &cSize,&sSize);

  dec = atof(decStr);
  int decSign = 1; // handle negative declinations (stupid)!!!
  if (strncmp(decStr,minusSign,1) == 0) // If the first character is a minus...
    decSign = -1;

  switch (rn)
    {
    case 11: // We got the whole smack
      {
	// template 15 21 06.880 +24 57 28.872 -0.029 -0.019 0.00587 1.05  0.1
	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec,pmRA,pmDec,plx);

	if (calSize)
	  *calSize = cSize;
	if (sigmaSize)
	  *sigmaSize = sSize;

	break;
      }

    case 9: // We got full coordinates but no calibrator info
      {
	// template 15 34 41.268 +26 42 52.895  0.135 -0.089 0.04365

	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec,pmRA,pmDec,plx);

	// If we were _expecting_ calibrator information, generate an error
	if (calSize && sigmaSize)
	  {
	    (void) fprintf(stderr,(char *) "Fatal error parsing calibration script\n");
	    exit(-1);
	  }

	break;
      } // end rn = 9

    case 8: // We got positions and calibrator info
      {
	// Template 16 01 02.662 +33 18 12.634 0.70 0.1
	// Note this case is degenerate with coords, pm, no plx, but I
	// have to assume something...
	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec);

	// The calibrator info got read into the pm variables!!!
	if (calSize)
	  *calSize = pmRA;
	if (sigmaSize)
	  *sigmaSize = pmDec;

	break;
      } // end rn = 8

    case 6: // We got positions but no calibrator info
      {
	// Template 16 01 02.662 +33 18 12.634 
	ra += (ram / 60.0) + (ras / 3600.0);  // Convert RA/Dec into radians
	ra *= 15.0 * DEG_TO_RAD;
	dec += decSign * ((decm / 60.0) + (decs / 3600.0));
	dec *= DEG_TO_RAD;

	wrk = new coords(ra,dec);

	// If we were _expecting_ calibrator information, generate an error
	if (calSize && sigmaSize)
	  {
	    (void) fprintf(stderr,(char *) "Fatal error parsing calibration script\n");
	    exit(-1);
	  }

	break;
      } // end rn = 6

    case 4: // We got decimal degrees and calibrator information
      {
	// Template 331.75 25.33 1.0 0.05
	ra *= DEG_TO_RAD;
	dec *= DEG_TO_RAD;
	wrk = new coords(ra,dec);

	if (calSize)
	  *calSize = cSize;
	if (sigmaSize)
	  *sigmaSize = sSize;

	break;
      } // end rn = 4

    case 2: // We got decimal degrees and no calibrator information
      {
	// Template 331.75 25.33
	ra *= DEG_TO_RAD;
	dec *= DEG_TO_RAD;
	wrk = new coords(ra,dec);

	// If we were _expecting_ calibrator information, generate an error
	if (calSize && sigmaSize)
	  {
	    (void) fprintf(stderr,(char *) "Fatal error parsing calibration script\n");
	    exit(-1);
	  }

	break;
      } // end rn = 2

    } // end switch (rn)

  return wrk;

/**********  end parseCoordLine  ****************************************/
}


/************************************************************************
*  readScript -- local to calib.cc (7/27/97)				*
*									*
*  A.  Purpose:	The function readScript reads the input calibration     *
*       script and parses the input on the input.			*
*									*
*  B.  Background and use: readScript reads the input user calibration  *
*       script, parses the contents, and loads a sequence of global     *
*       variables used in subsequent processing.                        *
*         First up are the calibration targets.  The code reads the     *
*       names of the calibration targets to be used in this calibration *
*       run, along with the possibility of pseudonyms.  (Astronomy is   *
*       nothing if not replete with pseudonyms.)  Optionally, target    *
*       positions are read for two reasons: one, to weight the impact   *
*       of calibrators by proximity, and two, to compute the baseline   *
*       projection on the target (u-v, HA) during the output stage --   *
*       this is added as a convenience for some users, as it is         *
*       slightly orthogonal to the main mission of calib.               *
*         Next up are the calibrators.  These are (of course) presumed  *
*       to be single stars, and their names, coordinates, angular sizes *
*       and uncertainties are read into global arrays.                  *
*         Mod 11/10/98 AB -- extended read logic to read full           *
*       astrometric state (pos, PM, plx) for all objects.               *
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        filename -- the (pointer to the) character string containing   *
*         the name of the calibration script to process (char *).       *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*                  nTargets, nTargetNames, targetName, targetV;         *
*                  nCalibrators, calName, calSize, sigmaSize, calibV.   *
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration readScript  ***************************/
void readScript(char *filename)
{
/**********  local declarations readScript  *****************************/
  char line[LINE_LENGTH];
  char *separator = ((char *) ":,;| \t\n"),*tmpstr,*tName;
  int rn;
  float rah,ram,ras,decd,decm,decs;
  double ra,dec;
  FILE *fptr;

  line[0] = '\0';
  
/**********  begin readScript *******************************************/
  // Open/check on the health of the input script filename
  if (!(fptr = fopen(filename,(char *) "r")))
    {
      (void) fprintf(stderr,(char *) " Fatal Error -- Can't open calibration script file %s\n",filename);
      exit(-1);
    }

  // Read the input information on the target stars
  tmpstr = fgets(line,LINE_LENGTH,fptr);
  while ((!feof(fptr))
	 && (nTargets < MAX_TARGETS)
	 && (strncmp(line,(char *) "---",3) != 0))
    if (line[0] != '#') // If this isn't a comment
      {
	// Get the target star name(s)
	nTargetNames[nTargets] = 0;
	while ((tName = strtok(tmpstr,separator))
	       && (nTargetNames[nTargets] < MAX_PSEUDONYMS))
	  if (tName[0] != '#')
	    {
	      tmpstr = NULL;
	      targetName[nTargets][nTargetNames[nTargets]] = strdup(tName);
	      nTargetNames[nTargets]++;
	    } // end parsing target names

	// Write out an informational message on the names
	if (verbose)
	  (void) printf((char *) "# >> Calibrating star %s ",targetName[nTargets][0]);
	if (nTargetNames[nTargets] > 1)
	  {
	    if (verbose)
	      (void) printf((char *) "(and pseudonyms ");
	    for (int tmp = 1; tmp < nTargetNames[nTargets]; tmp++)
	      if (verbose)
		(void) printf((char *) "%s ",targetName[nTargets][tmp]);
	    if (verbose)
	      (void) printf((char *) ") ");
	  } // end pseudonym loop

	// Check to see if a position has been given for the target
	(void) fgets(line,LINE_LENGTH,fptr);
	if ( (line[0] != '#')
	      && (((rn = sscanf(line,(char *) "%f%f%f%f%f%f",
			       &rah,&ram,&ras,&decd,&decm,&decs)) == 6)
		  || ((rn = sscanf(line,(char *) "%lf%lf",&ra,&dec)) == 2)))
	  {
	    targetCoord[nTargets] = parseCoordLine(line,0,0);
	    targetV[nTargets]
	      = new Vector(targetCoord[nTargets]->topocentricPos(MJD_2000,PTI_LATITUDE,PTI_LONGITUDE,PTI_ALTITUDE));

	    if (verbose)
	      (void) printf((char *) " at %s\n",targetCoord[nTargets]->print());

	    // If we were successful in interpreting this line as
	    // coordinates, read in the next one (presumably the next
	    // target name(s) or comment
	    tmpstr = fgets(line,LINE_LENGTH,fptr);

	  } // end read/decode target coordinates
	else
	  {
	    targetCoord[nTargets] = (coords *) NULL;
	    targetV[nTargets] = (Vector *) NULL;
	  }

	if (verbose)
	  (void) printf((char *) "\n");
	nTargets++;
      } // end processing target/while target loop
    else // This was a comment, echo it and get next line
      {
	if (verbose > NORMAL)
	  printf((char *) "%s",line);
	tmpstr = fgets(line,LINE_LENGTH,fptr);
      } // end comment processing

  if (feof(fptr))
    {
      (void) fprintf(stderr,(char *) " Fatal Error -- premature eof encountered in %s\n",filename);
      exit(-1);
    }
  else
    if (verbose)
      (void) printf((char *) "# >> with respect to ");



  // Get the calibrators
  // Read the next line for the number of calibrators
  (void) fgets(line,LINE_LENGTH,fptr);
  if (sscanf(line,(char *) "%d",&nCalibrators) != 1)
    {
      (void) fprintf(stderr,(char *) " Fatal Error -- Error reading calibration star information from %s\n");
      exit(-1);
    }
  else
    if (verbose)
      (void) printf((char *) "%d calibrators: \n",nCalibrators);

  int i = 0;
  while (!feof(fptr) && (i < nCalibrators)) // loop over calibrator read
    {
      (void) fgets(line,LINE_LENGTH,fptr);
      // Check the line for a comment
      if (line[0] != '#')
	{
	  // Check the read for format
	  if ((rn = sscanf(line,(char *) "%s",&calName[i])) == 1)
	    {

	      // Skip over the calibrator name
	      int tmp = 0;
	      while (line[tmp] != ' '        // Check for spaces,
		     && line[tmp] != '\t'    //  tabs,
		     && line[tmp] != '\0')   //  or terminators
		tmp++;

	      calibCoord[i] = parseCoordLine(&line[tmp],&calSize[i],&sigmaSize[i]);
	      calibV[i]
		= new Vector(calibCoord[i]->topocentricPos(MJD_2000,PTI_LATITUDE,PTI_LONGITUDE,PTI_ALTITUDE));
	      if (verbose)
		(void) printf((char *) "# >>  star %s  at %s  size %g  error %g",
			      calName[i],calibCoord[i]->print(),calSize[i],sigmaSize[i]);

	      calSize[i] *= MAS_TO_RAD;
	      sigmaSize[i] *= MAS_TO_RAD;

	      if (nTargets == 1)
		if (targetV[0])
		  if (verbose)
		    (void) printf((char *) "  %3.2g degrees away from target",
				  angle_between(*calibV[i],*targetV[0]) * RAD_TO_DEG);

	      if (verbose)
		(void) printf((char *) "\n");

	      if (useDoCalibrators) {  // treat calibrators as target stars
		// Get the calibration star name (to be treated as a target) -- assume just one name -- no pseudos
		nTargetNames[nTargets] = 0;
		tName = strtok(line,separator);
		if (tName[0] != '#')
		  {
		    tmpstr = NULL;
		    targetName[nTargets][nTargetNames[nTargets]] = strdup(tName);
		    nTargetNames[nTargets]++;
		  } // end parsing target name
		
		// Write out an informational message on the names
		if (verbose)
		  (void) printf((char *) "# >> cal star treated as target:  %s ",targetName[nTargets][0]);
		if (nTargetNames[nTargets] > 1)
		  {
		    if (verbose)
		      (void) printf((char *) "(and pseudonyms ");
		    for (int tmp = 1; tmp < nTargetNames[nTargets]; tmp++)
		      if (verbose)
			(void) printf((char *) "%s ",targetName[nTargets][tmp]);
		    if (verbose)
		      (void) printf((char *) ") ");
		  } // end pseudonym loop
		
		targetCoord[nTargets] = calibCoord[i];
		targetV[nTargets]
		  = new Vector(targetCoord[nTargets]->topocentricPos(MJD_2000,PTI_LATITUDE,PTI_LONGITUDE,PTI_ALTITUDE));
		
		if (verbose)
		  (void) printf((char *) " at %s\n",targetCoord[nTargets]->print());
		
		if (verbose)
		  (void) printf((char *) "\n");
		nTargets++;
	      } // end useDoCalibrators

	      i++; // increment calibrator counter
	    } // end successful read of calibrator name/info

	} // end check comment field
      else // otherwise echo the comment
	if (verbose > NORMAL)
	  printf((char *) "%s",line);
    } // end i/read calibrator loop

  if (i != nCalibrators)
    {
      if (verbose)
	{
	  (void) printf((char *) "# Warning: only %d calibrators read from %s (%d expected)\n",
			i,filename,nCalibrators);
	  (void) fprintf(stderr,(char *) "# Warning: only %d calibrators read from %s (%d expected)\n",
			 i,filename,nCalibrators);
	} // checking verbosity
      nCalibrators = i;
    }

  // Close the script file
  (void) fclose(fptr);

/**********  end readScript  ********************************************/
}


/************************************************************************
*  printRunDirectives -- local to calib.cc (A.B. 4/11/98)		*
*									*
*  A.  Purpose:	The function printRunDirectives prints a report of the  *
*       calibration directives for this run.				*
*									*
*  B.  Background and use: As calib acquires options, it becomes        *
*       increasingly necessary to document the settings of the program  *
*       for a given dataset.  This function dumps a short report on the *
*       settings used in this run to stdout (where the rest of the      *
*       output is also going). 						*
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration printRunDirectives  *******************/
void printRunDirectives()
{
/**********  begin printRunDirectives  **********************************/
  // Reports on some operational (used to be preprocessor) directives
  (void) fprintf(stdout,(char *) "\n# wbCalib version %s\n",(char *) VERSION);
  (void) fprintf(stdout,(char *) "# wbCalib calibration code settings for this run: \n");

  switch (usingWL)
    {
    case 'A' :
      {
	(void) fprintf(stdout,(char *) "# > Calibrating all wavebands\n");
	break;
      }
    case 'K' :
      {
	(void) fprintf(stdout,(char *) "# > Calibrating only K-band (rejecting others at read)\n");
	break;
      }
    case 'H' :
      {
	(void) fprintf(stdout,(char *) "# > Calibrating only H-band (rejecting others at read)\n");
	break;
      }
    } // end waveband switch

  if (usingFR == 0)
    (void) fprintf(stdout,(char *) "# > Calibrating all frame rates\n");
  else
    (void) fprintf(stdout,(char *) "# > Calibrating %fHz frame rate (%dms frame time) data (rejecting others at read)\n",
		   usingFR,((int) rint(1000.0 / (usingFR))));
    
  if (usingIncoherentV2)
    (void) printf((char *) "# > Calibrating incoherent ");
  else
    (void) printf((char *) "# > Calibrating coherent ");

  if (usingSpec)
    (void) printf((char *) "spectrometer ");
  else
    (void) printf((char *) "white-light ");

  (void) printf((char *) "V2 data\n");

  if (usingJitterCorrection)
    (void) printf((char *) "# > Applying jitter correction, coefficient %g\n",
		  jitterCoeff);

  if (usingRatioCorrection)
    (void) printf((char *) "# > Applying ratio correction\n");

  (void) printf((char *) "# > Using calibration scans within +/- %g hr of the target scan\n",
		calibTimeWindow * 24.0);

  if (usingTempErrorProp)
    {
      (void) printf((char *) "# > Weighting calibration scans WRT temporal proximity,");
      (void) printf((char *) " variance doubles in %g hr\n",timeErrorPropCoeff / 24.0);
    }
  else
    (void) printf((char *) "# > No weighting of calibration scans WRT temporal proximity\n");

  if (usingSkyErrorProp)
    {
      (void) printf((char *) "# > Weighting calibration scans WRT sky proximity,");
      (void) printf((char *) " variance doubles in %g deg\n",RAD_TO_DEG / angleErrorPropCoeff);
    }
  else
    (void) printf((char *) "# > No weighting of calibration scans WRT sky proximity\n");

  if (usingMinimumScanUncertainty)
    {
      (void) printf((char *) "# > Applying minimum calibrated scan uncertainty of %g\n",minScanUncertainty);
    }
  else
    (void) printf((char *) "# > No minimum calibrated scan uncertainty\n");

  if (useRootN)
    (void) printf((char *) "# > Weighting computed scan errors by root-Nsamples\n");
  else
    (void) printf((char *) "# > Not weighting computed scan errors by root-Nsamples\n");

  (void) printf((char *) "# > Using %g s scan time threshold\n",(scanThreshold * 3600.0 * 24.0));

  (void) printf((char *) "# > Flagging calibration warnings when multiple calibrators disagree by more than %g sigma\n",
		sqrt(chiWarning));
  (void) printf((char *) "# > Rejecting the scan when multiple calibrators disagree by more than %g sigma\n",
		sqrt(chiReject));

  if (usingWLScale)
    (void) printf((char *) "# > Scaling operating wavelengths by %g\n",wlScale);

  if (delayCheck)
    (void) printf((char *) "# > Comparing model/measured delays with %3g m tolerance\n",
		  delayCheckTol);
  else
    (void) printf((char *) "# > Not comparing model/measured delays\n");

  // added with new option by hjb 5/2005 ... just following along
  if(!baselineFile.empty())
    (void) printf((char *) "# > Using baseline file %s\n", baselineFile.c_str());

  if(usingFluxBiasCorrection) {
      if(!fluxBiasFile.empty())
         (void) printf((char *) "# > Using Flux Bias Coefficients from file %s\n", fluxBiasFile.c_str());
      else
         (void) printf((char *) "# > Using built-in Flux Bias Coefficients.\n");
  } // if(usingFluxBiasCorrection)

  (void) printf((char *) "\n");


/**********  end printRunDirectives  ************************************/
}

/************************************************************************
*  blineInsert -- local to wbCalib.cc (jsh 8/2008)                      *
*                                                                       *
*  A.  Purpose: The function blineInsert adds baselines from file fptr  *
*      to the bline map. Used by readBaseline().                        *
*                                                                       *
*  B.  Background and use:                                              *
*                                                                       *
*  C.  Called by:                                                       *
*                                                                       *
*  D.  Input/output files:                                              *
*                                                                       *
*  E.  Parameter description:                                           *
*       Input -                                                         *
*       Output -                                                        *
*                                                                       *
*  F.  Detail function description:                                     *
*             1.  Local Functions:                                      *
*             2.  Externals:                                            *
*             3.  Globals and Nonlocals used:                           *
*             4.  Global and Nonlocal variables modified:               *
*             5.  Function flow is as follows:                          *
*             6.  Functions called by                                   *
*                                                                       *
*  G.  Function value: This function returns no useable value.          *
*                                                                       *
*************************************************************************/
void blineInsert(FILE *fptr, map<string,baseline*> &blines, char work[LINE_LENGTH]) {
    while(1) {
        baseline *bl = new baseline(fptr);
        if(! bl->error())
           blines.insert(map<string,baseline*>::value_type(bl->baselineName(), bl));
        else
           break;
    }

    if(blines.empty()) {
        (void) printf("# Error: no valid baselines read from baseline file: %s\n", work);
    } else {
        map<string,baseline*>::iterator pos;
        for(pos = blines.begin(); pos != blines.end(); ++pos)
          (void) printf("# Found baseline name %s in baseline file %s \n", pos->first.c_str(), work);
    }

/**********  end blineInsert  ************************************/
}


/************************************************************************
*  readBaseline -- local to wbCalib.cc (A.B. 3/2000)			*
*									*
*  A.  Purpose:	The function readBaseline instantiates a baseline       *
*       object to go along with the sum file being read/to be read.     *
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
//baseline *readBaseline(char *sumFilename, char *baselinef)
// updated function interface 2/06, hjb
// updated 8/08, jsh
map<string,baseline*> readBaseline(char *sumFilename, char *baselinef)
{
/**********  local declarations readBaseline  ***************************/
  char work[LINE_LENGTH], *ptr;

  FILE *fptr = NULL;
  baseline *bLine = NULL;
  map<string,baseline*> blines;
  
/**********  begin readBaseline  ****************************************/
  // One.  Extract trailing suffix and append default baseline suffix
  (void) strncpy(work,sumFilename,LINE_LENGTH);
  if ((ptr = strstr(work,(const char *) ".sum")) != NULL)
    (void) strcpy(ptr, (const char *) ".baseline\0");
  
  // One.Five hjb 5/2005
  // working in the baseline override
  // the idea is not to disturb any of the original code for this function
  // so we'll take the user supplied file name, check that, and if it's readable,
  // slip it into the "work" variable used by the rest of the fucntion. if it's 
  // not readable, we're going to fall back the the original search strategy for
  // finding a baseline
  if(baselinef[0] != '\0') {
    // make sure the file is available, accessible, around for reading
    // if not, emit warnings and do nothing, thereby falling back to the 
    // standard search for baseline file (nearly transparent to user)
    if((fptr = fopen(baselinef,"r")) == NULL) {
       (void) printf("# Error reading user supplied baseline file %s\n", baselinef);
       (void) printf("# Attempting to open baseline file %s\n", baselinef);
    } else {
      (void) strncpy(work, baselinef, sizeof(work));
      work[sizeof(work)] = '\0';
      blineInsert(fptr, blines, work);
      if(!blines.empty())
         return blines;
    }
  }

  // if user supplied baseline file is missing or not valid, try to read from 
  // the xxx.baseline file
  if((fptr == NULL) || (blines.empty())) { 
      if((fptr = fopen(work, "r")) == NULL) {
         (void) printf("# Error trying to open baseline file %s\n", work);
      }
      else {
         blineInsert(fptr, blines, work);
         if(!blines.empty())
             return blines;
      }
  }

  // if xxx.baseline file is missing or not valid, try to read from
  // the xxx.bline file
  if((fptr == NULL) || (blines.empty())) {
      (void) printf("# Attempting to open bline file\n");
      if( (ptr = strstr(work, (const char *) ".baseline")) != NULL)
        (void) strcpy(ptr, (const char *) ".bline\0");
      if((fptr = fopen(work, "r")) == NULL) {
        (void) printf("# Error trying to open bline file %s\n", work);
        (void) printf("# Error Could not read any baseline specification\n");
        return blines;
      }
      blineInsert(fptr, blines, work);
  }
  return blines;

  // note - if we can open the user supplied file for reading, but we can't create
  // a good baseline out of it, we just fall back to .baseline, not .bline. this
  // isn't designed or intentional, but it seems okay, given that if our user supplied
  // file doesn't work, it's hard to imagine the bline could be okay

  // Two.  Check to see we can in fact read the file.  If so, create a baseline obj
 //  if( (fptr = fopen(work,(char *) "r")) != NULL) {
//     //(void) fclose(fptr);
//     baseline *thisBaseline = new baseline(fptr);
//     if(!thisBaseline->error()) {
//       if(verbose > NORMAL)
//      (void) printf((char *) "## reading baseline model from %s (%s)\n", work, thisBaseline->baselineName());

//       //return thisBaseline;
//     } else {
//       if(verbose > NORMAL)
//      (void) printf((char *) "\n## Error detected in %s", work);
//       delete thisBaseline;

//       // Attempt to find/read the RT system baseline file... .baseline
//       (void) strncpy(work, sumFilename, LINE_LENGTH);
//       if( (ptr = strstr(work, (const char *) ".sum")) != NULL)
//      (void) strcpy(ptr, (const char *) ".baseline\0");

//       if((fptr = fopen(work,(char *) "r")) != NULL) {
//      (void) fclose(fptr);
//      baseline *thisBaseline = new baseline(work);
//      if(!thisBaseline->error()) {
//        if (verbose > NORMAL)
//          (void) printf((char *) "; reading baseline model from %s (%s)\n", work,thisBaseline->baselineName());
//        return thisBaseline;
//      } else {
//        // The .baseline file failed as well -- return default baseline
//        if(verbose > NORMAL)
//          (void) printf((char *) " and %s; provisionally using default baseline (%s)\n", work,defBaseline->baselineName());
//        delete thisBaseline;
//        return defBaseline;
//      } // failed .baseline file
//       } else {
//      // Couldn't _find_ .baseline file
//      if(verbose > NORMAL)
//        (void) printf((char *) " and %s not found; provisionally using default baseline (%s)\n", work,defBaseline->baselineName());
//      return defBaseline;
//       }
//     } // end couldn't parse/error in .bline file
//   } else { // end reading .bline baseline file
//     // Three.  Couldn't _find_ .bline file -- try .baseline
//     if(verbose > NORMAL)
//       (void) printf((char *) "## %s not found",work);
//     // Attempt to find/read the RT system baseline file...
//     (void) strncpy(work,sumFilename,LINE_LENGTH);
//     if( (ptr = strstr(work,(const char *) ".sum")) != NULL)
//       (void) strcpy(ptr, (const char *) ".baseline\0");

//     if( (fptr = fopen(work,(char *) "r")) != NULL) {
//       (void) fclose(fptr);
//       baseline *thisBaseline = new baseline(work);
//       if(!thisBaseline->error()) {
//      if (verbose > NORMAL)
//        (void) printf((char *) "; reading baseline model from %s (%s)\n", work,thisBaseline->baselineName());
//      return thisBaseline;
//       } else {
//      // The .baseline file failed as well -- return default baseline
//      if (verbose > NORMAL)
//        (void) printf((char *) " and %s has errors; provisionally using default baseline (%s)\n", work,defBaseline->baselineName());
//      delete thisBaseline;
//      return defBaseline;
//       } // failed .baseline file
//     } else { // end finding .baseline file
//       // Couldn't _find_ .baseline file
//       if (verbose > NORMAL)
//      (void) printf((char *) " and %s not found; provisionally using default baseline (%s)\n", work,defBaseline->baselineName());
//       return defBaseline;
//     }
//   } // end couldn't find .bline




  /**********  end readBaseline  ******************************************/
}



/************************************************************************
*  readSumFile -- local to calib.cc (A.B. 7/27/97)			*
*									*
*  A.  Purpose:	The function readSumFile reads the input summary file   *
*       and extracts the relevant information for storage in the        *
*       linked-list star data structure.				*
*									*
*  B.  Background and use:						*
*         readSumFile is designed to be called multiple times with      *
*       multiple sumfile arguments to construct the database to be used *
*       in this calibration run.                                        *
*         Input to the routine consists of the name of the sumfile to   *
*       be read and parsed, and the (pointer to the) database of        *
*       visibility information that is being constituted.  On output    *
*       information from scans successfully read-in from the input      *
*       sumfile have been entered into the database.  All this          *
*       database manipulation is accomplished via a set of class        *
*       libraries described in StarList.h (and StarList.cc).            *
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files: filename argument				*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        filename -- the (pointer to the) sumfile filename to be        *
*         processed (char *).                                           *
*        starList -- the (pointer to the) starlist database object      *
*         being constructed for later use (StarList *).                 *
*	Output -							*
         starList -- contents modified.                                 *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the number of valid data   *
*       lines that were successfully read from the input file (int).    *
*									*
*************************************************************************/
int readSumFile(char *filename,char *baselinef, char *fluxcoeffs, StarList *starList)
{
/**********  local declarations readSumFile  ****************************/
// hjb added MIN/MAX JITTER 07/2007 to match jitter tests in nbCalib
#define MIN_JITTER 0.0
#define MAX_JITTER 2.0

#define MIN_RATIO 0.99
#define MAX_RATIO 5.0
#define DEF_FRAME_RATE 100

  char line[LINE_LENGTH],starName[132],ratioStr[12],dummyStr[12];
  char wlRatioStr[12],specRatioStr[12],blDesig[132],camera[256];
  int nRead,month,day,year,hour,minute,second,dataLines = 0;
  int nFrames=0,nLocks=0,calFlag=0;
  double jd,ut,delay,longDelay=0.0,frameRate=0.0,wlPhot,wlV2I,wlV2C,wlWavelength,
    specPhot,specWavelength,specV2I,specV2C,jitter,raDeg,decDeg;
  
  double fracLock=0.0,dummy;
 
  FILE *fptr = NULL;
  FILE *fptr_coeff = NULL;
  line[0] = '\0';
  starName[0] = '\0';
  ratioStr[0] = '\0';
  dummyStr[0] = '\0';
  wlRatioStr[0] = '\0';
  specRatioStr[0] = '\0';
  blDesig[0] = '\0';

  // Try to open this summary file
  if (!(fptr = fopen(filename,(char *) "r"))) {
    if (verbose)
      (void) printf((char *) "# >> Error opening summary file %s\n",filename);
    return 0;
  }

  // Read/retrieve the baseline/camera corresponding to this sum file (or use
  // default baseline if missing or error)

  // hjb 2/06 ... use the new constructor in libbaseline which allows for comments
  // and multi-record baseline files
  // in the future this needs update to allow for time associate of baseline to data record
  // baseline *thisBaseline = readBaseline(filename, baselinef);
  map<string,baseline*> blines = readBaseline(filename, baselinef);

  if(blines.empty()) {
    (void) printf("# Warning ... continuing with built-in baseline values\n");
    if(usingFluxBiasCorrection) {
	(void) printf("# Warning: Because default baselines are being used, no flux bias correction will be applied.\n");
	usingFluxBiasCorrection = 0;
    }
  } else {
    map<string,baseline*>::iterator pos;
    for(pos = blines.begin(); pos != blines.end(); ++pos) {
      (void) printf("# Using baseline %s ", pos->second->baselineName());
      (void) printf("(%f, %f, %f, %f) \n", pos->second->localBaseline()[0],
 		    pos->second->localBaseline()[1], pos->second->localBaseline()[2],
		    pos->second->bias());
    }
    (void) printf("\n");
  }

    // Create a map of baselines/camera names that are found in the sum file
    // but which do not exist in the baselineFile. This is used to print
    // warnings a single time rather than once for each sum line that has
    // a baseline/camera name missing in the baseline file 
    map<string,int> blWarningMap; 

    // Create a map of camera names that are found in the sum file but which
    // do not exist in the fluxFile. This is used to print warnings a single
    // time for each unique camera name in the sum file that does not exist in
    // the flux file
    map<string,int> camWarningMap;

  /* 7/28/97.  Okay, we're reading Mark's summary file, which
     according to the current vis.txt looks like this:
>this is basic data format
>    date     star  time(hrs) delay(m)   WL: N(DN)  V2I   V2C         spec: N(DN) V2C    V2I    std   
SUM: 07/18/97       705   6.001597 -18.0477393     293.1  0.271  0.241      14.4  0.631  0.725  0.70   0.57   0.55    36.4   333.2   642.5   37.8   52.9   60.7  2.180   0.97   1

  One extra tidbit.  I've added a check to see if a data line has
  been commented out by a pound sign.  This allows for some degree
  of user intervention in the instance of bogus data.

  Added 11/99.  We sometimes have vis failures in the calibration
  fields like ratios and forgrounds.  I've modified the logic to
  accept NaNs in the calibrations.  Caveat emptor.

  You also notice the 50 at the end of the line

SUM: 11/11/99 HDC78418    12.237483  -12.2892359   665.3  0.084  0.079   29.0  0.125  0.155  0.66   0.00    NaN   125.4     0.0   521.0  154.2    0.0  179.2  2.200   0.35  11  0  OK  9.14648056 26.62911224  50 


  */

  /* 12/00 AFB and RLA
     wbCalib at KI will read a slightly different SUM line format as given here
     SUM date (m/d/y/h:m:s) UThr sourceName delay ...
  */

  while (!feof(fptr)) {
    frameRate = (double)DEF_FRAME_RATE;
    (void) fgets(line,LINE_LENGTH,fptr);
    // Read data and check for wavelength match
    if (!(feof(fptr)) && (line[0] != '#')
	&& (((nRead = sscanf(line,
			     (char *) "SUM: %d/%d/%d %s %lf%lf%lf%lf%lf%lf%lf%lf%lf%s%s%s%lf%lf%lf%lf%lf%lf%s%s%s%s%lf%lf%lf%lf%lf%s",
			     &month,&day,&year,&starName,&ut,&delay,
			     &wlPhot,&wlV2I,&wlV2C,
			     &specPhot,&specV2C,&specV2I,
			     &jitter,&ratioStr,
			     &dummyStr,&dummyStr,&dummy,&dummy,&dummy,&dummy,&dummy,
			     &specWavelength,&dummyStr,&dummyStr,&dummyStr,&dummyStr,
			     &raDeg,&decDeg,&frameRate, &dummy, &dummy, &blDesig)) >= 25)
	    || ((nRead = sscanf(line,
			     (char *) "SUM: %d/%d/%d %s %lf%lf%lf%lf%lf%lf%lf%lf%lf%s%s%s%lf%lf%lf%lf%lf%lf%s%s%s%s%lf%lf%lf",
			     &month,&day,&year,&starName,&ut,&delay,
			     &wlPhot,&wlV2I,&wlV2C,
			     &specPhot,&specV2C,&specV2I,
			     &jitter,&ratioStr,
			     &dummyStr,&dummyStr,&dummy,&dummy,&dummy,&dummy,&dummy,
			     &specWavelength,&dummyStr,&dummyStr,&dummyStr,&dummyStr,
			     &raDeg,&decDeg,&frameRate)) == 22)
	    || ((nRead = sscanf(line,
				(char *) "SUM %d/%d/%d/%d:%d:%d %lf %s %lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%d%d%lf%lf%s%s%d%s",
				&month,&day,&year,&hour,&minute,&second,
				&ut,&starName,&delay,&longDelay,
				&wlPhot,&wlV2I,&wlV2C,&wlWavelength,
				&specPhot,&specV2I,&specV2C,&specWavelength,
				&jitter,&nFrames,&nLocks,&fracLock,&frameRate,
				&wlRatioStr,&specRatioStr,&calFlag,&blDesig)) == 27)))
      if ((usingWL == 'A')  // Added logic to read only specific bands  A.B. 11/99
	  || ((usingWL == 'K') && (fabs(specWavelength - K_wl) < K_width))
	  || ((usingWL == 'H') && (fabs(specWavelength - H_wl) < H_width))   )
	if ((usingFR == 0)  // Checking frame rate logic.  A.B. 11/99
	    || ((nRead < 29) && (usingFR == 100))
	    || ((nRead == 29) && (frameRate == usingFR))) {
	  dataLines++;  // If all tests passed, we've got a data line
	  int ratioKickout = 0; // Check for failing ratio
	  int jitterKickout = 0;

	  // Turn dates and UTs into JDs
	  if (year < 96)
	    year += 100;
	  if (year < 1900)
	    year += 1900;
	  jd = jdFromUTCal(ut,year,month,day);

	  // Apply corrections for jitter (and ratio) as indicated by
	  // the operational (used to be preprocessor) directives
	  if (usingJitterCorrection) {
	    // hjb 07/2007, adding in min/max jitter test to match nbCalib
	    if( (jitter > MIN_JITTER) && (jitter < MAX_JITTER) ) {
	      jitterKickout = 0;
	      // Current form of the jitter correction
	      double jitterCorrection = exp(jitterCoeff * jitter * jitter);
	    
	      wlV2I *= jitterCorrection;
	      wlV2C *= jitterCorrection;
	      specV2C *= jitterCorrection;
	      specV2I *= jitterCorrection;
	    } else {
	      jitterKickout = 1;
	      if(verbose)
		(void) fprintf(stdout, (char *) "# Warning: jitter %g out of range for %s at UT %f; data not used\n",
			       jitter, starName, ut);
	    }
	  } // end usingJitterCorrection
	  
	  if (usingRatioCorrection) // New logic must check for NaN
	    if (strncmp(line,"SUM: ",5) == 0) { // PTI line format
	      
	      double ratioCorrection = atof(ratioStr);
	      if ((ratioCorrection) &&
		  (ratioCorrection >= MIN_RATIO)
		  && (ratioCorrection <= MAX_RATIO)) {
		ratioKickout = 0;
		wlV2I *= ratioCorrection;
		wlV2C *= ratioCorrection;
		specV2C *= ratioCorrection;
		specV2I *= ratioCorrection;
	      } // end apply ratio correction
	      else { // str conversion fail or out of reasonable bounds
		ratioKickout = 1;
		if (verbose)
		  (void) fprintf(stderr,(char *) "# Ratio string %s in sumfile %s (at UT %f) out of range -- SUM line rejected\n", ratioStr,filename, ut);
	      }
	    } // end PTI format
	    else {// KI line format
	      double wlRatioCorrection = atof(wlRatioStr);
	      double specRatioCorrection = atof(specRatioStr);
	      
	      if ( (wlRatioCorrection) &&  // HJB: test if clause against double??
		   (wlRatioCorrection >= MIN_RATIO)
		   && (wlRatioCorrection <= MAX_RATIO)
		   && (specRatioCorrection)
		   && (specRatioCorrection >= MIN_RATIO)
		   && (specRatioCorrection <= MAX_RATIO)) {
		ratioKickout = 0;
		wlV2I *= wlRatioCorrection;
		wlV2C *= wlRatioCorrection;
		specV2C *= specRatioCorrection;
		specV2I *= specRatioCorrection;
	      } // end apply ratio correction
	      else { // str conversion fail or out of reasonable bounds
		ratioKickout = 1;
		if (verbose)
		  (void) fprintf(stderr,(char *) "# Ratio strings %s and %s in sumfile %s (at UT %f) out of range -- SUM line rejected\n", wlRatioStr,specRatioStr,filename, ut);
	      }
	      
	    } // end KI line format/usingRatioCorrection


	  if (strncmp(line,"SUM: ",5) == 0) {	// PTI vis-format line
	    //  -- set wlWavelength & longDelay
	    wlWavelength = specWavelength;
	    longDelay = 0.0;
	  } // end SUM:/PTI format
	      
	  
	  if (usingWLScale) {
	    wlWavelength *= wlScale;
	    specWavelength *= wlScale;
	  } // end applying wavelength scale

	  if(usingFluxBiasCorrection) {

            float WLFluxBiasCoeff;
            float SpecFluxBiasCoeff;
            float Nzero;
            float NormFlux;

	    // per rla, this correction applies to KI only
	    if(strncmp(line, "SUM " , 4) == 0) {

              // blDesig must match name in blines which contains "KI_" prefix
	      // fluxBiasMap does not (usually) have the "KI_" prefix. If KI_ 
	      // prefix is present, strip it from blDesig to get the camera name
              if (sscanf(blDesig, "KI_%s", camera) == 0)
		   strcpy(camera, blDesig);
              map<string, fluxBiasStruct>::iterator pos = fluxBiasMap.find(camera);
              if(pos != fluxBiasMap.end()) {
     		  WLFluxBiasCoeff = pos->second.WLFluxBias;
                  SpecFluxBiasCoeff = pos->second.SpecFluxBias;
                  Nzero = pos->second.NZero;
                  NormFlux = pos->second.Normflux;
		      
		  // Skip these warnings if cameraName read from fluxBiasFile
		  // They only apply if using built-in defaults
		  if(pos->second.usingDefaults == 1) {
		      map<string,int>::iterator pos2 = camWarningMap.find(camera);
		      // only print warning the first time camName encountered 
		      if(pos2 == camWarningMap.end()) {
		 	 // the "1" in the pair is arbitrary
		         camWarningMap.insert(std::make_pair(camera,1));
                         if((strcmp(camera,"KI_K1K2") == 0) || (strcmp(camera,"K1K2") == 0))
                            (void) printf("# Warning: Using built-in primary camera defaults for %s: WLFluxBias = %f  SpecFluxBias = %f  Nzero = %f  NormFlux = %f\n\n", camera, WLFluxBiasCoeff, SpecFluxBiasCoeff, Nzero, NormFlux);
                         else
		            (void) printf("# Warning: Using built-in camera defaults for %s: WLFluxBias = %f  SpecFluxBias = %f  Nzero = %f  NormFlux = %f\n\n", camera, WLFluxBiasCoeff, SpecFluxBiasCoeff, Nzero, NormFlux);
		      } // if(pos2 == camWarningMap.end())
                  } // if(pos->second.usingDefaults == 1)

                  // received 6/07 from rla/mc
                  // visibilities should be corrected for flux bias by the formula
                  // correction = 1 - k log10 ( (Ni+Nzero)/(Nnorm + Nzero))
                  // where Ni is this record's flux (spec or wl), Nzero about 20dN (80 phot) and the normalization
                  // level set to 1000 dN
                  // k_wl = WLFluxBiasCoeff ~= 0.063
                  // k_spec = SpecFluxBiasCoeff ~= 0.16
                  double wl_correction = 1 - WLFluxBiasCoeff * log10( (wlPhot + Nzero) / ( NormFlux + Nzero) );
                  wlV2I *= wl_correction;
                  wlV2C *= wl_correction;

                  double spec_correction = 1 - SpecFluxBiasCoeff * log10( (specPhot + Nzero) / (NormFlux + Nzero) );
                  specV2I *= spec_correction;
                  specV2C *= spec_correction;
              } // if(pos != fluxBiasMap.end())
              else { // camera name not found in fluxBiasMap
		 map<string,int>::iterator pos2 = camWarningMap.find(camera);
                 // only print warnings once for each unique cameraName
                 if(pos2 == camWarningMap.end()) {
                    // the "1" in the pair is arbitrary
                    camWarningMap.insert(std::make_pair(camera,1));
  		    if(fluxcoeffs[0] != '\0') {
                       (void) printf("# Warning: Baseline/Camera name %s in SUM record of file %s not found in flux bias coefficients file %s. No flux bias correction will be applied.\n\n", camera, filename, fluxcoeffs);
                    }
                    else {
 		       (void) printf("# Warning: Baseline/Camera name %s in SUM record of file %s not recognized in default list of flux bias coefficients. No flux bias correction will be applied.\n\n", camera, filename);
                    }
                 } // if (pos2 == camWarningMap.end())
              } // else
	    } // if(strncmp(line, "SUM " , 4) == 0)
	  } // if(usingFluxBiasCorrection) 

	  if ((strncmp(line,"SUM: ",5) == 0) && (nRead < 29)) // If PTI format degenerate read
	    frameRate = 100.0; // Assume 10ms data

	  // KI Kvis-format line -- set baseline fron designation
	  // if no input baseline model (so using default)
	  //if((strncmp(line,"SUM ",4) == 0) &&	
	  //   (thisBaseline == defBaseline)) {
	  // hjb updated if clause, to allow for pti records w/ baseline tags, from imel pipeline (2/06)
	  // map<string,baseline*>::iterator pos = blines.find(blDesig);
	  // baseline *thisBaseline;
	  // if( pos == blines.end())
	  // thisBaseline = defBaseline;
	  // else
	  // thisBaseline = pos->second;
          // }

	  baseline *thisBaseline;
 	  int missingBl = 0;
	  if(blDesig[0] != '\0') {
	    map<string,baseline*>::iterator pos = blines.find(blDesig);
	    if( pos == blines.end()) {
	      thisBaseline = defBaseline;
	      missingBl= 1;
            }	
	    else
	      thisBaseline = pos->second;
	  } else {
	    if(blines.size() == 1) 
	      thisBaseline = blines.begin()->second;
	    else {
	      thisBaseline = defBaseline;
              missingBl = 1;	
            }	
	  }

	  // this checks addresses, not actual member equality.
	  if( thisBaseline == defBaseline ) {
	    // case of NULL blDesig should fall through to the end
            if(strncmp(blDesig, (char *) "PTI_NS",6) == 0)
	      thisBaseline = &PTI_NS;
            else if(strncmp(blDesig, (char *) "PTI_NW",6) == 0)
	      thisBaseline = &PTI_NW;
	    // hjb added PTI_SW baseline 2/06
            else if(strncmp(blDesig, (char *) "PTI_SW",6) == 0)
	      thisBaseline = &PTI_SW;
            else if(strcmp(blDesig, (char *) "KI_K1K2S") == 0)
              thisBaseline = &KI_K1K2S;
            else if(strcmp(blDesig, (char *) "KI_K1K2") == 0)
	      thisBaseline = &KI_K1K2;
            else if(strcmp(blDesig, (char *) "KI_S1S2") == 0)
	      thisBaseline = &KI_S1S2;
	    // jsh added Camera_FT baselines 11/10
            else if(strcmp(blDesig, (char *) "KI_Camera_FT_PRIM") == 0)
              thisBaseline = &KI_Camera_FT_PRIM;
            else if(strcmp(blDesig, (char *) "KI_Camera_FT_SEC") == 0)
              thisBaseline = &KI_Camera_FT_SEC;
            else if(strcmp(blDesig, (char *) "KI_Camera_FT_LBAND") == 0)
              thisBaseline = &KI_Camera_FT_LBAND;
	    else {
	      thisBaseline = defBaseline;
            }

            if(missingBl == 1) {
               map<string,int>::iterator pos = blWarningMap.find(blDesig);	

	       // only print warning the first time unique blDesig encountered
               if(pos == blWarningMap.end()) {
                  // the "1" in the pair is arbitrary
                  blWarningMap.insert(std::make_pair(blDesig,1));
  	          if(baselinef[0] != '\0')
		     (void) printf("# Warning: Baseline name %s from SUM record in file %s not found in baseline file %s.\n", blDesig, filename, baselinef);
	 	  (void) printf("# Warning: Using built-in default %s baseline components E:%f N:%f U:%f C:%f for U,V and delay calculations.\n\n", thisBaseline->baselineName(), thisBaseline->localBaseline()[0], thisBaseline->localBaseline()[1],  thisBaseline->localBaseline()[2], thisBaseline->bias());
                  (void) printf("\n");
               }
            } // missingBl == 1
	  } // end set baseline/camera from designation

 	 // (void) printf("# [in readSumFile] thisBaseline is %s with bias %f\n", thisBaseline->baselineName(), thisBaseline->bias());

	  // Create a data node with this data contents
	  if ( (! ratioKickout) && (! jitterKickout) ) {

	    WBdata *wbNode = new WBdata(starName,0,month,day,year,
					jd,ut,delay,longDelay,
					wlPhot,wlWavelength,wlV2I,wlV2C,
					specPhot,specWavelength,
					specV2I,specV2C,
					jitter,frameRate,
					thisBaseline,
					blDesig);

	    // Find or create a star node corresponding to this star
	    if (!starList->findStar(starName))
	      starList->insert(starName);
	    
	    // Insert the data from this line into the linked-list
	    // structure for this star.  Note it is being
	    // insertion-sorted into the list based on jd.  Important
	    // safety tip -- the proper functioning of avgV2Data
	    // depends on the list being time-sorted.
	    starList->getStarData()->insert(wbNode,jd,INSERT_SORT);
	    //(void) printf("# Created node for %s, with baseline %s at bias %f\n", wbNode->get_starName(), wbNode->get_baseline()->baselineName(), 
	    //	  wbNode->get_baseline()->bias());
	  } else
	    dataLines--; // ratio conversion failure

	} // end read/line format/WL check
  } // end while loop

  (void) fclose(fptr); // Close the file

  return dataLines;

}


/************************************************************************
*  avgV2Data -- local to calib.cc (A.B. 7/27/97)			*
*									*
*  A.  Purpose:	The function avgV2Data computes the average of the V2   *
*       data output by Mark's vis program for a single scan.            *
*									*
*  B.  Background and use:						*
*         avgV2Data sounds simple -- you're just averaging (and         *
*       computing sample standard deviations) for a sequence of scans.  *
*       But in point of fact avgV2Data is in fact a little trickier     *
*       in that it is the agent charged with walking the constituent    *
*       linked list using the functionality in the linked list template *
*       class library dataList (dataList.h).  On input, the current     *
*       pointer in the starData list points to the first record in the  *
*       sequence for a particular scan.  avgV2Data then walks the list  *
*       acculating pertinent sums on all the records for that scan, as  *
*       judged by being within a time window (SCAN_THRESH) of this      *
*       first record.  On output, current should now be pointing to the *
*       first record in the next scan on the target.  Clearly this      *
*       methodology requires the record list to be arranged in          *
*       increasing time order.  This is how the database is constructed *
*       in readSumFile (new data records are inserted into the list     *
*       using the INSERT_SORT option).                                  *
*         On output averages and variances are compted for the V2, and  *
*       averages are computed for jd, ut, jitter, and ratioC.           *
*									*
*  C.  Called by: sysVisSC & calib					*
*									*
*  D.  Input/output files: none						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        starData -- the (pointer to the) linked-list object for the    *
*         object we're working on (dataList<WBdata>).                   *
*	Output -							*
*        jdAvg, utAvg, v2Avg, jitterAvg, ratioCAvg -- the averaged      *
*         quantities for this scan (various dimensions) (double *).     *
*        v2Var -- the estimated error variance in v2Avg                 *
*         (dimensionless) (double *).                                   *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the number of data entries *
*       found in this scan (int).                                       *
*									*
*************************************************************************/
int avgV2data(dataList<WBdata> *starData,
	      int *month,int *day,int *year,
	      double *jdAvg,double *utAvg,double *delayAvg,double *wlAvg,
	      double *v2Avg,double *v2Var,double *jitterAvg,
	      double *frameRate)
{
/**********  local declarations avgV2data  ******************************/
  // #define SCAN_THRESH (150.0 / 3600.0 / 24.0)

  string firstBaseline = starData->getCurrent()->get_baseline()->baselineName();
  double firstUT = starData->getCurrent()->get_ut();
  double firstJD = starData->getCurrent()->get_jd();
  *month = starData->getCurrent()->get_month();
  *day = starData->getCurrent()->get_day();
  *year = starData->getCurrent()->get_year();
  int sumTerms = 0;
  double myFrameRate = 0.0;
  double sumV2 = 0.0;
  double sumV2_2 = 0.0;
  double sumJD = 0.0;
  double sumUT = 0.0;
  double sumDelay = 0.0;
  double sumWL = 0.0;
  double sumJitter = 0.0;
  double sumRatioC = 0.0;

  int alreadyWarned = 0; 

/**********  begin avgV2data  *******************************************/
  // Retrieve the data for this scan (accumulate the appropriate
  // sums).  Subtle but important.  Note that this loop is doing the
  // linked-list iteration.  When it kicks out all the records within
  // SCAN_THRESH of the first record on routine invocation has been read.
  do
    {
      if (usingSpec) {
	if (usingIncoherentV2) {

	    sumV2 += starData->getCurrent()->get_specV2I();
	    sumV2_2 += starData->getCurrent()->get_specV2I()
	      * starData->getCurrent()->get_specV2I();
	    sumWL += starData->getCurrent()->get_specWavelength();
	} // end spectrometer IncoherentV2
	else {
	    sumV2 += starData->getCurrent()->get_specV2C();
	    sumV2_2 += starData->getCurrent()->get_specV2C()
	      * starData->getCurrent()->get_specV2C();
	    sumWL += starData->getCurrent()->get_specWavelength();
        }
      }
      else { // using white-light
	if (usingIncoherentV2) {
	    sumV2 += starData->getCurrent()->get_wlV2I();
	    sumV2_2 += starData->getCurrent()->get_wlV2I()
	      * starData->getCurrent()->get_wlV2I();
	    sumWL += starData->getCurrent()->get_wlWavelength();
	} // end white-light IncoherentV2
	else {
	    sumV2 += starData->getCurrent()->get_wlV2C();
	    sumV2_2 += starData->getCurrent()->get_wlV2C()
	      * starData->getCurrent()->get_wlV2C();
	    sumWL += starData->getCurrent()->get_wlWavelength();
	} // end using white-light CoherentV2
      }

      sumDelay += starData->getCurrent()->get_delay()
	- starData->getCurrent()->get_longDelay();
      sumJD += starData->getCurrent()->get_jd();
      sumUT += starData->getCurrent()->get_ut();
      sumJitter += starData->getCurrent()->get_jitter();
      myFrameRate = starData->getCurrent()->get_frameRate();
      sumTerms++;

      if(!alreadyWarned) {
          string currentBaseline = starData->getCurrent()->get_baseline()->baselineName();
          if(currentBaseline != firstBaseline) {
              (void) printf((char *) "\n# Warning: Baseline name changed from %s to %s within the averaging block starting at %f. Only Baseline %s will be used for this average.\n", firstBaseline.c_str(), currentBaseline.c_str(), firstUT, firstBaseline.c_str());
	      alreadyWarned = 1;
          }
      } // if(!alreadyWarned)
    } //do
  while ((starData->iterate())
	 && (fabs(starData->getCurrent()->get_jd() - firstJD) < scanThreshold));


  // Compute the output values.
  double denom = 1.0 / (double) sumTerms;
  *v2Avg = sumV2 * denom;
  *v2Var = (sumV2_2 - (sumV2 * sumV2 * denom)) * denom; //  / ((double) (sumTerms - 1));
  if (useRootN)  // Add root-N scaling (N in variance) to computed error bars A.B. 7.02
    *v2Var *= denom;

  *delayAvg = sumDelay * denom;
  *wlAvg = sumWL * denom;
  *jdAvg = sumJD * denom;
  *utAvg = sumUT * denom;
  *jitterAvg = sumJitter * denom;
  *frameRate = myFrameRate;

  // Return the number of records we processed for this scan
  return sumTerms;

/**********  end avgV2data  *********************************************/
}


/************************************************************************
*  v2Pred -- local to calib.cc (A.B. 7/27/97)				*
*									*
*  A.  Purpose:	The function v2Pred computes the expected v2 value (and *
*       variance) for a given calibrator and projected normalized       *
*       baseline length (aka inverse fringe spacing).			*
*									*
*  B.  Background and use:						*
*         v2Pred is pretty simple: it computes the expected V2 for an   *
*       ideal interferometer observing a simple star -- presumably a    *
*       calibrator.  The ratio of what we got to what we expected is    *
*       an estimate of the system visibility.                           *
*									*
*  C.  Called by: sysVisSC						*
*									*
*  D.  Input/output files: none						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        baseline -- the sky-projected baseline length (m) (double).    *
*        lambda -- the wavelength we're working at (m) (double).        *
*        aSize -- the estimated angular size of object (rad) (double).  *
*        aSigma -- the error in aSize (rad) (double).                   *
*	Output -							*
*        varV2Pred -- the variance in the output V2 prediction based    *
*         on the input angular size error (dimensionless) (double *).   *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the expected V2 value      *
*       (dimensionless) (double).                       		*
*									*
*************************************************************************/
double v2Pred(double baseline,double lambda,
	      double aSize,double aSigma,double *varV2Pred)
{
/**********  begin v2Pred  **********************************************/
  // The fringe spacing on the sky (units of radians)
  double fringeSP = baseline / lambda;
  double arg = PI * fringeSP * aSize;
  double dArgdaSize = PI * fringeSP;
  double v,dvdArg;

  // Trap the (annoying) point source case
  if (arg > 0)
    {
      v = 2.0 * j1(arg) / arg;
      dvdArg = 2.0 * (0.5 * arg * (j0(arg) - jn(2,arg)) - j1(arg)) / arg / arg;
    }
  else
    {
      v = 1.0;
      dvdArg = 0.0;
    }

  // Use the chain rule (otherwise I never get this right).
  double dv2dv = 2.0 * v;

  *varV2Pred = dv2dv * dvdArg * dArgdaSize * aSigma;
  *varV2Pred *= *varV2Pred;

  return (v * v);
/**********  end v2Pred  ************************************************/
}


/************************************************************************
*  sysVisSC -- local to calib.cc (A.B. 7/27/97)				*
*									*
*  A.  Purpose: The function sysVisSC estimates the system visibility   *
*       (and the error in that estimate) with respect to a particular   *
*       calibrator at a particular jd.					*
*									*
*  B.  Background and use:						*
*         sysVisSC mechanizes the estimate of the system visibility at  *
*       a particular epoch based on a particular calibrator.  In the    *
*       calib scheme, each calibrator is considered independently (this *
*       routine), and then these results are (weighted) averaged to     *
*       derive the system visibility estimate for the target (done      *
*       in sysVisEnsemble).                                             *
*         sysVisSC does its work by extracting all calibrator scans     *
*       within a given time window (CALIB_WINDOW) around the input      *
*       time.  From each calibrator scan a system visibility estimate   *
*       (and error estiamte) is derived.  These system visibility       *
*       estimates are then (weighted) averaged to determine the best    *
*       estimate for the system visibility at the input time.  Note:    *
*       these system visibility estimates _can_ be weighted with        *
*       respect to temporal proximity to the desired output time.       *
*									*
*  C.  Called by: sysVisEnsemble					*
*									*
*  D.  Input/output files: stdout & stderr				*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the Julian Data for the system visibility estimate.      *
*        calibData -- the (pointer to the) data list object for this    *
*         calibrator (dataList<WBdata> *).                              *
*        calSize -- the estimated angular size for this calibrator      *
*         (rad) (double).                                               *
*        sigmaSize -- the estimated error in calSize (rad) (double).    *
*        star -- the (pointer to the) unit 3-vector sky position for    *
*         this calibrator (dimensionless) (Vector *).                   *
*	Output -							*
*        varSysVis -- the estimated error variance in the system        *
*         visibility (squared) estimate (dimensionless) (double *).     *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the estimate of system     *
*       visibility (squared) at the time of interest derived from a     *
*       particular calibrator (dimensionless) (double).		        *
*									*
*************************************************************************/
/**********  function declaration sysVisSC  *****************************/
double sysVisSC(double jd,double targFrameRate,
		dataList<WBdata> *calibData,char *calName,
		double calSize,double sigmaSize,
		Vector *star,double *varsysVisSC)
{
/**********  local declarations sysVisSC  *******************************/

  int month,day,year,calHits = 0;
  double calFrameRate = 0.0;
  double jdAvg = 0.0,utAvg = 0.0,V2Avg = 0.0,V2sig = 0.0;
  double weight,w2,weightSum = 0.0,weightSum2 = 0.0,weightSumVar = 0.0;
  double expV2,varExpV2;
  double jdScan,utScan,delayScan,wlScan,V2Scan,V2ScanVar,jitterScan,ratioCScan;

/**********  begin sysVisSC  ********************************************/
  // Loop over the designated calibration targets
  (void) calibData->gotoHead();
  if (calibData->numberNodes())
    {
      // Skip forward to the interesting time
      while (!calibData->endOfList()
	     && ((jd - calibData->getCurrent()->get_jd()) > calibTimeWindow))
	calibData->iterate();

      // Extract relevant calibration information from this
      // calibration source
      while ((!calibData->endOfList())
	     && (fabs(jd - calibData->getCurrent()->get_jd()) < calibTimeWindow))
	{
	  baseline *Baseline = calibData->getCurrent()->get_baseline();
	  if ((avgV2data(calibData,&month,&day,&year,&jdScan,&utScan,
			 &delayScan,&wlScan,&V2Scan,
			 &V2ScanVar,&jitterScan,&calFrameRate) > 1)
	      && (calFrameRate == targFrameRate))
	    {
	      calHits++;

	      if (delayCheck)
		{
		  // New 11/30/97.  Check the delay value for this calibration object
		  // double delayModel = Baseline->delay(jdScan,*star);
		  double tmp = delayScan - Baseline->delay(jdScan,*star);
		  if (fabs(tmp) > delayCheckTol)
		    if (verbose)
		      {
			(void) printf((char *) "# Warning: at %10.4f, %4g m measured/expected delay mismatch on calibrator %s\n",jdScan,tmp,calName);
			(void) fprintf(stderr,(char *) "Warning: at %10.4f, %4g m measured/expected delay mismatch on calibrator %s\n",jdScan,tmp,calName);
		      } // end verbose/delay value check
		} // end delayCheck

	      // Compute the expected V2 value at this epoch, sky
	      // position, baseline projection, wavelength...
	      // New 12/20/97 -- use sumfile-averge WL
	      expV2 = v2Pred(Baseline->skyProj(jdScan,*star),
			     wlScan * 1.0e-6,
			     calSize,sigmaSize,&varExpV2);

	      if (usingTempErrorProp)
		{
		  // If defined, propagate the error in the system visibility
		  //  to the time of the target observation.  Assuming linear
		  //  (quadratic) increase in the uncertainty (variance).
		  V2ScanVar *= 1.0 + pow(timeErrorPropCoeff * fabs(jd - jdScan),1.0);
		}

	      // Weighting.  Weight by variance, jitter
	      weight = 1.0 / V2ScanVar;
	      weight *= exp(- jitterScan * jitterScan);
	      weightSum += weight;
	      weightSum2 += w2 = weight * weight;

	      V2Avg += weight * (V2Scan / expV2);
	      weightSumVar += w2 * (V2ScanVar / expV2 / expV2);
	      weightSumVar += w2 * pow((V2Scan / expV2 / expV2),2.0) * varExpV2;
	    } 
	  else {  
	    (void) printf("\n# Warning: Calibration scan on %s, near %f, has only 1 SUM rec. Skipping scan.\n", calName, utScan);
	  }
	    
	} // calibration scan

      if (calHits) // If we've found good calibration hits
	{
	  *varsysVisSC = weightSumVar / weightSum2;
	  return (V2Avg / weightSum);
	}
      else // Otherwise, return a stupid value to flag calib 
	return -1.0;

    } // end calibData node count check
  else // no data for this calibrator
    return -1.0;

/**********  end sysVisSC  **********************************************/
}


/************************************************************************
*  sysVisEnsemble -- local to calib.cc (A.B. 10/25/97)			*
*									*
*  A.  Purpose:	sysVisEnsemble computes the system visibility estimate  *
*       from the full ensemble of calibrators selected for this         *
*       calibration run.						*
*									*
*  B.  Background and use:						*
*         sysVisEnsemble mechanizes the computation of an average       *
*       system visibility estimate from the composite of single         *
*       calibrator system visibility estimates (computed in sysVisSC).  *
*       sysVisEnsemble also performs system visibility estimate         *
*       inner-consistency checks among multiple calibrators.            *
*         In the case of multiple calibrators, the cross-consistency    *
*       of derived system visibility estimates is checked, and          *
*       mismatches at two levels are flagged for the user as a          *
*       diagnostic that there might be something wrong with either the  *
*       input calibrator information or the calibration object itself.  *
*									*
*  C.  Called by: calib							*
*									*
*  D.  Input/output files: stdout & stderr				*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the Julian Data for the system visibility estimate       *
*         (double).                                                     *
*        targV -- the (pointer to the) unit 3-vector sky position of    *
*         the target (dimensionless) (Vector *).  Note: this is okay    *
*         if NULL, it just means there can be no angular proximity      *
*         weighting in the system visibility estimate computation.      *
*        starList -- the (pointer to the) visibility database object    *
*         (StarList *).                                                 *
*	Output -							*
*        sysVis -- the composite estimated system visibility at the     *
*         desired time (dimensionless) (double *).                      *
*        varSysVis -- the estimated error variance in sysVis            *
*         (dimensionless) (double *).                                   *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration sysVisEnsemble  ***********************/
int sysVisEnsemble(double jd,Vector *targV,
		   double targFrameRate,StarList *starList,
		   double *sysVis,double *varSysVis)
{
/**********  local declarations sysVisEnsemble  *************************/
  int i;
  double sV2[MAX_CALIBRATORS],varSV2[MAX_CALIBRATORS];

/**********  begin sysVisEnsemble  **************************************/
  // First loop over the calibrators.
  // Determine the estimated system visibility at the time of
  // this scan based on each calibrator...
  for (i = 0; i < nCalibrators; i++)
    if (starList->findStar(calName[i]))
      {
	dataList<WBdata> *calibData = starList->getStarData();
	// Evaluate the calibrator apparent position
	// *(calibV[i]) = calibCoord[i]->topocentricPos(jd,PTI_LONGITUDE,PTI_LATITUDE,PTI_ALTITUDE);
	// Bug alert: No way to ensure that this baseline/camera is valid for this time (quant effects small)
	baseline *Baseline = calibData->getCurrent()->get_baseline();
	*(calibV[i]) = calibCoord[i]->topocentricPos(jd,Baseline->latitude(),
						     Baseline->longitude(),
						     Baseline->altitude());
	sV2[i] = sysVisSC(jd,targFrameRate,
			  calibData,calName[i],calSize[i],
			  sigmaSize[i],calibV[i],&varSV2[i]);
      } // end first loop over calibrators
    else
      sV2[i] = -1.0;

  // Second loop over the calibrators.
  // Compute the weighted average of the system visibility
  // calibrations, and check the interconsistency of the
  // calibrations at this time.
  int calHits = 0;
  double weight,w2,weightSum = 0.0,v2Sum = 0.0;
  for (i = 0; i < nCalibrators; i++)
    if (sV2[i] > 0.0) // if there was a non-trivial calibration
      {
	calHits++;

	// Add angular proximity weighting
	if (usingSkyErrorProp)
	  if (targV)
	    varSV2[i] *= 1.0 + angleErrorPropCoeff * angle_between(*calibV[i],*targV);

	weightSum += weight = 1.0 / varSV2[i];
	w2 = weight * weight;
	v2Sum += weight * sV2[i];
	
	// Check for calibration inconsistencies
	for (int j = i + 1; j < nCalibrators; j++)
	  if (sV2[j] > 0.0)
	    {
	      double chi = pow(sV2[i] - sV2[j],2.0) / (varSV2[i] + varSV2[j]);
	      // Inconsistencies are flagged at two thresholds.  The
	      // first is just warning...
	      if (chi > chiWarning)
		{
		  if (chi < chiReject)
		    {
		      if (verbose > NORMAL)
			{
			  (void) printf((char *) "# Warning: at %10.4f sysVis estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
					jd,calName[i],calName[j],sqrt(chi));
			  (void) fprintf(stderr,(char *) "Warning: at %10.4f system visibility estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
					 jd,calName[i],calName[j],sqrt(chi));
			} // verbose check
		    } // end warning
		  else // The second is an outright rejection of the scan...
		    {
		      (void) printf((char *) "# Reject: at %10.4f sysVis estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
				    jd,calName[i],calName[j],sqrt(chi));
		      (void) fprintf(stderr,(char *) "Reject: at %10.4f system visibility estimates between calibrators %s and %s are inconsistent at the %3.1f sigma level\n",
				     jd,calName[i],calName[j],sqrt(chi));
		      // return a null-calibration
		      return 0;
		    }
		} // end chi test
	    } // end cross-calibration check loop
      } // end weight-average loop

  // If we've gotten a good calibration, compute the system visibility.
  if (calHits)
    {
      *sysVis = v2Sum / weightSum;
      *varSysVis = 1.0 / weightSum;
    }

  // Return the number of calibrators used in this system calibration
  return calHits;

/**********  end sysVisEnsemble  ****************************************/
}



/************************************************************************
*  calib -- local to calib.cc (A.B. 7/27/97)				*
*									*
*  A.  Purpose:	The function calib drives the calibration of V2 data    *
*       output by Mark's vis program.					*
*									*
*  B.  Background and use:						*
*         Other than database construction, calib is where it all       *
*       happens.  calib loops over the targets to be calibrated, the    *
*       list of pseudonyms for each target, and the scans recorded      *
*       under each pseudonym.  For each of these scans the average      *
*       data is computed (avgV2Data), and a system visibility estimate  *
*       is derived from the average of system visibility estimates      *
*       computed from multiple independent calibration objects          *
*       (sysVisEnsemble).                                               *
*         In the case of multiple calibrators, the cross-consistency    *
*       of derived system visibility estimates is checked, and          *
*       mismatches at two levels are flagged for the user as a          *
*       diagnostic that there might be something wrong with either the  *
*       input calibrator information or the calibration object itself.  *
*									*
*  C.  Called by: main							*
*									*
*  D.  Input/output files: stdout & stderr				*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        starList -- the (pointer to the) visibility database object    *
*         (StarList *).                                                 *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration calib  ********************************/
void calib(StarList *starList)
{
/**********  local declarations calib  **********************************/
  int i,month,day,year,numScans = 0;
  double targFrameRate = 0.0;
  double jdAvg,utAvg,delayAvg,wlAvg,V2Avg,V2Var,jitterAvg,ratioCAvg;
  double sV2[MAX_CALIBRATORS],varSV2[MAX_CALIBRATORS],sV4;


  // Output a status line indicating the column contents in the
  // calibfile format
  if(verbose) {
    (void) printf((char *) "# Calibration data format:\n");
    (void) printf((char *) "# name\t#VisPts\tMJD\tm/d/y/h:m:s\t\tUT\tDelay(m)\tWL(um)\tCalVis\tErr\tRawVis\tErr\tSysVis\tErr\t#Cal\tU(m)\tV(m)\tHA(hr)\tFrameRate\tBL/CameraName\n");
  } // end verbose/template line

  // Now do the calibration

  // Loop over the calibration targets
  for(int nTarg = 0; nTarg < nTargets; nTarg++) {
    numScans = 0;
    (void) printf((char *) "\n# Target %s:\n",targetName[nTarg][0]);

    // Loop over the pseudonyms for each target
    for(int pseudonym = 0; pseudonym < nTargetNames[nTarg]; pseudonym++)
      // If we've got hits on the target under this name...
      if(starList->findStar(targetName[nTarg][pseudonym])) {
	int nDataName = 0;
	dataList<WBdata> *targetData = starList->getStarData();
	(void) targetData->gotoHead();
	if(targetData->numberNodes())
	  do {
	    int st, nCal;
	    double tFrameRate = targetData->getCurrent()->get_frameRate();
	    double sV2m,varSV2m;
	    baseline *Baseline = targetData->getCurrent()->get_baseline();

	    // Compute the averaged V2 data for the calibration
	    // target at this time.  If we've got a non-trivial
	    // number of target scans...
	    if ((st = avgV2data(targetData,&month,&day,&year, &jdAvg,&utAvg,&delayAvg,&wlAvg,
				&V2Avg,&V2Var,&jitterAvg, &targFrameRate)) > 1) {
	      // Compute the ensemble system visibility estimate
	      // at this time.  If we've got a good visibility
	      // calibration...
	      // store the current pointer, so we can reset after calibration
	      // this is necessary for --doCalibrators option
	      WBdata* resetNode = targetData->getCurrent();
	      if (nCal = sysVisEnsemble(jdAvg,targetV[nTarg], targFrameRate,starList, &sV2m,&varSV2m)) {
		// ...compute and output the calibrated scan.
		double varTV2 = V2Var / sV2m / sV2m + pow(V2Avg / sV2m / sV2m,2.0) * varSV2m;

		// New 11/6/97  Check the delay value for the target object
		if ((targetCoord[nTarg]) && (delayCheck)) {
		  Vector targ // The instantaneous direction of target
		    = targetCoord[nTarg]->topocentricPos(jdAvg,	Baseline->latitude(), Baseline->longitude(), Baseline->altitude());
		  double delay = Baseline->delay(jdAvg,targ);
		  double delayDiff = delayAvg - delay;

		  if(fabs(delayDiff) > delayCheckTol) {
		    if(verbose) {
		      (void) printf((char *) "# Warning: at %10.4f, %g m measured/expected delay mismatch on target %s\n",
				    jdAvg,delayDiff,targetName[nTarg][pseudonym]);
		      (void) fprintf(stderr,(char *) "Warning: at %10.4f, %g m measured/expected delay mismatch on target %s\n",
				     jdAvg,delayDiff,targetName[nTarg][pseudonym]);
		    } // end delay value/verbose check
		  }
		}

		  // New 4/27/98  Check/apply minimum possible scan uncertainty
		  double scanSigma = sqrt(varTV2);
		  if((usingMinimumScanUncertainty) && (scanSigma < minScanUncertainty)) {
		    scanSigma = minScanUncertainty;
		    if(verbose)
		      (void) printf((char *) "# Minimum scan uncertainty applied at %11.5f\n",jdAvg);
		  }
		
		  char ut_str[LINE_LENGTH];
		  formatted_time(utAvg,ut_str);  // convert utAvg to hr:min:sec with defined format in ut_str

		  // (Finally) Printout the data line
		  (void) printf((char *) "%s\t%d\t%11.5f\t%d/%d/%d/%s",
				targetName[nTarg][pseudonym], st,jdAvg,month,day,year,ut_str);

		  (void) printf((char *) "\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%d",
				utAvg,delayAvg, wlAvg,V2Avg / sV2m,scanSigma,
				V2Avg,sqrt(V2Var), sV2m,sqrt(varSV2m),nCal);

		  // And if we've got a position on the
		  // target, output u-v coordinates and the
		  // like to help our consumers...
		  if(targetCoord[nTarg]) {
		    Vector targ // The instantaneous direction of target
		      = targetCoord[nTarg]->topocentricPos(jdAvg, Baseline->latitude(), Baseline->longitude(), Baseline->altitude());
		    Vector SV = Baseline->skyProjV(jdAvg,targ);
		    (void) printf((char *) "\t%10.6f\t%10.6f\t%g",SV[0],SV[1], Baseline->hourAngle(jdAvg,targ));
		  }

		  // Print out the frame rate and the baseline/camera name
		  if(targetData->getCurrent()->get_cameraName()[0] != '\0') 
  		      (void) printf((char *) "\t%.2f\t%s\n", tFrameRate,targetData->getCurrent()->get_cameraName());
		  else // if PTI there won't be a camera name so use baseline
 		      (void) printf((char *) "\t%.2f\t%s\n", tFrameRate,targetData->getCurrent()->get_baseline()->baselineName());	
		  numScans++;
		  nDataName++;
		} else { // Otherwise, calibration failure...notify the user.
		  if(st > 1) {
		    ///if(verbose)
		    (void) printf("# No suitable calibration found for %10.4f\n",jdAvg);
		  }
		} 
	      targetData->resetCurrent(resetNode);
	    } else {
	      (void) printf("# Warning: Target scan on %s, near UT %f, has only 1 SUM record. Will not calibrate.\n", targetName[nTarg][pseudonym], utAvg);					
	    }	
	  } while (!targetData->endOfList());

	// Target list is finished, printout a summary line for this target name
	(void) printf((char *) "# %d calibrated scans output under name %s\n\n\n",
		      nDataName,targetName[nTarg][pseudonym]);

      } // end psedonym list/data check
	// Pseudonym list is finished, printout a summary line on this target
    (void) printf((char *) "# %d total calibrated scans output on target %s\n\n\n", numScans,targetName[nTarg][0]);
  }

}

int readFluxBiasFile(char *fluxcoeffs) {
  FILE *fptr_coeff = NULL;
  map<string, fluxBiasStruct>::iterator pos;

     if(fluxcoeffs[0] != '\0') {
       if( !(fptr_coeff = fopen(fluxcoeffs, (char *) "r"))) {
          (void) printf((char *) "# >> Error opening flux bias coefficients file: %s\n", fluxcoeffs);
          (void) printf((char *) "# > Correcting for flux dependent bias with built in values\n");
       } else {
          // sscanf the line out of the coefficients file
          // Line format:
          // Baseline/CameraName WLFluxBiasCoeff SpecFluxBiasCoeff NzeroFlux NormFlux
	  // Baseline/CameraName can be KI_K1K2, KI_K1K2S, Camera_FT_PRIM, 
 	  // Camera_FT_SEC, or Camera_FT_LBAND
          // A line in the file should look like:
          // Camera_FT_PRIM 0.063 0.16 20 1000
          char cameraName[256];
          float wlcoeff, speccoeff, nzero, norm;
          int ret;
          char *line, line2[1024], linebuf[1024];

          while(fgets(linebuf, sizeof(linebuf), fptr_coeff) != NULL) {
             strcpy(line2, linebuf);
             line = line2;
	     if(strncmp(line,"#",1) == 0)
		 continue;
             if( (ret = sscanf(line, "%s %f %f %f %f", cameraName, &wlcoeff, &speccoeff, &nzero, &norm)) != 5) {
                (void) printf((char *) "# Error - Could not read flux dependent bias coefficients from file %s for the following line:\n", fluxcoeffs);
		(void) printf((char *) "#       %s\n", line);
		(void) printf((char *) "# Line format should be:\n");
		(void) printf((char *) "Baseline/Camera Name(string) WLFluxBiasCoeff(float) SpecFluxBiasCoeff(float) NzeroFlux(float) NormFlux(float)\n");
		(void) printf((char *) "# e.g. Camera_FT_PRIM 0.063 0.16 20 1000\n");
	  	(void) printf((char *) "# Please fix the line and re-run or run without -fluxBiasCorrection to use the built-in default values.\n\n");
             } else {
		// strip off any "KI_" prefix from the baseline/camera names
		char cam[256];
                if(sscanf(cameraName, "KI_%s", cam) == 1) {
                   strcpy(cameraName, cam);
                }

		// look for the baseline/camera name in the current map
		pos = fluxBiasMap.find(cameraName);

		// if not found, add it to the map
		if(pos == fluxBiasMap.end()) {
		   fluxBiasStruct newBias; 
		   newBias.WLFluxBias = wlcoeff;
		   newBias.SpecFluxBias = speccoeff;
		   newBias.NZero = nzero;
		   newBias.Normflux = norm;
		   newBias.usingDefaults = 0;
	           fluxBiasMap.insert(std::make_pair(cameraName,newBias));   	
		} 
		else {
 		   pos->second.WLFluxBias = wlcoeff;
		   pos->second.SpecFluxBias = speccoeff;
 		   pos->second.NZero = nzero;
		   pos->second.Normflux = norm;
		   pos->second.usingDefaults = 0;
                }
		(void) printf((char *) "# > Found camera name %s in %s\n", cameraName, fluxcoeffs);
             }
            } // while(fgets...)
          fclose(fptr_coeff);

          (void) printf((char *) "# > Correcting flux dependent bias with the following values from %s:\n", fluxcoeffs);
          (void) printf((char *) "# >%15s %13s %11s %9s %10s\n", "Baseline/Camera Name", "WL_Coeff", "Spec_Coeff", "Nzero", "NormFlux");
          for(pos=fluxBiasMap.begin(); pos != fluxBiasMap.end(); ++pos) {
              if(pos->second.usingDefaults == 0)
                 (void) printf((char *) "%20s %15.4f %10.4f %10.1f %10.1f\n", pos->first.c_str(), pos->second.WLFluxBias, pos->second.SpecFluxBias, pos->second.NZero, pos->second.Normflux);
          }
          printf("\n");

          return 0;
       }
     } // if(fluxcoeffs[0] != '\0')

     else { // using default flux bias coeffs
        (void) printf((char *) "# > No flux bias coefficients file specified with -fluxBiasCorrection.\n");
     }

     (void) printf((char *) "# >     Correcting flux dependent bias with the following built-in default values:\n");
     (void) printf((char *) "# >%15s %13s %11s %9s %10s\n", "Baseline/Camera Name", "WL_Coeff", "Spec_Coeff", "Nzero", "NormFlux");
     for(pos=fluxBiasMap.begin(); pos != fluxBiasMap.end(); ++pos) {
          (void) printf((char *) "%20s %15.4f %10.4f %10.1f %10.1f\n", pos->first.c_str(), pos->second.WLFluxBias, pos->second.SpecFluxBias, pos->second.NZero, pos->second.Normflux);
     }
     printf("\n");

     return 0;
} // readFluxBiasFile

/**********  main routine calib  ****************************************/
main (int argc,char **argv)
{
  char baseFile[LINE_LENGTH];
  char fluxFile[LINE_LENGTH];

  // initialize baselineFile and fluxBiasFile globals to blank
  baselineFile.clear();
  fluxBiasFile.clear();

  // Process any command-line options
  int argIndex = processCmdLineOpts(argc,argv);

  // copy baseline & flux file strings into c-strings to pass to functions
  strcpy(fluxFile, fluxBiasFile.c_str());
  strcpy(baseFile, baselineFile.c_str());

  // Trap trivial/bogus invocations
  if (argc < 3)
    (void) printHelpMessage();

  // Read the calibration script on the stars we're going to care
  // about
  (void) readScript(argv[argIndex]);

  // Report to the user on the run directives
  if (verbose)
    (void) printRunDirectives();

  if(usingFluxBiasCorrection) {

      // KI_K1K2 will use the default values for FT_PRIM
      //fluxBiasMap.insert(std::make_pair("K1K2",k1k2));
      fluxBiasMap.insert(std::make_pair("K1K2",ftPrim));
      fluxBiasMap.insert(std::make_pair("K1K2S",k1k2s));
      fluxBiasMap.insert(std::make_pair("Camera_FT_PRIM",ftPrim));
      fluxBiasMap.insert(std::make_pair("Camera_FT_SEC",ftSec));
      fluxBiasMap.insert(std::make_pair("Camera_FT_LBAND",ftLband));
      readFluxBiasFile(fluxFile);
  }

  // Initialize the data structure and read in the data from the
  // summary file(s) listed in the command line
  StarList *starList = new StarList();
  for (int i = argIndex + 1; i < argc; i++)
    {
      int nData = readSumFile(argv[i], baseFile, fluxFile, starList);
      if (verbose)
	(void) printf((char *) "# %d data lines read from %s\n",nData,argv[i]);
    }

  // Now we do the calibration.  Output is to stdout (to be redirected
  // as desired.  Isn't unix great...?)
  (void) calib(starList);

/**********  end program calib  ****************************************/
}
