//<PLAINTEXT>
/* Copyright (c) 1994, California Institute of Technology.  U.S. Government */
/* sponsorship under NASA Contract NAS7-918 is acknowledged.  */

/*++

File:  Vector.cc

Version: 1.00

Date: 4/96

Theme:  Coordinate Vectors

Author:  A. Boden from R.J. Weidner

Modules:
    double operator^(const Vector &a, const Vector &b) // dotproduct
    Vector operator+(const Vector &a, const Vector &b)
    Vector operator-(const Vector &a, const Vector &b)
    Vector operator*(const Vector &a, const Vector &b)
    Vector operator*(const Vector &a, const double b)
    Vector operator*(const double a, const Vector &b)
    Vector operator/(const Vector &a, const Vector &b)
    Vector operator/(const Vector &a, const double b)
    Vector operator/(const double a, const Vector &b)
    Vector Vector::perp(const Vector &w) const
    Vector Vector::operator>>(const Vector &u) const
    Vector Vector::operator<<(const Vector &u) const
    void Vector::order(int *iorder) const
    void Vector::reorder(int *iorder)
    Radians separation(const Vector &u,const Vector &v)
    Radians signed_separation(const Vector &u,const Vector &v,
	const Vector &axis)
    Vector Vector::rotate( const Vector &axis, const Radians &theta) const
    Vector Vector::rotate(Axes axis, const Radians &theta) const
    Vector Vector::rotate(const Euler &e) const
    Vector Vector::rotate(const Quaternion &q) const
    HMatrix operator|(const Vector &a,const Vector &b) // Outer Product

--*/

#include "Vector.h"

/* Vector-Vector operators */
/* dotproduct */
double operator^(const Vector &a,const Vector &b)
  {
    return(a[0]*b[0]+a[1]*b[1]+a[2]*b[2]);
  }

Vector operator+(const Vector &a,const Vector &b)
  {
    return(Vector(a[0]+b[0],a[1]+b[1],a[2]+b[2]));
  }

Vector operator-(const Vector &a,const Vector &b)
  {
    return(Vector(a[0]-b[0],a[1]-b[1],a[2]-b[2]));
  }

Vector operator*(const Vector &a,const Vector &b)
  {
    return(Vector(a[0]*b[0],a[1]*b[1],a[2]*b[2]));
  }


/* Vector-scalar operators */
Vector operator*(const Vector &a,const double b)
  {
    return(Vector(a[0]*b,a[1]*b,a[2]*b));
  }

Vector operator*(const double a,const Vector &b)
  {
    return(Vector(b[0]*a,b[1]*a,b[2]*a));
  }

Vector operator/(const Vector &a,const Vector &b)
  { 
    if (b[0]==0||b[1]==0||b[2]==0)
      return(Vector(0,0,0));
    else
      return(Vector(a[0]/b[0],a[1]/b[1],a[2]/b[2]));
  }

Vector operator/(const Vector &a,const double b)
  { 
    if (b == ZERO)
      return(Vector(ZERO,ZERO,ZERO));
    else
      return(Vector(a[0]/b,a[1]/b,a[2]/b));
  }

Vector operator/(const double a,const Vector &b)
  { 
    if((b[0]  == ZERO) || (b[1] == ZERO) || (b[2] == ZERO)) 
      return(Vector(ZERO,ZERO,ZERO));
    else
      return(Vector(a/b[0],a/b[1],a/b[2]));
  }

Vector Vector::perp(const Vector &w) const
  {

// The components of this vector perpendicular to w
// (There is a unique decomposition of this vector 
// into a sum r + s, where r is
// parallel to w and s is orthogonal to w.  We want the
// component s.)


    Vector temp;
    double bigv = biggest();
    double bigw = w.biggest();
    if(bigv != ZERO)
    {
	if(bigw == ZERO) temp = *this;
	else
	{
	    Vector t = *this / bigv;
	    Vector r = w / bigw;
	    temp = bigv * (t-(t>>r));
	    return(temp);
	}
    }
    else 
    {
	if(bigw != ZERO) temp = w;
	else return(temp);
    }
   
    // Just form any perpendicular
    int i;
    for(i=0;i<3;i++) {if(temp[i] == bigv)break;}
    int j = (i+1) % 3; 
    int k = (j+1) % 3; 
    if(temp[k] > temp[j]) j = k;
    double s = temp[j]; 
    temp[j] = -temp[i];
    temp[i] = temp[j];
    temp[k] = ZERO;

    return(temp);
}

Vector Vector::operator>>(const Vector &u) const
{
    // This vector projected onto u

    double bigv = biggest();
    double bigu = u.biggest();
    if((bigu == ZERO)||(bigv==ZERO)) return(*this);
    
    Vector r = u / bigu;
    Vector t = *this /bigv;
    double dotv = t ^ r;
    double dotu = r ^ r;
    double scale;
    if(dotv ==  ZERO)scale = ZERO;
    else scale = dotv*bigv/dotu;

    Vector temp = r * scale;
    return(temp);
}

Vector Vector::operator<<(const Vector &u) const
{
    // u projected onto this vector

    double bigv = biggest();
    double bigu = u.biggest();
    if((bigu == ZERO)||(bigv==ZERO)) return(u);
    
    Vector r = *this /bigv;
    Vector t = u / bigu;
    double dotu = t ^ r;
    double dotv = r ^ r;
    double scale;
    if(dotu ==  ZERO)scale = ZERO;
    else scale = dotu*bigu/dotv;

    Vector temp = t * scale;
    return(temp);
}


double angle_between(const Vector &u,const Vector &v)
  {
    /* Returns the angle between two vectors */
    return(acos(u.unit()^v.unit()));
  }

