//<PLAINTEXT>
/* Copyright (c) 1994, California Institute of Technology.  U.S. Government */
/* sponsorship under NASA Contract NAS7-918 is acknowledged.  */

/*++

File:   vector.h

Version: 1.00

Date: 4/96

Theme:  Vector Class

Author:  A. Boden adapted from hvector R.J. Weidner

Items:
    class Vector

--*/
#include <math.h>

#ifndef VECTORS_H
#define VECTORS_H
#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif


#define ZERO 0.0

/*class Matrix; // Do not include matrix.h here */

class Vector
{
public:

// Data

//Constructors

    Vector(void)
      { v[0] = v[1] = v[2] = ZERO;}
    Vector(const double a,const double b,const double c)
      { v[0] = a; v[1] = b; v[2] = c;}
    Vector(const double *a)
      { v[0] = *a; v[1] = *(a+1); v[2] = *(a+2);}
    Vector(const double a)
      { v[0] = v[1] = v[2] = a;}
    Vector(const Vector &h)
      { v[0] = h[0];v[1] = h[1];  v[2] = h[2];}
    Vector(const Vector *h)
      { v[0] = (*h)[0];v[1] = (*h)[1];  v[2] = (*h)[2];}
    Vector(const float *a)
      { v[0] = (double)*a;v[1] = (double)*(a+1);v[2] = (double)*(a+2);}

    Vector(const double ra,const double dec)
      { 
	v[0] = cos(ra) * cos(dec);
	v[1] = sin(ra) * cos(dec);
	v[2] = sin(dec);
      }

    Vector(const float ra,const float dec)
      { 
	v[0] = cos(ra) * cos(dec);
	v[1] = sin(ra) * cos(dec);
	v[2] = sin(dec);
      }


//Operators

    // element access

    double &operator[](int i){return(v[i]);} 
    const double &operator[](int i) const {return(v[i]);} 
    const double *data(const int i = 0) const {return(v + i);}
    double *data(const int i = 0) {return(v + i);}

    Vector &operator=(const Vector &h)
    { 
	v[0] = h[0]; v[1] = h[1];  v[2] = h[2];
	return(*this);
    }  
    Vector &operator=(const double d)
    {
	v[0] = d; v[1] = d; v[2] = d;
	return(*this);
    } 

    // unary minus

    Vector operator-(void) const
    { return(Vector(-v[0],-v[1],-v[2]));}

    // scalar combination

    Vector &operator+=(const double a)
      { v[0] +=a; v[1] += a; v[2] += a; return(*this);}  
    Vector &operator-=(const double a)
      { v[0] -=a; v[1] -= a; v[2] -= a; return(*this);}  
    Vector &operator*=(const double a)
      { v[0] *=a; v[1] *= a; v[2] *= a; return(*this);}  
    Vector &operator/=(const double a)
      { 
	v[0] /= a;
	v[1] /= a;
	v[2] /= a;
	return(*this);
      }  

    // Vector combination

    Vector &operator+=(const Vector &a)
    {
	v[0] = v[0]+a[0];
	v[1] = v[1]+a[1];
	v[2] = v[2]+a[2];
	return(*this);
    }
    Vector &operator-=(const Vector &a)
    {
	v[0] = v[0]-a[0];
	v[1] = v[1]-a[1];
	v[2] = v[2]-a[2];
	return(*this);
    }
    Vector &operator*=(const Vector &a)
    {
	v[0] *= a[0];
	v[1] *= a[1];
	v[2] *= a[2];
	return(*this);
    }
    Vector &operator/=(const Vector &a)
    {
      v[0] /= a[0];
      v[1] /= a[1];
      v[2] /= a[2];
      return(*this);
    }

    // Boolean operators

    int operator==(const Vector &a) const
    { return ((v[0]==a.v[0])
	&&(v[1]==a.v[1])&&(v[2]==a.v[2]));}
    int operator>(const Vector &a) const
    { return ((v[0]>a.v[0])
	&&(v[1]>a.v[1])&&(v[2]>a.v[2]));}
    int operator>=(const Vector &a) const
    { return ((v[0]>=a.v[0])
	&&(v[1]>=a.v[1])&&(v[2]>=a.v[2]));}
    int operator<(const Vector &a) const
    { return((v[0]<a.v[0])
	&&(v[1]<a.v[1])&&(v[2]<a.v[2]));}
    int operator<=(const Vector &a) const
    { return ((v[0]<=a.v[0])
	&&(v[1]<=a.v[1])&&(v[2]<=a.v[2]));}
    int operator!=(const Vector &a) const
    { return((v[0]!=a.v[0])
	||(v[1]!=a.v[1])||(v[2]!=a.v[2]));}

    // Cross Product 

    Vector operator%(const Vector &a) const 
    { return(Vector((v[1]*a[2] - v[2]*a[1]), 
		(v[2]*a[0] - v[0]*a[2]),
		(v[0]*a[1] - v[1]*a[0])));}

    // Projection

    Vector operator>>(const Vector &u) const ; // v onto u 
    Vector operator<<(const Vector &u) const ; // u onto v

//Member Functions

    void set(const double a,const double b,const double c)
    {
	v[0] = a; v[1] = b; v[2] = c;
    }
    void set(const double *a)
    {
	v[0] = *a; v[1] = *(a+1); v[2] = *(a+2);
    }
    void set(const float *a)
    {
	v[0] = (double)*a; v[1] = (double)*(a+1); v[2] = (double)*(a+2);
    }

    double normalize(void)
    {
	double t = sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
	v[0] /= t; v[1] /= t; v[2] /= t;
	return(t);
    }

    double norm(void) const
    {
	return(sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]));
    }

    Vector unit(void) const
    {
	double a = v[0];double b=v[1];double c=v[2];
	double t = sqrt(a*a + b*b + c*c);
	a /= t; b /= t; c /= t;
	return(Vector(a,b,c));
    }

    Vector bracket(const Vector &min,
		   const Vector &max) const
    {
	double a = v[0];double b=v[1];double c=v[2];
	if(a < min[0]) a = min[0];
	else if(a > max[0]) a = max[0];
	if(b < min[1]) b = min[1];
	else if(b > max[1]) b = max[1];
	if(c < min[2]) c = min[2];
	else if(c > max[2]) c = max[2];
	return(Vector(a,b,c));
    }

    //sphere extracts the spherical coordinates of a Vector (radians)

    void sphere(double *theta, double *phi)
      {
	*theta = acos(v[2]/sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]));
	*phi = atan2(v[1],v[0]);
      }

    //ra_dec extracts the RA,dec coordinates of a Vector (radians)

    void ra_dec(double *ra, double *dec)
      {
	double t = sqrt(v[0]*v[0] + v[1]*v[1]);
	*ra = atan2(v[1],v[0]);
	*dec = atan2(v[2],t);
    }

    // Returns the elements of this Vector perpendicular to w
    Vector perp(const Vector &w) const; 

    // Metrics 
    double min(void) const
      {    
	double m = v[0];
	if (m > v[1]) m=v[1];
	if (m > v[2]) m=v[2];
	return(m);
      }
    double max(void) const
      {
	double m = v[0];
	if (m < v[1]) m=v[1];
	if (m < v[2]) m=v[2];
	return(m);
      }
    double smallest(void) const
      {
	double m = fabs(v[0]);
	if(m > fabs(v[1])) m=fabs(v[1]);
	if(m > fabs(v[2])) m=fabs(v[2]);
	return(m);
      }
    double biggest(void) const
      {
	double m = fabs(v[0]);
	if(m < fabs(v[1]))m=fabs(v[1]);
	if(m < fabs(v[2]))m=fabs(v[2]);
	return(m);
      }

private:
    double v[3];

//friends
// dotproduct
    friend double operator^(const Vector &a,const Vector &b);

// a + b
    friend Vector operator+(const Vector &a,const Vector &b); 

 // a - b
    friend Vector operator-(const Vector &a,const Vector &b);

 // a * b
    friend Vector operator*(const Vector &a,const Vector &b);

 // a / b
    friend Vector operator/(const Vector &a,const Vector &b);

 // a * b -- scalar
    friend Vector operator*(const Vector &a,const double b);

 // a * b -- scalar
    friend Vector operator*(const double a,const Vector &b);
 // a / b -- scalar
    friend Vector operator/(const Vector &a,const double b);

 // a / b
    friend Vector operator/(const double a,const Vector &b);
 // Outer Product
      /*    friend Matrix operator|(const Vector &a,const Vector &b); */

    friend double angle_between(const Vector &u,const Vector &v);


};

#endif
