/*
   NOVAS-C

   Naval Observatory Vector Astrometry Subroutines
   C Version 1.0
   June, 1996

   U. S. Naval Observatory
   Astronomical Applications Dept.
   3450 Massachusetts Ave., NW
   Washington, DC  20392-5420
*/


#ifndef _NOVAS_
   #include "novas.h"
#endif

#include "math.h"

/********bary_star */

short int bary_star(double tjd,fk5_entry *star,
                    double *ra, double *dec)
/*
------------------------------------------------------------------------

   PURPOSE: Computes the apparent place of a star as viewed from the
      solar system barycenter at date 'tjd', given its mean place,
      proper motion, parallax, and radial velocity for J2000.0.

   REFERENCES:
      Kaplan, G. H. et. al. (1989). Astron. Journ. Vol. 97, 
         pp. 1197-1210.
      Kaplan, G. H. "NOVAS: Naval Observatory Vector Astrometry
         Subroutines"; USNO internal document dated 20 Oct 1988;
         revised 15 Mar 1990.

   INPUT
   ARGUMENTS:
      tjd (double)
         TDT Julian date for apparent place.
      *star (struct fk5_entry)
         Pointer to catalog entry structure (defined in novas.h).

   OUTPUT
   ARGUMENTS:
      *ra (double)
         Apparent right ascension in hours, referred to true equator
         and equinox of date 'tjd'.
      *dec (double)
         Apparent declination in degrees, referred to true equator
         and equinox of date 'tjd'.

   RETURNED
   VALUE:
      (short int)
          0...Everything OK.
         >0...Error code from function 'solarsystem'.

   GLOBALS
   USED:
      T0, FN0

   FUNCTIONS
   CALLED:
      get_earth       novas.c
      starvectors     novas.c
      proper_motion   novas.c
      geocentric      novas.c
      sun_field       novas.c
      aberration      novas.c
      precession      novas.c
      nutate          novas.c
      vector2radec    novas.c

   VER./DATE/
   PROGRAMMER:
      V1.0/01-93/TKB (USNO/NRL Optical Interfer.) Translate Fortran.
      V1.1/07-93/WTH (USNO/AA) Update to C standards.
      V1.2/10-95/WTH (USNO/AA) Added call to 'get_earth'.

   NOTES:
      1. This function is the "C" version of Fortran NOVAS routine
      'apstar'.

------------------------------------------------------------------------ */
{
   double tdb,time2,peb[3],veb[3],pes[3],ves[3],pos1[3],pos2[3],pos3[3],
          pos4[3],pos5[3],pos6[3],pos7[3],vel1[3];
   double dummy,secdiff;

   short int error=0;

/*
   Compute apparent place
*/

   /* 
      Compute the TDB julian date corresponding to tjd
   */
   (void) convert_tdb2tdt(tjd,&dummy,&secdiff);
   tdb = tjd + secdiff / 86400.0;

   /* Compute the star's 3-space position relative to the barycenter at epoch */
   starvectors(star,pos1,vel1);
   proper_motion(T0,pos1,vel1,tdb,pos2);

   /* No barycentric to geocentric (parallax) correction 
      geocentric(pos2,peb, pos3,&time2); */

   /* No Solar monopole correction
      sun_field (pos3,pes, pos4); */

   /* No stellar aberation
      aberration (pos4,veb,time2, pos5); */

   /* No precession and nutation
      precession (T0,pos5,tdb, pos6);
      nutate (tdb,FN0,pos6, pos7); */

   vector2radec(pos2,ra,dec);

   return 0;
}

