#include <stdio.h>
#include <string.h>
#include <math.h>


#include "timeUtil.h"
#include "novas.h"

#include "baseline.h"
#include "coords.h"
#include "PTI.h"
#include "KI.h"



/************************************************************************
*  erToLocal -- local to baseline.cc (A.B. 6/23/98)			*
*									*
*  A.  Purpose:	erToLocal computes a earth-reference (greenwhich-prime) *
*       to local (East, North, Up) rotation matrix.  This rotation      *
*       matrix is temporally static and only a function of our position *
*       on the Earth.                           			*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        longitude, latitude -- the Earth position for the local site   *
*         (radians) (double).                                           *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the Earth-referenced to    *
*       local East-North-Up rotation matrix (RMatrix).		        *
*									*
*************************************************************************/
RMatrix erToLocal(double longitude,double latitude)
{
/**********  begin erToLocal  *******************************************/
  // Vector up the rotational pole
  Vector zhat(0.0,0.0,1.0);

  // Compute and normalize the earth-referenced location vector
  Vector location(longitude,latitude);
  (void) location.normalize();

  // Define earth-referenced unit vectors for East, North, and Up
  Vector East = zhat % location;
  (void) East.normalize();
  Vector Up = location.unit();
  Vector North = Up % East;

  // Rotation matrix -- earth-referenced to location-local
  return RMatrix(East,North,Up);
/**********  end erToLocal  *********************************************/
}


/************************************************************************
*  initBLName -- local to baseline.cc (A.B. 10/2000)			*
*									*
*  A.  Purpose:	initBLName scans our BL database to see whether we      *
*       recognize the 3-vector baseline we've been given.		*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        lBaseline -- ENU baseline 3-vector to check (m) (Vector).      *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration initBLName  ***************************/
char *initBLName(Vector lBaseline)
{
#define MATCH_TOL 0.02 // Require 2% geometry match
/**********  begin initBLName  ******************************************/
  // This is just a sequence of pairwise checks with an unknown at the end
  // Normalize the input vector to simplify comparison
  (void) lBaseline.normalize();

  // Check PTI_NS
  if (fabs(1 - (lBaseline ^ PTI_Baseline_NS) / PTI_Baseline_NS.norm())
      < MATCH_TOL)
    return strdup((const char *) "PTI_NS");

  // Check PTI_NW
  if (fabs(1 - (lBaseline ^ PTI_Baseline_NW) / PTI_Baseline_NW.norm())
      < MATCH_TOL)
    return strdup((const char *) "PTI_NW");

  if (fabs(1 - (lBaseline ^ PTI_Baseline_SW) / PTI_Baseline_SW.norm())
      < MATCH_TOL)
    return strdup((const char *) "PTI_SW");

  // Check KI_K1K2
  if (fabs(1 - (lBaseline ^ KI_Baseline_K1K2) / KI_Baseline_K1K2.norm()) < MATCH_TOL)
    return strdup((const char *) "KI_K1K2");

  // Check KI_S1S2
  if (fabs(1 - (lBaseline ^ KI_Baseline_S1S2) / KI_Baseline_S1S2.norm()) < MATCH_TOL)
    return strdup((const char *) "KI_S1S2");

  // If we haven't hit a match list as "unknown"
  return strdup((const char *) "unknown");

/**********  end initBLName  ********************************************/
}


/************************************************************************
*  baseline constructor -- local to baseline.cc (A.B. 8/97)		*
*									*
*  A.  Purpose:	This baseline constructor makes a baseline object from  *
*       input 3-vector location and local baseline vector.		*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        location -- Earth location 3-vector -- assumes zero altitude   *
*         (arbitrary units) (Vector).                                   *
*        lBaseline -- the specification of the baseline in local ENU    *
*         coordinates (nominally meters) (Vector).                      *
*        dBias -- the interferometer delay bias (nominally meters)      *
*         (double).                                                     *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
baseline::baseline(Vector location,Vector lBaseline,
		   double dBias,const char *Designation)
{
/**********  begin baseline constructor  ********************************/
  // Make the local assignment
  localBline = lBaseline;

  // Initialize the baseline designation
  if (strncmp(Designation,(char *) "Unknown",7) == 0)
    designation = initBLName(lBaseline);
  else
    designation = strdup(Designation);
  
  // Vector up the rotational pole
  // Vector zhat(0.0,0.0,1.0);

  // Normalize the earth-referenced location vector
  (void) location.normalize();
  bPosition = (EARTH_RADIUS) * location;
  siteLongitude = atan2(location[1],location[0]);
  siteLatitude = asin(location[2]);

  // Define earth-referenced unit vectors for East, North, and Up
  // Vector East = zhat % location;
  // (void) East.normalize();
  // Vector Up = location.unit();
  // Vector North = Up % East;

  // Rotation matrix -- earth-referenced to location-local
  // Rotate the baseline vector into the earth-referenced system
  //  Note we are multiplying by the transpose of the earth-centered
  //  to location-local rotation matrix
  LocalToERef = erToLocal(siteLongitude,siteLatitude).transpose();
  earthRefBline = LocalToERef ^ localBline;
  upEarthRef = Vector(siteLongitude,siteLatitude).unit();

  // Make the delay bias assignment
  delayBias = dBias;

  bCov = (double **) 0;

  // Initialize the error flag
  errorFlag = 0;


/**********  end baseline constructor  **********************************/
}


/************************************************************************
*  baseline constructor -- local to baseline.cc (A.B. 8/97)		*
*									*
*  A.  Purpose:	This baseline constructor makes a baseline object from  *
*       the input longitude and latitude of the site, and the local     *
*       baseline vector.						*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        longitude -- the baseline location Earth longitude measured    *
*         positive from the prime meridian East (radians) (double).     *
*        latitude -- the baseline location Earth latitude measured      *
*         positive from the equator North (radians) (double).           *
*        lBaseline -- the specification of the baseline in local ENU    *
*         coordinates (nominally meters) (Vector).                      *
*        dBias -- the interferometer delay bias (nominally meters)      *
*         (double).                                                     *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
baseline::baseline(double longitude,double latitude,Vector lBaseline,
		   double dBias,const char *Designation)
{
/**********  begin baseline constructor  ********************************/
  // Make the local assignment
  localBline = lBaseline;

  // Initialize the baseline designation
  if (strncmp(Designation,(char *) "Unknown",7) == 0)
    designation = initBLName(lBaseline);
  else
    designation = strdup(Designation);

  // Vector up the rotational pole
  // Vector zhat(0.0,0.0,1.0);

  // Make a location vector from the input site latitude and longitude
  // Vector location(longitude,latitude);

  // Normalize the earth-referenced location vector (shouldn't be
  // necessary, but just to be safe...)
  // (void) location.normalize();
  siteLongitude = longitude; // atan2(location[1],location[0]);
  siteLatitude = latitude; // asin(location[2]);
  bPosition = (EARTH_RADIUS) * Vector(siteLongitude,siteLatitude).unit();

  // Define earth-referenced unit vectors for East, North, and Up
  // Vector East = zhat % location;
  // (void) East.normalize();
  // Vector Up = location.unit();
  // Vector North = Up % East;

  // Rotation matrix -- earth-referenced to location-local
  // RMatrix erToLocal(East,North,Up);

  // Rotate the baseline vector into the earth-referenced system
  //  Note we are multiplying by the transpose of the earth-centered
  //  to location-local rotation matrix
  LocalToERef = erToLocal(siteLongitude,siteLatitude).transpose();
  earthRefBline = LocalToERef ^ localBline;
  upEarthRef = Vector(siteLongitude,siteLatitude).unit();

  // Make the delay bias assignment
  delayBias = dBias;

  bCov = (double **) 0;

  // Initialize the error flag
  errorFlag = 0;

/**********  end baseline constructor  **********************************/
}


/************************************************************************
*  baseline constructor -- local to baseline.cc (A.B. 8/97)		*
*									*
*  A.  Purpose:	This baseline constructor makes a baseline object from  *
*       the input longitude and latitude of the site, and the local     *
*       baseline vector.						*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        longitude -- the baseline location Earth longitude measured    *
*         positive from the prime meridian East (radians) (double).     *
*        latitude -- the baseline location Earth latitude measured      *
*         positive from the equator North (radians) (double).           *
*        altitude -- the baseline location altitude measured positive   *
*         from the nominal Earth radius up from the Earth center        *
*         (meters) (double).                                            *
*        lBaseline -- the specification of the baseline in local ENU    *
*         coordinates (nominally meters) (Vector).                      *
*        dBias -- the interferometer delay bias (nominally meters)      *
*         (double).                                                     *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
baseline::baseline(double longitude,double latitude,double altitude,
		   Vector lBaseline,double dBias,
		   const char *Designation)
{
/**********  begin baseline constructor  ********************************/
  // Make the local baseline assignment
  localBline = lBaseline;

  // Initialize the baseline designation
  if (strncmp(Designation,(char *) "Unknown",7) == 0)
    designation = initBLName(lBaseline);
  else
    designation = strdup(Designation);

  // Normalize the earth-referenced location vector (shouldn't be
  // necessary, but just to be safe...)
  // (void) location.normalize();
  siteLongitude = longitude; // atan2(location[1],location[0]);
  siteLatitude = latitude; // asin(location[2]);
  siteAltitude = altitude;
  bPosition = (EARTH_RADIUS + siteAltitude) * Vector(siteLongitude,siteLatitude).unit();

  // Rotate the baseline vector into the earth-referenced system
  //  Note we are multiplying by the transpose of the earth-centered
  //  to location-local rotation matrix
  LocalToERef = erToLocal(siteLongitude,siteLatitude).transpose();
  earthRefBline = LocalToERef ^ localBline;
  upEarthRef = Vector(siteLongitude,siteLatitude).unit();

  // Make the delay bias assignment
  delayBias = dBias;

  // No baseline covariance matrix
  bCov = (double **) 0;

  // Initialize the error flag
  errorFlag = 0;

/**********  end baseline constructor  **********************************/
}


/************************************************************************
*  baseline constructor -- local to baseline.cc (A.B. 8/97)		*
*									*
*  A.  Purpose:	This baseline constructor makes a baseline object from  *
*       the contents of the input filename.                             *
*									*
*  B.  Background and use:						*
*       current baseline file format:                                   *
* # Earth baseline position (long/latitude/altitude): -116.8633   33.3567  1680
* # Best-fit model baseline (ENUbias) (m):  -37.117229  -103.264482  3.318304  -12.928528
* # RMS residual delay (m): 1.76047e-05  (10 out of 258 delays excluded in 14 iterations)
* # baseline error est (m): 2.5666e-05  1.89944e-05  0.000187795  0.000185815
* # Full covariance matrix (m^2):
* # cov  6.58746e-10  -9.66053e-11  -4.05506e-09  4.03286e-09  
* # cov  -9.66053e-11  3.60787e-10  8.5615e-10  -8.54825e-10  
* # cov  -4.05506e-09  8.5615e-10  3.52671e-08  -3.48918e-08  
* # cov  4.03286e-09  -8.54825e-10  -3.48918e-08  3.45272e-08  
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        filename -- the (pointer to the) character filename to open    *
*         and read for the baseline contents.                           *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
baseline::baseline(char *filename,const char *Designation)
{
/**********  local declarations baseline constructor  *******************/
#define LINE_LENGTH 132
#define MIN_TOTAL 20
#define MIN_ACCEPT 0.6

  char line[LINE_LENGTH],bogus[LINE_LENGTH];
  int exclusions,total;
  double rms,dDouble,east,north,up,bias;
  FILE * fptr;

/**********  begin baseline constructor  ********************************/
  // Initialize the error flag
  errorFlag = 0;

  if (!(fptr = fopen(filename,"r")))
    {
      // (void) fprintf(stderr,"### Error -- baseline file %s not found\n",filename);
      errorFlag = 1;
    } // end file open/check

  // Currently read my bFit baseline file A.B. 7/23/98
  // Add reading of .baseline PTI RT system file A.B. 11/00

  // Template PTI .baseline file looks like:
  // BASELINE:  11/13/100   1.490556 X:  -81.684519 Y:  -28.213514 Z:    3.103883 C:    0.026365 (North,East,Up)

  // Added reading of KI .baseline files A.B. 11/24/02
  // Template KI .baseline file:
  // BASELINE: 11/21/2002 0.00 E: 51.703989 N: 67.455287 U: 0.001524 C: 40.51


  // Retrieve and store the baseline earth position
  (void) fgets(line,LINE_LENGTH,fptr);

  int tmp;
  if (((tmp = sscanf(line,"BASELINE:  %s%lf X:%lf Y:%lf Z:%lf C:%lf (North,East,Up",
		     bogus,&dDouble,&east,&north,&up,&bias)) == 6)
      || ((tmp = sscanf(line,"BASELINE: %s%lf E:%lf N:%lf U:%lf C:%lf",
			bogus,&dDouble,&east,&north,&up,&bias)) == 6))
    { // We've got a PTI or KI RT baseline file
      localBline[0] = east;
      localBline[1] = north;
      localBline[2] = up;
      delayBias = bias;

      if ((strchr(line,'X') != NULL)  // Decide if we have a PTI or KI .baseline file
	  && (strchr(line,'Y') != NULL)  // PTI files use X Y Z labels
	  && (strchr(line,'Z') != NULL)) // KI files use E N U labels
	{
	  siteLongitude = PTI_LONGITUDE; // The RT file contains _only_ the 3-vector and bias
	  siteLatitude = PTI_LATITUDE;   //  so we are left to hardwire the location
	  siteAltitude = PTI_ALTITUDE;
	}
      else
	{
	  siteLongitude = KI_LONGITUDE; // The RT file contains _only_ the 3-vector and bias
	  siteLatitude = KI_LATITUDE;   //  so we are left to hardwire the location
	  siteAltitude = KI_ALTITUDE;
	}

      bPosition = (EARTH_RADIUS + siteAltitude) * Vector(siteLongitude,siteLatitude).unit();

      // Rotate the local baseline vector into the earth-referenced system
      // Note we are multiplying by the transpose of the earth-centered to
      // location-local rotation matrix
      LocalToERef = erToLocal(siteLongitude,siteLatitude).transpose();
      earthRefBline = LocalToERef ^ localBline;
      upEarthRef = Vector(siteLongitude,siteLatitude).unit();

      // No covariance (of course)
      bCov = (double **) NULL;

    } // end PTI RT baseline file
  else // Otherwise, processing bFit output file
    {

      // If we're still around, parse a bFit file...
      if (sscanf(line,"# Earth baseline position (long/latitude/altitude): %lf  %lf  %lf\n",
		 &siteLongitude,&siteLatitude,&siteAltitude) != 3)
	{
	  // (void) fprintf(stderr," Error reading baseline file %s\n",filename);
	  errorFlag = 1;
	}
      siteLongitude *= DEG_TO_RAD;
      siteLatitude *= DEG_TO_RAD;
      bPosition = (EARTH_RADIUS + siteAltitude) * Vector(siteLongitude,siteLatitude).unit();


  // Retrieve the baseline vector
      (void) fgets(line,LINE_LENGTH,fptr);
      if (sscanf(line,"# Best-fit model baseline (%s (m):  %lf  %lf  %lf  %lf\n",
		 bogus,&localBline[0],&localBline[1],&localBline[2],&delayBias) != 5)
	{
	  // (void) fprintf(stderr," Error reading baseline file %s\n",filename);
	  errorFlag = 1;
	}

      // Rotate the local baseline vector into the earth-referenced system
      // Note we are multiplying by the transpose of the earth-centered to
      // location-local rotation matrix
      LocalToERef = erToLocal(siteLongitude,siteLatitude).transpose();
      earthRefBline = LocalToERef ^ localBline;
      upEarthRef = Vector(siteLongitude,siteLatitude).unit();

      // Read the rejection rate for this baseline estimate
      // # RMS residual delay (m): 1.76047e-05  (10 out of 258 delays excluded in 14 iterations)
      (void) fgets(line,LINE_LENGTH,fptr);
      if (sscanf(line,"# RMS residual delay (m): %lf (%d out of %d delays excluded",
		 &rms,&exclusions,&total) != 3)
	{
	  // (void) fprintf(stderr," Error reading baseline file %s\n",filename);
	  errorFlag = 1;
	}
      if ((total < MIN_TOTAL)
	  || ((1.0 - ((double) exclusions / (double) total)) < MIN_ACCEPT))
	{
	  // (void) fprintf(stderr,"# Baseline solution contained in %s of poor quality\n",filename);
	  errorFlag = 1;
	}


      // Retrieve the baseline estimate covariance.  Skip over some
      // uninteresting lines (from the computational perspective)
      int i;
      for (i = 0; i < 2; i++)
	(void) fgets(line,LINE_LENGTH,fptr);
 
      bCov = new double *[4];
      for (i = 0; i < 4; i++)
	{
	  (void) fgets(line,LINE_LENGTH,fptr);
	  bCov[i] = new double[4];
	  if (sscanf(line,"# cov %lf%lf%lf%lf",
		     &bCov[i][0],&bCov[i][1],&bCov[i][2],&bCov[i][3]) != 4)
	    {
	      // (void) fprintf(stderr," Error reading baseline file %s\n",filename);
	      errorFlag = 1;
	    } // end if sscanf
	} // end for loop

    } // end else reading bFit file

  (void) fclose(fptr); // Close the file

  // Initialize the baseline designation
  if (strncmp(Designation,(char *) "Unknown",7) == 0)
    designation = initBLName(localBline);
  else
    designation = strdup(Designation);
}

/************************************************************************
*  baseline constructor -- (hjb 6/05)	                         	*
* takes a pointer to an open .binfo or .baseline file                   *
* users should call this constructor repeatedly until the baseline      *
* object they get back has a non-zero errorFlag                         *
*									*
* this should work for .binfo files of all epochs, old style pti        *
* .baseline files, and the one keck .baseline format i every knew about *
*************************************************************************/
baseline::baseline(FILE *fptr)
{
#define LINE_LENGTH 132
#define MIN_TOTAL 20
#define MIN_ACCEPT 0.6

  enum Sites {PTI, KI, UNKNOWN};
  enum btypes {BASELINE, BLINE};
  Sites site = UNKNOWN;
  char* t = NULL;

  char line[LINE_LENGTH], date[LINE_LENGTH], desig[LINE_LENGTH], bogus[LINE_LENGTH];
  int exclusions,total;
  double rms, time, east, north, up, bias;
  
  // initialize the site long so we know if we set it already or not
  siteLongitude = -99.0;
  
  // more initialization
  errorFlag = 0;
  btypes filetype = BASELINE;
  desig[0] = '\0';

  if(fptr == NULL)
    errorFlag = 1;

  int tmp;
  while(1) {
    
    if( (t = fgets(line,LINE_LENGTH,fptr)) == NULL) {
      errorFlag = 1;
      // this will be eof
      break;
    }

    if((tmp = sscanf(line, "BASELINE: %s%lf E:%lf N:%lf U:%lf C:%lf %s",
		     date, &time, &east, &north, &up, &bias, desig)) == 7) {
      // STANDARD .BASELINE FILE
      // figure out the telescope by the designator
      break;
    } else if((tmp = sscanf(line, "BASELINE:  %s%lf X:%lf Y:%lf Z:%lf C:%lf (North,East,Up",
			    date, &time, &east, &north, &up, &bias)) == 6) {
      // OLD STYLE PTI .BASELINE FILE
      // As always, the (North, East, Up string is just plain wrong: ignore
      site = PTI;
      // Do a coordinates look up to find designator string
      //            X(E)  Y(N)  Z(U)
      // PTI_NS    -37  -103    3
      // PTI_NW    -81  -28     3
      // PTI_SW    -44   75     0
      if(north > 0)
	strncpy(desig, "PTI_SW", sizeof(desig));
      else if(east < -70)
	strncpy(desig, "PTI_NW", sizeof(desig));
      else 
	strncpy(desig, "PTI_NS", sizeof(desig));
      break;
    } else if((tmp = sscanf(line,"# Earth baseline position (long/latitude/altitude): %lf  %lf  %lf",
			    &siteLongitude,&siteLatitude,&siteAltitude)) == 3) {
      // OLD STYLE BLINE FILE
      // this is the tricky one to figure out a desig for
      // first figure out the site from the long/lat/alt
      // then try to match the baseline coords
      // it is possible that this is a bad fit - e.g. fit between
      // two baselines, and so it should generate a science error (see below)
      filetype = BLINE;
      break;
    } else if((tmp=sscanf(line, "# Baseline Name: %s", desig)) == 1) {
      // NEW STYLE BLINE FILE (POSSIBLE MULTI-BASELINE)
      (void) fgets(line, LINE_LENGTH, fptr);
      sscanf(line, "# Date (UTC): %s", date);
      (void) fgets(line, LINE_LENGTH, fptr);
      sscanf(line, "# Time (hr): %lf", &time);
      (void) fgets(line, LINE_LENGTH, fptr);
      if((tmp=sscanf(line,"# Earth baseline position (long/latitude/altitude): %lf  %lf  %lf",
		     &siteLongitude,&siteLatitude,&siteAltitude)) != 3) 
	errorFlag = 1;
      filetype = BLINE;
      break;
    } // else {
      // we found a comment or some other not relevant thing. Try again;
    //}
  }

  if(! errorFlag) {
    if(filetype == BLINE) {
      siteLongitude *= DEG_TO_RAD;
      siteLatitude *= DEG_TO_RAD;
      
      // Retrieve the baseline vector ... next line
      (void) fgets(line,LINE_LENGTH,fptr);
      // this line might say either NEUbias or ENUbias in the first string position.
      // NEUbias is always WRONG - the files are always in ENUbias order.Generally
      // just assume everything is in ENUbias and you'll be mostly right.
      if((tmp=sscanf(line,"# Best-fit model baseline (%s (m):  %lf  %lf  %lf  %lf\n",
		     &bogus, &east, &north, &up, &bias)) != 5) {
	errorFlag = 1;
      }
      
      // Read the rejection rate for this baseline estimate
      (void) fgets(line,LINE_LENGTH,fptr);
      if((tmp=sscanf(line,"# RMS residual delay (m): %lf (%d out of %d delays excluded",
		     &rms,&exclusions,&total)) != 3) {
	errorFlag = 1;
      }
      if((total < MIN_TOTAL) ||
	 ((1.0 - ((double) exclusions / (double) total)) < MIN_ACCEPT)) {
	// This is a science case for an error - not a technical error
	// but they're probably going to continue being treated as if they
	// were technical errors
	errorFlag = 2;
      }    
      
      // skip two lines
      (void) fgets(line, LINE_LENGTH, fptr);
      (void) fgets(line, LINE_LENGTH, fptr);
      // get baseline estimate covariance
      bCov = new double *[4];
      for(int i = 0; i < 4; i++) {
	(void) fgets(line,LINE_LENGTH,fptr);
	bCov[i] = new double[4];
	if((tmp=sscanf(line,"# cov %lf%lf%lf%lf", 
		       &bCov[i][0],&bCov[i][1],&bCov[i][2],&bCov[i][3])) != 4) {
	  errorFlag = 1;
	}
      }
    } else
    // No covariance in .baseline files
      bCov = (double **) NULL;


    localBline[0] = east;
    localBline[1] = north;
    localBline[2] = up;
    delayBias = bias;
    
    if(desig[0] == '\0') {
      designation = initBLName(localBline);
    } else {
      designation = strdup(desig);
    }
  
    if(site == UNKNOWN) {
      // try to figure site from designator
      if(strncmp(desig, "PTI", 3) == 0)
	site = PTI;
      else if (strncmp(desig, "KI", 2) == 0)
	site = KI;
    }

    if(site == PTI) {  // These are all in radians already
      if(siteLongitude == -99.0) {
	siteLongitude = PTI_LONGITUDE;
	siteLatitude = PTI_LATITUDE;
	siteAltitude = PTI_ALTITUDE;
      }
    } else if (site == KI) {
      if(siteLongitude == -99.0) {
	siteLongitude = KI_LONGITUDE; 
	siteLatitude = KI_LATITUDE;   
	siteAltitude = KI_ALTITUDE;
      }
    }

    bCov = (double **) NULL;
    
    bPosition = (EARTH_RADIUS + siteAltitude) * Vector(siteLongitude,siteLatitude).unit();

    // Rotate the local baseline vector into the earth-referenced system
    // Note we are multiplying by the transpose of the earth-centered to
    // location-local rotation matrix
    LocalToERef = erToLocal(siteLongitude,siteLatitude).transpose();
    earthRefBline = LocalToERef ^ localBline;
    upEarthRef = Vector(siteLongitude,siteLatitude).unit();
  }

}


/************************************************************************
*  baseline destructor -- local to baseline.cc (9/2/98)			*
*									*
*  A.  Purpose:	This destructor takes care of any memory allocation     *
*       made in baseline construction.					*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  declaration destructor baseline  ***************************/
baseline::~baseline(void)
{
/**********  begin destructor baseline  *********************************/
  // At this point the only thing we need worry about is the memory
  // allocated (if any) for the baseline covariance matrix
  if (bCov)
    {
      for (int i = 0; i < 4; i++)
	delete bCov[i];
      delete bCov;
    }

/**********  end destructor baseline  ***********************************/
}




// Utility functions


// Greenwich apparent siderial time function
double gast(double jd)
{
  double mobl,tobl,eq,dpsi,deps,gastTmp;

  // Compute the _apparent_ siderial time, complete with precession

  // Make the optimal fractional split for maximum precision in the
  // siderial time routine
  double jdInt = ((double) ((long) jd));
  double jdFrac = jd - jdInt;

  /* Test/fix if this is a Modified Julian date */
  jdInt = fixJD(jdInt);
  jd = fixJD(jd);

  // Compute the terms for the equation of equinoxes
  (void) earthtilt(jd,&mobl,&tobl,&eq,&dpsi,&deps);
  // ...And pass them to the NOVAS siderial time function
  (void) sidereal_time(jdInt,jdFrac,eq,&gastTmp);

  return gastTmp;
} // end gast
  


/************************************************************************
*  ER_TO_INERTIAL -- local to baseline.cc (A.B. archival).		*
*									*
*  A.  Purpose:	The function ER_TO_INERTIAL computes the instantaneous  *
*       Earth-referenced-to-inertial coordinate rotation matrix at a    *
*       given epoch specified by an input Julian date.			*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the input (possibly modified) Julian date (double).      *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the Earth-referenced-to    *
*       inertial rotation matrix (RMatrix)              		*
*									*
*************************************************************************/
/**********  function declaration  ER_TO_INERTIAL  **********************/
RMatrix ER_TO_INERTIAL(double jd)
{
  // Get the Earth rotation angle (Greenwich apparent siderial time)
  // at this Julian date
  double epAngle = gast(jd) * 15.0 * DEG_TO_RAD;

  // return the rotation matrix to go from earth-referenced to
  // quantities inertial at this epoch
  return RMatrix(cos(epAngle),-sin(epAngle),0.0,
		 sin(epAngle),cos(epAngle),0.0,
		 0.0,0.0,1.0);

} // end ER_TO_INERTIAL



/************************************************************************
*  ER_TO_INERTIAL_DOT -- local to baseline.cc (A.B. archival).		*
*									*
*  A.  Purpose:	The function ER_TO_INERTIAL computes the instantaneous  *
*       Earth-referenced-to-inertial coordinate rotation matrix at a    *
*       given epoch specified by an input Julian date.			*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the input (possibly modified) Julian date (double).      *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the Earth-referenced-to    *
*       inertial rotation matrix time rate of change (RMatrix) (1/sec). *
*									*
*************************************************************************/
/**********  function declaration  ER_TO_INERTIAL_DOT  ******************/
RMatrix ER_TO_INERTIAL_DOT(double jd)
{
#define SEC_PER_DAY (24.0 * 60.0 * 60.0) // Approximate right now

  // Get the Earth rotation angle (Greenwich apparent siderial time)
  // at this Julian date
  double epAngle = gast(jd) * 15.0 * DEG_TO_RAD;
  double dAngledSec = TWO_PI / SEC_PER_DAY;

  // return the rotation matrix to go from earth-referenced to
  // quantities inertial at this epoch
  // return RMatrix(cos(epAngle),-sin(epAngle),0.0,
  //                sin(epAngle),cos(epAngle),0.0,
  //                0.0,0.0,1.0);

  return RMatrix(-dAngledSec * sin(epAngle),-dAngledSec * cos(epAngle),0.0,
		 dAngledSec * cos(epAngle),-dAngledSec * sin(epAngle),0.0,
		 0.0,0.0,0.0);

} // end ER_TO_INERTIAL_DOT



// Baseline class member functions

// Baseline projection member functions

/************************************************************************
*  member function inertial -- local to baseline.cc (A.B. 8/19/97)	*
*									*
*  A.  Purpose:	The member function inertial computes the inertial      *
*       baseline 3-vector at the input Julian date.			*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the input (possibly modified) Julian date (double).      *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the inertial 3-vector      *
*       of the baseline at the input epoch (meters) (Vector).		*
*									*
*************************************************************************/
/**********  member function declaration inertial  **********************/
Vector baseline::inertial(double jd)
{
/**********  begin inertial *********************************************/
  // Essentially trivial computation
  return ER_TO_INERTIAL(jd) ^ earthRefBline;

/**********  end inertial  **********************************************/
}


/************************************************************************
*  member function inertialDot -- local to baseline.cc (A.B. 8/19/97)	*
*									*
*  A.  Purpose:	The member function inertial computes the inertial      *
*       baseline 3-vector time rate of change at the input Julian date. *
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the input (possibly modified) Julian date (double).      *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the inertial 3-vector      *
*       time rate of change of the baseline at the input epoch          *
*       (meters/sec) (Vector).	                                        *
*									*
*************************************************************************/
/**********  member function declaration inertialDot  *******************/
Vector baseline::inertialDot(double jd)
{
/**********  begin inertialDot ******************************************/
  // Essentially trivial computation
  return ER_TO_INERTIAL_DOT(jd) ^ earthRefBline;

/**********  end inertialDot  *******************************************/
}


/************************************************************************
*  member function skyProjV -- local to baseline.cc (A.B. 8/97)		*
*									*
*  A.  Purpose:	The function skyProjV computes the projection of the    *
*       baseline vector on the sky at a given location in local ra,dec  *
*       (East, North, Up, or u,v,w) coordinates.			*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the input (possibly modified) Julian date (double).      *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the local ra,dec           *
*       projection vector for the instantaneous baseline vector (m)     *
*       (Vector).  The are directly related to the u-v-w coordinates 	*
*       for the observation (one must divide by lambda to get to uvw).  *
*									*
*************************************************************************/
/**********  member function declaration skyProjV  **********************/
Vector baseline::skyProjV(double jd,Vector star)
{
/**********  begin skyProjV  ********************************************/
  // Just to be safe, normalize the input intertial star direction...
  (void) star.normalize();

  // Compute the inertial baseline vector at this epoch
  Vector myInertial = inertial(jd);

  // Calculate inertial unit vectors in the direction of increasing ra
  // (local East) and dec (local North)...
  Vector zhat(0.0,0.0,1.0);

  // Compute the other two unit vectors 
  Vector ra = zhat % star;
  (void) ra.normalize();
  Vector dec = star % ra;

  // Return the result in ra,dec (East, North/uvw) coordinates...
  return Vector(myInertial ^ ra,myInertial ^ dec,myInertial ^ star);

/**********  end skyProjV  **********************************************/
}


// Angle member functions

/************************************************************************
*  member function localSiderial -- local to baseline.cc (A.B. 8/97)	*
*									*
*  A.  Purpose: The member function localSiderial computes the          *
*       instantaneous local siderial time at an input Julian date.	*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the input (possibly modified) Julian date (double).      *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the local siderial time at *
*       the input Julian date (hr) (double).           		        *
*									*
*************************************************************************/
/**********  member function declaration localSiderial  *****************/
double baseline::localSiderial(double jd)
{
/**********  begin localSiderial  ***************************************/
  // Compute the local siderial time, mapped into the principle range
  double tmp = gast(jd) + (siteLongitude * RAD_TO_DEG / 15.0);

  // Make it a positive-definite quantity
  if (tmp < 0)
    tmp += 24.0;

  return tmp;

/**********  end localSiderial  *****************************************/
}


/************************************************************************
*  member function hourAngle -- local to baseline.cc (8/97)		*
*									*
*  A.  Purpose:	The member function hourAngle computes the hourAngle    *
*       for the input sky location at the input Julian date.		*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the input (possibly modified) Julian date (double).      *
*        star -- the intertial unit vector in the direction of the      *
*         star (dimensionless) (Vector).                                *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the hourAngle for the      *
*       given sky location at the given Julian date (hrs) (double).     *
*									*
*************************************************************************/
/**********  member function declaration hourAngle  *********************/
double baseline::hourAngle(double jd,Vector star)
{
/**********  begin hourAngle  *******************************************/
  // Star's right ascention in hours
  double starRA = atan2(star[1],star[0]);
  if (starRA < 0)
    starRA += TWO_PI;
  starRA *= RAD_TO_DEG / 15.0;

  // the difference is the return value (negative if the star is pre-transit)
  double dif = localSiderial(jd) - starRA;

  // Map into the range -12 <= ha <= 12
  if (dif > 12.0)
    dif -= 24.0;
  if (dif < -12.0)
    dif += 24.0;

  return dif;

/**********  end hourAngle  *********************************************/
}


/************************************************************************
*  member function zenithAngle -- local to baseline.cc (A.B. 8/97)	*
*									*
*  A.  Purpose: The member function zenithAngle computes the apparent   *
*       zenith angle for a given sky location at a particular Julian    *
*       date.  This calculation DOES NOT take atmospheric refraction    *
*       into account.							*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the input (possibly modified) Julian date (double).      *
*        star -- the intertial unit vector in the direction of the      *
*         star (dimensionless) (Vector).                                *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the apparent zenith angle  *
*       for an input sky location at a given Julian date (rad) (double).*
*									*
*************************************************************************/
/**********  member function declaration zenithAngle  *******************/
double baseline::zenithAngle(double jd,Vector star)
{
/**********  begin zenithAngle  *****************************************/
  // Inertial up vector at this epoch
  Vector upI = ER_TO_INERTIAL(jd) ^ upEarthRef;

  // Return the angular separation (no atmospheric refraction!)
  (void) star.normalize();
  return acos(upI ^ star);

/**********  end zenithAngle  *******************************************/
}


/************************************************************************
*  member function parallacticAngle -- local to baseline.cc (A.B. 8/98) *
*									*
*  A.  Purpose:	The member function paralacticAngle computes the        *
*       instantaneous parallactic angle for a given intertial position  *
*       at a given jd. 							*
*									*
*  B.  Background and use: I stole the code for this routine from Pat   *
*       Wallace's routine slaPa (pa.c) in his slalib package.	Here	*
*       are his notes for the routine:                                  *
*                                                                       *
*>  HA, Dec to Parallactic Angle.                                       *
*>                                                                      *
*>  (double precision)                                                  *
*>                                                                      *
*>  Given:                                                              *
*>     ha     d     hour angle in radians (geocentric apparent)         *
*>     dec    d     declination in radians (geocentric apparent)        *
*>     phi    d     observatory latitude in radians (geodetic)          *
*>                                                                      *
*>  The result is in the range -pi to +pi                               *
*>                                                                      *
*>  Notes:                                                              *
*>                                                                      *
*>  1)  The parallactic angle at a point in the sky is the position     *
*>      angle of the vertical, i.e. the angle between the direction to  *
*>      the pole and to the zenith.  In precise applications care must  *
*>      be taken only to use geocentric apparent HA,Dec and to consider *
*>      separately the effects of atmospheric refraction and telescope  *
*>      mount errors.                                                   *
*>                                                                      *
*>  2)  At the pole a zero result is returned.                          *
*>                                                                      *
*>  P.T.Wallace   Starlink   1 April 1994                               *
*                                                                       *
*   In case Pat's description isn't completely clear, the parallactic   *
*   angle can be thought of as the position angle of the local zenith   *
*   as measured from the instantaneous apparent position of the star.   *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the input (possibly modified) Julian date (double).      *
*        star -- the intertial unit vector in the direction of the      *
*         star (dimensionless) (Vector).                                *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the computed value of the  *
*       parallactic angle (radians) (double).           		*
*									*
*************************************************************************/
/**********  member function parallacticAngle  **************************/
double baseline::parallacticAngle(double jd,Vector star)
{
/**********  begin parallacticAngle  ************************************/

  double dec = asin(star[2]);
  double ha = hourAngle(jd,star) * 15.0 * DEG_TO_RAD;
 
  double clat = cos(siteLatitude);
  double sqsz = clat * sin(ha);
  double cqsz = sin(siteLatitude) * cos(dec) - sin(dec) * clat * cos(ha);

  if ((sqsz == 0.0) && (cqsz == 0.0))
    cqsz = 1.0;
  return atan2(sqsz,cqsz);

/**********  end parallacticAngle  **************************************/
}


/************************************************************************
*  delay -- local to baseline.cc (A.B. 11/30/97)			*
*									*
*  A.  Purpose:	The function delay computes the expected                *
*       interferometric delay at an input epoch based on input          *
*       coordinates.			                                *
*									*
*  B.  Background and use: This function is used to compute the model   *
*       delay value.  It assumes no atmospheric refraction effects, and *
*       therefore implies a plane-parallel atmospheric model.           *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the (possibly modified) julian date at epoch.            *
*        star -- the inertial apparent position vector of the star in   *
*         question (Vector).                                            *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the expected delay value   *
*       (meters) (double).                                   		*
*									*
*************************************************************************/
/**********  function declaration delay  ********************************/
double baseline::delay(double jd,Vector star)
{
/**********  begin delay  ***********************************************/
  // Pretty simple, huh?
  return ((star ^ inertial(jd)) + delayBias);

/**********  end delay  *************************************************/
}


/************************************************************************
*  delayRate -- local to baseline.cc (A.B. 12/99)			*
*									*
*  A.  Purpose:	The function delayRate computes the expected            *
*       interferometric delay rate at an input epoch based on input     *
*       coordinates.			                                *
*									*
*  B.  Background and use: This function is used to compute the model   *
*       delay value.  It assumes no atmospheric refraction effects, and *
*       therefore implies a plane-parallel atmospheric model.           *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the (possibly modified) julian date at epoch.            *
*        star -- the inertial apparent position vector of the star in   *
*         question (Vector).                                            *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the expected delay rate    *
*       value (meters) (double).                                   	*
*									*
*************************************************************************/
/**********  function declaration delay  ********************************/
double baseline::delayRate(double jd,Vector star)
{
/**********  begin delayRate  *******************************************/
  // Pretty simple, huh?
  return (star ^ inertialDot(jd));

/**********  end delayRate  *********************************************/
}



/************************************************************************
*  member function locationValocity -- local to baseline.cc (A.B. 9/98)	*
*									*
*  A.  Purpose:	The member function locationVelocity computes the       *
*       instantaneous observatory velocity in inertial (ra,dec)         *
*       coordinates.							*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the (possibly modified) julian date at epoch.            *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the instantaneous          *
*       observatory velocity vector (m/sec) in inertial coordinates     *
*       (relative to an instantaneous ra,dec system) (Vector).		*
*       In the present implementation the calcaulation considers only   *
*       earth rotation.                                                 *
*									*
*************************************************************************/
/**********  member function declaration locationVelocity  **************/
Vector baseline::locationVelocity(double jd)
{
/**********  begin locationVelocity  ************************************/

  jd = fixJD(jd);

  // Earth rotational velocity
  Vector omega(0,0,EARTH_OMEGA);

  // Instantaneous baseline inertial position
  Vector bPos = ER_TO_INERTIAL(jd) ^ bPosition;

  // Motion of the observatory relative to earth center is given by
  // the cross product
  Vector bVel = omega % bPos;


  // Compute the earth barycentric kinematics through a novas call
  double tmp;
  Vector earthPos,earthVel,dum1,dum2;
  (void) get_earth(jd,3,&tmp,&(earthPos[0]),&(earthVel[0]),&(dum1[0]),&(dum2[0]));
  earthVel *= 1.49597870e+11 / (3600 * 24); // Go from AU/day to m/sec

  // (For now) the result is just the earth rotation part
  return earthVel;

/**********  end locationVelocity  **************************************/
}
