#define _BASELINE_

#ifndef _ANGLES_
#include "angles.h"
#endif

#ifndef VECTORS_H
#include "Vector.h"
#endif

#ifndef RMATRIX_H
#include "RMatrix.h"
#endif


#define EARTH_RADIUS 6.378140E6      // Earth equitoral radius in meters
#define EARTH_OMEGA (TWO_PI / 24.0 / 3600.0)  // Earth angular velocity in rad/sec

// #define MJD_2000 51544.5  // MJD of J2000 reference epoch

// The baseline class collects functions and utilities for computing
// interesting and useful things for ground-based interferometers.


class baseline
{
private:
  char *designation;
  int errorFlag;
  double siteLongitude,siteLatitude,siteAltitude,delayBias;
  Vector bPosition,localBline,earthRefBline,upEarthRef;
  RMatrix LocalToERef;
  double **bCov;


public:

  // Baseline constructors

  // Constructor from earth location vector, ENU baseline vector,
  // optional bias
  baseline(Vector location,Vector lBaseline,
	   double dBias = 0.0,
	   const char *blDesignation = (const char *) "Unknown");

  // Constructor from earth longitude/latitude pair, ENU baseline
  // vector, optional bias
  baseline(double longitude,double latitude,Vector lBaseline,
	   double dBias = 0.0,
	   const char *blDesignation = (const char *) "Unknown"); 

  // Constructor from earth longitude/latitude/altitude triplet, ENU
  // baseline vector, optional bias
  baseline(double longitude,double latitude,double altitude,
	   Vector lBaseline,double dBias = 0.0,
	   const char *blDesignation = (const char *) "Unknown"); 

  // Constructor from file (name)
  baseline(char *filename,
	   const char *blDesignation = (const char *) "Unknown");

  // Constructor from pointer to file
  baseline(FILE *fptr);


  // Baseline destructor
  ~baseline(void);



  // Member functions
  // baseline projections (vector and scalar)
  Vector inertial(double jd);
  Vector inertialDot(double jd);
  Vector skyProjV(double jd,Vector star);
  double skyProj(double jd,Vector star)
    { Vector SV = skyProjV(jd,star); SV[2] = 0.0;  return SV.norm(); }

  // angular functions
  double localSiderial(double jd);
  double hourAngle(double jd,Vector star);
  double zenithAngle(double jd,Vector star);
  double parallacticAngle(double jd,Vector star);

  // interferometric delay calculations
  double delay(double jd,Vector star);
  double delayRate(double jd,Vector star);


  // Other geometric calculations
  // Location velocity (for stellar aberation calculations)
  Vector locationVelocity(double jd);


  // Member accessor functions
  char *baselineName(void) {return designation;}
  double longitude(void) {return siteLongitude;}
  double latitude(void) {return siteLatitude;}
  double altitude(void) {return siteAltitude;}
  double bias(void) {return delayBias;}

  Vector localBaseline(void) {return localBline;}
  double **localCov(void) {return bCov;}
  double localCov(int i,int j)
    {
      if (bCov)
	return bCov[i][j];
      else
	return 0.0;
    }

  int error(void) {return errorFlag;}


};

// Associated utility functions

RMatrix erToLocal(double longitude,double latitude);

RMatrix ER_TO_INERTIAL(double jd);

// RMatrix ER_TO_INERTIAL_DOT(double jd);  // Not sure if anyone needs this one
