#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "coords.h"


#define DEFAULT_DELTAT 63.0504  // DeltaT @ Jan 1 2000

/************************************************************************
*	Copyright (c) 1997-2001 California Institute of Technology      *
*                     U.S. Government Sponsorship                       *
* Under contract with the National Aeronautics and Space Administration *
*************************************************************************/


/******************************************
 *
 * Coordinate class constructor 
 *
 * ra, dec should be in radians
 * pmRa, pmDec should be in arcsec/yr
 *  and pmRA should be REAL PROPER MOTION
 *  (not divided by cos(dec))
 * plx should be in arcsec
 * (input units modified A.B. 11/01)
 *
 *******************************************/
coords::coords(double ra, double dec, double pmRa, double pmDec, double plx, double epoch)
{
  earth = 3; // which planet we are on.
  deltaT = DEFAULT_DELTAT; // deltaT = TDT - JD
  rV = 0.0; // for lack of a better number

  RA = ra;
  DEC = dec;
  pmRA = pmRa;
  pmDEC = pmDec;
  PLX = plx;

  if (epoch == 1991.25)
    {
      // Assume in HIPPARCOS system and translate.
      RA += 8.75*(pmRA*4.8481368e-9);
      DEC += 8.75*(pmDEC*4.8481368e-9);
    }
  
  fk5pos.starnumber = 0;

  fk5pos.ra = RA * RAD_TO_DEG / 15.0;  // Convert from radians into hours
  fk5pos.dec = DEC * RAD_TO_DEG;  // Convert from radians into decimal degrees

  fk5pos.promora = pmRA * 100.0 / 15.0; // Convert from arcsec/yr to time sec/century
  fk5pos.promodec = pmDEC * 100.0;  // Convert from arcsec/yr to arcsec/century

  fk5pos.parallax = PLX;  // Already in arcsec

  fk5pos.radialvelocity = rV;
  fk5pos.visualmag = 0.0; // This is after all supposed to deal with coordinates only

}

/*****************************************
 *
 * coordinate class constructor from PTI
 * GUI-formatted line assumed to be 
 * J2000 coordinates
 *
 * Sample GUI line (without and with parallax -- 64 Psc):
 * HDC4676 00 48 58.708 +16 56 26.318 -0.002 -0.202 5.1 3.7 F8V...  xxx   xxx  trg
 * HDC4676 00 48 58.708 +16 56 26.318 -0.002 -0.202 0.0418 5.1 3.7 F8V...  xxx   xxx  trg
 *
 ****************************************/

coords::coords(char *guiline)
{
  int rah, ram, decd, decm;
  double ras, decs, pmRa, pmDec, plx;
  char starName[12];
  char line[500];

  earth = 3; // which planet we are on.
  deltaT = DEFAULT_DELTAT; // DeltaT = TDT - JD
  rV = 0.0; // for lack of a better number

  memcpy(line, guiline, strlen(guiline)+1 );
  char* str_ptr = strtok(line, " ");
  sscanf(str_ptr,"%s",starName);
  str_ptr = strtok('\0'," ");
  sscanf(str_ptr,"%d",&rah);
  str_ptr = strtok('\0'," ");
  sscanf(str_ptr,"%d",&ram);
  str_ptr = strtok('\0'," ");
  sscanf(str_ptr,"%lf",&ras);
  str_ptr = strtok('\0'," ");
  RA = hmsToRad(rah,ram,ras); // Units of positions in RA hrs, Dec degrees

  // DEC is harder - deal with sign separately
  char sign;
  sscanf(str_ptr,"%c",&sign);
  int decSign = 1;
  if (sign == '-') {
    decSign = -1;
    str_ptr++;
  }
  else if (sign == '+')
         str_ptr++;
  sscanf(str_ptr,"%2d",&decd);
  str_ptr = strtok('\0'," ");
  sscanf(str_ptr,"%2d",&decm);
  str_ptr = strtok('\0'," ");
  sscanf(str_ptr,"%lf",&decs);
  str_ptr = strtok('\0'," ");
       
  DEC = dmsToRad(decSign,decd,decm,decs);

  sscanf(str_ptr,"%lf", &pmRa);
  str_ptr = strtok('\0'," ");
  sscanf(str_ptr,"%lf", &pmDec);
  str_ptr = strtok('\0'," ");
  pmRA = pmRa; // Assuming units of PM in arcsec per year
  pmDEC = pmDec;

  PLX = plx = 0.0;
  if(str_ptr != NULL) {
    if (sscanf(str_ptr,"%lf",&plx) == 1)
      // Check if this parallax makes sense (it might be a magnitude)
      //  Assuming units of plx coming in in arcseconds
      if (plx < 1.0)
	PLX = plx;
  }

  fk5pos.starnumber = 0;

  fk5pos.ra = RA * RAD_TO_DEG / 15.0; // Convert from radians into decimal hours
  fk5pos.dec = DEC * RAD_TO_DEG; // Convert from radians into decimal degrees

  // Converting PM to byzantine units (thanks George...)
  fk5pos.promora = pmRA * 100.0 / 15.0;  // Convert from arcsec/yr to time sec/century
  fk5pos.promodec = pmDEC * 100.0; // Convert from arcsec/yr to arcsec/century

  fk5pos.parallax = PLX; // Already in arcsec

  fk5pos.radialvelocity = rV;
  fk5pos.visualmag = 0.0; // This is after all supposed to deal with coordinates only

}


/******************************************
 *
 *  geoCentIP/apparentPos - given a JD it returns the 
 *  apparent ra, dec, referred to the true 
 *  equator and equinox of JD.
 *
 *
 *
 *
 ******************************************/

Vector coords::apparentPos(double jd)
{
  short int err;
  double ra, dec;

  jd = fixJD(jd); // Fix the JD if necessary
  double tdt = dynamicalTimeJD(jd,&deltaT); // Dynamical time for app_star

  err = app_star(jd, earth, &fk5pos, &ra, &dec);  
  if (err != 0)
    fprintf(stderr,"app_star returned error %2d in geoCentIP.\n", err);

  // Bug Fix.  app_star ra/dec return values are in decimal hours and
  // degrees.  Convert to radians  A.B. 7/14/98

  Vector result(ra * 15.0 * DEG_TO_RAD,dec * DEG_TO_RAD);
  return result;
} // end apparentPos




/******************************************
 *
 * topoCentIP - give JD and location it returns
 * topocentric ra, dec referred to the true 
 *  equator and equinox of JD.
 *
 *
 ******************************************/

Vector coords::topocentricPos(double jd, double latitude, double longitude, double altitude)
{

  // latitude, longitude -- observer coordinates come in radians
  // altitude -- meters

  short int err;
  double ra, dec;

  jd = fixJD(jd); // Fix the JD if necessary
  double tdt = dynamicalTimeJD(jd,&deltaT); // Dynamical time for topo_star

  site_info site;
  site.latitude = latitude * RAD_TO_DEG; // convert to degrees
  site.longitude = longitude * RAD_TO_DEG; // convert to degrees
  site.height = altitude;

  err = topo_star(tdt, earth, deltaT, &fk5pos, &site, &ra, &dec);

  if(err != 0){
    fprintf(stderr,"topo_star returned error %2d in localIP.\n", err);
  }

  // Bug Fix.  topo_star ra/dec return values are in decimal hours and
  // degrees.  Convert to radians  A.B. 7/14/98
  //printf("  topocentricPos:  ra=%f  dec=%f\n", ra, dec);

  Vector result(ra * 15.0 * DEG_TO_RAD,dec * DEG_TO_RAD);
  return result;
} // end topocentricPos




/******************************************
 *
 *  geoCentIP/apparentPos - given a JD it returns the 
 *  apparent ra, dec, referred to the true 
 *  equator and equinox of JD.
 *
 *
 *
 *
 ******************************************/

Vector coords::virtualPos(double jd)
{
  short int err;
  double ra, dec;

  jd = fixJD(jd); // Fix the JD if necessary
  double tdt = dynamicalTimeJD(jd,&deltaT); // Dynamical time for virtual_star

  err = virtual_star(tdt, earth, &fk5pos, &ra, &dec);  
  if (err != 0)
    fprintf(stderr,"virtual_star returned error %2d in geoCentIP.\n", err);

  // Bug Fix.  virtual_star ra/dec return values are in decimal hours and
  // degrees.  Convert to radians  A.B. 7/14/98

  Vector result(ra * 15.0 * DEG_TO_RAD,dec * DEG_TO_RAD);
  return result;
} // end virtualPos



/******************************************
 *
 * localPos - give JD and location it returns
 * topocentric ra, dec referred to the true 
 *  equator and equinox of JD.
 *
 *
 ******************************************/

Vector coords::localPos(double jd, double latitude, double longitude, double altitude)
{

  // latitude, longitude -- observer coordinates come in radians
  // altitude -- meters

  short int err;
  double ra, dec;

  jd = fixJD(jd); // Fix the JD if necessary
  double tdt = dynamicalTimeJD(jd,&deltaT); // Dynamical time for local_star

  site_info site;
  site.latitude = latitude * RAD_TO_DEG;  // convert to degrees
  site.longitude = longitude * RAD_TO_DEG; // convert to degrees
  site.height = altitude;

  err = local_star(tdt, earth, deltaT, &fk5pos, &site, &ra, &dec);

  if(err != 0){
    fprintf(stderr,"local_star returned error %2d in localIP.\n", err);
  }

  // Bug Fix.  local_star ra/dec return values are in decimal hours and
  // degrees.  Convert to radians  A.B. 7/14/98

  Vector result(ra * 15.0 * DEG_TO_RAD,dec * DEG_TO_RAD);
  return result;
} // end localPos



/******************************************
 *
 *  barycentricPos - given a JD it returns the 
 *  barycentric apparent ra, dec, referred to the true 
 *  equator and equinox of JD.
 *
 *
 *
 *
 ******************************************/

Vector coords::barycentricPos(double jd)
{
  short int err;
  double ra, dec;

  jd = fixJD(jd); // Fix the JD if necessary
  double tdt = dynamicalTimeJD(jd,&deltaT); // Dynamical time for bary_star
  
  err = bary_star(tdt,&fk5pos,&ra,&dec);  
  if (err != 0)
    fprintf(stderr,"bary_star returned error %2d in barycentricPos.\n", err);

  // Bug Fix.  app_star ra/dec return values are in decimal hours and
  // degrees.  Convert to radians  A.B. 7/14/98

  Vector result(ra * 15.0 * DEG_TO_RAD,dec * DEG_TO_RAD);
  return result;
} // end barycentricPos






/**********************************************
 *
 *   Utility functions
 *
 *
 *********************************************/

char* coords::print(int pm,int plx)
{
  char *result,work[256];
  result = new char[256];
  
  int rah, ram, sign, decd, decm;
  double ras, decs, pmra, pmdec;
  
  hms(&rah, &ram, &ras);
  dms(&sign, &decd, &decm, &decs);

  char mySign = '+';
  if (sign == -1)
    mySign = '-';

  //   sprintf(result,"%02d %02d %06.3lf %+03d %02d %06.3lf ",
  //           rah, ram, ras, decd, decm, decs);

  sprintf(result,"%02d %02d %06.3lf %c%02d %02d %06.3lf ",
	  rah, ram, ras, mySign, abs(decd), decm, decs);

  if (pm)
    {
      sprintf(work," %5.3f  %5.3f ",pmRA,pmDEC);
      (void) strcat(result,work);
    }

  if (plx)
    {
      sprintf(work," %6.4f ",PLX);
      (void) strcat(result,work);
    }


  return result;
}

/*******************************************/

void coords::hms(int *h, int *m, double *s)
{
  // Conversion from floating point hours to int hours, min, (floating-point) seconds
  double ra = fk5pos.ra;
  *h = (int) ra;
  *m = (int) (60.0 * (ra - ((double) (*h))));
  *s = 60.0 * (60.0 * (ra - ((double) (*h))) - ((double) (*m)));
  return;
}

/*******************************************/
  
void coords::dms(int *sign, int *d, int *m, double *s)
{
  // Floating-point degrees to int deg, min, seconds conversion
  if (fk5pos.dec<0.0)
  *sign = -1;
  else
    *sign = +1;
  double dec = fabs(fk5pos.dec);
  *d = (int) dec;
  *m = (int) (60.0 * (dec - ((double) (*d))));
  *s = 60.0 * (60.0 * (dec - ((double) (*d))) - (double) (*m));

  return;
}  

/*******************************************/

double coords::ra()
{
  return RA;
}

/*******************************************/

double coords::dec()
{
  return DEC;
}

/* general utility functions */

/******************************************/

double hmsToRad(int h, int m, double s)
{
  double ra = h + (m/60.0) + (s/3600.0);
  ra *= 15.0; // Convert to degrees
  ra *= DEG_TO_RAD; // Convert to radians
  return ra;
}

/******************************************/

double dmsToRad(int sign, int d, int m, double s)
{
  double dec;
  dec = sign*(d + (m/60.0) + (s/3600.0));
  dec *= DEG_TO_RAD;
  return dec;
}

/******************************************/

char *threeVectorPrint(Vector vec)
{
  char *result;
  result = new char[256];
  sprintf(result,"%lf %lf %lf", vec[0], vec[1], vec[2]);
  return result;
}



/************************************************************************
*  precessV -- local to coords.cc (A.B. 1/6/99)				*
*									*
*  A.  Purpose: The function precessV computes the precession           *
*       transformation of an input position 3-vector from the input     *
*       reference time to an output reference time.			*
*									*
*  B.  Background and use: This function is just a wrapper for the      *
*       NOVAS function precession.					*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd1 -- the Julian date of the initial reference time (double). *
*        jd2 -- the Julian data of the final reference time (double).   *
*        arg -- the 3-vector to compute the precessed representation    *
*         (Vector).                                                     *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the 3-vector result of the *
*       precession calculation (Vector).		                *
*									*
*************************************************************************/
Vector precessV(double jd1, double jd2, Vector arg)
{
/**********  begin precessV  ********************************************/
  Vector result;
  double dummy,secdiff;

  // Convert to barycentric dynamical time (what NOVAS precession wants)
  jd1 = fixJD(jd1); // Correct JDs in case we are working in MJD
  (void) convert_tdb2tdt(jd1,&dummy,&secdiff);
  double jd1TDB = jd1 + secdiff / 86400.0;

  jd2 = fixJD(jd2); // Correct JDs in case we are working in MJD
  (void) convert_tdb2tdt(jd2,&dummy,&secdiff);
  double jd2TDB = jd2 + secdiff / 86400.0;

  if (precession(jd1TDB,&(arg[0]),jd2TDB,&(result[0])))  // NOVAS invocation
    (void) fprintf(stderr," Error in precession\n");

  return result;

/**********  end precessV  **********************************************/
}



/************************************************************************
*  nutateMeanV -- local to coords.cc (A.B. 1/7/99)			*
*									*
*  A.  Purpose: The function nutateMeanV computes the nutation          *
*       coordinate transformation of an input apparent (true) position  *
*       3-vector from an input reference time to its mean position at   *
*       T0/J2000.					                *
*									*
*  B.  Background and use: This function is just a wrapper for the      *
*       NOVAS function nutate.  As we have written it, we have made no  *
*       allowances for bi-directional transformation (the NOVAS nutate  *
*       allows both forward and backward transformations).	        *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the Julian date of the reference time (double).          *
*        arg -- the 3-vector to compute the nutated representation      *
*         (Vector).                                                     *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the 3-vector result of the *
*       nutation calculation (Vector).		                        *
*									*
*************************************************************************/
Vector nutateMeanV(double jd, Vector arg)
{
/**********  begin nutateMeanV  *****************************************/
  Vector result;
  double dummy,secdiff;

  jd = fixJD(jd); // Correct JDs in case we are working in MJD

  // Convert to barycentric dynamical time (what NOVAS nutate wants)
  (void) convert_tdb2tdt(jd,&dummy,&secdiff);
  double jdTDB = jd + secdiff / 86400.0;

  if (nutate(jdTDB,1,&(arg[0]),&(result[0]))) // NOVAS invocation, from apparent to mean
    (void) fprintf(stderr," Error in nutate\n");

  return result;

/**********  end nutateMeanV  *******************************************/
}



/************************************************************************
*  nutateTrueV -- local to coords.cc (A.B. 1/7/99)			*
*									*
*  A.  Purpose: The function nutateTrueV computes the nutation          *
*       coordinate transformation of an input mean (J2000) position     *
*       3-vector to the apparent (true) position at an input reference  *
*       time.                                                           *
*									*
*  B.  Background and use: This function is just a wrapper for the      *
*       NOVAS function nutate.  As we have written it, we have made no  *
*       allowances for bi-directional transformation (the NOVAS nutate  *
*       allows both forward and backward transformations).	        *
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the Julian date of the desired reference time (double).  *
*        arg -- the 3-vector to compute the nutated representation      *
*         (Vector).                                                     *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the 3-vector result of the *
*       nutation calculation (Vector).		                        *
*									*
*************************************************************************/
Vector nutateTrueV(double jd, Vector arg)
{
/**********  begin nutateTrueV  *****************************************/
  Vector result;
  double dummy,secdiff;

  jd = fixJD(jd); // Correct JDs in case we are working in MJD

  // Convert to barycentric dynamical time (what NOVAS nutate wants)
  (void) convert_tdb2tdt(jd,&dummy,&secdiff);
  double jdTDB = jd + secdiff / 86400.0;

  if (nutate(jdTDB,0,&(arg[0]),&(result[0]))) // NOVAS invocation, from mean to apparent
    (void) fprintf(stderr," Error in nutate\n");

  return result;

/**********  end nutateTrueV  *******************************************/
}


/************************************************************************
*  distance                                                             *
*                                                                       *
*  Helper method that takes two sets of RA/DEC values in RADIANS and    *
*  returns the delta RA/DEC.  Designed to be used on the results        *
*  of the topocentricPos method.  Used by the AO and OFS sequencers.    *
*									*
*************************************************************************/
void distance(double ra1, double dec1, double ra2, double dec2,
              double *deltara, double *deltadec)
{
/**********  begin distance  *****************************************/

  *deltadec = dec1 - dec2;
  *deltara = (ra1 - ra2) * 
      cos( (((dec1*RAD_TO_DEG) + (dec2*RAD_TO_DEG)) / 2.0) * (PI/180.0) );

/**********  end distance  *******************************************/
}
