#include <string.h>
#include <stdio.h>
#include <math.h>

#include "histogram.h"
#include "bsearch.h"


/************************************************************************
*  construct_1dhist -- local to hist_functions.c (A.B. 6/27/94)	        *
*									*
*  A.  Purpose:	The 1dhist constructor creates a 1d histogram           *
*       object, loads it with the appropriate values, and               *
*       returns a pointer to it.  This version of the constructor       *
*       creates a histogram structure and fills it based on the input   *
*       arguments.                      				*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        title -- the title string for the histogram (char *).          *
*        num_bins -- the number of bins in the histogram (int).         *
*        low_edge -- the lower edge of the histogram range (float).     *
*        upp_edge -- the upper edge of the histogram range (float).     *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns a pointer to the created   *
*       1d histogram data structure, or a NIL pointer if there is a     *
*       memory allocation error.	                        	*
*									*
*************************************************************************/
/**********  function declaration construct_1dhist  *********************/
histogram_1d::histogram_1d(const char *title, int nbins,
			   float ledge, float uedge)
{
/**********  local declarations construct_1dhist  ***********************/
  int i;
  float bin_size,hbin_size;

/**********  begin construct_1dhist  ************************************/
  if (!(hist_title = strdup(title))
      || !(bin_val = new float[nbins])
      || !(data_val = new float[nbins]))
    {
      (void) printf("Error allocating memory -- construct_1dhist\n");
      return;
    }
  num_bins = nbins;
  overflow = underflow = entries = 0;
  lower_edge = ledge;
  upper_edge = uedge;
  bin_size = (uedge - ledge) / (float) nbins;
  hbin_size = bin_size / 2.0;
  for (i = 0; i < num_bins; i++)
    {
      // bin_val are the bin centers!!!
      bin_val[i] = lower_edge + hbin_size + (float) (i * bin_size);
      data_val[i] = 0.0;
    }
  sum_weight = sum_val = sum_val2 = 0.0;
/**********  end construct_1dhist  **************************************/
}


/************************************************************************
*  construct_1dhist -- local to hist_functions.c (A.B. 6/27/94)	        *
*									*
*  A.  Purpose:	The 1dhist constructor creates a 1d histogram           *
*       object, loads it with the appropriate values, and               *
*       returns a pointer to it.  This version of the constructor reads *
*       the contents of the histogram from an input filename.		*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        filename -- the (pointer to the) character string containing   *
*         the file name to read the histogram from (char *).            *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns a pointer to the created   *
*       1d histogram data structure, or a NIL pointer if there is a     *
*       memory allocation error.	                        	*
*									*
*************************************************************************/
/**********  function declaration construct_1dhist  *********************/
histogram_1d::histogram_1d(const char *filename)
{
/**********  local declarations construct_1dhist  ***********************/
#define LINE_LENGTH 132

char title[LINE_LENGTH];
int i;
float bin_size,hbin_size,dummy;
FILE *fptr;

/**********  begin construct_1dhist  ************************************/
if (!(fptr = fopen(filename,"r")))
  {
    (void) printf("Error reading histogram file %s -- construct_1dhist\n",filename);
    return;
  }
if ((fscanf(fptr,"### %s\n",title) != 1)
    || (fscanf(fptr,"#   1d -- Num Bins: %d\t\tNum Entries: %d\n",&num_bins,&entries) != 2)
    || (fscanf(fptr,"#   Lower Edge: %g\tUpper Edge: %g\n",&lower_edge,&upper_edge) != 2)
    || (fscanf(fptr,"#   Underflows: %d\t\tOverflows: %d\n",&underflow,&overflow) != 2))
  {
    (void) printf("Error reading histogram file %s -- construct_1dhist\n",filename);
    (void) fclose(fptr);
    return;
  }
if (!(hist_title = strdup(title))
    || !(bin_val = new float[num_bins])
    || !(data_val = new float[num_bins]))
  {
  (void) printf("Error allocating memory -- construct_1dhist\n");
  return;
  }
bin_size = (upper_edge - lower_edge) / (float) num_bins;
hbin_size = bin_size / 2.0;
for (i = 0; i < num_bins; i++)
  {
    bin_val[i] = lower_edge + hbin_size + (float) (i * bin_size);
    if (fscanf(fptr,"%g\t%g\n",&dummy,&data_val[i]) != 2)
      {
	(void) printf("Error reading histogram file %s -- construct_1dhist\n",filename);
	(void) fclose(fptr);
	return;
      }
  } /* end i loop */
(void) fclose(fptr);
/**********  end construct_1dhist  **************************************/
}


/************************************************************************
*  construct_1dhist -- local to hist_functions.c (A.B. 6/27/94)	        *
*									*
*  A.  Purpose:	The 1dhist constructor creates a 1d histogram           *
*       object, loads it with the appropriate values, and               *
*       returns a pointer to it.  This version of the constructor       *
*       creates a histogram structure and fills it based on an input    *
*       preexisting histogram.                 				*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        ref_hist -- the 1d histogram the new histogram is built from   *
*         (histogram_1d).                                               *
*        bin_scale -- the scale for the bin sizes for the new histogram *
*         in relation to the old one (float).                           *
*        val_scale -- the scale of the new histogram data values in     *
*         relation to the old data values (float).                      *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns a pointer to the created   *
*       1d histogram data structure, or a NIL pointer if there is a     *
*       memory allocation error.	                        	*
*									*
*************************************************************************/
/**********  function declaration construct_1dhist  *********************/
histogram_1d::histogram_1d(histogram_1d ref_hist,
                           float bin_scale,float val_scale)
{
/**********  local declarations construct_1dhist  ***********************/
int i,j,min_new_bin,max_new_bin;
float bin_size,hbin_size,ref_bin_size,ref_ledge,ref_uedge,
      new_ledge,new_uedge,cover_frac;

/**********  begin construct_1dhist  ************************************/
num_bins = (int) (ref_hist.num_bins / bin_scale);
if (!(hist_title = strdup(ref_hist.hist_title))
    || !(bin_val = new float[num_bins])
    || !(data_val = new float[num_bins]))
  {
  (void) printf("Error allocating memory -- construct_1dhist\n");
  return;
  }
overflow = ref_hist.overflow;
underflow = ref_hist.underflow;
entries = 0;
lower_edge = ref_hist.lower_edge;
upper_edge = ref_hist.upper_edge;
bin_size = (upper_edge - lower_edge) / (float) num_bins;
hbin_size = bin_size / 2.0;
for (i = 0; i < num_bins; i++)
  {
  bin_val[i] = lower_edge + hbin_size + (float) (i * bin_size);
  data_val[i] = 0.0;
  }
ref_bin_size = (ref_hist.upper_edge - ref_hist.lower_edge)
                / (float) ref_hist.num_bins;
for (i = 0; i < ref_hist.num_bins; i++)
  {
    if (i == 0)
      ref_ledge = ref_hist.lower_edge;
    else
      ref_ledge = (ref_hist.bin_val[i - 1] + ref_hist.bin_val[i]) / 2.0;
    if (i == ref_hist.num_bins - 1)
      ref_uedge = ref_hist.upper_edge;
    else
      ref_uedge = (ref_hist.bin_val[i] + ref_hist.bin_val[i + 1]) / 2.0;
    min_new_bin = hist_binary_search(bin_val,num_bins,ref_ledge);
    max_new_bin = hist_binary_search(bin_val,num_bins,ref_uedge);
    for (j = min_new_bin; j <= max_new_bin; j++)
      {
	if (j == 0)
	  new_ledge = lower_edge;
	else
	  new_ledge = (bin_val[j - 1] + bin_val[j]) / 2.0;
	if (j == num_bins - 1)
	  new_uedge = upper_edge;
	else
	  new_uedge = (bin_val[j] + bin_val[j + 1]) / 2.0;
	if (ref_ledge > new_ledge)
	  cover_frac = (new_uedge - ref_ledge) / ref_bin_size;
	else
	  cover_frac = (new_uedge - new_ledge) / ref_bin_size;
	if (new_uedge > ref_uedge)
	  cover_frac -= (new_uedge - ref_uedge) / ref_bin_size;
	(void) enter(bin_val[j],val_scale * cover_frac * ref_hist.data_val[i]);
      } /* end j loop */
  } /* end i loop */
/**********  end construct_1dhist  **************************************/
}


/************************************************************************
*  destruct_1dhist -- local to hist_1d.cc (A.B. 6/27/94)	        *
*									*
*  A.  Purpose:	The 1dhist destructor nukes a 1d histogram object.      *
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description: None					*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useful value.           *
*									*
*************************************************************************/
/**********  function declaration destruct_1dhist  **********************/
histogram_1d::~histogram_1d()
{
/**********  begin destruct_1dhist  *************************************/
delete hist_title;
delete bin_val;
delete data_val;
/**********  end destruct_1dhist  ***************************************/
}


/************************************************************************
*  erase_1dhist -- local to hist_1d.cc (A.B. 6/27/94)		        *
*									*
*  A.  Purpose:	The function erase_1dhist erases the data contents of   *
*       an input 1d histogram, leaving the remainder of the data        *
*       structure intact.                                               *
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        hist_ptr -- the (pointer to the) 1d histogram whoes data       *
*         contents are to be erased (struct histogram_1d *).        *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no usable value as of yet. *
*									*
*************************************************************************/
/**********  function declaration erase_1dhist  *************************/
int histogram_1d::erase(void)
{
/**********  local declarations erase_1dhist  ***************************/
int i;

/**********  begin erase_1dhist  ****************************************/
for (i = 0; i < num_bins; i++)
  data_val[i] = 0.0;
return 1;
/**********  end erase_1dhist  ******************************************/
}


/************************************************************************
*  enter_1dhist -- local to hist_1d.cc (A.B. 6/27/93)		        *
*									*
*  A.  Purpose:	The function enter_1dhist places an entry into a        *
*       given 1d histogram.						*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        ind_value -- the independent value of the data pair to be      *
*         entered into the histogram (float).                           *
*        data_value -- the dependent value of the data pair to be       *
*         entered into the histogram (float).                           *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
int histogram_1d::enter(float ind_value, float data_value)
{
/**********  local declarations enter_1dhist  ***************************/
int bin;

/**********  begin enter_1dhist  ****************************************/
if (ind_value < lower_edge)
  {
  underflow++;
  return 0;
  }
if (ind_value > upper_edge)
  {
  overflow++;
  return 0;
  }
bin = hist_binary_search(bin_val,num_bins,ind_value);
data_val[bin] += data_value;
sum_val += ind_value * data_value;
sum_val2 += ind_value * ind_value * data_value;
sum_weight += data_value;
entries++;
return 1;
/**********  end enter_1dhist  ******************************************/
}


/************************************************************************
*  write_1dhist -- local to hist_functions.c (A.B. 12/10/93)		*
*									*
*  A.  Purpose: The function write_1dhist writes the contents of the    *
*       given 1D histogram out to a given ASCII file.  The format is    *
*       chosen to be used with GNUPLOT.					*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        output_file -- the (pointer to the) file the histogram is      *
*         written to (FILE *).                                          *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration write_1dhist  *************************/
void histogram_1d::write(FILE *output_file)
{
/**********  local declarations write_1dhist  ***************************/
int i;

/**********  begin write_1dhist  ****************************************/
if (!output_file)
  {
  (void) printf("Error writing histogram %s\n",hist_title);
  return;
  }
(void) fprintf(output_file,"### %s\n",hist_title);
(void) fprintf(output_file,"#   1d -- Num Bins: %d\t\tNum Entries: %d\n",
	       num_bins,entries);
(void) fprintf(output_file,"#   Lower Edge: %g\tUpper Edge: %g\n",
	       lower_edge,upper_edge);
(void) fprintf(output_file,"#   Underflows: %d\t\tOverflows: %d\n",
	       underflow,overflow);
(void) fprintf(output_file,"#   Mean: %g\t\tStdDev: %g\n",
	       mean(),sigma());

for (i = 0; i < num_bins; i++)
  (void) fprintf(output_file,"%g\t%g\n",bin_val[i],data_val[i]);

(void) fprintf(output_file,"\n\n\n");

/**********  end write_1dhist  ******************************************/
}

/* Sample Statistics Routines */

float histogram_1d::mean(void)
{
  if (sum_weight != 0.0)
    return ((float) (sum_val / sum_weight));
  else
    {
      (void) printf("Error -- empty histogram statistics\n");
      return 0.0;
    }
};

float histogram_1d::sigma(void)
{
  if (sum_weight != 0.0)
    {
      double variance = sum_weight / (sum_weight - 1.0)
	* ((sum_val2 / sum_weight) - pow(mean(),2.0));
      return (float) sqrt(variance);
    }
  else
    {
      (void) printf("Error -- empty histogram statistics\n");
      return 0.0;
    }
};

