#include <string.h>
#include <stdio.h>
#include <math.h>

#include "histogram.h"
#include "bsearch.h"


/************************************************************************
*  construct_2dhist -- local to hist_2d.cc (A.B. 6/27/94)		*
*									*
*  A.  Purpose:	The function construct_2dhist constructs a 2d histogram *
*       object, and loads it with the appropriate values.               *
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        title -- the title string for the histogram (char *).          *
*        num_bins -- the number of bins in the histogram (int).         *
*        low_edge -- the lower edge of the histogram range (float).     *
*        upp_edge -- the upper edge of the histogram range (float).     *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns a pointer to the created   *
*       2d histogram object, or a NIL pointer if there is a memory      *
*       allocation error.	                        	        *
*									*
*************************************************************************/
/**********  function declaration construct_2dhist  *********************/
histogram_2d::histogram_2d(const char *title,
			   int xbins, float xledge, float xuedge,
			   int ybins, float yledge, float yuedge)
{
/**********  local declarations construct_2dhist  ***********************/
  int i,j;
  float bin_size,hbin_size;

/**********  begin construct_2dhist  ************************************/
  if (!(hist_title = strdup(title))
      || !(xbin_val = new float[xbins])
      || !(data_val = new float *[xbins])
      || !(ybin_val = new float[ybins]))
    {
      (void) printf("Error allocating memory -- construct_2dhist\n");
      return;
    }
  for (i = 0; i < xbins; i++)
    if (!(data_val[i] = new float[ybins]))
      {
	(void) printf("Error allocating memory -- construct_2dhist\n");
	return;
      }
  num_xbins = xbins;
  num_ybins = ybins;
  xoverflow = xunderflow = 0;
  yoverflow = yunderflow = 0;
  entries = 0;
  xlower_edge = xledge;
  xupper_edge = xuedge;
  bin_size = (xupper_edge - xlower_edge) / (float) num_xbins;
  hbin_size = bin_size / 2.0;
  for (i = 0; i < num_xbins; i++)
    {
      xbin_val[i] = xlower_edge + hbin_size + (float) (i * bin_size);
      for (j = 0; j < ybins; j++)
	data_val[i][j] = 0.0;
    }
  ylower_edge = yledge;
  yupper_edge = yuedge;
  bin_size = (yupper_edge - ylower_edge) / (float) num_ybins;
  hbin_size = bin_size / 2.0;
  for (i = 0; i < num_ybins; i++)
    ybin_val[i] = ylower_edge + hbin_size + (float) (i * bin_size);
  sum_weight = sum_xval = sum_xval2 = sum_yval = sum_yval2 = sum_xyval = 0.0;
/**********  end construct_2dhist  **************************************/
}


/************************************************************************
*  destruct_2dhist -- local to hist_2d.cc (A.B. 6/27/94)		*
*									*
*  A.  Purpose:	The function destruct_2dhist nukes a 2d histogram       *
*       object.                                                         *
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description: None					*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useful value            *
*									*
*************************************************************************/
/**********  function declaration destruct_2dhist  **********************/
histogram_2d::~histogram_2d()
{
/**********  local declarations destruct_2dhist  ************************/
int i;

/**********  begin destruct_2dhist  *************************************/
delete hist_title;
delete xbin_val;
delete ybin_val;
for (i = 0; i < num_xbins; i++)
  delete data_val[i];
/**********  end destruct_2dhist  ***************************************/
}


/************************************************************************
*  erase_2dhist -- local to hist_2d.cc (A.B. 6/27/94)		        *
*									*
*  A.  Purpose:	The function erase_1dhist erases the data contents of   *
*       an input 2d histogram, leaving the remainder of the data        *
*       structure intact.                                               *
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description: None					*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no usable value as of yet. *
*									*
*************************************************************************/
/**********  function declaration erase_2dhist  *************************/
int histogram_2d::erase()
{
/**********  local declarations erase_2dhist  ***************************/
int i,j;

/**********  begin erase_2dhist  ****************************************/
for (i = 0; i < num_xbins; i++)
  for (j = 0; j < num_ybins; j++)
    data_val[i][j] = 0.0;
return 1;
/**********  end erase_2dhist  ******************************************/
}


/************************************************************************
*  enter_2dhist -- local to hist_2d.cc (A.B. 6/27/94)		        *
*									*
*  A.  Purpose:	The function enter_2dhist places an entry into a        *
*       given 2d histogram.						*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        hist_ptr -- the (pointer to the) histogram structure to be     *
*         updated with this new information (struct histogram_2d).  *
*        xvalue,yvalue -- the independent orderd pair to be             *
*         entered into the histogram (float).                           *
*        data_value -- the dependent value of the data pair to be       *
*         entered into the histogram (float).                           *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration enter_2dhist  *************************/
int histogram_2d::enter(float xvalue, float yvalue, float data_value)
{
/**********  local declarations enter_2dhist  ***************************/
int binx,biny;

/**********  begin enter_2dhist  ****************************************/
if (xvalue < xlower_edge)
  {
  xunderflow++;
  return 0;
  }
if (xvalue > xupper_edge)
  {
  xoverflow++;
  return 0;
  }
if (yvalue < ylower_edge)
  {
  yunderflow++;
  return 0;
  }
if (yvalue > yupper_edge)
  {
  yoverflow++;
  return 0;
  }
binx = hist_binary_search(xbin_val,num_xbins,xvalue);
biny = hist_binary_search(ybin_val,num_ybins,yvalue);
data_val[binx][biny] += data_value;
entries++;
sum_weight += data_value;
sum_xval += xvalue * data_value;
sum_xval2 += xvalue * xvalue * data_value;
sum_yval += yvalue * data_value;
sum_yval2 += yvalue * yvalue * data_value;
sum_xyval += xvalue * yvalue * data_value;
return 1;
/**********  end enter_2dhist  ******************************************/
}


/************************************************************************
*  write_2dhist -- local to hist_2d.cc (A.B. 6/27/94)		        *
*									*
*  A.  Purpose: The function write_2dhist writes the contents of the    *
*       given 2d histogram out to a given ASCII file.  The format is    *
*       chosen to be used with GNUPLOT.					*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        output_file -- the (pointer to the) file the histogram is      *
*         written to (FILE *).                                          *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration write_2dhist  *************************/
void histogram_2d::write(FILE *output_file)
{
/**********  local declarations write_2dhist  ***************************/
int i,j;

/**********  begin write_2dhist  ****************************************/
if (!output_file)
  {
  (void) printf("Error writing histogram %s\n",hist_title);
  return;
  }
(void) fprintf(output_file,"### %s\n",hist_title);
(void) fprintf(output_file,"#   2d -- Num X Bins: %d\tNum Y Bins: %d\tNum Entries: %d\n",
	       num_xbins,num_ybins,entries);
(void) fprintf(output_file,"#   X Lower Edge: %g\tX Upper Edge: %g\n",
	       xlower_edge,xupper_edge);
(void) fprintf(output_file,"#   X Underflows: %d\t\tX Overflows: %d\n",
	       xunderflow,xoverflow);
(void) fprintf(output_file,"#   Y Lower Edge: %g\tY Upper Edge: %g\n",
	       ylower_edge,yupper_edge);
(void) fprintf(output_file,"#   Y Underflows: %d\t\tY Overflows: %d\n",
	       yunderflow,yoverflow);
for (i = 0; i < num_xbins; i++)
  for (j = 0; j < num_ybins; j++)
    (void) fprintf(output_file,"%g\t%g\t%g\n",
		   xbin_val[i],ybin_val[j],data_val[i][j]);
/**********  end write_2dhist  ******************************************/
}

/* Sample Statistics Routines */


float histogram_2d::xmean(void)
{
  if (sum_weight != 0.0)
    return (float) (sum_xval / sum_weight);
  else
    {
      (void) printf("Error -- empty histogram statistics\n");
      return 0.0;
    }
};

float histogram_2d::xsigma(void)
{
  if (sum_weight != 0.0)
    return (float) sqrt((sum_xval2 + xmean() * xmean() * sum_weight - 2 * xmean() * sum_xval) / sum_weight);
  else
    {
      (void) printf("Error -- empty histogram statistics\n");
      return 0.0;
    }
};


float histogram_2d::ymean(void)
{
  if (sum_weight != 0.0)
    return (float) (sum_yval / sum_weight);
  else
    {
      (void) printf("Error -- empty histogram statistics\n");
      return 0.0;
    }
};

float histogram_2d::ysigma(void)
{
  if (sum_weight != 0.0)
    return (float) sqrt((sum_yval2 + ymean() * ymean() * sum_weight - 2 * ymean() * sum_yval) / sum_weight);
  else
    {
      (void) printf("Error -- empty histogram statistics\n");
      return 0.0;
    }
};


float histogram_2d::covariance(void)
{
  if (sum_weight != 0.0)
    return (float) ((sum_xyval + xmean() * ymean() * sum_weight - xmean() * sum_yval - ymean() * sum_xval) / sum_weight);
  else
    {
      (void) printf("Error -- empty histogram statistics\n");
      return 0.0;
    }
};


float histogram_2d::correlation(void)
{
  return covariance() / (xsigma() * ysigma());
};
