package KIcoordinates;

require 5.000;

use Exporter();

@ISA = qw(Exporter);
@EXPORT = qw(RAhms RAdeg DECdms DECdeg);


#########################################################################
#     Interferometry Science Center (ISC) -- http://isc.caltech.edu     #
#	Copyright (c) 1997-2002 California Institute of Technology      #
#                     U.S. Government Sponsorship                       #
# Under contract with the National Aeronautics and Space Administration #
#########################################################################



### coordinates.pm ###
## coordinates.pm contains a set of utilities for dealing with
## coordinate formats


# RAhms produces an HR MIN SEC RA representation from a decimal degree input
sub RAhms
{

    my $raDeg = shift @_;
    if ($raDeg > 360.0)
    {
	$raDeg -= 360.0;
    }

    my $sigFigs = 1;
    if (scalar @_)
    {
	$sigFigs = int (shift @_);
    }

    my $raSign = "";  ## handle negative arguments (possible for hour angles)
    if ($raDeg < 0)
    {
	$raSign = "-";
	$raDeg = -$raDeg;
    }

    my $hr = int ($raDeg / 15.0);
    $raDeg -= 15.0 * $hr;

    if ($hr < 10)
    {
	$hr = sprintf "0%1d", $hr;
    }
    else
    {
	$hr = sprintf "%2d", $hr;
    }

    my $min = int (60.0 / 15.0 * $raDeg);
    $raDeg -= 15.0 / 60.0 * $min;

    if ($min < 10)
    {
	$min = sprintf "0%1d", $min;
    }
    else
    {
	$min = sprintf "%2d", $min;
    }

    if ($sigFigs < 0)
    {
	$raSign .= $hr;
	return join " ", $raSign, $min;
    }


    ## If we're still around, do the seconds
    my $sec = 3600.0 / 15.0 * $raDeg;
    my $formatStr;
    if ($sec < 10)
    {
	# $sec = sprintf "0%4.3f", $sec;
	$formatStr = join '',"0%",$sigFigs+1,".",$sigFigs,"f";
    }
    else
    {
	$formatStr = join '',"%",$sigFigs+2,".",$sigFigs,"f";
    }
    $sec = sprintf "$formatStr", $sec;

    ## Final format tests/wrap tests
    if ($sec =~ m/^010/)
    {
	$sec =~ s/010/10/;
    }
    if ($sec =~ m/^60/)
    {
	$sec =~ s/60/00/;
	$min += 1;
	if ($min =~ m/60/)
	{
	    $min = "00";
	    $hr += 1;
	    if ($hr > 23)
	    {
		$hr = "00";
	    } ## end day wrap/ hr format
	} ## end hr wrap/min format
    } ## end min wrap/sec format


    $raSign .= $hr;
    return join " ", $raSign,$min,$sec;

} #end RAhms


# RAdeg produces a decimal degree RA from HR MIN SEC input
sub RAdeg
{
    my $raHMS = shift @_;

    my @ra = split /\:|h|m|s|_|\s+/, $raHMS;

    my $raReturn = 0.0;

    if ($ra[0])
    {
	$raReturn = 15.0 * $ra[0];
    }
    if ($ra[1])
    {
	$raReturn += 15.0 / 60.0 * $ra[1];
    }
    if ($ra[2])
    {
	$raReturn += 15.0 / 3600.0 * $ra[2];
    }

    return $raReturn;

} #end RAdeg

    

# DECdms produces a DEG MIN SEC Dec representation from a decimal degree input
sub DECdms
{

    my $decDeg = shift @_;
    my $sigFigs = 1;
    if (scalar @_)
    {
	$sigFigs = int (shift @_);
    }

    if ($decDeg < 0)
    {
	$sign = '-';
	$decDeg *= -1.0;
    }
    else
    {
	$sign = '+';
    }

    my $deg = int ($decDeg);
    $decDeg -= $deg;

    my $min = int (60.0 * $decDeg);
    $decDeg -= $min / 60.0;

    my $sec;
    if ($sigFigs != -1)
    {
	$sec = 3600.0 * $decDeg;

	my $formatStr;
	if ($sec < 10)
	{
	    # $sec = sprintf "0%4.3f", $sec;
	    $formatStr = join '',"0%",$sigFigs+1,".",$sigFigs,"f";
	}
	else
	{
	    $formatStr = join '',"%",$sigFigs+2,".",$sigFigs,"f";
	}
	$sec = sprintf "$formatStr", $sec;

	if ($sec >= 60.0)
	{
	    $min += 1;
	    $sec -= 60.0;

	    if ($min >= 60)
	    {
		$deg += 1;
		$min -= 60.0;
		$formatStr = join '',"0%",$sigFigs+1,".",$sigFigs,"f";
		$sec = sprintf "$formatStr", $sec;
	    } ## enc check deg wrap
	} ## end check sec wrap



    }


    if ($min < 10)
    {
	$min = sprintf "0%1d", $min;
    }
    if ($deg < 10)
    {
	$deg = sprintf "0%1d", $deg;
    }
    $deg = join "", $sign,$deg;


    if ($sigFigs != -1)
    {
	return join " ", $deg,$min,$sec;
    }
    else
    {
	return join " ", $deg,$min;
    }



} ## end DECdms



# DECdeg produces a decimal degree Dec output from a DEG MIN SEC string input
sub DECdeg
{

    my $decDMS = shift @_;

    my @dec = split /\:|d|m|s|_|\s+/, $decDMS;

    my $decSign = 1.0;  ## Logic for negative declination  A.B. 7/2000
    # hjb added match for "minus-zero" 6/2007
    if( ($dec[0] < 0.0) || ($dec[0] =~ m/^-0/) )
    {
	$decSign = -1.0;
	$dec[0] =~ s/-//g; ## mask-off the negative sign
    }

    my $decVal = $dec[0];
    if ($dec[1])
    {
	$decVal += $dec[1] / 60.0;
    }
    if ($dec[2])
    {
	$decVal += $dec[2] / 3600.0;
    }

    return ($decSign * $decVal);


} #end DECdeg



# return ok for the require statement
return 1;

