package hist1d;



use bsearch;



##***********************************************************************
#  construct_1dhist -- local to hist1d.pm (A.B. 5/02)	                *
#									*
#  A.  Purpose:	The 1dhist constructor creates a 1d histogram           *
#       object, loads it with the appropriate values, and               *
#       returns a pointer to it.  This version of the constructor       *
#       creates a histogram structure and fills it based on the input   *
#       arguments.                      				*
#									*
#  B.  Background and use:						*
#									*
#  C.  Called by:							*
#									*
#  D.  Input/output files:						*
#									*
#  E.  Parameter description:						*
#	Input -								*
#        title -- the title string for the histogram (char *).          *
#        num_bins -- the number of bins in the histogram (int).         *
#        low_edge -- the lower edge of the histogram range (float).     *
#        upp_edge -- the upper edge of the histogram range (float).     *
#									*
#  F.  Detail function description:					*
#	      1.  Local Functions:					*
#	      2.  Externals:						*
#	      3.  Globals and Nonlocals used:				*
#	      4.  Global and Nonlocal variables modified:		*
#	      5.  Function flow is as follows:				*
#	      6.  Functions called by					*
#									*
#  G.  Function value: This function returns a pointer to the created   *
#       1d histogram data structure, or a NIL pointer if there is a     *
#       memory allocation error.	                        	*
#									*
##***********************************************************************/
##********#  function declaration construct_1dhist  *********************/
sub new
{
##********#  local declarations construct_1dhist  ***********************/

    my $class = shift @_;  ## yank-off the class designation

    ## Create the hash with the rest of the stuff from the call argument
    ##  list
    my $objRef = { @_ };

##********#  begin construct_1dhist  ************************************/
    ## Now work through some of the most important fields
    if (!($objRef->{title}))  ## Assign a dummy title
    {
	$objRef->{title} = "Dummy Title";
    }

    if (!($objRef->{numBins}))
    {
	die "histogram numBins must be defined for $objRef->{title}\n";
    }

    if (!(grep /lowerEdge/, keys %$objRef)
	|| !(grep /upperEdge/, keys %$objRef))
    {
	die "histogram lowerEdge and upperEdge must be defined for $objRef->{title}\n";
    }


    $objRef->{overflow} = $objRef->{underflow} = $objRef->{entries} = 0;

    my (@bin_val,@data_val);
    

    if ($objRef->{binSpacing} =~ m/log/)
    {
	my $ratio = $objRef->{upperEdge} / $objRef->{lowerEdge};
	$ratio = $ratio**(1.0 / $objRef->{numBins});

	for (my $i = 0; $i < $objRef->{numBins}; $i++)
	{
	    ## bin_val are the bin centers!!!
	    $bin_val[$i] = $objRef->{lowerEdge}
	                    * $ratio**($i + 0.5);
	    $data_val[$i] = 0.0;
	} ## end $i loop
    }  ## log spacing
    else  ## default linear spacing
    {
	my $binSize
	    = ($objRef->{upperEdge} - $objRef->{lowerEdge}) / $objRef->{numBins};
	my $hbinSize = $binSize / 2.0;

	for (my $i = 0; $i < $objRef->{numBins}; $i++)
	{
	    ## bin_val are the bin centers!!!
	    $bin_val[$i] = $objRef->{lowerEdge}
	                    + $hbinSize
	                    + ($i * $binSize);
	    $data_val[$i] = 0.0;
	} ## end $i loop
    } ## end linear spacing

    $objRef->{binVal} = [];
    @{ $objRef->{binVal} } = @bin_val;
    $objRef->{dataVal} = [];
    @{ $objRef->{dataVal} } = @data_val;
    
    $objRef->{sumWeight} = $objRef->{sumVal} = $objRef->{sumVal2} = 0.0;

    $objRef->{minEntry} = 0.0;  ## Extrema initial values
    $objRef->{maxEntry} = 0.0;

    bless $objRef;

    return $objRef;

##********#  end construct_1dhist  **************************************/
}



##***********************************************************************
#  enter_1dhist -- local to hist_1d.cc (A.B. 6/27/93)		        *
#									*
#  A.  Purpose:	The function enter_1dhist places an entry into a        *
#       given 1d histogram.						*
#									*
#  B.  Background and use:						*
#									*
#  C.  Called by:							*
#									*
#  D.  Input/output files:						*
#									*
#  E.  Parameter description:						*
#	Input -								*
#        ind_value -- the independent value of the data pair to be      *
#         entered into the histogram (float).                           *
#        data_value -- the dependent value of the data pair to be       *
#         entered into the histogram (float).                           *
#									*
#  F.  Detail function description:					*
#	      1.  Local Functions:					*
#	      2.  Externals:						*
#	      3.  Globals and Nonlocals used:				*
#	      4.  Global and Nonlocal variables modified:		*
#	      5.  Function flow is as follows:				*
#	      6.  Functions called by					*
#									*
#  G.  Function value: This function returns no useable value.		*
#									*
##***********************************************************************/
sub enter  ##(float ind_value, float data_value)
{
##*********  local declarations enter_1dhist  ***************************/

    my $this = shift @_;

    my $indVal = shift @_;
    my $dataVal;
    if (scalar @_)
    {
	$dataVal = shift @_;
    }
    else  ## default data value
    {
	$dataVal = 1.0;
    }

##*********  begin enter_1dhist  ****************************************/
    ## Start by testing for bogosity on input
    if (($indVal !~ m/\d/) || ($dataVal !~ m/\d/))
    {
	return;
    }

    ## Update sums for statistics
    $this->{sumVal} += $indVal * $dataVal;
    $this->{sumVal2} += $indVal * $indVal * $dataVal;
    $this->{sumWeight} += $dataVal;


    if ($this->{entries} == 0)  ## If this is the first entry
    {
	$this->{minEntry} = $this->{maxEntry} = $indVal;
    }
    else
    {
	$this->{minEntry} = $indVal if ($indVal < $this->{minEntry});
	$this->{maxEntry} = $indVal if ($indVal > $this->{maxEntry});
    }


    ## Check bounds
    if ($indVal < $this->{lowerEdge})
    {
	$this->{underflow}++;
	return;
    }
    if ($indVal > $this->{upperEdge})
    {
	$this->{overflow}++;
	return;
    }

    ## Make the entry
    my $bin = bsearch::hist_binary_search($indVal,@{ $this->{binVal} });
    ${ $this->{dataVal} }[$bin] += $dataVal;


    $this->{entries} += $dataVal;


##*********  end enter_1dhist  ******************************************/
}



##************************************************************************
#  write_1dhist -- local to hist_functions.c (A.B. 12/10/93)		*
#									*
#  A.  Purpose: The function write_1dhist writes the contents of the    *
#       given 1D histogram out to a given ASCII file.  The format is    *
#       chosen to be used with GNUPLOT.					*
#									*
#  B.  Background and use:						*
#									*
#  C.  Called by:							*
#									*
#  D.  Input/output files:						*
#									*
#  E.  Parameter description:						*
#	Input -								*
#        output_file -- the (pointer to the) file the histogram is      *
#         written to (FILE *).                                          *
#									*
#  F.  Detail function description:					*
#	      1.  Local Functions:					*
#	      2.  Externals:						*
#	      3.  Globals and Nonlocals used:				*
#	      4.  Global and Nonlocal variables modified:		*
#	      5.  Function flow is as follows:				*
#	      6.  Functions called by					*
#									*
#  G.  Function value: This function returns no useable value.		*
#									*
#************************************************************************/
#**********  function declaration write_1dhist  *************************/
sub write
{
##*********  local declarations write_1dhist  ***************************/

    my $this = shift @_;
    my $summaryOnly = shift @_;

    my $output = "";

##*********  begin write_1dhist  ****************************************/

    $output .= sprintf "### Histogram Summary %s\n", $this->{title};
    $output .= sprintf "#   1d -- Num Bins: %d\t\tNum Entries: %d\n",
	$this->{numBins},$this->{entries};

    $output .= sprintf "#   Lower Edge: %g\tUpper Edge: %g ",
       $this->{lowerEdge},$this->{upperEdge};
    if ($this->{binSpacing} =~ m/log/)
    {
	$output .= " logarithmic bin spacing\n";
    }
    else
    {
	my $binSize = ($this->{upperEdge} - $this->{lowerEdge}) / ($this->{numBins});
	$output .= " linear bin spacing -- bin size $binSize\n";
    }

    $output .= sprintf "#   Underflows: %d\t\tOverflows: %d\n",
       $this->{underflow},$this->{overflow};
    $output .= sprintf "#   Minimum Entry: %g\tMaximum Entry: %g\n",
       $this->{minEntry},$this->{maxEntry};
    $output .= sprintf "#   Mode: %g\tMaximum Value: %g\n",
       $this->mode(),$this->maxVal();
    $output .= sprintf "#   Mean: %g\t\tStdDev: %g\n",
       $this->mean(),$this->sigma();

    $output .= "\n\n\n" if ($summaryOnly);
    return $output if ($summaryOnly);

    for (my $i = 0; $i < $this->{numBins}; $i++)
    {
	$output .= sprintf "%g\t%g\n",
	    ${ $this->{binVal} }[$i],${ $this->{dataVal} }[$i];
    }

    $output .= sprintf "\n\n\n";


    return $output;

##*********  end write_1dhist  ******************************************/
} 



##************************************************************************
#  read_1dhist -- local to hist_functions.c (A.B. 12/10/93)		*
#									*
#  A.  Purpose: The function read_1dhist reads the contents of the      *
#       given 1D histogram from an input ASCII string.  The format      *
#       matches that written by write_1dhist.				*
#									*
#  B.  Background and use:						*
#									*
#  C.  Called by:							*
#									*
#  D.  Input/output files:						*
#									*
#  E.  Parameter description:						*
#	Input -								*
#        output_file -- the (pointer to the) file the histogram is      *
#         written to (FILE *).                                          *
#									*
#  F.  Detail function description:					*
#	      1.  Local Functions:					*
#	      2.  Externals:						*
#	      3.  Globals and Nonlocals used:				*
#	      4.  Global and Nonlocal variables modified:		*
#	      5.  Function flow is as follows:				*
#	      6.  Functions called by					*
#									*
#  G.  Function value: This function returns no useable value.		*
#									*
#************************************************************************/
#**********  function declaration read_1dhist  **************************/
sub read
{
##*********  local declarations read_1dhist  ****************************/

    my $inputStream = shift @_;

##*********  begin read_1dhist  *****************************************/

    my @streamLines = split /\n/, $inputStream;

    my @comments = grep /^\#/, @streamLines;

    my ($title,$numBins,$numEntries,$dummy,$lowerEdge,$upperEdge,
	$binSpacing,$underflows,$overflows,$minEntry,$maxEntry);

    foreach my $line (@comments)
    {
	if ($line =~ m/Histogram Summary/)
	{
	    $title = $line;
	    $title =~ s/. Histogram Summary //;
	}

	if ($line =~ m/\#   1d -- Num Bins: \d+\t\tNum Entries: \d+/)
	{
	    $line =~ s/\#   1d -- Num Bins: //;
	    ($numBins,$dummy) = split /\s+/, $line, 2;
	}

	if ($line =~ m/^\#   Lower Edge: /)
	{
	    $line =~ s/^\#   Lower Edge: //g;
	    $line =~ s/Upper Edge: //g;
	    chomp $line;
	    ($lowerEdge,$upperEdge,$binSpacing) = split /\s+/, $line, 3;
	    if ($binSpacing =~ m/log/)
	    {
		$binSpacing = "log";
	    }
	    else
	    {
		$binSpacing = "linear";
	    }
	}

	if ($line =~ m/^\#   Underflows: /)
	{
	    $line =~ s/^\#   Underflows: //g;
	    $line =~ s/Overflows: //g;
	    chomp $line;
	    ($underflows,$overflows) = split /\s+/, $line;
	} ## 

	if ($line =~ m/^\#   Minimum Entry: /)
	{
	    $line =~ s/^\#   Minimum Entry: //g;
	    $line =~ s/Maximum Entry: //g;
	    chomp $line;
	    ($minEntry,$maxEntry) = split /\s+/, $line;
	} ## min/max entry
	

    } ## end foreach

    ## Now create a histogram object...
    $hist = new(hist1d,
		title => $title,
		numBins => $numBins,
		lowerEdge => $lowerEdge,
		upperEdge => $upperEdge,
		binSpacing => $binSpacing);

    ## ...and insert the data
    my @dataLines = grep {!/^\#/} @streamLines;
    foreach my $line (@dataLines)
    {
	chomp $line;
	my ($x,$y) = split /\s+/, $line, 2;
	$hist->enter($x,$y) if ($y);
    } ## end foreach

    ## Set some extra stuff
    $hist->{underflow} = $underflow;
    $hist->{overflow} = $overflow;
    $hist->{minEntry} = $minEntry;
    $hist->{maxEntry} = $maxEntry;

    return $hist;


##*********  end read_1dhist  *******************************************/
} 




## Histogram statistics functions
sub mean
{
    my $this = shift @_;
    if ($this->{sumWeight})
    {
	return ($this->{sumVal} / $this->{sumWeight});
    }
    else
    {
	print STDERR "Warning -- $this->{title} empty histogram statistics\n";
	return 0.0;
    }
} ## end mean


sub sigma
{
    my $this = shift @_;
    if ($this->{sumWeight})
    {
	my $variance = $this->{sumWeight} / ($this->{sumWeight} - 1.0)
	    * (($this->{sumVal2} / $this->{sumWeight}) - $this->mean() * $this->mean());
	if ($variance > 0)
	{
	    return sqrt($variance);
	}
	else
	{
	    return 0;
	}
    }
    else
    {
	print STDERR "Warning -- $this->{title} empty histogram statistics\n";
	return 0.0;
    }
} ## end sigma


sub percentile
{
    my $this = shift @_;
    my $dataVal = shift @_;

    if ($this->{entries} == 0)
    {
	print STDERR "Warning -- $this->{title} empty histogram statistics\n";
	return 0.0;
    } ## end empty histogram check

    if ($dataVal < $this->{lowerEdge})
    {
	return 0.0;
    } ## end below lowerEdge check

    if ($dataVal > $this->{upperEdge})
    {
	return 100.0;
    } ## end above upperEdge check

    my $binIndex = 0;
    my $ltEntries = 0;
    while (($binIndex < $this->{numBins})
	   && (${ $this->{binVal} }[$binIndex] < $dataVal))
    {
	$ltEntries += ${ $this->{dataVal} }[$binIndex];
        $binIndex++;
    } ## end while

    return 100.0 * ($ltEntries / $this->{entries});
	

} ## end percentile




sub maxBin
{

    my $this = shift @_;

    ## Find peak bin and value
    my $peakBin = 0;
    my $peakVal = ${ $this->{dataVal} }[$peakBin];
    for (my $i = 1; $i < $this->{numBins}; $i++)
    {
	if (${ $this->{dataVal} }[$i] > $peakVal)
        {
             $peakBin = $i;
	     $peakVal = ${ $this->{dataVal} }[$i];
        } ## end if
    } ## end for

    return ($peakBin,$peakVal);
}  ## end maxBin



sub maxVal
{
    my $this = shift @_;
    my ($peakBin,$peakVal) = $this->maxBin();
    return $peakVal;
}


sub mode
{
    my $this = shift @_;
    my ($peakBin,$peakVal) = $this->maxBin();
    return ${ $this->{binVal} }[$peakBin];
}





## Member accessor/manipulation functions
sub title
{
    my $this = shift @_;
    if (@_) { $this->{title} = shift @_ }
    return $this->{title};
}

sub numBins
{
    my $this = shift @_;
    ## if (@_) { $this->{numBins} = shift @_ }  ## No support for manipulating bins
    return $this->{numBins};
}


sub lowerEdge
{
    my $this = shift @_;
    ## if (@_) { $this->{lowerEdge} = shift @_ }  ## No support for manipulating bins
    return $this->{lowerEdge};
}

sub upperEdge
{
    my $this = shift @_;
    ## if (@_) { $this->{upperEdge} = shift @_ }  ## No support for manipulating bins
    return $this->{upperEdge};
}


sub bounds
{
    my $this = shift @_;
    ## if (@_) { $this->{upperEdge} = shift @_ }  ## No support for manipulating bins
    return ($this->{lowerEdge},$this->{upperEdge});
}


sub binVal
{
    my $this = shift @_;

    if (@_)
    {
	my $index = shift @_;
	return ${ $this->{binVal} }[$index];
    }
    else
    {
	return @{ $this->{binVal} };
    } 
}



sub dataVal
{
    my $this = shift @_;

    if (@_)
    {
	my $index = shift @_;
	return ${ $this->{dataVal} }[$index];
    }
    else
    {
	return @{ $this->{dataVal} };
    } 
}






return 1;
