package hist2d;



use bsearch;



##***********************************************************************
#  construct_2dhist -- local to hist2d.pm (A.B. 5/02)	                *
#									*
#  A.  Purpose:	The 2dhist constructor creates a 2d histogram           *
#       object, loads it with the appropriate values, and               *
#       returns a pointer to it.  This version of the constructor       *
#       creates a histogram structure and fills it based on the input   *
#       arguments.                      				*
#									*
#  B.  Background and use:						*
#									*
#  C.  Called by:							*
#									*
#  D.  Input/output files:						*
#									*
#  E.  Parameter description:						*
#	Input -								*
#        title -- the title string for the histogram (char *).          *
#        num_bins -- the number of bins in the histogram (int).         *
#        low_edge -- the lower edge of the histogram range (float).     *
#        upp_edge -- the upper edge of the histogram range (float).     *
#									*
#  F.  Detail function description:					*
#	      1.  Local Functions:					*
#	      2.  Externals:						*
#	      3.  Globals and Nonlocals used:				*
#	      4.  Global and Nonlocal variables modified:		*
#	      5.  Function flow is as follows:				*
#	      6.  Functions called by					*
#									*
#  G.  Function value: This function returns a pointer to the created   *
#       2d histogram data structure, or a NIL pointer if there is a     *
#       memory allocation error.	                        	*
#									*
##***********************************************************************/
##********#  function declaration construct_2dhist  *********************/
sub new
{
##********#  local declarations construct_2dhist  ***********************/

    my $class = shift @_;  ## yank-off the class designation

    ## Create the hash with the rest of the stuff from the call argument
    ##  list
    my $objRef = { @_ };

##********#  begin construct_2dhist  ************************************/
    ## Now work through some of the most important fields
    if (!($objRef->{title}))  ## Assign a dummy title
    {
	$objRef->{title} = "Dummy Title";
    }

    if (!($objRef->{numXBins}))
    {
	die "histogram numXBins must be defined for $objRef->{title}\n";
    }

    if (!($objRef->{numYBins}))
    {
	die "histogram numYBins must be defined for $objRef->{title}\n";
    }

    if (!(grep /lowerXEdge/, keys %$objRef)
	|| !(grep /upperXEdge/, keys %$objRef))
    {
	die "histogram lowerXEdge and upperXEdge must be defined for $objRef->{title}\n";
    }


    if (!(grep /lowerYEdge/, keys %$objRef)
	|| !(grep /upperYEdge/, keys %$objRef))
    {
	die "histogram lowerYEdge and upperYEdge must be defined for $objRef->{title}\n";
    }


    ## Initialize various fields
    $objRef->{overflowX} = $objRef->{underflowX} = $objRef->{overflowY}
     = $objRef->{underflowY} = $objRef->{entries} = 0;


    ## Initialize the bin structures
    my (@Xbin_val,@Ybin_val);

    ## X
    if ($objRef->{XbinSpacing} =~ m/log/)
    {
	my $ratio = $objRef->{upperXEdge} / $objRef->{lowerXEdge};
	$ratio = $ratio**(1.0 / $objRef->{numXBins});

	for (my $i = 0; $i < $objRef->{numXBins}; $i++)
	{
	    ## bin_val are the bin centers!!!
	    $Xbin_val[$i] = $objRef->{lowerXEdge}
	                    * $ratio**($i + 0.5);
	} ## end $i loop
    }  ## log spacing
    else  ## default linear spacing
    {
	my $binSize
	    = ($objRef->{upperXEdge} - $objRef->{lowerXEdge}) / $objRef->{numXBins};
	my $hbinSize = $binSize / 2.0;

	for (my $i = 0; $i < $objRef->{numXBins}; $i++)
	{
	    ## bin_val are the bin centers!!!
	    $Xbin_val[$i] = $objRef->{lowerXEdge}
	                    + $hbinSize
	                    + ($i * $binSize);
	} ## end $i loop
    } ## end linear spacing

    $objRef->{XbinVal} = [];
    @{ $objRef->{XbinVal} } = @Xbin_val;


    ## Y
    if ($objRef->{YbinSpacing} =~ m/log/)
    {
	my $ratio = $objRef->{upperYEdge} / $objRef->{lowerYEdge};
	$ratio = $ratio**(1.0 / $objRef->{numYBins});

	for (my $i = 0; $i < $objRef->{numYBins}; $i++)
	{
	    ## bin_val are the bin centers!!!
	    $Ybin_val[$i] = $objRef->{lowerYEdge}
	                    * $ratio**($i + 0.5);
	} ## end $i loop
    }  ## log spacing
    else  ## default linear spacing
    {
	my $binSize
	    = ($objRef->{upperYEdge} - $objRef->{lowerYEdge}) / $objRef->{numYBins};
	my $hbinSize = $binSize / 2.0;

	for (my $i = 0; $i < $objRef->{numYBins}; $i++)
	{
	    ## bin_val are the bin centers!!!
	    $Ybin_val[$i] = $objRef->{lowerYEdge}
	                    + $hbinSize
	                    + ($i * $binSize);
	} ## end $i loop
    } ## end linear spacing

    $objRef->{YbinVal} = [];
    @{ $objRef->{YbinVal} } = @Ybin_val;


    ## Initialize data array
    ## $objRef->{dataVal};
    for (my $i = 0; $i < $objRef->{numXBins}; $i++)
    {
	for (my $j = 0; $j < $objRef->{numYBins}; $j++)
	{
	    ${ $objRef->{dataVal} }[$i][$j] = 0.0;
        }
    } ## end data initialization loop


    ## Initialize statistics counters
    $objRef->{sumWeight} = $objRef->{sumXVal} = $objRef->{sumXVal2} = 0.0;
    $objRef->{sumYVal} = $objRef->{sumYVal2} = 0.0;
    $objRef->{sumXYVal} = 0.0;
    $objRef->{minXEntry} = 0.0;  ## Extrema initial values
    $objRef->{maxXEntry} = 0.0;
    $objRef->{minYEntry} = 0.0;  ## Extrema initial values
    $objRef->{maxYEntry} = 0.0;

    bless $objRef;

    return $objRef;

##********#  end construct_2dhist  **************************************/
}



##***********************************************************************
#  enter_2dhist -- local to hist_2d.cc (A.B. 6/27/93)		        *
#									*
#  A.  Purpose:	The function enter_2dhist places an entry into a        *
#       given 2d histogram.						*
#									*
#  B.  Background and use:						*
#									*
#  C.  Called by:							*
#									*
#  D.  Input/output files:						*
#									*
#  E.  Parameter description:						*
#	Input -								*
#        ind_value -- the independent value of the data pair to be      *
#         entered into the histogram (float).                           *
#        data_value -- the dependent value of the data pair to be       *
#         entered into the histogram (float).                           *
#									*
#  F.  Detail function description:					*
#	      1.  Local Functions:					*
#	      2.  Externals:						*
#	      3.  Globals and Nonlocals used:				*
#	      4.  Global and Nonlocal variables modified:		*
#	      5.  Function flow is as follows:				*
#	      6.  Functions called by					*
#									*
#  G.  Function value: This function returns no useable value.		*
#									*
##***********************************************************************/
sub enter(float ind_value, float data_value)
{
##*********  local declarations enter_2dhist  ***************************/

    my $this = shift @_;

    my $xVal = shift @_;
    my $yVal = shift @_;
    my $dataVal;
    if (scalar @_)
    {
	$dataVal = shift @_;
    }
    else  ## default data value
    {
	$dataVal = 1.0;
    }

##*********  begin enter_2dhist  ****************************************/
    ## Update sums for statistics
    $this->{sumXVal} += $xVal * $dataVal;
    $this->{sumXVal2} += $xVal * $xVal * $dataVal;
    $this->{sumYVal} += $yVal * $dataVal;
    $this->{sumYVal2} += $yVal * $yVal * $dataVal;
    $this->{sumXYVal} += $xVal * $yVal * $dataVal;
    $this->{sumWeight} += $dataVal;


    if ($this->{entries} == 0)  ## If this is the first entry
    {
	$this->{minXEntry} = $this->{maxXEntry} = $xVal;
	$this->{minYEntry} = $this->{maxYEntry} = $yVal;
    }
    else
    {
	$this->{minXEntry} = $xVal if ($xVal < $this->{minXEntry});
	$this->{maxXEntry} = $xVal if ($xVal > $this->{maxXEntry});
	$this->{minXEntry} = $yVal if ($yVal < $this->{minXEntry});
	$this->{maxXEntry} = $yVal if ($yVal > $this->{maxXEntry});
    }


    ## Check bounds
    if ($xVal < $this->{lowerXEdge})
    {
	$this->{Xunderflow}++;
	return;
    }
    if ($xVal > $this->{upperXEdge})
    {
	$this->{Xoverflow}++;
	return;
    }
    if ($yVal < $this->{lowerYEdge})
    {
	$this->{Yunderflow}++;
	return;
    }
    if ($yVal > $this->{upperYEdge})
    {
	$this->{Yoverflow}++;
	return;
    }


    ## Make the entry
    my $xBin = bsearch::hist_binary_search($xVal,@{ $this->{XbinVal} });
    my $yBin = bsearch::hist_binary_search($yVal,@{ $this->{YbinVal} });
    ${ $this->{dataVal} }[$xBin][$yBin] += $dataVal;


    $this->{entries}++;


##*********  end enter_2dhist  ******************************************/
}



##************************************************************************
#  write_2dhist -- local to hist_functions.c (A.B. 12/10/93)		*
#									*
#  A.  Purpose: The function write_2dhist writes the contents of the    *
#       given 2D histogram out to a given ASCII file.  The format is    *
#       chosen to be used with GNUPLOT.					*
#									*
#  B.  Background and use:						*
#									*
#  C.  Called by:							*
#									*
#  D.  Input/output files:						*
#									*
#  E.  Parameter description:						*
#	Input -								*
#        output_file -- the (pointer to the) file the histogram is      *
#         written to (FILE *).                                          *
#									*
#  F.  Detail function description:					*
#	      1.  Local Functions:					*
#	      2.  Externals:						*
#	      3.  Globals and Nonlocals used:				*
#	      4.  Global and Nonlocal variables modified:		*
#	      5.  Function flow is as follows:				*
#	      6.  Functions called by					*
#									*
#  G.  Function value: This function returns no useable value.		*
#									*
#************************************************************************/
#**********  function declaration write_2dhist  *************************/
sub write
{
##*********  local declarations write_2dhist  ***************************/

    my $this = shift @_;
    my $summaryOnly = shift @_;

    my $output = "";

##*********  begin write_2dhist  ****************************************/

    $output .= sprintf "### Histogram Summary %s\n", $this->{title};
    $output .= sprintf "#   2d -- Num X Bins: %d\tNum Y Bins: %d\tNum Entries: %d\n",
	$this->{numXBins},$this->{numYBins},$this->{entries};
    $output .= sprintf "#   Lower X Edge: %g\tUpper X Edge: %g\n",
       $this->{lowerXEdge},$this->{upperXEdge};
    $output .= sprintf "#   Lower Y Edge: %g\tUpper Y Edge: %g\n",
       $this->{lowerYEdge},$this->{upperYEdge};
    $output .= sprintf "#   X Underflows: %d\tX Overflows: %d\n",
       $this->{Xunderflow},$this->{Xoverflow};
    $output .= sprintf "#   Y Underflows: %d\tY Overflows: %d\n",
       $this->{Yunderflow},$this->{Yoverflow};
    $output .= sprintf "#   Minimum X Entry: %g\tMaximum X Entry: %g\n",
       $this->{minXEntry},$this->{maxXEntry};
    $output .= sprintf "#   Minimum Y Entry: %g\tMaximum Y Entry: %g\n",
       $this->{minXEntry},$this->{maxXEntry};
    my ($x,$y,$maxV) = $this->mode();
    $output .= sprintf "#   X,Y Mode: %g,%g\tMaximum Value: %g\n",
       $x,$y,$maxV;
    $output .= sprintf "#   MeanX: %g\t\tStdDevX: %g\n",
       $this->meanX(),$this->sigmaX();
    $output .= sprintf "#   MeanY: %g\t\tStdDevY: %g\n",
       $this->meanY(),$this->sigmaY();
    $output .= sprintf "#   X-Y correlation: %g\t\tX-Y covariance: %g\n",
       $this->correlation(),$this->covXY();

    $output .= "\n\n\n" if ($summaryOnly);
    return $output if ($summaryOnly);

    for (my $i = 0; $i < $this->{numXBins}; $i++)
    {
	for (my $j = 0; $j < $this->{numYBins}; $j++)
	{
	    $output .= sprintf "%g\t%g\t%g\n",
	     ${ $this->{XbinVal} }[$i],
	     ${ $this->{YbinVal} }[$j],
	     ${ $this->{dataVal} }[$i][$j];
        } ## end $j loop
        $output .= "\n";
    } ## end $i loop

    $output .= sprintf "\n\n\n";


    return $output;

##*********  end write_2dhist  ******************************************/
} 




## Histogram statistics functions
sub meanX
{
    my $this = shift @_;
    if ($this->{sumWeight})
    {
	return ($this->{sumXVal} / $this->{sumWeight});
    }
    else
    {
	print "Error -- empty histogram statistics\n";
	return 0.0;
    }
} ## end meanX


sub meanY
{
    my $this = shift @_;
    if ($this->{sumWeight})
    {
	return ($this->{sumYVal} / $this->{sumWeight});
    }
    else
    {
	print "Error -- empty histogram statistics\n";
	return 0.0;
    }
} ## end meanY


sub sigmaX
{
    my $this = shift @_;
    if ($this->{sumWeight})
    {
	my $variance = $this->{sumWeight} / ($this->{sumWeight} - 1.0)
	    * (($this->{sumXVal2} / $this->{sumWeight}) - $this->meanX() * $this->meanX());
	return sqrt($variance);
    }
    else
    {
	print "Error -- empty histogram statistics\n";
	return 0.0;
    }
} ## end sigmaX


sub sigmaY
{
    my $this = shift @_;
    if ($this->{sumWeight})
    {
	my $variance = $this->{sumWeight} / ($this->{sumWeight} - 1.0)
	    * (($this->{sumYVal2} / $this->{sumWeight}) - $this->meanY() * $this->meanY());
	return sqrt($variance);
    }
    else
    {
	print "Error -- empty histogram statistics\n";
	return 0.0;
    }
} ## end sigmaY


sub covXY
{
    my $this = shift @_;
    if ($this->{sumWeight})
    {
	my $covariance = $this->{sumWeight} / ($this->{sumWeight} - 1.0)
	    * (($this->{sumXYVal} / $this->{sumWeight}) - $this->meanX() * $this->meanY());
	return $covariance;
    }
    else
    {
	print "Error -- empty histogram statistics\n";
	return 0.0;
    }
	
} ## end covXY


sub correlation
{
    my $this = shift @_;
    if ($this->{sumWeight})
    {
	return ($this->covXY() / ($this->sigmaX() * $this->sigmaY()));
    }
} ## end correlation




sub maxBin
{
    my $this = shift @_;
    ## Find peak bin and value
    my $peakX = 0;
    my $peakY = 0;
    my $peakVal = ${ $this->{dataVal} }[$peakX][$peakY];
    for (my $i = 1; $i < $this->{numXBins}; $i++)
    {
	for (my $j = 0; $j < $this->{numYBins}; $j++)
	{
	    if (${ $this->{dataVal} }[$i][$j] > $peakVal)
            {
		$peakX = $i;
		$peakY = $j;
		$peakVal = ${ $this->{dataVal} }[$i][$j];
            } ## end if
        } ## end $j loop
    } ## end $i loop

    return ($peakX,$peakY,$peakVal);
}  ## end maxBin



sub maxVal
{
    my $this = shift @_;
    my ($peakX,$peakY,$peakVal) = $this->maxBin();
    return ($peakX,$peakY);
}


sub mode
{
    my $this = shift @_;
    my ($peakX,$peakY,$peakVal) = $this->maxBin();
    return (${ $this->{XbinVal} }[$peakX],
            ${ $this->{YbinVal} }[$peakY],$peakVal);
}




## Member accessor/manipulation functions
sub title
{
    my $this = shift @_;
    if (@_) { $this->{title} = shift @_ }
    return $this->{title};
}

sub numXBins
{
    my $this = shift @_;
    ## if (@_) { $this->{numBins} = shift @_ }  ## No support for manipulating bins
    return $this->{numXBins};
}

sub numYBins
{
    my $this = shift @_;
    ## if (@_) { $this->{numBins} = shift @_ }  ## No support for manipulating bins
    return $this->{numYBins};
}


sub lowerXEdge
{
    my $this = shift @_;
    ## if (@_) { $this->{lowerEdge} = shift @_ }  ## No support for manipulating bins
    return $this->{lowerXEdge};
}

sub upperXEdge
{
    my $this = shift @_;
    ## if (@_) { $this->{upperEdge} = shift @_ }  ## No support for manipulating bins
    return $this->{upperXEdge};
}

sub lowerYEdge
{
    my $this = shift @_;
    ## if (@_) { $this->{lowerEdge} = shift @_ }  ## No support for manipulating bins
    return $this->{lowerYEdge};
}

sub upperYEdge
{
    my $this = shift @_;
    ## if (@_) { $this->{upperEdge} = shift @_ }  ## No support for manipulating bins
    return $this->{upperYEdge};
}


sub Xbounds
{
    my $this = shift @_;
    ## if (@_) { $this->{upperEdge} = shift @_ }  ## No support for manipulating bins
    return ($this->{lowerXEdge},$this->{upperXEdge});
}

sub Ybounds
{
    my $this = shift @_;
    ## if (@_) { $this->{upperEdge} = shift @_ }  ## No support for manipulating bins
    return ($this->{lowerYEdge},$this->{upperYEdge});
}


sub XbinVal
{
    my $this = shift @_;

    if (@_)
    {
	my $Xindex = shift @_;
	return ${ $this->{XbinVal} }[$Xindex];
    }
    else
    {
	return @{ $this->{XbinVal} };
    }
}


sub YbinVal
{
    my $this = shift @_;

    if (@_)
    {
	my $Yindex = shift @_;
	return ${ $this->{YbinVal} }[$Yindex];
    }
    else
    {
	return @{ $this->{YbinVal} };
    }
}



sub dataVal
{
    my $this = shift @_;

    if (@_)
    {
	my $Xindex = shift @_;
	my $Yindex = shift @_;
	return ${ $this->{dataVal} }[$Xindex][$Yindex];
    }
    else
    {
	return @{ $this->{dataVal} };
    } 
}






return 1;
