#include <math.h>

#include "novas.h"

extern double djl(int kyear,int kmonth,int kday);

/* macro to detect and fix modified Julian dates into real Julian dates */

/* Note -- I think this should actually be 2400000.5 -- see
   Astronomical Almanac 1986 p B4 */

#define BASE_JD 2400000.5
#define fixJD(jd) ((jd) > (BASE_JD) ? (jd) : (jd + BASE_JD))

/* MJD of 0 UT Jan 1, 2000 */
#define MJD_J2000 51544.0


#define MEAN_SIDERIAL_DAY 0.99726956634   /* Units of mean solar days */


/************************************************************************
*  gmstZeroUT -- local to myTime.c (A.B. 7/10/97)			*
*									*
*  A.  Purpose:	The function gmstZeroUT computes the Greenwhich mean    *
*       siderial time at zero hours UT on the specified JD.		*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration gmstZeroUT  ***************************/
double gmstZeroUT(double jd)
{
/**********  local declarations gmstZeroUT  *****************************/
#define C1 24110.54841
#define C2 8640184.812866
#define C3 0.093104
#define C4 -6.2e-6

  double Tu,gmstZUT;

/**********  begin gmstZeroUT  ******************************************/
  jd = fixJD(jd);
  /* Compute the gmst at 0 hrs UT (formula from Astronomical Almanac
     1986 p B6) */
  Tu = (jd - 2451545.0) / 36525.0;
  gmstZUT = C1 + (C2 * Tu) + (C3 * Tu * Tu) + (C4 * Tu * Tu * Tu);
  gmstZUT /= 3600.0;

  /* Map it into the principle range */
  while (gmstZUT < 0)
    gmstZUT += 24.0;

  while (gmstZUT > 24.0)
    gmstZUT -= 24.0;

  return gmstZUT;

/**********  end gmstZeroUT  ********************************************/
}


/************************************************************************
*  utFromJD -- local to myTime.c (A.B. 7/7/97)				*
*									*
*  A.  Purpose:								*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration utFromJD  *****************************/
double utFromJD(double jd)
{
  double jdInt,jdFrac;
  double gmst,ut;

/**********  begin utFromJD  ********************************************/
  /* Test/fix if this is a Modified Julian date */
  jd = fixJD(jd);
  /* Compute the current Greenwhich mean siderial time (gmst) */
  jdInt = ((double) ((long) jd));
  jdFrac = jd - jdInt;
  (void) sidereal_time(jdInt,jdFrac,0.0,&gmst);

  /* Compute UT */
  ut = gmst - gmstZeroUT(jd);

  if (ut < 0)
    ut += 24.0;

  /* ut *= MEAN_SIDERIAL_DAY; Into units of solar days */

  return ut;
/**********  end utFromJD  **********************************************/
}



/************************************************************************
*									*
*									*
*  A.  Purpose:								*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
/**********  function declaration jdFromUTCal  **************************/
double jdFromUTCal(double ut,int year,int month,int day)
{
/**********  begin jdFromUTCal  *****************************************/
  double jd = djl(year,month,day);

#ifdef RETURN_MODIFIED_JD
  jd -= BASE_JD;
#endif

  jd += ut / 24.0;
  return jd;

/**********  end jdFromUTCal  *******************************************/
}


/************************************************************************
*  jdFromUTYday -- local to myTime.c (A.B. 7/11/97)			*
*									*
*  A.  Purpose: jdFromUTYday is a trivial utility to compute the Julian *
*       date from the input ut, year and yearday.		        *
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*	Output -							*
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns no useable value.		*
*									*
*************************************************************************/
double jdFromUTYday(double ut,int year,int yday)
{
  int month,day;

  (void) intoMonthDay(year,yday,&month,&day);
  return jdFromUTCal(ut,year,month,day);
}



/************************************************************************
*  approximateDeltaT -- local to timeUtil.c (A.B. 11/01)		*
*									*
*  A.  Purpose: The function approximateDeltaT computes an              *
*       _approximate_ Julian date to dynamical time correction based    *
*       on a linear fit to the predicted deltaT correction published    *
*       by the Naval Observatory at the URL                             *
* 	ftp://maia.usno.navy.mil/ser7/deltat.preds			*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- Julian Date (possibly MJD)                               *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the computed deltaT        *
*       estimate (days) (sec)                                           *
*									*
*************************************************************************/
/**********  function declaration approximateDeltaT  ********************/
double approximateDeltaT(double jd)
{
#define DT_SLOPE 0.67547
#define DT_INTERCEPT 63.0504

  double dT;

/**********  begin approximateDeltaT  ***********************************/
  jd = fixJD(jd); /* Determine whether we have to (and) fix this JD */
  jd -= BASE_JD;  /* Redundant, but guarantees we are in MJD */
  jd -= MJD_J2000; /* Days since Jan 1, 2000 */

  dT = DT_INTERCEPT + DT_SLOPE * (jd / 365.00); /* DeltaT in sec */

  return dT;

/**********  end approximateDeltaT  *************************************/
}



/************************************************************************
*  dynamicalTimeJD -- local to timeUtil.c (A.B. 11/01)                  *
*									*
*  A.  Purpose:								*
*									*
*  B.  Background and use:						*
*									*
*  C.  Called by:							*
*									*
*  D.  Input/output files:						*
*									*
*  E.  Parameter description:						*
*	Input -								*
*        jd -- the Julian date (possibly MJD) of epoch (double).        *
*	Output -							*
*        dT -- the deltaT JD/TDT correction (sec) (double).             *
*									*
*  F.  Detail function description:					*
*	      1.  Local Functions:					*
*	      2.  Externals:						*
*	      3.  Globals and Nonlocals used:				*
*	      4.  Global and Nonlocal variables modified:		*
*	      5.  Function flow is as follows:				*
*	      6.  Functions called by					*
*									*
*  G.  Function value: This function returns the calculated TDT         *
*       dynamical time tjd (days) (double).		                *
*									*
*************************************************************************/
/**********  function declaration dynamicalTimeJD  **********************/
double dynamicalTimeJD(double jd,double *dT)
{

  double tdt;

/**********  begin dynamicalTimeJD  *************************************/
  *dT = approximateDeltaT(jd); /* Compute dynamical time correction */
  tdt = jd + (*dT / 86400.0); /* Dynamical time for topo_star */

  return tdt;

/**********  end dynamicalTimeJD  ***************************************/
}

